%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(dates, [de_enum_date/4,
		  de_dates/2,
		  date_normal/2,
		  date_w3c/2,
		  date_year/2,
		  date_ymd/4]).

:- encoding(utf8).

%%%% 
%%%% Format of parsed date:
%%%% 
%%%% ymd(Y,M,D)
%%%% ym(Y,M)
%%%% y(Y)
%%%% md(M,D)
%%%% ya(Y,A)  
%%%% a(A)
%%%% d(D)
%%%% m(M)
%%%%
%%%% Y,M,D are integers
%%%% A is an atom representing a movable day
%%%%   

%%%% 
%%%% Note: To detect a date, a delimitating character before and after the
%%%% date are required, two delimitation characters between dates
%%%% 

%%%%
%%%% enum_date(+Codes, D, P)
%%%%
%%%% Enumerates Date and After-Position of parsed dates in Codes,
%%%% in position order from left to right. 
%%%% 
de_enum_date(Codes, D, P) :-
	length(Codes, L),
	phrase(de_enum_date_1(D), Codes, R),
	length(R, L1),
	P is L - L1.

%%%%
%%%% enum_date(+Codes, D, P1, P2)
%%%%
%%%% Enumerates Date, Befor- and After-Position of parsed dates in Codes,
%%%% in position order from left to right. 
%%%% 
de_enum_date(Codes, D, P1, P2) :-
	length(Codes, L),
	phrase(de_enum_date_2(D, P1R, P2R), Codes, _),
	P1 is L - P1R,
	P2 is L - P2R.

de_enum_date_2(D, P1, P2, A, F) :-
	de_skip(A, B),
	de_date(E, B, C),
	!,
	G=C,
	( D=E,
	  F=G,
	  length(B, P1),
	  length(C, P2)
	; de_enum_date_2(D, P1, P2, G, F)
	).

de_enum_date_1(D) -->
	de_skip,
	de_date(D1),
	!,
	( { D = D1 }
	; de_enum_date_1(D)
	).

%%%% 
%%%% List of parsed Dates in Codes, in position order
%%%% 
de_dates(Codes, Dates) :-
	phrase(de_dates_1(Dates), Codes, _).

de_dates_1([D|Ds]) -->
	de_skip,
	de_date(D),
	!,
	de_dates_1(Ds).
de_dates_1([]) -->
	``.

de_skip --> ``.
de_skip --> [_], de_skip.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

de_in_year_range(Y) :-
	Y >= 1000,
	Y =< 3000,
	!.
de_in_year_range(Y) :-
	Y < 100.

de_in_soleyear_range(Y) :-
	Y >= 1000,
	Y =< 3000,
	!.

de_num_month(1) --> `01`.
de_num_month(1) --> `1`.
de_num_month(2) --> `02`.
de_num_month(2) --> `2`.
de_num_month(3) --> `03`.
de_num_month(3) --> `3`.
de_num_month(4) --> `04`.
de_num_month(4) --> `4`.
de_num_month(5) --> `05`.
de_num_month(5) --> `5`.
de_num_month(6) --> `06`.
de_num_month(6) --> `6`.
de_num_month(7) --> `07`.
de_num_month(7) --> `7`.
de_num_month(8) --> `08`.
de_num_month(8) --> `8`.
de_num_month(9) --> `09`.
de_num_month(9) --> `9`.
de_num_month(10) --> `10`.
de_num_month(11) --> `11`.
de_num_month(12) --> `12`.

de_txt_month(1) --> `Januar`.
de_txt_month(1) --> `Jan`.
de_txt_month(1) --> `jan`.
de_txt_month(1) --> `Hartung`.
de_txt_month(1) --> `Jänner`.
de_txt_month(1) --> `Jaenner`.
de_txt_month(1) --> `Jenner`.
de_txt_month(1) --> `Hartmond`.
de_txt_month(1) --> `Hartmonat`.
de_txt_month(2) --> `Februar`.
de_txt_month(2) --> `Febr`.
de_txt_month(2) --> `Feb`.
de_txt_month(2) --> `feb`.
de_txt_month(2) --> `Feber`.
de_txt_month(2) --> `Hornung`.
de_txt_month(3) --> `März`.
de_txt_month(3) --> `Maerz`.
de_txt_month(3) --> `Mrz`.
de_txt_month(3) --> `mrz`.
de_txt_month(3) --> `Märzen`.
de_txt_month(3) --> `Maerzen`.
de_txt_month(3) --> `Lenzing`.
de_txt_month(3) --> `Lenz`.
de_txt_month(3) --> `Lenzmond`.
de_txt_month(3) --> `Lenzmonat`.
de_txt_month(4) --> `April`.
de_txt_month(4) --> `Aprill`.
de_txt_month(4) --> `Apr`.
de_txt_month(4) --> `apr`.
de_txt_month(4) --> `Launing`.
de_txt_month(5) --> `Mai`.
de_txt_month(5) --> `mai`.
de_txt_month(5) --> `May`.
de_txt_month(5) --> `Maÿ`.
de_txt_month(5) --> `Wonnemond`.
de_txt_month(5) --> `Wonnemonat`.
de_txt_month(6) --> `Juni`.
de_txt_month(6) --> `Juny`.
de_txt_month(6) --> `Juno`.
de_txt_month(6) --> `Jun`.
de_txt_month(6) --> `jun`.
de_txt_month(6) --> `Brachmond`.
de_txt_month(6) --> `Brachmonat`.
de_txt_month(7) --> `Juli`.
de_txt_month(7) --> `July`.
de_txt_month(7) --> `Julÿ`.
de_txt_month(7) --> `Jul`.
de_txt_month(7) --> `jul`.
de_txt_month(7) --> `Heumond`.
de_txt_month(7) --> `Heumonat`.
de_txt_month(8) --> `August`.
de_txt_month(8) --> `Aug`.
de_txt_month(8) --> `aug`.
de_txt_month(8) --> `Ernting`.
de_txt_month(8) --> `Erntemond`.
de_txt_month(8) --> `Erntemonat`.
de_txt_month(9) --> `September`.
de_txt_month(9) --> `Septemb`.
de_txt_month(9) --> `Septem`.
de_txt_month(9) --> `Septbr`.
de_txt_month(9) --> `Sept`.
de_txt_month(9) --> `Sep`.
de_txt_month(9) --> `sep`.
de_txt_month(9) --> `7br`.
de_txt_month(9) --> `7ber`.
de_txt_month(9) --> `VIIbr`.
de_txt_month(9) --> `VIIber`.
de_txt_month(9) --> `Herbstmond`.
de_txt_month(9) --> `Herbstmonat`.
de_txt_month(9) --> `Scheiding`.
de_txt_month(9) --> `Holzing`.
de_txt_month(9) --> `Holzmond`.
de_txt_month(10) --> `Oktober`.
de_txt_month(10) --> `Oktob`.
de_txt_month(10) --> `October`.
de_txt_month(10) --> `Octob`.
de_txt_month(10) --> `Oktbr`.
de_txt_month(10) --> `Octbr`.
de_txt_month(10) --> `Okt`.
de_txt_month(10) --> `Oct`.
de_txt_month(10) --> `oct`.
de_txt_month(10) --> `okt`.
de_txt_month(10) --> `8br`.
de_txt_month(10) --> `8ber`.
de_txt_month(10) --> `VIIIbr`.
de_txt_month(10) --> `VIIIber`.
de_txt_month(10) --> `Weinmond`.
de_txt_month(10) --> `Weinmonat`.
de_txt_month(10) --> `Gilbhart`.
de_txt_month(10) --> `Gilbhard`.
de_txt_month(11) --> `November`.
de_txt_month(11) --> `Novbr`.
de_txt_month(11) --> `Nov`.
de_txt_month(11) --> `nov`.
de_txt_month(11) --> `9br`.
de_txt_month(11) --> `9ber`.
de_txt_month(11) --> `IXbr`.
de_txt_month(11) --> `IXber`.
de_txt_month(11) --> `Windmond`.
de_txt_month(11) --> `Windmonat`.
de_txt_month(11) --> `Nebelung`.
de_txt_month(12) --> `Dezember`.
de_txt_month(12) --> `Dezem`.
de_txt_month(12) --> `Dezemb`.
de_txt_month(12) --> `December`.
de_txt_month(12) --> `Decem`.
de_txt_month(12) --> `Decemb`.
de_txt_month(12) --> `Decbr`.
de_txt_month(12) --> `Dezbr`.
de_txt_month(12) --> `Dez`.
de_txt_month(12) --> `Dec`.
de_txt_month(12) --> `dez`.
de_txt_month(12) --> `dec`.
de_txt_month(12) --> `10br`.
de_txt_month(12) --> `10ber`.
de_txt_month(12) --> `Xbr`.
de_txt_month(12) --> `Xber`.
de_txt_month(12) --> `Julmond`.
de_txt_month(12) --> `Christmond`.
de_txt_month(12) --> `Christmonat`.

de_daymonth(6,24) --> `Johannis`.
de_daymonth(9,29) --> `Michaelis`.
de_daymonth(12,24) --> `Heiligabend`.
de_daymonth(12,31) --> `Sylvester`.
de_daymonth(12,31) --> `Silvester`.
de_daymonth(1,1) --> `Neujahr`.


de_month(M) -->
	de_num_month(M),
	de_skip_delimiters_before_numend,
	de_numend_token.
de_month(M) -->
	de_txt_month(M),
	de_delimiter.

de_daymov(good_friday) --> `Karfreitag`.
de_daymov(good_friday) --> `Charfreitag`.
de_daymov(ascension_thursday) --> `Himmelfahrt`.
de_daymov(pentecost_sunday) --> `Pfingsten`.
de_daymov(pentecost_sunday) --> `Pfingstsonntag`.
de_daymov(pentecost_monday) --> `Pfingstmontag`.
de_daymov(easter_sunday) --> `Ostern`.
de_daymov(easter_sunday) --> `Ostersonntag`.
de_daymov(easter_monday) --> `Ostermontag`.

de_numend_token_before_month --> de_numend_token.
de_numend_token_before_month --> ` `.	

de_numend_token --> `.`.
de_numend_token --> `/`.	
de_numend_token --> `tn`.
de_numend_token --> `ten`.
de_numend_token --> `sten`.
de_numend_token --> `ster`.
de_numend_token --> `ter`.

de_year_intro --> `Jahr`.
de_year_intro --> `Jahre`.
de_year_intro --> `Jahren`.
de_year_intro --> `bis`.
de_year_intro --> `von`.
de_year_intro --> `seit`.
de_year_intro --> `Anno`.
de_year_intro --> `anno`.
de_year_intro --> `Frühling`.
de_year_intro --> `Frühjahr`.
de_year_intro --> `Sommer`.
de_year_intro --> `Herbst`.
de_year_intro --> `Herbste`.
de_year_intro --> `Winter`.

de_day_intro --> `am`.
de_day_intro --> `zum`.
de_day_intro --> `vom`.
de_day_intro --> `Am`.
de_day_intro --> `bis`.
de_day_intro --> `dem`.
de_day_intro --> `den`.
de_day_intro --> `seit`.

de_md(M,D) -->
	de_delimiter,
	de_day_proper_before_month(D),
	de_skip_inner_delimiters,
	de_month(M).
de_md(M,D) -->
	de_daymonth(M,D).

de_date(ymd(Y,M,D)) -->
	de_delimiter,
	de_normdate(Y,M,D),
	de_peek_delimiter.
de_date(X) -->
	de_md(M, D),
	( de_skip_inner_delimiters,
	  de_year(Y),
	  de_peek_delimiter,
	  !,
	  { X = ymd(Y,M,D) }
	| { X = md(M,D) }
	).
de_date(ym(Y,M)) -->
	de_delimiter,
	de_month(M),
	de_skip_inner_delimiters,
	de_soleyear(Y),
	de_peek_delimiter.
de_date(X) -->
	de_delimiter,
	de_daymov(A),
	de_inner_delimiter,
	( de_skip_inner_delimiters,
	  de_year(Y),
	  de_peek_delimiter,
	  !,
	  { X = ya(Y, A) }
	; { X = a(A) }
	).
de_date(y(Y)) -->
	de_year_intro,
	de_inner_delimiter,
	de_skip_inner_delimiters,
	de_soleyear(Y),
	de_peek_delimiter.
de_date(y(Y)) -->
	de_hyphen_or_paren,
	de_soleyear(Y),
	de_peek_delimiter.
de_date(y(Y)) -->
	de_delimiter,
	de_soleyear(Y),
	de_peek_hyphen.
de_date(X) -->
	de_delimiter,
 	de_day_intro,
	de_date(X).
de_date(d(D)) -->
	de_delimiter,
 	de_day_intro,
 	de_inner_delimiter,
 	de_day_proper(D),
	de_soleday_after.
de_date(m(M)) -->
	de_txt_month(M),
	de_delimiter.

de_skip_delimiters_before_numend -->
	``.
de_skip_delimiters_before_numend -->
	de_delimiter,
	de_skip_delimiters_before_numend.

de_skip_delimiters -->
	de_delimiter,
	!,
	de_skip_delimiters.
de_skip_delimiters -->
	``.

de_skip_inner_delimiters -->
	de_inner_delimiter,
	!,
	de_skip_inner_delimiters.
de_skip_inner_delimiters -->
	``.

de_soleday_after -->
	de_skip_delimiters,
	[ C ],
	{ code_type(C, lower) }.

de_delimiter -->
	[C],
	{ de_delimiter(C) }.

de_inner_delimiter -->
	[C],
	{ de_inner_delimiter(C) }.

de_peek_delimiter([A|C], [A|C]) :-
	de_delimiter(A).

de_delimiter(X) :- code_type(X, punct), !.
de_delimiter(X) :- code_type(X, space), !.

de_inner_delimiter(0'-) :- !.
de_inner_delimiter(0'/) :- !.
de_inner_delimiter(0'.) :- !.
de_inner_delimiter(0':) :- !.
de_inner_delimiter(0'~) :- !. % nonbreaking space LaTeX syntax
de_inner_delimiter(160) :- !. % nonbreaking space Unicode %
de_inner_delimiter(X) :- code_type(X, space), !.

de_peek_hyphen([A|C], [A|C]) :-
	A = 0'-.

de_day_proper(D) -->
	natnum(D),
	{ D > 0,
	  D =< 31
	},
	de_skip_delimiters_before_numend,
	de_numend_token.

de_day_proper_before_month(D) -->
	natnum(D),
	{ D > 0,
	  D =< 31
	},
	de_skip_delimiters_before_numend,
	de_numend_token_before_month.

de_year(Y) -->
	natnum(Y),
	{ de_in_year_range(Y)
	}.

de_soleyear(Y) -->
	natnum(Y),
	{ de_in_soleyear_range(Y)
	}.

natnum(D) -->
	digit(D0),
	digits(T),
	{ number_codes(D, [D0|T])
	}.

digit(D) --> [D], { code_type(D, digit)	}.
digits([D|T]) --> digit(D), !, digits(T).
digits([]) --> [].

de_hyphen_or_paren -->
	`-`.
de_hyphen_or_paren -->
	`(`.

de_normdate(Y,M,D) -->
	digit(Y1),
	digit(Y2),
	digit(Y3),
	digit(Y4),
	`-`,
	{ atom_codes(YA, [Y1,Y2,Y3,Y4]),
	  atom_number(YA, Y)
	},
	digit(M1),
	digit(M2),
	`-`,
	{ atom_codes(MA, [M1,M2]),
	  atom_number(MA, M),
	  1 =< M, M =< 12
	},
	digit(D1),
	digit(D2),
	{ atom_codes(DA, [D1,D2]),
	  atom_number(DA, D),
	  1 =< D, M =< 31
	}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	 
date_normal(ymd(Y,M,D), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),
	format(atom(A), '~w~|~`0t~d~4+-~|~`0t~d~2+-~|~`0t~d~2+', [S,Y1,M,D]).
date_normal(md(M,D), A) :-
	!,
	format(atom(A), '????-~|~`0t~d~2+-~|~`0t~d~2+', [M,D]).
date_normal(d(D), A) :-
	!,
	format(atom(A), '????-??-~|~`0t~d~2+', [D]).
date_normal(ym(Y,M), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+-~|~`0t~d~2+', [S,Y1,M]).
date_normal(y(Y), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+', [S,Y1]).
date_normal(ya(Y,X), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+-~w', [S,Y1,X]).
date_normal(a(A), A) :-
	!. 
date_normal(m(M), A) :-
 	!, 
 	format(atom(A), '????-~|~`0t~d~2+', [M]).
date_normal(_, A) :-
	 A = '????-??-??'.
	 
date_year(ymd(Y,_,_), Y).
date_year(ym(Y, _), Y).
date_year(ya(Y), Y).
date_year(y(Y), Y).	 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

date_ymd(ymd(Y,M,D), Y, M, D).
date_ymd(md(M,D),    nil, M, D).
date_ymd(ym(Y,M),    Y, M, nil).
date_ymd(y(Y),       Y, nil, nil).
date_ymd(ya(Y,_),    Y, nil, nil).
% date_ymd(a(A),       nil, nil, nil).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

date_w3c(ymd(Y,M,D), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),
	format(atom(A), '~w~|~`0t~d~4+-~|~`0t~d~2+-~|~`0t~d~2+', [S,Y1,M,D]).
date_w3c(md(M,D), A) :-
	!,
	format(atom(A), '--~|~`0t~d~2+-~|~`0t~d~2+', [M,D]).
date_w3c(d(D), A) :-
	!,
	format(atom(A), '---~|~`0t~d~2+', [D]).
date_w3c(ym(Y,M), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+-~|~`0t~d~2+', [S,Y1,M]).
date_w3c(y(Y), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+', [S,Y1]).
date_w3c(ya(Y,X), A) :-
	!,
	( Y < 0 -> S='-', Y1 is -Y ; S='', Y1 = Y ),	  
	format(atom(A), '~w~|~`0t~d~4+-~w', [S,Y1,X]).
date_w3c(a(A), A) :-
	!. 
date_w3c(m(M), A) :-
	!, 
	format(atom(A), '--~|~`0t~d~2+', [M]).
date_w3c(_, A) :-
	 A = '?'.
