%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(dewiki, [dewiki_link/2,
		   linked_persons_1/2,
		   linked_persons_2/2,
%		   install_dewiki_to_gnd/0,
%		   gnd_to_dewiki/2,
%		   dewiki_to_gnd/2,
		   install_dewiki_links/0,
		   prepare_dewiki_links/1]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(utils)).
:- use_module(library('semweb/rdf_turtle')).

:- use_module(kbset_base(gnd)).

:- multifile dewiki_link/2.

dewiki_pagelinks_file(BornBeforeYear, F) :-
	filename(dewiki_page_links_pattern, F0),
	format(atom(F), F0, [BornBeforeYear]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% gnd_to_dewiki/2
%%%% dewiki_to_gnd/2
%%%%
%%%%    about 275.000 facts
%%%%    source: dewiki, seems more complete than the links to Wikipedia
%%%%            in the GND (August 2014)
%%%%    not fully 1-to-1:
%%%%      gnd_to_dewiki/2 yield multiple wiki items for about 80 GND numbers
%%%%      dewiki_to_gnd/2 yields a three numbers for 1 wiki item
%%%%    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% :- dynamic gnd_to_dewiki/2.
% :- dynamic dewiki_to_gnd/2.
% 
% install_dewiki_to_gnd :-
% 	abolish(dewiki_id_gnd_aux/1),
% 	retractall(gnd_to_dewiki(_ ,_)),
% 	retractall(dewiki_to_gnd(_ ,_)),
% 	filename(dewiki_id_gnd, F),
% 	info(10, 'Reading dewiki/id/gnd source: ~w', [F]),
% 	consult(F),
% 	info(10, 'Processing dewiki/id/gnd source', [F]),
% 	findall(IG, dewiki_id_gnd_aux(IG), IGs),
% 	process_igs(IGs),
% 	abolish(dewiki_id_gnd_aux/1).
% 
% process_igs([Id, GND|IGs]) :-
% 	clean_dewiki_ttl_id(Id, Id1),
% 	atom_concat('gnd:', GND, GND1),
% 	assert(dewiki_to_gnd(Id1, GND1)),
% 	assert(gnd_to_dewiki(GND1, Id1)),
% 	process_igs(IGs).
% process_igs([]).
% 
% clean_dewiki_ttl_id(A, A1) :-
% 	atom(A),
% 	sub_atom(A, 0, L, _, '<http://de.dbpedia.org/resource/'),
% 	sub_atom(A, L, _, 1, A1),
% 	!.
% clean_dewiki_ttl_id(A, _) :-
% 	err('No dbpedia resource: ~q', [A]).
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic seen_link/2.

%%%% 
%%%% about 8-10 min
%%%% 
prepare_dewiki_links(Year) :-
	info(10, 'Converting: DEWiki links (preparing)'),
	install_gnd_persons_bornbefore(Year),
	install_gnd_subjects,
	retractall( seen_link(_, _) ),
	info(10, 'Converting: DEWiki links'),
	verify_predicates([wiki_to_gnd/2, gnd_person/1, gnd_setting/2]),
	filename(dewiki_page_links_ttl, In),
	% gnd_setting(born_before_year, Year),
	dewiki_pagelinks_file(Year, Out),
	atomic_concat(Out, '.pl', Out1),
	info(10, 'Reading ~w', [In]),
	info(10, 'Writing ~w', [Out1]),
	info_progress_start,
	onto_file( from_file(rdf_process_turtle(current_input,
						process_dewiki_pagelinks_triples,
						[]),
			     In),
		   Out1 ).

process_dewiki_pagelinks_triples(Triples, _) :-
	info_progress(10, 100000),
	( member(T, Triples),
	  process_dewiki_pagelink_triple(T),
	  fail
	; true
	).

%%%% 
%%%% There may different possibilities: count just links among people born
%%%% before or count links of one such person to anything
%%%% 
process_dewiki_pagelink_triple(T) :-
	T = rdf(S, 'http://dbpedia.org/ontology/wikiPageWikiLink', O),
	clean_dewiki_id(S, S1),
	wiki_to_gnd(S1, GNDS),
	gnd_person(GNDS),
	clean_dewiki_id(O, O1),
	wiki_to_gnd(O1, GNDO),
	gnd_person(GNDO),
	( sort([S1,O1], [A,B]),
	  \+ seen_link(A,B),
	  assertz( seen_link(A,B) ),
	  writeq(dewiki_link(A,B)), writeln('.'),
	  writeq(dewiki_link(B,A)), writeln('.'),
	  fail
	; true
	).

clean_dewiki_id(A, A1) :-
	atom(A),
	sub_atom(A, 0, L, _, 'http://de.dbpedia.org/resource/'),
	sub_atom(A, L, _, 0, A1),
	!.
clean_dewiki_id(A, _) :-
	err('No dbpedia resource: ~q', [A]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic dewiki_link/2.

install_dewiki_links :-
	info(10, 'Loading: DEWiki links'),
	verify_predicates([gnd_setting/2, gnd_person/1]),
	gnd_setting(born_before_year, Year),
	dewiki_pagelinks_file(Year, In),
	consult(In).

map_flip([A-B|ABs], [B-A|BAs]) :-
	map_flip(ABs, BAs).
map_flip([], []).

linked_persons_1(X, Ys) :-
	gnd_to_wiki(X, _, X1),
	!,
	findall(Y, (dewiki_link(X1, Y1),
		    wiki_to_gnd(Y1, Y),
		    gnd_person(Y)), Ys1),
	sort(Ys1, Ys).
linked_persons_1(_, []).

linked_persons_2(X, Ys) :-
	gnd_to_wiki(X, _, X1),
	!,
	findall(Y, (dewiki_link(X1, Y1),
		    ( wiki_to_gnd(Y1, Y),
		      gnd_person(Y)
		    ; dewiki_link(Y1, Z1),
		      wiki_to_gnd(Z1, Y),
		      gnd_person(Y)
		    )),
		Ys1),
	sort(Ys1, Ys).
linked_persons_2(_, []).
