%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(filename, [filename/2,
		     pure_file_base_name/2,
		     register_file_path/2,
		     register_file/2,
		     set_docdir/1,
		     set_docdir/0,
		     docdir/1,
		     restricted_expand_filename/2
		    ]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).

:- dynamic filename_1/2.
:- dynamic filename_2/2.

register_file_path(Pattern, Path) :-
	retractall( filename_1(Pattern, _) ),
	( atom(Path) ->
	  restricted_expand_filename(Path, Path1)
	; Path1 = Path
	),
	assert( filename_1(Pattern, Path1) ).

register_file(Shortname, FileName) :-
	retractall( filename_2(Shortname, _) ),
	( atom(FileName) ->
	  restricted_expand_filename(FileName, FileName1)
	; FileName1 = FileName
	),
	assert( filename_2(Shortname, FileName1) ).

filename(Spec, FileName) :-
	atom(Spec),
	!,
	( filename_2(Spec, Pattern) ->
	  filename(Pattern, FileName)
	; FileName = Spec
	).
filename(Pattern, FileName) :-
	arg(1, Pattern, FileName1),
	filename_1(Pattern, Path),
	!,
	format(atom(FileName), '~w~w', [Path, FileName1]).
filename(X, _) :-
	err('No match for file spec ~q', [X]).

pure_file_base_name(FileName, PureBase) :-
	file_base_name(FileName, Base),
	file_name_extension(PureBase, _, Base),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic docdir/1.

set_docdir(Spec) :-
	!,
	retractall( docdir(_) ),
	( Spec = load_context ->
	  prolog_load_context(directory, Dir)
	; restricted_expand_filename(Spec, Dir)
	),
	ensure_trailing_dir_sep(Dir, DocDir),
	info(10, 'Setting document directory to ~w', [DocDir]),
	register_file_path(docdir(_), DocDir),
	asserta( docdir(DocDir) ).

set_docdir :-
	retractall( docdir(_) ),
	current_prolog_flag(argv, Argv),
	( member(Arg, Argv),
	  atom_concat('--docdir=', DocDir1, Arg) ->
	  info(10, 'Determined document directory from invocation argument')
	; getenv('KBSET_DOCDIR', DocDir1), DocDir1 \= '' ->
	  info(10, 'Determined document directory from environment variable')
	; predicate_property(user:kbset_config(_, _), file(DocDir0)),
	  file_directory_name(DocDir0, DocDir1) ->
	  info(10, 'Determined document_directory from location of config file')
	;  err('Failed to determine document directory - is a kbset_config file loaded?')
	),
	ensure_trailing_dir_sep(DocDir1, DocDir),
	info(10, 'Setting document directory to ~w', [DocDir]),
	register_file_path(docdir(_), DocDir),
	asserta( docdir(DocDir) ).

restricted_expand_filename(FileOrDirSpec, DirName) :-
	( sub_atom(FileOrDirSpec, 0, 1, _, '/') ->
	  DirName = FileOrDirSpec
	; sub_atom(FileOrDirSpec, 0, 2, _, '~/'),
	  getenv('HOME', Home),
	  Home \= '' ->
	  ensure_trailing_dir_sep(Home, Home1),
	  sub_atom(FileOrDirSpec, 2, _, 0, FileOrDirSpec1),
	  atom_concat(Home1, FileOrDirSpec1, DirName)
	; sub_atom(FileOrDirSpec, 1, 1, _, ':') ->
	  %% windows file name
	  DirName = FileOrDirSpec
	; docdir( DocDir ) ->
	  atom_concat(DocDir, FileOrDirSpec, DirName)
	; err('Failed to expand file spec: ~q', [FileOrDirSpec])
	).

ensure_trailing_dir_sep(X, X1) :-
	( sub_atom(X, _, 1, 0, '/') ->
	  X1 = X
	; atom_concat(X, '/', X1)
	).
