%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(geo,
	  [install_geonames/0,
	   convert_geonames_to_prolog/0,
	   geo_name_to_id/2,
	   geo_city/4,
	   geonames/19,
	   geo_feature_class_pretty/3,
	   geo_closest_big/3,
%	   geo_name/2,
	   geo_localized_name/2,
	   geo_localized_wiki_url/2,
	   geo_has_url/1,
	   geo_latlon/2,
	   geo_distance/3
	   ]).
/*

install_geonames

*/
:- use_module(library(csv)).
:- use_module(swilib(info)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_base(localization)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(wikinames)).
:- use_module(kbset_base(localization)).

:- dynamic geonames/19.

:- flag(cnt, _, 0).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Operates on the GEONAMES knowledge base that is obtained from reading in
%%%% directly a GEONAMES data file in TSV format and asserting them as
%%%% geonames/19 predicate.
%%%%
%%%% Such a file can be obtained by concatenating relevant country files
%%%% such as DE.txt and PL.txt (DE.zip and PL.zip unzipped) from the
%%%% geonames Website.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% geonames(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S).
%% 
%% A geonameid         : integer id of record in geonames database
%% B name              : name of geographical point (utf8) varchar(200)
%% C asciiname         : name of geographical point in plain ascii characters,
%%                       varchar(200)
%% D alternatenames    : alternatenames, comma separated, ascii names 
%%                       automatically transliterated, convenience attribute
%% 		         from alternatename table, varchar(8000)
%% E latitude          : latitude in decimal degrees (wgs84)
%% F longitude         : longitude in decimal degrees (wgs84)
%% G feature class     : see http://www.geonames.org/export/codes.html, char(1)
%% H feature code      : see http://www.geonames.org/export/codes.html,
%%                       varchar(10)
%% I country code      : ISO-3166 2-letter country code, 2 characters
%% J cc2               : alternate country codes, comma separated, ISO-3166
%%                       2-letter country code, 60 characters
%% K admin1 code       : fipscode (subject to change to iso code), see
%%                       exceptions below, see file admin1Codes.txt for
%% 		         display names of this code; varchar(20)
%% L admin2 code       : code for the second administrative division, a
%%                       county in the US, see file admin2Codes.txt;
%% 		         varchar(80) 
%% M admin3 code       : code for third level administrative division,
%%                       varchar(20)
%% N admin4 code       : code for fourth level administrative division,
%%                       varchar(20)
%% O population        : bigint (8 byte int) 
%% P elevation         : in meters, integer
%% Q dem               : digital elevation model, srtm3 or gtopo30, average
%%                       elevation of 3''x3'' (ca 90mx90m) or 30''x30''
%% 		         (ca 900mx900m) area in meters, integer.
%%                       srtm processed by cgiar/ciat.
%% R timezone          : the timezone id (see file timeZone.txt) varchar(40)
%% S modification date : date of last modification in yyyy-MM-dd format
%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- dynamic is_established_geo/1.
:- dynamic issue/1.

:- dynamic geo_city_cache/3.

geo_city_cache_threshold(10000).

install_geo_city_cache :-
	info(20, 'Caching geonames cities'),
	geo_city_cache_threshold(Thresh),
	retractall(geo_city_cache(_, _, _)),
	( geodirect_city(A, Thresh, LA-LO, Pop),
	  assert(geo_city_cache(A, LA-LO, Pop)),
	  fail
	; true
	).

geodirect_city(A, MinPop, LA-LO, Pop) :-
	geonames(A,_,_,_,LA1,LO1,'P',FC,_,_,_,_,_,_,O,_,_,_,_),
	FC \= 'PPLF',
	FC \= 'PPLH',
	FC \= 'PPLW',
	FC \= 'PPLX',
	FC \= 'PPLQ',
	seems_value(O),
	atom_number(O, Pop),
	Pop >= MinPop,
	atom_number(LA1, LA),
	atom_number(LO1, LO).

geo_city(A, MinPop, LL, Pop) :-
	geo_city_cache_threshold(Thresh),
	( MinPop < Thresh ->
	  geodirect_city(A, MinPop, LL, Pop)
	; geo_city_cache(A, LL, Pop),
	  Pop >= MinPop
	).

:- dynamic aux/1.

geo_localized_name(A, N) :-
	user_language(Lang),
	geo_altname(A, Lang, N1),
	!,
	N1 = N.
geo_localized_name(A, N) :-
	geo_name(A, N1),
	!,
	N1 = N.

geo_has_url(A) :-
	geo_altname(A, link, _),
	!.

geo_localized_wiki_url(A, URL) :-
	geo_altname(A, link, URL1),
	sub_atom(URL1, 0, _, _, 'http://en.wikipedia.org/'),
	!,
	localized_wiki_url(URL1, URL).

geo_name(A, N) :-
	geonames(A,N,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
	!.

geo_features(A, G, H) :-
	geonames(A,_,_,_,_,_,G,H,_,_,_,_,_,_,_,_,_,_,_),
	!.

geo_feature_class_pretty(X, Y, Z) :-
	user_language(Lang),
	( Lang = de ->
	  geo_feature_class_pretty_de(X, Y, Z)
	; geo_feature_class_pretty_en(X, Y, Z)
	).

geo_feature_class_pretty_en('A', _, 'administrative division') :- !.
geo_feature_class_pretty_en('H', _, 'waters') :- !.
geo_feature_class_pretty_en('L', _, 'area') :- !.
geo_feature_class_pretty_en('P', 'PPLX', 'city section') :- !.
geo_feature_class_pretty_en('P', _, 'city/village') :- !.
geo_feature_class_pretty_en('R', _, 'traffic') :- !.
geo_feature_class_pretty_en('S', _, 'spot') :- !.
geo_feature_class_pretty_en('T', _, 'mountain/hill') :- !.
geo_feature_class_pretty_en('U', _, 'undersea') :- !.
geo_feature_class_pretty_en('V', _, 'forest/heath') :- !.
geo_feature_class_pretty_en(_, _, 'unknown').

geo_feature_class_pretty_de('A', _, 'Staat/Verwaltungsbezirk') :- !.
geo_feature_class_pretty_de('H', _, 'Gewässer') :- !.
geo_feature_class_pretty_de('L', _, 'Gebiet') :- !.
geo_feature_class_pretty_de('P', 'PPLX', 'Stadtbezirk') :- !.
geo_feature_class_pretty_de('P', _, 'Ort') :- !.
geo_feature_class_pretty_de('R', _, 'Verkehr') :- !.
geo_feature_class_pretty_de('S', _, 'Punkt') :- !.
geo_feature_class_pretty_de('T', _, 'Berg/Anhöhe') :- !.
geo_feature_class_pretty_de('U', _, 'Unterwasser') :- !.
geo_feature_class_pretty_de('V', _, 'Wald/Heide') :- !.
geo_feature_class_pretty_de(_, _, 'Unbekannt').


:- dynamic geo_closest_big_cache/3.

%%%% 
%%%% Id1 is closest city/village with population >= MinPop 
%%%%
geo_closest_big(Id, MinPop, Id1) :-
	geo_closest_big_cache(Id, MinPop, Id1),
	!.
geo_closest_big(Id, MinPop, Id1) :-
	findall(aux(A,LL), geo_city(A, MinPop, LL, _), Aux),
	geo_latlon(Id, LL0),
	map_add_aux_dist(Aux, LL0, Aux1),
	keysort(Aux1, [_-Id1|_]),
	assert(geo_closest_big_cache(Id, MinPop, Id1)).

map_add_aux_dist([aux(A,LL1)|Xs], LL0, [D-A|Xs1]) :-
	haversine_distance(LL1, LL0, D),
	map_add_aux_dist(Xs, LL0, Xs1).
map_add_aux_dist([], _, []).


% elim_duplicates(Ids, Ids1) :-
% 	elim_duplicates(Ids, 10, Ids1),
% 	elim_duplicates(Ids, 5, Ids5),
% 	( Ids1 = Ids5 ->
% 	  true
% 	; register_issue(duplicates(Ids))
% 	).

elim_duplicates(Ids, Threshold, Ids1) :-
	sort(Ids, Ids2),
	map_add_geo_quality_key(Ids2, Ids3),
	keysort(Ids3, Ids4),
	map_val(Ids4, Ids5),
	map_elim_nearby(Ids5, [], Threshold, Ids6),
	sort(Ids6, Ids1).

map_elim_nearby([Id|Ids], SoFar, Thresh, Ids1) :-
	member(Id1, SoFar),
	geo_distance(Id, Id1, D),
	D =< Thresh,
	!,
	map_elim_nearby(Ids, SoFar, Thresh, Ids1).
map_elim_nearby([Id|Ids], SoFar, Thresh, Ids1) :-
	map_elim_nearby(Ids, [Id|SoFar], Thresh, Ids1).
map_elim_nearby([], SoFar, _, SoFar).

geo_latlon(A, LA-LO) :-
	geonames(A,_,_,_,LA1,LO1,_,_,_,_,_,_,_,_,_,_,_,_,_),
	!,
	atom_number(LA1, LA),
	atom_number(LO1, LO).
	
geo_distance(A1, A2, D) :-
	geo_latlon(A1, LL1),
	geo_latlon(A2, LL2),
	haversine_distance(LL1, LL2, D).
		
date_invnum(D, N) :-
	sub_atom(D,0,4,_,Y),
	sub_atom(D,5,2,_,M),
	sub_atom(D,8,2,_,T),
	atomic_list_concat([Y,M,T], A),
	atom_number(A, N1),
	N is 99999999 - N1.

seems_value(X) :-
	\+ seems_no_value(X).

seems_no_value('0').
seems_no_value('').
seems_no_value('-9999').

geo_quality_key(Id, Key) :-
	%% to compare multiple entries for seemingly same location,
	%% lower key = better quality
	Key = k(Estab,
		NoArea,
		SpecPopul,
		SpecElev,
		SpecDElev,
		SpecModDate),
	geonames(Id,_,_,_,_,_,G,_,_,_,_,_,_,_,O,P,Q,_,S),
	( is_established_geo(Id) -> Estab = 0 ; Estab = 1),
	( G = 'A' -> NoArea = 1 ; NoArea = 0 ),
	( seems_no_value(O) -> SpecPopul = 1 ; SpecPopul = 0 ),
	( seems_no_value(P) -> SpecElev = 1 ; SpecElev = 0 ),
	( seems_no_value(Q) -> SpecDElev = 1 ; SpecDElev = 0 ),
	date_invnum(S, SpecModDate).

map_add_geo_quality_key([X|Xs], [K-X|Xs1]) :-
	geo_quality_key(X, K),
	map_add_geo_quality_key(Xs, Xs1).
map_add_geo_quality_key([], []).


supported_geo_altname_language(de).
supported_geo_altname_language(en).
supported_geo_altname_language(link).
	  
:- dynamic geo_altname/3.

install_geonames_altnames :-
	retractall(geo_altname/3),
	flag(cnt, _, 0),
	filename(geonames_altnames, F),
	info(10, 'Reading geonames alternateNames source: ~w', [F]),
	( csv_read_file_row(F, X, [convert(false),
				   functor(row),
				   % arity(8),
				   separator(0'\t)]),
	  X = row(_, Id, Lang, Name, Pref, _, _, _),
	  supported_geo_altname_language(Lang),
	  ( Lang = link ->
	    true
	  ; Pref = '1' ->
	    retractall( geo_altname(Id, Lang, _) )
	  ; true
	  ),
	  Y = geo_altname(Id, Lang, Name),
	  assert(Y),
	  flag(cnt, N, N+1),
	  info_progress(20, 1000, N),
	  fail
	; info_progress_done,
	  flag(cnt, N, N),
	  info(10, 'Read ~w geonames alternateNames entries', [N])
	).

install_geonames_from_tsv :-
	install_geonames_altnames,
	retractall(geonames(_,_,_,_,_,
			    _,_,_,_,_,
			    _,_,_,_,_,
			    _,_,_,_)),
	flag(cnt, _, 0),
	filename(geonames, F),
	info(10, 'Reading geonames source: ~w', [F]),
	( csv_read_file_row(F, X, [convert(false),
				   functor(geonames),
				   ignore_quotes(true),
%				   arity(19),
				   separator(0'\t)]),
	  functor(X, _, 19),
	  assert(X),
	  flag(cnt, N, N+1),
	  info_progress(20, 10000, N),
	  fail
	; info_progress_done,
	  flag(cnt, N, N),
	  info(10, 'Read ~w geonames entries', [N])
	),
	install_geo_city_cache,
	install_geo_name_to_id,
	retractall(geo_closest_big_cache(_, _, _)).

install_geonames :-
	filename(geonames_prolog, F),
	consult(F),
	install_geo_city_cache,
	install_geo_name_to_id,
	retractall(geo_closest_big_cache(_, _, _)).

convert_geonames_to_prolog :-
	install_geonames_from_tsv,
	filename(geonames_prolog, F),
	atom_concat(F, '.pl', F1),
	info(10, 'Writing ~w', [F1]),
	onto_file((listing(geonames/19), listing(geo_altname/3)), F1),
	abolish(geonames/19),
	abolish(geo_altname/3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% 
%%%% 
haversine_distance(Lat1-Lon1, Lat2-Lon2, D) :-
	%% 0.017453292519943295 = pi/180 (conversion to radians)
	%% 6371.0 = Earth's mean radius in km
	Lat1R is Lat1*0.017453292519943295,
	Lat2R is Lat2*0.017453292519943295,
	DLat is (Lat2 - Lat1)*0.017453292519943295,
	DLon is (Lon2 - Lon1)*0.017453292519943295,
	A is sin(DLat/2)^2 + cos(Lat1R) * cos(Lat2R) * sin(DLon/2)^2,
	D is 6371.0 * 2 * atan(sqrt(A), sqrt(1 - A)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic geo_name_to_id/2.

install_geo_name_to_id :-
	verify_predicates([geonames/19]),
	retractall( geo_name_to_id(_, _) ),
	( geonames(A,B,_,D,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_),
	  atomic_list_concat(Names, ',', D),
	  sort([B|Names], Names1),
	  ( member(N, Names1),
	    assert( geo_name_to_id(N, A) ),
	    fail
	  ; true
	  ),
	  fail
	; true
	).
