%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019, 2021 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(gnd_prepare, [convert_gnd_to_prolog/1,
			prepare_gnd_subject_name/0,
			prepare_gnd_persons_bornbefore/1]).

:- encoding(utf8).

:- use_module(swilib(err)).
:- use_module(swilib(fromonto)).
:- use_module(swilib(info)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(utils)).
:- use_module(library(zlib)).

%%%%
gnd_rdfns('https://d-nb.info/ddc/class/', 'gndClass:').
gnd_rdfns('https://d-nb.info/gnd/','gnd:').
gnd_rdfns('https://d-nb.info/standards/elementset/gnd#','gndo:').
gnd_rdfns('https://d-nb.info/standards/vocab/gnd/gender#', 'gndGender:').
gnd_rdfns('https://d-nb.info/standards/vocab/gnd/geographic-area-code#', 'gndGeo:').
gnd_rdfns('https://d-nb.info/standards/vocab/gnd/gnd-sc#', 'gndSC:').
%%%%
gnd_rdfns('http://viaf.org/viaf/', 'viaf:').
gnd_rdfns('http://www.wikidata.org/entity/', 'wde:').
gnd_rdfns('http://creativecommons.org/publicdomain/zero/', 'cc:').
gnd_rdfns('http://dbpedia.org/resource/', 'dbpedia:').
gnd_rdfns('https://de.wikipedia.org/wiki/', 'wikiDE:').
gnd_rdfns('https://en.wikipedia.org/wiki/', 'wikiEN:').
gnd_rdfns('http://isni.org/isni/', 'isni:').
gnd_rdfns('http://id.loc.gov/', 'ilg:').
%%%%
gnd_rdfns('http://schema.org/', 'schema:').
gnd_rdfns('https://d-nb.info/standards/elementset/gnd#', 'gndo:').
gnd_rdfns('http://purl.org/library/', 'lib:').
gnd_rdfns('http://www.w3.org/2002/07/owl#', 'owl:').
gnd_rdfns('http://www.w3.org/2001/XMLSchema#', 'xsd:').
gnd_rdfns('http://www.w3.org/2004/02/skos/core#', 'skos:').
gnd_rdfns('http://www.w3.org/2000/01/rdf-schema#', 'rdfs:').
gnd_rdfns('https://ns.editeur.org/thema/', 'editeur:').
gnd_rdfns('http://www.opengis.net/ont/geosparql#', 'geo:').
gnd_rdfns('http://umbel.org/umbel#', 'umbel:').
gnd_rdfns('http://rdaregistry.info/Elements/u/', 'rdau:').
gnd_rdfns('http://www.opengis.net/ont/sf#', 'sf:').
gnd_rdfns('http://id.loc.gov/ontologies/bflc/', 'bflc:').
gnd_rdfns('http://purl.org/dc/terms/', 'dcterms:').
gnd_rdfns('http://vivoweb.org/ontology/core#', 'vivo:').
gnd_rdfns('http://iflastandards.info/ns/isbd/elements/', 'isbd:').
gnd_rdfns('http://xmlns.com/foaf/0.1/', 'foaf:').
gnd_rdfns('http://purl.org/ontology/mo/', 'mo:').
gnd_rdfns('http://id.loc.gov/vocabulary/relators/', 'marcRole:').
gnd_rdfns('https://d-nb.info/standards/elementset/agrelon#', 'agrelon:').
gnd_rdfns('http://purl.org/dc/dcmitype/', 'dcmitype:').
gnd_rdfns('http://dbpedia.org/property/', 'dbp:').
gnd_rdfns('https://d-nb.info/standards/elementset/dnb#', 'dnbt:').
gnd_rdfns('http://www.loc.gov/mads/rdf/v1#', 'madsrdf:').
gnd_rdfns('http://dnb.de/', 'dnb_intern:').
gnd_rdfns('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'rdf:').
gnd_rdfns('http://www.w3.org/2006/vcard/ns#', 'v:').
gnd_rdfns('http://www.w3.org/2007/05/powder-s#', 'wdrs:').
gnd_rdfns('http://www.ebu.ch/metadata/ontologies/ebucore/ebucore#', 'ebu:').
gnd_rdfns('http://purl.org/ontology/bibo/', 'bibo:').
gnd_rdfns('http://purl.org/ontology/gbv/', 'gbv:').
gnd_rdfns('http://purl.org/dc/elements/1.1/"', 'dc:').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% convert_gnd_to_prolog
%%%%
%%%% Convert the GND RDF file to Prolog. Different formats, asserted or
%%%% printed. If printed, each fact is in a single line, allowing
%%%% preprocessing with GREP. URNs are abbreviated according to
%%%% gnd_rdfns/2. Literals are represented by terms l/1 (instead of the
%%%% literal/1 returned by the SWI RDF parser).
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%

:- dynamic pgd_settings/3.
:- dynamic pgd_exclude/1.
:- dynamic pre_gnd_rdf/3.
:- dynamic pre_gnd_object/2.

convert_gnd_to_prolog(Options) :-
	%% If out_file ends with .gz, then a gzipped file is written
	( memberchk( in_file=In, Options ) -> true
	; filename(gnd_source_rdf_person, In)
	),
	( memberchk( out_file=OutPL, Options ) -> true ; OutPL=assert ),
	( memberchk( born_before=BBY, Options ) -> true ; BBY=none ),
	( memberchk( conversion=Conv, Options ) -> true ; Conv=flat ),
	( memberchk( exclude=PropsE, Options ) -> true ; PropsE=none ),
	retractall( pgd_settings(_, _, _) ),
	retractall( pgd_exclude(_) ),
	asserta( pgd_settings(Conv, OutPL, BBY) ),
	( atomic(PropsE), PropsE \= none ->
	  ( pgd_exclude_set(PropsE, P), assertz( pgd_exclude(P) ), fail ; true )
	; ( member(P, PropsE), assertz( pgd_exclude(P) ), fail ; true )
	),
	( OutPL = assert ->
	  retractall( pre_gnd_rdf(_, _, _) ),
	  retractall( pre_gnd_object(_, _) )
	; true
	),
	info(10, 'Converting from GND to Prolog'),
	info(10, 'Source ~w', [In]),
	info(10, 'Output ~w', [OutPL]),
	info_progress_start,
	( sub_atom(In, _, _, 0, '.gz') ->
	  setup_call_cleanup(gzopen(In, read, Stream, [type(binary)]),
			     ( OutPL=assert ->
			       process_rdf(Stream, process_gnd_triples, [])
			     ; onto_file(process_rdf(Stream,
						     process_gnd_triples,
						     []), OutPL)
			     ),
			     close(Stream))
	; onto_file(process_rdf(In, process_gnd_triples, []), OutPL)
	).


process_gnd_triples(Triples, _) :-
	pgd_settings(Conv, Out, BBY),
	info_progress(10, 1000),
	( BBY = none -> true
	; approx_year_of_birth(Triples, YOB), YOB < BBY
	),
	( Conv = flat ->
	  ( member(T, Triples),
	    clean_gnd_triple(T, rdf(S, P, O)),
	    \+ pgd_exclude(P),
	    ( Out = assert ->
	      assertz( pre_gnd_rdf(S, P, O) )
	    ; writeq( gnd_rdf(S, P, O) ),  writeln('.')
	    ),
	    fail
	  ; true
	  )
	; Conv = object ->
	  map_clean_gnd_triple(Triples, Triples1),
	  ( triples_to_object(Triples1, Id, Slots),
	    ( Out = assert ->
	      assertz( pre_gnd_object(Id, Slots) )
	    ; writeq( gnd_object(Id, Slots) ),
	      writeln('.')
	    ),
	    fail
	  ; true
	  )
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

triples_to_object(Triples, Id, Slots) :-
	triples_roots(Triples, Roots),
	member(Id, Roots),
	setof(P-V, V1^( member(rdf(Id, P, V1), Triples),
			\+ pgd_exclude(P),
			triples_value(V1, Triples, V )),
	      Slots).

triples_roots(T, R) :-
	setof(X, P^V^( member(rdf(X, P, V), T), \+ anonid(X) ), R).
	
triples_value(X, T, V) :-
	anonid(X),
	!,
	setof(P-V1, V2^( member( rdf(X, P, V2), T ),
			 \+ pgd_exclude(P),
			 triples_value(V2, T, V1) ),
	      V).
triples_value(X, _, X).

anonid(X) :-
	atom(X),
	sub_atom(X, 0, _, _, '_:'),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pgd_exclude_set(ex1, 'agrelon:hasAuntUncle').
pgd_exclude_set(ex1, 'agrelon:hasChild').
pgd_exclude_set(ex1, 'agrelon:hasFriend').
pgd_exclude_set(ex1, 'agrelon:hasGrandChild').
pgd_exclude_set(ex1, 'agrelon:hasGrandParent').
pgd_exclude_set(ex1, 'agrelon:hasNieceNephew').
pgd_exclude_set(ex1, 'agrelon:hasParent').
pgd_exclude_set(ex1, 'agrelon:hasSibling').
pgd_exclude_set(ex1, 'agrelon:hasSpouse').
pgd_exclude_set(ex1, 'agrelon:hasTeacher').
pgd_exclude_set(ex1, 'dcterms:license').
pgd_exclude_set(ex1, 'dcterms:modified').
pgd_exclude_set(ex1, 'dnbt:deprecatedUri').
pgd_exclude_set(ex1, 'gndo:gndIdentifier').
pgd_exclude_set(ex1, 'gndo:oldAuthorityNumber').
pgd_exclude_set(ex1, 'gndo:publication').
pgd_exclude_set(ex1, 'gndo:variantNameForThePerson').
pgd_exclude_set(ex1, 'wdrs:describedby').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% % % % %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Used to filter out persons not born before YOB
%%%% 
approx_year_of_birth(Triples, YOB) :-
	( memberchk(rdf(_, 'https://d-nb.info/standards/elementset/gnd#dateOfBirth', DL),
		    Triples) ->
	  raw_dateliteral_year(DL, YOB, _)
	; memberchk(rdf(_, 'https://d-nb.info/standards/elementset/gnd#dateOfBirthAndDeath', DL),
		    Triples) ->
	  raw_dateliteral_year(DL, YOB, _)
	; memberchk(rdf(_, 'https://d-nb.info/standards/elementset/gnd#dateOfDeath', DL),
		    Triples) ->
	  raw_dateliteral_year(DL, YOD, _),
	  YOB is YOD - 80
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

raw_dateliteral_year(L, Year, Accurracy) :-
	%% fails if no year could be determined
 	( L = literal(type('http://www.w3.org/2001/XMLSchema#gYear', Y1)) ->
	  atom_number(Y1, Year),
	  Accurracy = 1
	; L = literal(type('http://www.w3.org/2001/XMLSchema#gYearMonth', Y1)) ->
	  sub_atom(Y1, 0, 4, _, Y2),
	  atom_number(Y2, Year),
	  Accurracy = 1
	; L = literal(type('http://www.w3.org/2001/XMLSchema#date', Y1)) ->
	  sub_atom(Y1, 0, 4, _, Y2),
	  atom_number(Y2, Year),
	  Accurracy = 1
	; L = literal(type('http://www.w3.org/2001/XMLSchema#string', A)) ->
	  roughdate_year(A, Year),
	  Accurracy = 2
	).

dateliteral_year(L, Year, Accurracy) :-
	%% fails if no year could be determined
 	( L = l(type('xsd:gYear', Y1)) ->
	  atom_number(Y1, Year),
	  Accurracy = 1
	; L = l(type('xsd:gYearMonth', Y1)) ->
	  sub_atom(Y1, 0, 4, _, Y2),
	  atom_number(Y2, Year),
	  Accurracy = 1
	; L = l(type('xsd:date', Y1)) ->
	  sub_atom(Y1, 0, 4, _, Y2),
	  atom_number(Y2, Year),
	  Accurracy = 1
	; L = l(A), atom(A) ->
	  roughdate_year(A, Year),
	  Accurracy = 2
	).

roughdate_year(Atom, Year) :-
	atom_codes(Atom, Codes), 
	phrase( roughdate_year(Year1), Codes, _ ),
	!,
	( (sub_atom(Atom, _, _, _, 'v. Chr') ; sub_atom(Atom, _, _, _, 'v.Chr')) ->
	  Year is -Year1
	; Year is Year1
	).

roughdate_year(Y) --> `[um `, yearnum(Y).
roughdate_year(Y) --> `[ca. `, yearnum(Y).
roughdate_year(Y) --> `[ca.`, yearnum(Y).
roughdate_year(Y) --> `[circa `, yearnum(Y).
roughdate_year(Y) --> `[etwa `, yearnum(Y).		       
roughdate_year(Y) --> `[Um `, yearnum(Y).
roughdate_year(Y) --> `[fl. `, yearnum(Y).		       
roughdate_year(Y) --> `[nach `, yearnum(Y).
roughdate_year(Y) --> `[vor `, yearnum(Y).		  		       		       
roughdate_year(Y) --> `[geb. `, yearnum(Y).		  
roughdate_year(1000) --> `[Person des Mittelalters]`.
roughdate_year(-100) --> `[Person der Antike]`.
roughdate_year(1375) --> `[Person des Spätmittelalters]`.
roughdate_year(2000) --> `[Gegenwart]`.
roughdate_year(Y) --> `[`, yearnum(Y1), `. Jh.`, { Y is Y1 * 100 - 80 }.
roughdate_year(Y) --> `[`, yearnum(Y1), `. Jht.`, { Y is Y1 * 100 - 80 }.		       
roughdate_year(Y) --> `[Anfang `, yearnum(Y1), `. Jh.`, { Y is Y1 * 100 - 75 }.
roughdate_year(Y) --> `[Ende `, yearnum(Y1), `. Jh.`, { Y is Y1 * 100 - 25 }.
roughdate_year(Y) --> `[Ende `, yearnum(Y1), `.`, { Y1 =< 21, Y is Y1 * 100 - 25 }.		       
roughdate_year(Y) --> `[Mitte `, yearnum(Y1), `. Jh.`, { Y is Y1 * 100 - 50 }.		  		  
roughdate_year(Y) --> `[Anfang `, yearnum(Y1), `. Jht.`, { Y is Y1 * 100 - 75 }.
roughdate_year(Y) --> `[Ende `, yearnum(Y1), `. Jht.`, { Y is Y1 * 100 - 25 }.
roughdate_year(Y) --> `[Mitte `, yearnum(Y1), `. Jht.`, { Y is Y1 * 100 - 50 }.		  		  
roughdate_year(Y) --> `[`, yearnum(_), `./`, yearnum(Y1), `. Jh.`, { Y is Y1 * 100 - 50 }.
roughdate_year(Y) --> `[-`, yearnum(Y1), { Y is -Y1 }.
roughdate_year(Y) --> `[v`, yearnum(Y1), { Y is -Y1 }.		       
roughdate_year(Y) --> `[[`, yearnum(Y).		       
roughdate_year(Y) --> `[`, yearnum(Y).

yearnum(Y) -->
	ydigit(D1),
	ydigits(D2),
	{ 
	  number_codes(Y, [D1|D2])
	}.

ydigit(D) --> [D], { code_type(D, digit) }.
ydigit(48) --> `X`.
ydigits([D|T]) -->
	ydigit(D), !, ydigits(T).
ydigits([]) --> [].
		       	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

map_clean_gnd_triple([X|Xs], [X1|Xs1]) :-
	clean_gnd_triple(X, X1),
	map_clean_gnd_triple(Xs, Xs1).
map_clean_gnd_triple([], []).

clean_gnd_triple(rdf(S,P,O), rdf(S1,P1,O1)) :-
	clean_gnd_item(S, S1),
	clean_gnd_item(P, P1),
	clean_gnd_item(O, O1).

%%%% 
%%%% Note: In GND.rdf.gz GND from December 2014 so called precomposed
%%%% characters are used.  unicode_nfkc/2 converts them. It seems that the
%%%% earlier version of May [?] 2014 did not use precomposed characters.
%%%% 
clean_gnd_item(A, B) :-
	atom(A),
	gnd_rdfns(P, N),
	sub_atom(A, 0, L, _, P),
	!,
	sub_atom(A, L, _, 0, A1),
	atom_concat(N, A1, B).
clean_gnd_item(literal(type('http://www.w3.org/2001/XMLSchema#string',
			    A)),
	       l(A1)) :-
	!,
	clean_gnd_literal(A, A1).
clean_gnd_item(literal(type(T,A)), l(type(T1,A1))) :-
	atom(T),
	gnd_rdfns(P, N),
	sub_atom(T, 0, L, _, P),
	!,
	clean_gnd_literal(A, A1),
	sub_atom(T, L, _, 0, T2),
	atom_concat(N, T2, T1).
clean_gnd_item(literal(A), l(A1)) :-
	!,
	clean_gnd_literal(A, A1).
clean_gnd_item(A, A) :-
	atom(A),
	!.
clean_gnd_item(A, _) :-
	err('Item ~q', [A]).

clean_gnd_literal(A, A1) :-
	atom(A),
	!,
	unicode_nfkc(A, A1).
clean_gnd_literal(lang(L,A), lang(L,A1)) :-
	atom(A),
	!,
	unicode_nfkc(A, A1).
clean_gnd_literal(X, _) :-
	err('No method to clean literal: ~q', [X]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Conversion to Access Predicates
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% About 30 min for 1850
%%%% 
prepare_gnd_persons_bornbefore(BornBeforeYear) :-
	filename(gnd_source_rdf_person, In),
	filename(gnd_persons_bornbefore_pattern, F0),
	format(atom(Out), F0, [BornBeforeYear]),
	atom_concat(Out, '.pl', Out1), 
	convert_gnd_to_prolog([in_file=In,
			       out_file=assert,
			       conversion=object,
			       exclude=ex1,
			       born_before=BornBeforeYear]),
	onto_file( pre_gnd_write_access_base, Out1 ).

pre_gnd_write_access_base :-
	( pgd_settings(_, _, Year), number(Year) ->
	  format('~q.~n', [gnd_setting(born_before_year, Year)])
	; format('gnd_setting(_, _) :- false.~n')
	),
	pre_gnd_mk_gnd_person(Ids),
	length(Ids, LIds),
	info(10, 'Number of GND persons: ~w', [LIds]),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  format('~q.~n', [gnd_person(Id)]),
	  fail
	; true
	),
	info(10, 'Writing gnd_person_data'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_person_data(Id, Obj, Fact),
	  format('~q.~n', [Fact]),
	  fail
	; true
	),
	info(10, 'Writing gnd_person_name'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_enum_gnd_name(Id, Obj, Last, First, Extra),
	  format('~q.~n', [gnd_person_name(Id, Last, First, Extra)]),
	  fail
	; true
	),
	info(10, 'Writing gnd_name_to_person'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_enum_gnd_name(Id, Obj, Last, First, Extra),
	  format('~q.~n', [gnd_name_to_person(Last, First, Extra, Id)]),
	  fail
	; true
	),
	info(10, 'Writing gnd_person_info'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_gnd_person_info(Id, Obj, Fact),
	  format('~q.~n', [Fact]),
	  fail
	; true
	),
	info(10, 'Writing gnd_to_wiki'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_gnd_to_wiki(Id, Obj, Fact),
	  format('~q.~n', [Fact]),
	  fail
	; true
	),
	info(10, 'Writing wiki_to_gnd'),
	info_progress_start,
	( member(Id, Ids),
	  info_progress(10, 10000),
	  pre_gnd_object(Id, Obj),
	  pre_gnd_mk_wiki_to_gnd(Id, Obj, Fact),
	  format('~q.~n', [Fact]),
	  fail
	; true
	).

	  	  
pre_gnd_mk_gnd_person(Ids) :-
	findall(Id1, ( pre_gnd_object(Id1, Obj),
		       once((member('rdf:type'-Type, Obj),
			     gnd_differentiated_person_type(Type) )) ),
		Ids1 ),
	sort(Ids1, Ids).

gnd_differentiated_person_type('gndo:DifferentiatedPerson') :- !.
gnd_differentiated_person_type('gndo:Pseudonym') :- !.
gnd_differentiated_person_type('gndo:CollectivePseudonym') :- !.
gnd_differentiated_person_type('gndo:LiteraryOrLegendaryCharacter') :- !.
gnd_differentiated_person_type('gndo:Gods') :- !.
gnd_differentiated_person_type('gndo:RoyalOrMemberOfARoyalHouse') :- !.
gnd_differentiated_person_type('gndo:Spirits') :- !.
			   
pre_gnd_mk_enum_gnd_name(_Id, T, Last, First, Extra) :-
	findall( n(Last1, First1, Extra1),
		 enum_gnd_name_1(T, Last1, First1, Extra1),
		 GNs ),
	sort_gns(GNs, GNs1),
	member( n(Last, First, Extra), GNs1).

sort_gns(X, Y) :-
	sort(X, X1),
	map_rm_gns(X1, Y).

map_rm_gns([n(A,'',''), n(A,X,Y)|Xs], Xs1) :-
	!,
	map_rm_gns([n(A,X,Y)|Xs], Xs1).
map_rm_gns([n(A,X,''), n(A,X,Y)|Xs], Xs1) :-
	!,
	map_rm_gns([n(A,X,Y)|Xs], Xs1).
map_rm_gns([n(A,'',Y), n(A,X,Y)|Xs], Xs1) :-
	!,
	map_rm_gns([n(A,X,Y)|Xs], Xs1).
map_rm_gns([N|Xs], [N|Xs1]) :-
	map_rm_gns(Xs, Xs1).
map_rm_gns([], []).

enum_gnd_name_1(Obj, Last, First, Extra) :-
	( memberchk('gndo:preferredNameEntityForThePerson'-E, Obj)
	; member('gndo:variantNameEntityForThePerson'-E, Obj)
	),
	name_entity_components(E, Last, First, Extra).

name_entity_components(E, Last, First, Extra) :-
	%% For now: possibly not all useful combinations considered: %
	( memberchk('gndo:surname'-l(Last1), E) ->
	  rdf_ensure_atom(Last1, Last),
	  ( memberchk('gndo:forename'-l(First1), E) ->
	    rdf_ensure_atom(First1, First2),
	    ( memberchk('gndo:prefix'-l(Prefix1), E) ->
	      rdf_ensure_atom(Prefix1, Prefix2),
	      ( First = First2 ;  atomic_list_concat([First2, ' ', Prefix2], First) )
	    ; First = First2
	    )
	  ; First = ''
	  )
	; memberchk('gndo:personalName'-l(Last1), E) ->
	  rdf_ensure_atom(Last1, Last2),
	  ( memberchk('gndo:counting'-l(Count1), E) ->
	    rdf_ensure_atom(Count1, Count2),
	    ( Last = Last2 ; atomic_list_concat([Last2, ' ', Count2], Last) )
	  ; Last = Last2
	  ),
	  First = ''
	),
	( memberchk('gndo:nameAddition'-l(Extra1), E) ->
	  rdf_ensure_atom(Extra1, Extra)
	; Extra = ''
	).

pre_gnd_mk_person_data(Id, Obj, Fact) :-
	Fact = gnd_person_data(Id, Name, YoB, YoD, Gender, POs, YoBA, YoDA),
	( memberchk('gndo:preferredNameForThePerson'-l(Name1), Obj) ->
	  rdf_ensure_atom(Name1, Name)
	; Name = nil
	),
	( memberchk('gndo:dateOfBirth'-DoB, Obj), dateliteral_year(DoB, YoB, YoBA) ->
	  true
	; memberchk('gndo:dateOfBirthAndDeath'-DoB, Obj), dateliteral_year(DoB, YoB, YoBA) ->
	  true
	; YoB = nil, YoBA = 4
	),
	( memberchk('gndo:dateOfDeath'-DoD, Obj), dateliteral_year(DoD, YoD, YoDA) ->
	  true
	; YoD = nil, YoDA = 4
	),
	( memberchk('gndo:gender'-Gender1, Obj) ->
	  ( Gender1 = 'gndGender:male' -> Gender = m
	  ; Gender1 = 'gndGender:female' -> Gender = f
	  ; sub_atom(Gender1, _, _, 0, female) -> Gender = f
	  ; sub_atom(Gender1, _, _, 0, male) -> Gender = m
	  ; Gender = nil
	  )
	; Gender = nil
	),
	findall(PO, member('gndo:professionOrOccupation'-PO, Obj), POs1),
	sort(POs1, POs).

rdf_ensure_atom(type('xmls:string',X), X) :-
	!.
rdf_ensure_atom(X, X).

pre_gnd_mk_gnd_person_info(Id, Obj, Fact) :-
	Fact = gnd_person_info(Id, Lang, Inf),
	member('gndo:biographicalOrHistoricalInformation'-l(lang(Lang,Inf)), Obj).

pre_gnd_mk_wiki_to_gnd(Id, Obj, Fact) :-
	Fact = wiki_to_gnd(W, Id),
	findall(W1, ( pre_gnd_mk_gnd_to_wiki(Id, Obj, gnd_to_wiki(_, _, W1) ) ), Ws),
	sort(Ws, Ws1),
	member(W, Ws1).

pre_gnd_mk_gnd_to_wiki(Id, Obj, Fact) :-
	Fact = gnd_to_wiki(Id, L, W),
	member(P-W1, Obj),
	wiki_reference(P, W1, L, W).

wiki_reference('foaf:page', W, de, W1) :-
	sub_atom(W, 0, L, _, 'wikiDE:'),
	!,
	sub_atom(W, L, _, 0, W2),
	www_form_encode(W1, W2).
wiki_reference('foaf:page', W, de, W1) :-
	sub_atom(W, 0, L, _, 'https://de.wikipedia.org/wiki/'),
	!,
	sub_atom(W, L, _, 0, W2),
	www_form_encode(W1, W2).
wiki_reference('owl:sameAs', W, en, W1) :-
	sub_atom(W, 0, L, _, 'dbpedia:'),
	!,
	sub_atom(W, L, _, 0, W2),
	www_form_encode(W1, W2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prepare_gnd_subject_name :-
	filename(gnd_source_rdf_subject, GNDSachbegriff),
	filename(gnd_subject, Sachbegriff),
	atom_concat(Sachbegriff, '.pl', SachbegriffPL),
	prepare_gnd_subject_name(GNDSachbegriff, SachbegriffPL).

prepare_gnd_subject_name(GNDSachbegriff, OutFile) :-
	convert_gnd_to_prolog([in_file=GNDSachbegriff,
			       out_file=assert,
			       conversion=flat]),
	onto_file(( pre_gnd_rdf(Subject,
				'gndo:preferredNameForTheSubjectHeading',
				l(Name)),
		    format('~q.~n', [gnd_subject_name(Subject, Name)]),
		    fail
		  ; true
		  ),
		  OutFile).






