%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Destructive Hashtables
%%%%
%%%% 
%%%% NON-BACKTRACKABLE
%%%% 
%%%% - Implementation of hashtables that uses nb_setarg/3 and
%%%%   nb_linkarg/3
%%%% - Values are copied at insertion
%%%% - State of the table is not reversed at backtracking
%%%%
%%%% - Two Flavors:
%%%%   - ht:  associate a single value with a key
%%%%   - htm: asscoiate a multiset of values with a key
%%%%
%%%% BACKTRACKABLE
%%%%
%%%%   - htb: associate a single value with a key
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(hash_kbset, [mk_ht/2,
 		 ht_put/3,
		 ht_del/2,
 		 ht_get/3,
 		 ht_flag/4,
 		 ht_key/2,
		 ht_capacity/2,
		 mk_htm/2,
		 htm_add/3,
		 htm_del/3,
		 htm_get/3,
		 htm_key/2,
		 htm_capacity/2,
		 mk_htb/2,
		 htb_put/3,
		 htb_get/3,
		 htb_del/2,
		 htb_key/2,
		 htb_capacity/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% NON-BACKTRACKABLE BASIC HASHTABLES
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% mk_ht(+Capacity, -HashTable)
%%%% 
%%%% Currently there is no support for resizing.
%%%%
mk_ht(Capacity, HashTable) :-
	functor(Table, k, Capacity),
	init_table(Capacity, Table),
	HashTable = ht(Table, Capacity).

init_table(N, Table) :-
	N > 0,
	arg(N, Table, []),
	N1 is N - 1,
	init_table(N1, Table).
init_table(0, _).

ht_capacity(ht(_, Size), Size).

%%%% 
%%%% ht_get(+HashTable, +Key, ?Value)
%%%% 
ht_get(ht(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	memberchk(Key-Value1, Entries),
	Value1 = Value.

%%%% 
%%%% ht_put(+HashTable, +Key, +Value)
%%%% 
ht_put(ht(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	( member(KV, Entries), arg(1, KV, Key1), Key1 == Key ->
	  nb_setarg(2, KV, Value)  
	; Entries == [] ->
	  nb_setarg(Index, Table, [Key-Value])
	; duplicate_term([Key-Value|_], KVE1),
	  nb_linkarg(2, KVE1, Entries),
	  nb_linkarg(Index, Table, KVE1)
        ).

%%%% 
%%%% ht_flag(+HashTable, +Key, -OldValue, +NewValue)
%%%% 
%%%% Get/put of integer values in one step. Works like SWI-Prolog's flag/3,
%%%% e.g. to increment a value, ht_put(..., V, V+1) can be used. If Key
%%%% has not been put in previously, OldValue is 0.
%%%% 
ht_flag(ht(Table, Size), Key, OldValue, NewValueExpression) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	( member(KV, Entries), arg(1, KV, Key1), Key1 == Key ->
	  arg(2, KV, OldValue),
	  NewValue is NewValueExpression,	    
	  nb_setarg(2, KV, NewValue)  
	; Entries == [] ->
	  OldValue = 0,
	  NewValue is NewValueExpression,	    
	  nb_setarg(Index, Table, [Key-NewValue])
	; OldValue = 0,
	  NewValue is NewValueExpression,	    
	  duplicate_term([Key-NewValue|_], KVE1),
	  nb_linkarg(2, KVE1, Entries),
	  nb_linkarg(Index, Table, KVE1)
        ).

%%%% 
%%%% ht_del(+HashTable, +Key)
%%%% 
%%%% Delete the association with Key. Fails if there is no
%%%% such association.
%%%% 
ht_del(ht(Table, Size), Key) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	del_unifiable_nb_linkarg(Entries, Key-_, Index, Table).

%%%% 
%%%% ht_key(+HashTable, ?Key)
%%%% 
%%%% Enumerate the keys in the hashtable.
%%%% 
ht_key(ht(Table, Size), Key) :-
	ht_key_1(Table, Size, Key).

ht_key_1(T, N, K) :-
	N > 0,
	arg(N, T, E),
	member(K-_, E).
ht_key_1(T, N, K) :-
	N > 0,	
	N1 is N - 1,
	ht_key_1(T, N1, K).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% NON-BACKTRACKABLE HASHTABLES WITH MULTISETS OF VALUES 
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%% 
%%%% mk_htm(+Capacity, -HashTable)
%%%% 
mk_htm(Capacity, HashTable) :-
	functor(Table, k, Capacity),
	init_table(Capacity, Table),
	HashTable = htm(Table, Capacity).

htm_capacity(htm(_, Size), Size).

%%%%
%%%% htm_add(+MultisetHashTable, +Key, +Value)
%%%%
%%%% Adds an occurence of Value to Key.
%%%% 
htm_add(htm(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	( Entries == [] ->
	  nb_setarg(Index, Table, [Key-Value])
	; duplicate_term([Key-Value|_], KVE1),
	  nb_linkarg(2, KVE1, Entries),
	  nb_linkarg(Index, Table, KVE1)
        ).

%%%%
%%%% htm_get(+MultisetHashTable, +Key, ?Value)
%%%%
%%%% Enumerates the occurences of values of Key.
%%%% 
htm_get(htm(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	member(Key-Value, Entries).


%%%% 
%%%% htm_del(+MultisetHashTable, +Key, +Value)
%%%% 
%%%% Delete an occurence of Value from Key. Key and Value are identified
%%%% using =@=/2. Fails if there is no occurence of Key and Value in
%%%% the hashtable.
%%%% 
htm_del(htm(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	del_variant_nb_linkarg(Entries, Key-Value, Index, Table).

del_variant_nb_linkarg(L, E, PlaceArg, Place) :-
	( L == [] ->
	  fail
	; arg(1, L, Car),
	  arg(2, L, Cdr),
	  ( Car =@= E ->
	    nb_linkarg(PlaceArg, Place, Cdr)
	  ; del_variant_nb_linkarg(Cdr, E, 2, L)
	  )
	).

del_unifiable_nb_linkarg(L, E, PlaceArg, Place) :-
	( L == [] ->
	  fail
	; arg(1, L, Car),
	  arg(2, L, Cdr),
	  ( Car = E ->
	    nb_linkarg(PlaceArg, Place, Cdr)
	  ; del_unifiable_nb_linkarg(Cdr, E, 2, L)
	  )
	).

%%%% 
%%%% htm_key(+HashTable, ?Key)
%%%% 
%%%% Enumerate the keys in the multiset-hashtable.
%%%% 
htm_key(htm(Table, Size), Key) :-
	htm_key_1(Table, Size, Key).

htm_key_1(T, N, K) :-
	N > 0,
	arg(N, T, E),
	setof(K1, V^member(K1-V, E), Ks),
	member(K, Ks).
htm_key_1(T, N, K) :-
	N > 0,	
	N1 is N - 1,
	htm_key_1(T, N1, K).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% BACKTRACKABLE BASIC HASHTABLES
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% mk_ht(+Capacity, -HashTable)
%%%% 
%%%% Currently there is no support for resizing.
%%%%
mk_htb(Capacity, HashTable) :-
	functor(Table, k, Capacity),
	init_table(Capacity, Table),
	HashTable = htb(Table, Capacity).


htb_capacity(htb(_, Size), Size).

%%%% 
%%%% htb_get(+HashTable, +Key, ?Value)
%%%% 
htb_get(htb(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	memberchk(Key-Value1, Entries),
	Value1 = Value.

%%%% 
%%%% htb_put(+HashTable, +Key, +Value)
%%%% 
htb_put(htb(Table, Size), Key, Value) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	( member(KV, Entries), KV = Key-_ ->
	  setarg(2, KV, Value)  
	; Entries == [] ->
	  setarg(Index, Table, [Key-Value])
	; setarg(Index, Table, [Key-Value|Entries])
        ).


%%%% 
%%%% htb_del(+HashTable, +Key)
%%%% 
%%%% Delete the association with Key. Fails if there is no
%%%% such association.
%%%% 
htb_del(htb(Table, Size), Key) :-
	term_hash(Key, N),
	Index is 1 + (N mod Size),
	arg(Index, Table, Entries),
	( select_once(Key-_, Entries, Entries1) ->
	  setarg(Index, Table, Entries1)
	; fail
	).

select_once(X, [X|Y], Y) :-
	!.
select_once(X, [Y|Z], [Y|U]) :-
	select_once(X, Z, U).




%%%% htb_key(+HashTable, ?Key)
%%%% 
%%%% Enumerate the keys in the hashtable.
%%%% 
htb_key(htb(Table, Size), Key) :-
	htb_key_1(Table, Size, Key).

htb_key_1(T, N, K) :-
	N > 0,
	arg(N, T, E),
	member(K-_, E).
htb_key_1(T, N, K) :-
	N > 0,	
	N1 is N - 1,
	htb_key_1(T, N1, K).


