%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(location, [word_to_location/2,
		     location_feature_class/2,
		     location_feature_code/2,
		     location_population/2,
		     show_location/1,
		     show_location_nonl/1,
		     show_location_vw/1,
		     identify_location/2,

		     location_pretty/2,
		     location_url/2,

		     retract_supplemented_locations/0,
		     process_supplemented_location/2
		     ]).

:- use_module(kbset_base(geo)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_base(wikinames)).
:- use_module(kbset_base(localization)).

:- dynamic sup_name_to_location/2.

location_pretty(X, Y) :-
	geo_localized_name(X, Y).

location_url(X, URL) :-
	geo_localized_wiki_url(X, URL),
	!.
location_url(X, URL) :-
	geo_latlon(X, Lat-Lon),
	Options = [],
	geohack_url(Lat, Lon, Options, URL).

word_to_location(X, Y) :-
	ord_enum(Y, loc_name_to_id(X, Y)).

loc_name_to_id(X, Y) :-
	geo_name_to_id(X, Y).
loc_name_to_id(X, Y) :-
	sup_name_to_location(X, Y).

location_feature_class(A, G) :-
	geonames(A,_,_,_,_,_,G,_,_,_,_,_,_,_,_,_,_,_,_),
	!.

location_feature_code(A, H) :-
	geonames(A,_,_,_,_,_,_,H,_,_,_,_,_,_,_,_,_,_,_),
	!.

location_population(LID, Population) :-
	geonames(LID,_,_,_,_,_,_,_,_,_,_,_,_,_,O,_,_,_,_),
	!,
	atom_number(O, Population).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Pretty Printing Functions
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_location(X) :-
	show_geo(X).

show_geo([]) :- !.
show_geo([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo(Y), fail ; true).
show_geo(A) :-
	geo_localized_name(A, N),
	format('~w~n', [N]).

show_location_nonl([]) :- !.
show_location_nonl([X|Xs]) :- !, (member(Y,[X|Xs]), show_location_nonl(Y), fail ; true).
show_location_nonl(A) :-
	geo_localized_name(A, N),
	format('~w', [N]).

show_geo_1([]) :- !.
show_geo_1([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo_1(Y), fail ; true).
show_geo_1(A) :-
	geonames(A,B,C,D,E,F,G,H,I,_,_,_,_,_,O,_,_,_,S),
	format('A: ~w~nB: ~w~nC: ~w~nD: ~w~nE: ~w~nF: ~w~nG: ~w~nH: ~w~nI: ~w~nO: ~w~nS: ~w~n',
	       [A,B,C,D,E,F,G,H,I,O,S]).

show_geo_2([]) :- !.
show_geo_2([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo_2(Y), fail ; true).
show_geo_2(A) :-
	geonames(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S),
	format('A: ~w~nB: ~w~nC: ~w~nD: ~w~nE: ~w~nF: ~w~nG: ~w~nH: ~w~nI: ~w~nJ: ~w~nK: ~w~nL: ~w~nM: ~w~nN: ~w~nO: ~w~nP: ~w~nQ: ~w~nR: ~w~nS: ~w~n',
	       [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S]).

show_geo_line(A) :-
	geonames(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S),
	format('~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w\t~w~n',
	       [A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S]).

show_geo_4([]) :- !.
show_geo_4([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo_4(Y), fail ; true).
show_geo_4(A) :-
	geo_phrases(CloseTo, And),
	geonames(A,_N,_,_,_,_,G,H,_,_,_,_,_,_,O,_,_,_,_),
	geo_localized_name(A, N),
	geo_feature_class_pretty(G, H, Class),
	atom_number(O, Pop),
	( Pop < 10000 ->
	  geo_closest_big(A, 10000, C1),
	  %% currently quite slow: comparing with all candidates
	  geo_closest_big(A, 100000, C2),
	  geo_localized_name(C1, NC1),
	  ( C1 = C2 ->
	    format('~w (~w) ~w ~w~n', [N, Class, CloseTo, NC1])
	  ; geo_localized_name(C2, NC2),
	    format('~w (~w) ~w ~w ~w ~w~n', [N, Class, CloseTo, NC1, And, NC2])
	  )
	; Pop < 100000 ->
	  geo_closest_big(A, 100000, C2),
	  geo_localized_name(C2, NC2),
	  format('~w (~w) ~w ~w~n', [N, Class, CloseTo, NC2])
	; format('~w (~w)~n', [N, Class])
	).

show_geo_5([]) :- !.
show_geo_5([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo_5(Y), fail ; true).
show_geo_5(A) :-
	geo_phrases(CloseTo, _And),
	geonames(A,_N,_,_,_,_,G,H,_,_,_,_,_,_,O,_,_,_,_),
	geo_localized_name(A, N),
	geo_feature_class_pretty(G, H, Class),
	atom_number(O, Pop),
	( Pop < 100000 ->
	  geo_closest_big(A, 100000, C2),
	  geo_localized_name(C2, NC2),
	  format('~w (~w) ~w ~w~n', [N, Class, CloseTo, NC2])
	; format('~w (~w)~n', [N, Class])
	).

show_location_vw(X) :-
	show_geo_6(X).

show_geo_6([]) :- !.
show_geo_6([X|Xs]) :- !, (member(Y,[X|Xs]), show_geo_6(Y), fail ; true).
show_geo_6(A) :-
	geo_phrases(CloseTo, _And),
	geonames(A,_N,_,_,_,_,G,H,_,_,_,_,_,_,O,_,_,_,_),
	geo_localized_name(A, N),
	geo_feature_class_pretty(G, H, Class),
	atom_number(O, Pop),
	( Pop < 100000 ->
	  geo_closest_big(A, 100000, C2),
	  geo_localized_name(C2, NC2),
	  format('~w (~w) ~w ~w~n', [N, Class, CloseTo, NC2])
	; format('~w (~w)~n', [N, Class])
	),
	geo_latlon(A, Lat-Lon),
	geohack_url(Lat, Lon, [], URL1),
	openstreetmap_url(Lat, Lon, [], URL2),
	( geo_localized_wiki_url(A, URL0) ->
	  format('        ~w~n', [URL0])
	; true
	),
	format('        ~w~n', [URL1]),
	format('        ~w~n', [URL2]).


geo_phrases(X, Y) :-
        user_language(Lang),
	( Lang = de ->
	  geo_phrases_de(X, Y)
	; geo_phrases_en(X, Y)
	).

geo_phrases_en('close to', 'and').
geo_phrases_de('bei', 'und').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% WWW
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

www_geo(Id) :-
	www_geo_geohack(Id, []).

www_geo_map(Id) :-
	www_geo_openstreetmap(Id, []).

www_geo_openstreetmap(Id, Options) :-
	geo_latlon(Id, LL),
	www_latlon_openstreetmap(LL, Options).

www_latlon_openstreetmap(Lat-Lon, Options) :-
	openstreetmap_url(Lat, Lon, Options, URL),
	www_open_url(URL).

openstreetmap_url(Lat, Lon, Options, URL) :-
	%%
	%% Zoom: 1-18
	%% Layers: combination of letters:
	%% M=Mapnik, C=Cyclemap, T=Transport Map, Q=MapQuest,
	%% O=OpenStreetMap renderer, H=Humanitarian
	%%
	( memberchk(zoom=Zoom, Options) -> true ; Zoom=10 ),
	( memberchk(layers=Layers, Options) -> true ; Layers='M' ),
	format(atom(URL),
	       'http://www.openstreetmap.org/#map=~w/~6f/~6f&layers=~w',
	       [Zoom, Lat, Lon, Layers]).

www_geo_geohack(Id, Options) :-
	geo_latlon(Id, LL),
	www_latlon_geohack(LL, Options).

www_latlon_geohack(Lat-Lon, Options) :-
	geohack_url(Lat, Lon, Options, URL),
	www_open_url(URL).

geohack_url(Lat, Lon, Options, URL) :-
	%%
	( memberchk(language=Lang, Options) -> true
	; user_language(Lang)
	),
	( memberchk(type=Type, Options) -> true ; Type=adm3rd ),
	format(atom(URL),
	       'https://tools.wmflabs.org/geohack/geohack.php?language=~w&params=~w;~w_type:~w',
	       [Lang, Lat, Lon, Type]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

identify_location(PVs, [Id]) :-
	memberchk(id=Id, PVs),
	!.
identify_location(PVs, Ids) :-
	findall(Id1, identify_location_1(PVs, Id1), Ids1),
	sort(Ids1, Ids).

pvs_get(PVs, Key, Val) :-
	( memberchk(Key=Val, PVs) ->
	  true
	; true
	).


identify_location_1(PVs, Id) :-
	( memberchk(name=Name, PVs) ->
	  geo_name_to_id(Name, Id)
	; true
	),
	A = Id,
% 	Lat = E,
% 	Lon = F,
	pvs_get(PVs, feature_class, G),
	pvs_get(PVs, feature_code, H),
	pvs_get(PVs, country_code, I),
	pvs_get(PVs, preferred_name, N),
	geonames(A,N,_,_,_E,_F,G,H,I,_,_,_,_,_,O,_,_,_,_),
	atom_number(O, Pop),
	( memberchk(population_is_at_least=PopL, PVs) ->
	  ensure_number(PopL, PopLN),
	  PopLN =< Pop
	; true
	),
	( memberchk(population_is_at_most=PopM, PVs) ->
	  ensure_number(PopM, PopMN),
	  Pop =< PopMN
	; true
	).
	  
ensure_number(X, X) :-
	number(X),
	!.
ensure_number(X, Y) :-
	atom_number(X, Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

retract_supplemented_locations :-
	retractall( sup_name_to_location(_, _) ).

process_supplemented_location(Id, Sup) :-
	( member(name=Name, Sup),
	  assert( sup_name_to_location(Name, Id) ),
	  fail
	; true
	).
