%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(person, 
	  [person/1,
	   person_year_of_birth/2,
	   person_year_of_death/2,
	   person_gender/2,
	   person_occupation/2,
	   person_info/3,
	   person_last_name/2,
	   person_preferred_last_name/2,
	   person_preferred_name/2,
	   person_first_names/3,
	   person_first_names_initials/3,
	   person_extra_names/3,
	   person_extra_name/3,
	   person_is_in_dewiki/1,
	   person_is_in_wiki/1,
	   person_linked_to_person/2,
	   name_to_person/2,
	   person_pretty/2,
	   show_person/1,
	   show_person_nonl/1,
	   show_person_v/1,
	   show_person_vv/1,
	   show_person_vw/1,
	   www_person/1,
	   person_pretty_url/2,
	   person_url/2,
	   identify_person/2,
	   install_person_stuff/0,
	   register_ext_person/2,
	   ext_person/2,
	   retract_supplemented_persons/0,
	   process_supplemented_person/2]).

:- use_module(swilib(onto_atom)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).

:- use_module(kbset_base(gnd)).
:- use_module(kbset_base(dewiki)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_base(textutils)).
:- use_module(kbset_base(localization)).

:- dynamic sup_person_name/4.
:- dynamic sup_name_to_person/4.
:- dynamic sup_person_sup/2.

:- dynamic ext_person_data/6.
:- dynamic ext_person_name/4.
:- dynamic ext_name_to_person/4.
:- dynamic ext_person_url/2.
:- dynamic ext_person/2.


%%%% 
%%%% TODO: Means to add further persons that are not in the GND
%%%% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Accessor Predicates (One or no solution)
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% One or no solution
%%%% 
person_year_of_birth(PID, YoB) :-
	per_person_data(PID, _, YoB, _, _, _),
	!,
	number(YoB).

person_year_of_death(PID, YoD) :-	
	per_person_data(PID, _, _, YoD, _, _),
	!,
	number(YoD).

person_gender(PID, MF) :-
	per_person_data(PID, _, _, _, MF, _),
	!,
	memberchk(MF, [f,m]).

person_is_in_dewiki(PID) :-
	gnd_to_wiki(PID, de, _),
	!.

person_is_in_wiki(PID) :-
	gnd_to_wiki(PID, _, _),
	!.
	
person_preferred_last_name(PID, W) :-
	per_person_data(PID, Name, _, _, _, _),
	split_gnd_name(Name, W, _, _),
	!.

person_preferred_name(PID, Name) :-
	per_person_data(PID, Name, _, _, _, _),
	!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Accessor Predicates (Nondeterministic)
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

person_occupation(PID, Occupation) :-
	per_person_data(PID, _, _, _, _, Occs),
	!,
	map_subject_names(Occs, Occs1),
	sort(Occs1, Occs2),
	member(Occupation, Occs2).

person_last_name(PID, Name) :-
	findall(Name1,
		per_person_name(PID, Name1, _, _),
		Names),
	sort(Names, Names1),
	member(Name, Names1).

person_extra_name(PID, LastName, ExtraName) :-
	findall(Name1,
		per_person_name(PID, LastName, _, Name1),
		Names),
	sort(Names, Names1),
	member(ExtraName, Names1).


person_linked_to_person(PID, PID1) :-
	linked_persons_1(PID, PIDs),
	member(PID1, PIDs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


pp_year(nil) :-
	!,
	write('????').
pp_year(N) :-
	number(N),
	N >= 0,
	!,
	format('~|~`0t~d~4+', N).
pp_year(N) :-
	number(N),
	N < 0,
	!,
	N1 is -N,
	format('-~|~`0t~d~4+', N1).
pp_year(X) :-
	write(X).

person_pretty(X-_, Y) :-
	!,
	person_pretty(X, Y).
person_pretty(X, Y) :-
	onto_atom(show_person_nonl(X), Y).

show_person([]) :- !.
show_person([X|Xs]) :- !, ( member(Y, [X|Xs]), show_person(Y), fail ; true ).
show_person(X-_) :-
	!,
	show_person(X).
show_person(X) :-
	show_person_nonl(X),
	nl.

show_person_nonl(GNDId) :-
	per_person_data(GNDId, Name, YoB, YoD, _, _, YoBA, YoDA),
	!,
	write(Name),
	( (YoB = nil ; \+ integer(YoB)) ->
	  (( YoD = nil ; \+ integer(YoD)) ->
	    true
	  ; write(' ('),
	    write('-'),
	    pp_accuracy_year(YoDA, YoD),
	    write(')')
	  )
	; write(' ('),
	  pp_accuracy_year(YoBA, YoB),
	  write('-'),
	  (( YoD = nil ; \+ integer(YoD)) ->
	    true
	  ; pp_accuracy_year(YoDA, YoD)
	  ),
	  write(')')
	).
show_person_nonl(GNDId) :-
	format('Person with id ~w', [GNDId]).

pp_accuracy_year(1, Y) :- !, pp_year(Y).
pp_accuracy_year(0, Y) :- !, pp_year(Y).
pp_accuracy_year(_, Y) :- !, write('['), pp_year(Y), write(']').

show_person_v([]) :-
	!.
show_person_v([X|Xs]) :-
	!, ( member(Y, [X|Xs]), show_person_v(Y), fail ; true ).
show_person_v(X) :-
	show_person_v_1(X),
	nl.

show_person_v_1(GNDId) :-
	show_person_nonl(GNDId),
	per_person_data(GNDId, _, _, _, _, Ss),
	map_subject_names(Ss, Ss1),
	sort(Ss1, Ss2),
	( Ss2 = [] ->
	  true
	; format('~n        '),
	  pp_commalist(Ss2)
	).

show_person_vw([]) :-
	!.
show_person_vw([X|Xs]) :-
	!, ( member(Y, [X|Xs]), show_person_vw(Y), fail ; true ).
show_person_vw(X) :-
	show_person_v_1(X),
	print_person_urls(X, 8),
	nl.

show_person_vv(GNDId) :-
	show_person_v_1(GNDId),
	user_language(Lang),
	person_infos(GNDId, Lang, Infs),
	( member(Inf, Infs),
	  write('\n\n'),
	  write(Inf),
	  fail
	; true
	),
	nl.

map_subject_names([S|Ss], [S1|Ss1]) :-
	subject_name(S, S1),
	!,
	map_subject_names(Ss, Ss1).
map_subject_names([], []).

subject_name(X, Y) :-
	gnd_subject_name(X, Y),
	!.
subject_name(X, X).

pp_commalist([]) :-
	!.
pp_commalist([X]) :-
	!,
	write(X).
pp_commalist([X,Y|Ys]) :-
	write(X),
	write(', '),
	pp_commalist([Y|Ys]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

www_person(Id) :-
	person_url(Id, URL),
	www_open_url(URL).

person_dewiki_url(Id, URL) :-
	gnd_to_wiki(Id, de, W),
	!,
	www_form_encode(W, W1),
	format(atom(URL), 'http://de.wikipedia.org/wiki/~w', [W1]).
person_dewiki_url(Id, URL) :-
	gnd_to_wiki(Id, en, W),
	!,
	www_form_encode(W, W1),
	format(atom(URL), 'http://en.wikipedia.org/wiki/~w', [W1]).

person_dewiki_unencoded_url(Id, URL) :-
	gnd_to_wiki(Id, de, W),
	!,
	format(atom(URL), 'http://de.wikipedia.org/wiki/~w', [W]).
person_dewiki_unencoded_url(Id, URL) :-
	gnd_to_wiki(Id, en, W),
	!,
	format(atom(URL), 'http://en.wikipedia.org/wiki/~w', [W]).

person_gnd_url(Id, URL) :-
	atom_concat('gnd:', GNDId, Id),
	atom_concat('http://d-nb.info/gnd/', GNDId, URL).

person_url(Id, URL) :-
	person_dewiki_url(Id, URL),
	!.
person_url(Id, URL) :-
	ext_person_url(Id, URL),
	!.
person_url(Id, URL) :-
	person_gnd_url(Id, URL).

print_person_urls(Id, Indent) :-
	( person_dewiki_unencoded_url(Id, URL1) ->
	  nl, indent(Indent), write(URL1)
	; true
	),
	( person_gnd_url(Id, URL2) ->
	  nl, indent(Indent), write(URL2)
	; true
	),
	( ext_person_url(Id, URL3) ->
	  nl, indent(Indent), write(URL3)
	; true
	).
	
person_pretty_url(Id, URL) :-
	person_dewiki_unencoded_url(Id, URL),
	!.
person_pretty_url(Id, URL) :-
	person_gnd_url(Id, URL).

person_infos(Id, Lang, Atoms) :-
	findall(Inf, person_info(Id, Lang, Inf), Infs),
	sort(Infs, Atoms).

person_info(Id, Lang, Inf) :-
	gnd_person_info(Id, Lang, Inf).
person_info(Id, Lang, Inf) :-
	ext_person(Id, PropVals),
	member( biographicalOrHistoricalInformation=lang(Lang, Inf),
		PropVals).
person_info(Id, Lang, Inf) :-
	sup_person_sup(Id, PropVals),
	member( biographicalOrHistoricalInformation=lang(Lang, Inf),
		PropVals).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

name_to_person(LastName, PID) :-
	ord_enum(PID, per_name_to_person(LastName, _, _, PID)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
delete_ext_persons :-
	retractall(ext_person_data(_,_,_,_,_,_)),
	retractall(ext_person_name(_,_,_,_)),
	retractall(name_to_person(_,_,_,_)),
	retractall(ext_person_url(_,_)),
	retractall(ext_person(_,_)).

delete_ext_person(Id) :-
	retractall(ext_person_data(Id,_,_,_,_,_)),
	retractall(ext_person_name(Id,_,_,_)),
	retractall(ext_name_to_person(_,_,_,Id)),
	retractall(ext_person_url(Id,_)),
	retractall(ext_person(Id,_)).
	
register_ext_person(Id, PropVals) :-
	delete_ext_person(Id),
	( memberchk(preferredNameForThePerson=PrefName, PropVals) ->
	  true
	; err('No preferred name supplied for person ~q', [Id])
	),
	( memberchk(dateOfBirth=DoB, PropVals), local_date_year(DoB, YoB) ->
	  true
	; YoB = nil
	),
	( memberchk(dateOfDeath=DoD, PropVals), local_date_year(DoD, YoD) ->
	  true
	; YoD = nil
	),
	( memberchk(gender=male, PropVals) -> Gender=m
	; memberchk(gender=female, PropVals) -> Gender=f
	; Gender = nil
	),
	findall(Occ, member(professionOrOccupation=Occ, PropVals),
		OccupationIDs),
	assert(ext_person(Id, PropVals)),
	assert(ext_person_data(Id, PrefName, YoB, YoD, Gender, OccupationIDs)),
	( ( Name = PrefName
	  ; member(variantNameForThePerson=Name, PropVals)
	  ),
	  split_gnd_name(Name, NLast, NFirst, NExtra),
	  assert(ext_person_name(Id, NLast, NFirst, NExtra)),
	  assert(ext_name_to_person(NLast, NFirst, NExtra, Id)),
	  fail
	; true
	),
	( memberchk(url=URL, PropVals) ->
	  assert(ext_person_url(Id, URL))
	; true
	),
	reinstall_person_cache(Id).


per_person_data(Id, Name, YoB, YoD, Gender, OccupationIDs, YoBA, YoDA) :-
	gnd_person_data(Id, Name1, YoB1, YoD1, Gender1, OccupationIDs1,
			YoBA1, YoDA1),
	!,
	Name = Name1, YoB = YoB1, YoD = YoD1, Gender = Gender1,
	YoBA = YoBA1, YoDA = YoDA1,
	OccupationIDs = OccupationIDs1.
per_person_data(Id, Name, YoB, YoD, Gender, OccupationIDs, 1, 1) :-
	ext_person_data(Id, Name1, YoB1, YoD1, Gender1, OccupationIDs1),
	!,
	Name = Name1, YoB = YoB1, YoD = YoD1, Gender = Gender1,
	OccupationIDs = OccupationIDs1.


per_person_data(Id, Name, YoB, YoD, Gender, OccupationIDs) :-
	gnd_person_data(Id, Name1, YoB1, YoD1, Gender1, OccupationIDs1),
	!,
	Name = Name1, YoB = YoB1, YoD = YoD1, Gender = Gender1,
	OccupationIDs = OccupationIDs1.
per_person_data(Id, Name, YoB, YoD, Gender, OccupationIDs) :-
	ext_person_data(Id, Name1, YoB1, YoD1, Gender1, OccupationIDs1),
	!,
	Name = Name1, YoB = YoB1, YoD = YoD1, Gender = Gender1,
	OccupationIDs = OccupationIDs1.

per_person_name(Id, LastName, FirstName, ExtraName) :-
	gnd_person_name(Id, LastName1, FirstName1, ExtraName1),
	LastName=LastName1, FirstName=FirstName1, ExtraName = ExtraName1.
per_person_name(Id, LastName, FirstName, ExtraName) :-
	ext_person_name(Id, LastName1, FirstName1, ExtraName1),
	LastName=LastName1, FirstName=FirstName1, ExtraName = ExtraName1.
per_person_name(Id, LastName, FirstName, ExtraName) :-
	sup_person_name(Id, LastName1, FirstName1, ExtraName1),
	LastName=LastName1, FirstName=FirstName1, ExtraName = ExtraName1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

retract_supplemented_persons :-
 	retractall( sup_person_name(_,_,_,_) ),
 	retractall( sup_name_to_person(_,_,_,_) ),
 	retractall( sup_person_sup(_,_) ).

process_supplemented_person(Id, Sup) :-
	assert( sup_person_sup(Id, Sup) ),
	( member(name=Name, Sup),
	  split_gnd_name(Name, Last, First, Extra),
	  assert( sup_person_name(Id, Last, First, Extra) ),
	  assert( sup_name_to_person(Last, First, Extra, Id) ),
	  fail
	; true
	),
	reinstall_person_cache(Id).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

per_name_to_person(LastName, FN, EN, PID1) :-
	gnd_name_to_person(LastName, FN, EN, PID1).
per_name_to_person(LastName, FN, EN, PID1) :-
	ext_name_to_person(LastName, FN, EN, PID1).
per_name_to_person(LastName, FN, EN, PID1) :-
	sup_name_to_person(LastName, FN, EN, PID1).

local_date_year(Y, Y) :-
	number(Y),
	!.
local_date_year(D, Y) :-
	atom(D),
	atom_chars(D, D1),
	( D1 = [Y1,Y2,Y3,Y4,-,_,_,-,_,_] ->
	  true
	; D1 = [Y1,Y2,Y3,Y4,-,_,_] ->
	  true
	; D1 = [Y1,Y2,Y3,Y4] ->
	  true
	; D1 = [_,_,'.',_,_,'.',Y1,Y2,Y3,Y4] ->
	  true
	; D1 = [_,_,'.',Y1,Y2,Y3,Y4] ->
	  true
	; D1 = [Y1,Y2,Y3,Y4,_,_,_,_] ->
	  true
	),
	atom_chars(Y0, [Y1,Y2,Y3,Y4]),
	atom_number(Y0, Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

identify_person(PVs, [Id]) :-
	memberchk(id=Id, PVs),
	!.
identify_person(PVs, Ids) :-
	findall(Id1, identify_person_1(PVs, Id1), Ids1),
	sort(Ids1, Ids).

identify_person_1(PVs, Id) :-
	( memberchk(dateOfBirth=DoB, PVs), local_date_year(DoB, YoB) ->
	  true
	; true
	),
	( memberchk(dateOfDeath=DoD, PVs), local_date_year(DoD, YoD) ->
	  true
	; true
	),
	( memberchk(gender=male, PVs) -> Gender=m
	; memberchk(gender=female, PVs) -> Gender=f
	; true
	),
	findall(Id1,
		( member(name=Name, PVs),
		  split_gnd_name(Name, NLast0, NFirst0, NExtra0),
		  ( NLast0 \= '' -> NLast = NLast0 ; true ),
		  ( NFirst0 \= '' -> NFirst = NFirst0 ; true),
		  ( NExtra0 \= '' -> NExtra = NExtra0 ; true),
		  per_name_to_person(NLast, NFirst, NExtra, Id1),
		  
		  ( memberchk(preferredNameForThePerson=Fullname1, PVs) ->
		    split_gnd_name(Fullname1, NLast1, NFirst1, NExtra1),
		    per_name_to_person(NLast1, NFirst1, NExtra1, Id1)
		  ; true
		  ),
		  ( memberchk(variantNameForThePerson=Fullname2, PVs) ->
		    split_gnd_name(Fullname2, NLast2, NFirst2, NExtra2),
		    per_name_to_person(NLast2, NFirst2, NExtra2, Id1)
		  ; true
		  )
		),
		Ids1),
	sort(Ids1, Ids2),
	member(Id, Ids2),
	per_person_data(Id, _, YoB, YoD, Gender, Occups),
	aux_identify_persons_match_occupations(Occups, PVs).


aux_identify_persons_match_occupations(_, PVs) :-
	\+ memberchk(professionOrOccupation=_, PVs),
	!.
aux_identify_persons_match_occupations(Occups, PVs) :-
	member(professionOrOccupation=Occ, PVs),
	member(Occ1, Occups),
	subject_name(Occ1, Occ),
	!.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

person(X) :-
	ext_person(X, _).
person(X) :-
	gnd_person(X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic person_first_names_cache/3.
:- dynamic person_extra_names_cache/3.
:- dynamic person_first_names_initials_cache/3.

person_extra_names(X, Y, Z) :-
	person_extra_names_cache(X, Y, Z).

person_first_names(X, Y, Z) :-
	person_first_names_cache(X, Y, Z).

person_first_names_initials(X, Y, Z) :-
	person_first_names_initials_cache(X, Y, Z).

install_person_stuff :-
	info(10, 'Setting up: person caches'),
	flag(cnt, _, 0),
	retractall(person_first_names_cache(_,_,_)),
	retractall(person_extra_names_cache(_,_,_)),
	retractall(person_first_names_initials_cache(_,_,_)),
	verify_predicates([ext_person/2, gnd_person/1]),
	( person(X),
	  flag(cnt, N, N+1),
	  info_progress(20, 10000, N),
	  install_person_cache(X),
	  fail
	; true
	).

install_person_cache(PID) :-
	( ord_enum(LastName, person_last_name(PID, LastName)),
	  install_person_cache_1(PID, LastName),
	  fail
	; true
	).

install_person_cache_1(PID, LastName) :-
	( person_first_names_0(PID, LastName, Y1),
	  assert(person_first_names_cache(PID, LastName, Y1)),
	  fail
	; true
	),
	( person_extra_names_0(PID, LastName, Y2),
	  assert(person_extra_names_cache(PID, LastName, Y2)),
	  fail
	; true
	),
	( person_first_names_initials_0(PID, LastName, Y3),
	  assert(person_first_names_initials_cache(PID, LastName, Y3)),
	  fail
	; true
	).

	
reinstall_person_cache(PID) :-
	retractall( person_first_names_cache(PID, _, _) ),
	retractall( person_extra_names_cache(PID, _, _) ),
	retractall( person_first_names_initials_cache(PID, _, _) ),
	install_person_cache(PID).

%%%% 
person_first_names_0(PID, LastName, ListOfFirstNames) :-
	findall(ListOfFirstNames1,
		person_first_names_1(PID, LastName, ListOfFirstNames1),
		Names),
	sort(Names, Names1),
	member(ListOfFirstNames, Names1).
	
person_first_names_1(PID, LastName, ListOfFirstNames) :-
	per_person_name(PID, LastName, FirstNamesAtom, _),
	FirstNamesAtom \= '',
	atomic_list_concat(ListOfFirstNames1, ' ', FirstNamesAtom),
	check_first_names(ListOfFirstNames1),
	map(trim_right_fullstop, ListOfFirstNames1, ListOfFirstNames).

check_first_names(Names) :-
	%% Occurs often in the GND:
	memberchk('...', Names),
	!,
	fail.
check_first_names(_).
	
trim_right_fullstop(A, A1) :-
	( atom_concat(A1, '.', A) ->
	  true
	; A1 = A
	).

person_extra_names_0(PID, LastName, ListOfExtraNames) :-
	findall(ListOfExtraNames1,
		person_extra_names_1(PID, LastName, ListOfExtraNames1),
		Names),
	sort(Names, Names1),
	member(ListOfExtraNames, Names1).
	
person_extra_names_1(PID, LastName, ListOfExtraNames) :-
	per_person_name(PID, LastName, _, ExtraNamesAtom),
	ExtraNamesAtom \= '',
	atomic_list_concat(ListOfExtraNames1, ' ', ExtraNamesAtom),
	map(trim_right_fullstop, ListOfExtraNames1, ListOfExtraNames).

%%%% 
person_first_names_initials_0(PID, LastName, ListOfFirstNameInitial) :-
	findall(ListOfFirstNameInitial1,
		person_first_names_initials_1(PID, LastName, ListOfFirstNameInitial1),
		Inits),
	sort(Inits, Inits1),
	member(ListOfFirstNameInitial, Inits1),
	ListOfFirstNameInitial \= [].

person_first_names_initials_1(PID, LastName, ListOfFirstNameInitial) :-
	per_person_name(PID, LastName, FirstNamesAtom, _),
	FirstNamesAtom \= '',
	split_atom_at_punct_and_space(FirstNamesAtom, FirstNames),
	map(name_initial, FirstNames, ListOfFirstNameInitial),
	ListOfFirstNameInitial \= [v].

name_initial(von, v).
name_initial('v.', v).
name_initial(Name, _) :-
	\+ starts_with_upper_char(Name),
	!,
	fail.
name_initial(Name, Initial) :-
	atom_concat(Initial, '.', Name),
	!.
name_initial(Name, Initial) :-
	sub_atom(Name, 0, 1, _, Initial).
name_initial(Name, Initial) :-
	sub_atom(Name, 0, 2, _, Initial),
	memberchk(Initial, ['Ch', 'Th', 'Ph', 'Ps']).

indent(N) :-
	N > 0,
	!,
	write(' '),
	N1 is N - 1,
	indent(N1).
indent(0).
