%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(readwrite, [read_in_latex/2,
		      register_latex_command/3,
		      lxelems_write_src/1,
		      lxelems_plaintxt_codes/2,
		      lxelem_enum_plaintxt/2,
		      lxelem_enum_srctxt/2,
		      pretty_elemtype/2]).

:- encoding(utf8).

:- use_module(swilib(pretty)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(utils)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Limitations:
%%%%
%%%% - Only command arguments in "{}" (and options in "[]") are
%%%%   considered as such
%%%% - Optional arguments:
%%%%   Their content is taken as a single atom, delimited by "]"
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lxelems_plaintxt_codes(Es, Cs) :-
	findall(C, ( member(E, Es),
		     lxelem_enum_plaintxt(E, A),
		     atom_codes(A, Cs1),
		     member(C, Cs1) ),
		Cs).

%%%% 
%%%% These are nondeterministic predicates that enumerate
%%%% atoms constituting the text
%%%% 
lxelem_enum_srctxt(opq(Atom), Atom).
lxelem_enum_srctxt(wrd(Atom), Atom).
lxelem_enum_srctxt(com(Atom), Atom). % :- format(atom(Atom1), '%~w', [Atom]).
lxelem_enum_srctxt(pun(Atom), Atom).
lxelem_enum_srctxt(spc(Atom), Atom).
lxelem_enum_srctxt(cmd(Cmd,_), Atom) :- format(atom(Atom), '\\~w', [Cmd]).
lxelem_enum_srctxt(beg, '{').
lxelem_enum_srctxt(end, '}').
lxelem_enum_srctxt(ctr(Ctr), Atom) :- format(atom(Atom), '{~w}', [Ctr]).
lxelem_enum_srctxt(opt(Opt), Atom) :- format(atom(Atom), '[~w]', [Opt]).
lxelem_enum_srctxt(ebg, '{').
lxelem_enum_srctxt(een, '}').

%%%%
lxelem_enum_plaintxt(opq(_), '').
lxelem_enum_plaintxt(wrd(Atom), Atom).
lxelem_enum_plaintxt(com(_), '').
lxelem_enum_plaintxt(pun(Atom), Atom).
lxelem_enum_plaintxt(spc(Atom), Atom).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic rc/3.

%%%% 
%%%% Effects:
%%%% 
%%%% Argument patterns: List of items:
%%%% c - "control" argument (does not belong to plaintext)
%%%% t - "text" argument, belongs to plaintext
%%%% o - option argument, does not belong to plaintext
%%%% 
%%%% - at parsing: prevent wrong parsing
%%%%   (e.g. deciding whether the next item in "{}" is to be considered
%%%%   as argument)
%%%%
%%%% - at determining the plain text:
%%%%   specifies what arguments are considered as beloning to the
%%%%   plain text (argument pattern is not used to determine plain text,
%%%%   thus all entries for a command should have the same plain text).
%%%%
register_latex_command(Name, ArgumentPattern, PlainText) :-
	retractall(rc(Name, ArgumentPattern, _)),
	assert(rc(Name, ArgumentPattern, PlainText)).

latex_command(setcounter,[c,c],'').
latex_command(setlength,[c,c],'').
latex_command(newlength,[c],'').
latex_command(newcommand,[c,o,c],'').
latex_command(newcommand,[c,c],'').
latex_command(renewcommand,[c,o,c],'').
latex_command(renewcommand,[c,c],'').
latex_command(fancyhead,[c],'').
latex_command(fancyhead,[o,c],'').
latex_command(hspace,[c],'').
latex_command(vspace,[c],'').
latex_command('hspace*',[c],'').
latex_command('vspace*',[c],'').
latex_command('hfill',[],'').
latex_command('vfill',[],'').
latex_command(label,[c],'').
latex_command(ref,[c],'').
latex_command(cite,[c],'').
latex_command(cite,[o,c],'').
latex_command('\\',[o],'\n').
latex_command('\\',[],'\n').
latex_command('L',[],'Ł').
latex_command('l',[],'ł').
latex_command('&',[],'&').
latex_command('{',[],'{').
latex_command('}',[],'}').
latex_command('#',[],'#').
latex_command('_',[],'_').
latex_command('%',[],'%').
latex_command(textbackslash,[],'\\').
latex_command(begin,[c,c],'').
latex_command(begin,[c],'').
latex_command(end,[c],'').
latex_command(newblock,[],'').
latex_command(glqq,[],'„').
latex_command(grqq,[],'“').
latex_command(glq,[],'‚').
latex_command(grq,[],'‘').
latex_command(pagenumbering,[c],'').
latex_command('=',[],'-').
latex_command(rule,[c,c],'-').
latex_command(rule,[o,c,c],'-').

% latex_command('PAGE',[c],'').
% latex_command('secx',[t],'').

:- initialization(( latex_command(C, A, T),
		    register_latex_command(C, A, T),
		    fail
		  ; true
		  )).

rc(X, Y) :-
	rc(X, Y, _).

command_enum_pattern(Cmd, Pattern) :-
	findall(P, rc(Cmd, P), Ps),
	( Ps = [] ->
	  default_pattern(Pattern)
	; sort_patterns(Ps, Ps1),
	  member(Pattern, Ps1)
	).

sort_patterns(Ps, Ps1) :-
	map_pattern_add_sortkey(Ps, Ps2),
	keysort(Ps2, Ps3),
	map_val(Ps3, Ps1).

pattern_add_sortkey(P, k(L,P1)-P) :-
	length(P, L1),
	L is 1/(L1+1),
	map_pattern_item_sortkey(P, P1).

pattern_item_sortkey(t, 0).
pattern_item_sortkey(c, 1).
pattern_item_sortkey(o, 2).

map_pattern_add_sortkey([X|Xs], [X1|Xs1]) :-
	pattern_add_sortkey(X, X1),
	map_pattern_add_sortkey(Xs, Xs1).
map_pattern_add_sortkey([], []).

map_pattern_item_sortkey([X|Xs], [X1|Xs1]) :-
	pattern_item_sortkey(X, X1),
	map_pattern_item_sortkey(Xs, Xs1).
map_pattern_item_sortkey([], []).

default_pattern([c,c,t]).
default_pattern([c,o,c]).
default_pattern([c,t]).
default_pattern([o,t]).
default_pattern([o,c]).
default_pattern([t]).
default_pattern([c]).
default_pattern([o]).
default_pattern([]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lxelems_write_src(Es) :-
	( member(E, Es),
	  lxelem_enum_srctxt(E, A),
	  write(A),
	  fail
	; true
	).

lxelems_write_plain(Es) :-
	( member(E, Es),
	  lxelem_enum_plaintxt(E, A),
	  write(A),
	  fail
	; true
	).

cmd_plaintxt(Cmd, Atom) :-
	rc(Cmd, _, Atom),
	!.

cmd_srctxt(Cmd, Atom) :-
	format(atom(Atom), '\\~w', [Cmd]).

pretty_elemtype(F/N, A) :-
	pretty_elemtype_1(F/N, A),
	!.
pretty_elemtype(F/_, F).

pretty_elemtype_1(wrd/_, 'word').
pretty_elemtype_1(com/_, 'comment').
pretty_elemtype_1(pun/_, 'punctuation').
pretty_elemtype_1(spc/_, 'whitespace').
pretty_elemtype_1(wrd/_, 'word').
pretty_elemtype_1(cmd/_, 'meta:command').
pretty_elemtype_1(opq/_, 'meta:code').
pretty_elemtype_1(beg/_, 'meta:begin').
pretty_elemtype_1(end/_, 'meta:end').
pretty_elemtype_1(opt/_, 'meta:option').
pretty_elemtype_1(ctr/_, 'meta:control').
pretty_elemtype_1(ebg/_, 'meta:curly braces begin').
pretty_elemtype_1(een/_, 'meta:curly braces end').
pretty_elemtype_1(nil/_, 'unspecified').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Elems is a list of elements, where an element is one of:
%%%% 
%%%% opq(Atom)
%%%%     "opaque", latex code that is not decomposed
%%%% 
%%%% wrd(Atom)
%%%%     "word", word without space, punctuation and control characters
%%%% 
%%%% com(Atom) 
%%%%     "comment"
%%%% 
%%%% pun(Atom)
%%%%     "punctuation", just punctuation characters
%%%%
%%%% spc(Atom)
%%%%     "space", whitespace, control and possibly other characters
%%%%
%%%% cmd(Command, ArgumentPattern)
%%%% beg
%%%% end
%%%% opt
%%%% ctr
%%%% ebg
%%%% een
%%%%
%%%% nil
%%%%     ignored
%%%%			   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

read_in_latex(FileSpec, Elems) :-
	( FileSpec = [_|_] ->
	  Codes = FileSpec,
	  phrase(elems(Elems, []), Codes, R)
	; filename(FileSpec, File),
	  info(20, 'Reading ~w', [File]),
	  read_file_to_codes(File, Codes, [encoding(utf8)]),
	  ( is_xml_file(File) ->
	    phrase(rxml_elems(Elems, []), Codes, R)
	  ; phrase(elems(Elems, []), Codes, R)
	  )
	),
	( R \= [] ->
	  atom_codes(A, R),
	  sub_atom(A, 0, 60, _, Rest),
	  ( last(Elems, Last) -> true ; Last = none ),
	  err('Parsing failed at: "~w..."~nLast element: ~q', [Rest, Last]),
	  fail
	; FileSpec = [_|_] ->
	  true
	; length(Elems, L),
	  info(20, 'Read ~D elements', [L])
	).

%%% NEW

test(Codes, X, Y, Z) :-
	phrase(elems(X, Y), Codes, Z1),
	atom_codes(Z, Z1).

test(Codes, X, R) :-
	phrase(elems(X, []), Codes, R).



elems([X|Xs], Ys) -->
	opaque_elem(X),
	!,
	elems(Xs, Ys).
elems(Xs, Ys) -->
	cmd_seq(Xs, Zs),
	!,
	elems(Zs, Ys).
elems([X|Xs], Ys) -->
	elem(X),
	!,
	elems(Xs, Ys).
elems(X, X) -->
	[].

cmd_seq([cmd(Name,Pattern)|Xs], Ys) -->
	command_name(Name),
	{ command_enum_pattern(Name, Pattern) },
	cmd_args(Pattern, Xs, Ys).
cmd_seq([ebg|Xs], Ys) -->
	`{`,
	elems(Xs, [een| Ys]),
	`}`.

cmd_args([t|Pattern], Xs, Ys) -->
	skip_whitespace(Xs, [beg|Xs1]),
	`{`,
	elems(Xs1, [end|Zs]),
	`}`,
	cmd_args(Pattern, Zs, Ys). 
cmd_args([c|Pattern], Xs, Ys) -->
	skip_whitespace(Xs, [ctr(Control)|Xs1]),
	`{`,
	codes(Codes),
	`}`,
	{ atom_codes(Control, Codes) },
	cmd_args(Pattern, Xs1, Ys).
cmd_args([o|Pattern], Xs, Ys) -->
	skip_whitespace(Xs, [opt(Control)|Xs1]),
	`[`,
	codes(Codes),
	`]`,
	{ atom_codes(Control, Codes) },
	cmd_args(Pattern, Xs1, Ys).
cmd_args([], Xs, Xs) -->
	``.

skip_whitespace(Xs, Ys) -->
	skip_whitespace_1(Cs),
	{ ( Cs = [] ->
	    Ys = Xs 
	  ; atom_codes(A, Cs),
	    Xs = [spc(A)|Ys]
	  )
	}.

skip_whitespace_1([X|Xs]) -->
	[X],
	{ code_type(X, space) },
	!,
	skip_whitespace_1(Xs).
skip_whitespace_1([]) -->
	[].

codes(X) -->
	codes(X, []).

codes(X, X) -->
	``.
codes([C|Cs], Y) -->
	[C],
	{ C \= 0'{, C \= 0'}, C \= 0'[, C \= 0'] },
	codes(Cs, Y).	
codes([0'{|X], Y) -->
	`{`,
	codes(X, [0'}|Z]),
	`}`,
	codes(Z, Y).
codes([0'[|X], Y) -->
	`[`,
	codes(X, [0']|Z]),
	`]`,
	codes(Z, Y).
	
% elem(X) -->
%  	opaque_elem(X).
elem(X) -->
	word_elem(X).
elem(X) -->
	comment_elem(X).
elem(X) -->
	punct_elem(X).
elem(X) -->
	space_elem(X).

opaque_elem(opq(E)) -->
	`\\documentclass`,
	begin_document(E1),
	!,
	{ append(`\\documentclass`, E1, E2),
	  atom_codes(E, E2)
	}.

opaque_elem(opq(E)) -->
	`\\end{document}`,
	codes_until_eof(E1),
	{ append(`\\end{document}`, E1, E2),
	  atom_codes(E, E2)
	}.

%%%% 
%%%% Text in xopaque environment is preserved but excluded from processing
%%%%  
opaque_elem(opq(E)) -->
	`\\begin{xopaque}`,
	end_opaque(E1),
	!,
	{ append(`\\begin{xopaque}`, E1, E2),
	  atom_codes(E, E2)
	}.

end_opaque(`\\end{xopaque}`) -->
	`\\end{xopaque}`,
	!.
end_opaque([X|Xs]) -->
	[X],
	end_opaque(Xs).

begin_document(`\\begin{document}`) -->
	`\\begin{document}`,
	!.
begin_document([X|Xs]) -->
	[X],
	begin_document(Xs).
	
word_elem(wrd(E)) -->
	[X],
	{ \+ delimiter(X) },
	codes_until_delimiter(Y),
	{ atom_codes(E, [X|Y]) }.

comment_elem(com(X)) -->
	`%`,
	codes_until_eol(Y),
	{ atom_codes(X, [0'%|Y]) }.

punct_elem(pun(E)) -->
	[X],
	{ punctuation(X) },
	!,
	codes_until_no_punctuation(Y),
	{ atom_codes(E, [X|Y]) }.

space_elem(spc(E)) -->
	[X],
	{ spc_delimiter(X) },
	codes_until_no_delimiter(Y),
	{ atom_codes(E, [X|Y]) }.

%%%% 
%%%% A backslash escaped single delimiter is considered as command name:
%%%% 
command_name(X) -->
	`\\`,
	[X1],
	{ delimiter(X1) },
	!,
	{ atom_codes(X, [X1]) }.
command_name(X) -->
	`\\`,
	codes_until_delimiter(X1),
	{ atom_codes(X, X1) }.

spc_delimiter(X) :-
	delimiter(X),
	\+ x_delimiter(X),
	\+ punctuation(X).

x_delimiter(0'\\).
x_delimiter(0'}).
x_delimiter(0'{).
x_delimiter(0'%).

delimiter(0'*) :- !, false.
delimiter(X) :- code_type(X, space), !.
delimiter(X) :- code_type(X, cntrl), !.
delimiter(X) :- code_type(X, punct), !.

punctuation(X) :-
	code_type(X, punct),
	\+ x_delimiter(X),
	!.
	
codes_until_delimiter([X|Y]) -->
	[X],
	{ \+ delimiter(X) },
	!,
	codes_until_delimiter(Y).
codes_until_delimiter([]) -->
	[].

codes_until_no_delimiter([X|Y]) -->
	[X],
	{ spc_delimiter(X) },
	!,
	codes_until_no_delimiter(Y).
codes_until_no_delimiter([]) -->
	[].

codes_until_no_punctuation([X|Y]) -->
	[X],
	{ punctuation(X) },
	!,
	codes_until_no_punctuation(Y).
codes_until_no_punctuation([]) -->
	[].

codes_until_squareclose([X|Y]) -->
	[X],
	{ X \= 0'] },
	!,
	codes_until_squareclose(Y).
codes_until_squareclose([]) -->
	[].

codes_until_eol([X|Y]) -->
	[X],
	{ \+ code_type(X, end_of_line) },
	!,
	codes_until_eol(Y).
codes_until_eol([X]) -->
	[X],
	{ code_type(X, newline) }.
%% In case there is no newline at EOF:
codes_until_eol([]) -->
	[].

codes_until_eof([X|Y]) -->
	[X],
	!,
	codes_until_eof(Y).
codes_until_eof([]) -->
	[].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% RXML - "Rough XML" 
%%%%
%%%% Rudimentary support for XML documents. They are read in just as text,
%%%% except that everything in "<>" is treated as "opaque" element.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_xml_file(File) :-
	file_name_extension(_, Ext, File),
	memberchk(Ext, [xml, 'XML', tei, 'TEI']).

rxml_elems([X|Xs], Ys) -->
	rxml_opaque_elem(X),
	!,
	rxml_elems(Xs, Ys).
rxml_elems([X|Xs], Ys) -->
	elem(X),
	!,
	rxml_elems(Xs, Ys).
rxml_elems(X, X) -->
	[].

rxml_opaque_elem(opq(E)) -->
	`<`,
	rxml_end_opaque(E1),
	!,
	{ append(`<`, E1, E2),
	  atom_codes(E, E2)
	}.

rxml_end_opaque(`>`) -->
	`>`,
	!.
rxml_end_opaque([X|Xs]) -->
	[X],
	rxml_end_opaque(Xs).

