%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(textutils,
	  [starts_with_upper_char/1,
	   upcase_first_char/2,
	   is_upper_char/1,
	   is_upper_code/1,
	   downcase/2,
	   downcase_chars/2,
	   has_suffix/3,
	   has_suffix/2,
	   not_has_suffix/2,
	   alnum_prefix/2,
	   is_roman_number/1,
	   starts_with_roman_number/2,
	   split_atom_at_punct_and_space/2,
	   punct_stops_sentence/1,
	   plain_sortkey/2,
	   enum_umlaut_variants/2,
	   remove_diacritics_used_in_handwriting/2,
	   esc_latex/2,
	   esc_latex_pretty/2]).

%%%% 
%%%% The Prolog built-in downcase_atom/2 and char_type/2 seems to handle
%%%% Umlauts not properly. Perhaps we should add here also other special
%%%% characters e.g. from Skandinavian languages.
%%%%

downcase(A, A1) :-
	atom_chars(A, C),
	downcase_chars(C, C1),
	atom_chars(A1, C1).

downcase_chars([C|Cs], [C1|Cs1]) :-
	downcase_char(C, C1),
	downcase_chars(Cs, Cs1).
downcase_chars([], []).

upcase_char(C, C1) :-
	downcase_char(C1, C).

downcase_char('Ä', 'ä') :- !.
downcase_char('Ü', 'ü') :- !.
downcase_char('Ö', 'ö') :- !.
downcase_char('É', 'é') :- !.
downcase_char('Á', 'á') :- !.
downcase_char('Ñ', 'ñ') :- !.
downcase_char('Ã', 'ã') :- !.

downcase_char(C, C1) :-	char_type(C, to_upper(C1)), !.

upcase_first_char(W, W1) :-
	sub_atom(W, 0, 1, _, C),
	( is_upper_char(C) -> W1 = W
	; sub_atom(W, 1, _, 0, W2),
	  upcase_char(C, C1),
	  atomic_list_concat([C1,W2], W1)
	).

starts_with_upper_char(W) :-
	sub_atom(W, 0, 1, _, C),
	is_upper_char(C).

is_upper_char(C) :- char_type(C, upper), !.
is_upper_char('Ä') :- !.
is_upper_char('Ü') :- !.
is_upper_char('Ö') :- !.
is_upper_char('É') :- !.
is_upper_char('Á') :- !.
is_upper_char('Ñ') :- !.
is_upper_char('Ã') :- !.

is_upper_code(C) :- code_type(C, upper), !.
is_upper_code(0'Ä) :- !.
is_upper_code(0'Ü) :- !.
is_upper_code(0'Ö) :- !.
is_upper_code(0'É) :- !.
is_upper_code(0'Á) :- !.
is_upper_code(0'Ñ) :- !.
is_upper_code(0'Ã) :- !.

has_suffix(W, Suffix, W1) :-
	sub_atom(W, _, Len, 0, Suffix),
	sub_atom(W, 0, _, Len, W1),
	!.

has_suffix(W, Suffix) :-
	sub_atom(W, _, _, 0, Suffix).

not_has_suffix(W, Suffix) :-
	\+  sub_atom(W, _, _, 0, Suffix).

is_roman_number(X) :-
	atom_length(X, L),
	L > 0,
	atom_chars(X, Cs),
	\+ ( member(C, Cs), \+ roman_char(C) ).

roman_char('I').
roman_char('V').
roman_char('X').
roman_char('L').
roman_char('C').
roman_char('D').
roman_char('M').

starts_with_roman_number(W, N) :-
	atom_chars(W, Cs),
	swrn(Cs, N).

swrn(Cs, N) :-
	swrn_1(Cs, Cs1),
	Cs1 \= [],
	atom_chars(N, Cs1).

swrn_1(['.'|_], []) :- !.
swrn_1([' '|_], []) :- !.
swrn_1([C|Cs], [C|Cs1]) :-
	roman_char(C),
	!,
	swrn_1(Cs, Cs1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

alnum_prefix(A, A1) :-
	atom_chars(A, Cs),
	map_wp(Cs, Cs1),
	atom_chars(A1, Cs1).

map_wp([C|Cs], [C|Cs1]) :-
	char_type(C, alnum),
	!,
	map_wp(Cs, Cs1).
map_wp([C|Cs], [C|Cs1]) :-
	accented_alnum_char(C),
	!,
	map_wp(Cs, Cs1).
map_wp(_, []).

accented_alnum_char('Á').
accented_alnum_char('Ã').
accented_alnum_char('Ä').
accented_alnum_char('Ç').
accented_alnum_char('È').
accented_alnum_char('É').
accented_alnum_char('Ñ').
accented_alnum_char('Ö').
accented_alnum_char('Ü').
accented_alnum_char('ß').
accented_alnum_char('á').
accented_alnum_char('ã').
accented_alnum_char('ä').
accented_alnum_char('ç').
accented_alnum_char('è').
accented_alnum_char('é').
accented_alnum_char('ñ').
accented_alnum_char('ö').
accented_alnum_char('ü').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

punct_stops_sentence(',') :-
	!,
	fail.
punct_stops_sentence('.') :-
	!.
punct_stops_sentence('?') :-
	!.
punct_stops_sentence('!') :-
	!.
punct_stops_sentence(';') :-
	!.
punct_stops_sentence(':') :-
	!.
punct_stops_sentence(W) :-
	sub_atom(W, _, _, _, '.'),
	!.
punct_stops_sentence(W) :-
	sub_atom(W, _, _, _, '?'),
	!.
punct_stops_sentence(W) :-
	sub_atom(W, _, _, _, '!'),
	!.
punct_stops_sentence(W) :-
	sub_atom(W, _, _, _, ';'),
	!.
punct_stops_sentence(W) :-
	sub_atom(W, _, _, _, ':'),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

split_atom_at_punct_and_space(A, As) :-
	atom_chars(A, Cs),
	sanc(Cs, New, New, As).

is_punct_or_space(C) :-
	char_type(C, punct),
	!.
is_punct_or_space(C) :-
	char_type(C, space).

sanc([C|Cs], Prev, Rest, As) :-
	is_punct_or_space(C),
	!,
	( Prev == Rest ->
	  sanc(Cs, NewPrev, NewPrev, As)
	; Rest = [],
	  atom_chars(A, Prev),
	  As = [A|As1],
	  sanc(Cs, NewPrev, NewPrev, As1)
	).
sanc([C|Cs], Prev, [C|Rest], As) :-
	sanc(Cs, Prev, Rest, As).
sanc([], Prev, Rest, As) :-
	( Prev == Rest ->
	  As = []
	; Rest = [],
	  atom_chars(A, Prev),
	  As = [A]
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

enum_umlaut_variants(A, B) :-
	atom_chars(A, A1),
	enum_umlaut_variants_1(A1, B1),
	atom_chars(B, B1).

enum_umlaut_variants_1([X|Ys], [X|Ys1]) :-
	enum_umlaut_variants_1(Ys, Ys1).
enum_umlaut_variants_1([a,e|Xs],[ä|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1([o,e|Xs],[ö|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1([u,e|Xs],[ü|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1(['A',e|Xs],['Ä'|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1(['O',e|Xs],['Ö'|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1(['U',e|Xs],['Ü'|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1([s,s|Xs],['ß'|Xs1]) :-
	!,
	enum_umlaut_variants_1(Xs, Xs1).
enum_umlaut_variants_1([], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_diacritics_used_in_handwriting(A, B) :-
	atom_chars(A, A1),
	map_rmduih_1(A1, B1),
	atom_chars(B, B1).

map_rmduih_1([X|Xs], [X1|Xs1]) :-
	rmduih_1(X, X1),
	!,
	map_rmduih_1(Xs, Xs1).
map_rmduih_1([], []).

rmduih_1('ū', 'u') :- !.
rmduih_1('Ū', 'U') :- !.
rmduih_1('ÿ', 'y') :- !.
rmduih_1(X, X).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

plain_sortkey(W, W1) :-
	atom_chars(W, Xs),
	map_id_equivalent(Xs, Xs1),
	atom_chars(W1, Xs1).

map_id_equivalent([X|Xs], Xs1) :-
	id_equivalent(X, X1),
	append(X1, Xs2, Xs1),
	map_id_equivalent(Xs, Xs2).
map_id_equivalent([], []).

id_equivalent(C, [C]) :-
	char_type(C, ascii),
	char_type(C, alnum),
	!.
id_equivalent(C, [' ']) :-
	char_type(C, white),
	!.
id_equivalent(',', [',']) :-	!.
id_equivalent('.', ['.']) :-	!.
id_equivalent('Á', ['A']) :-	!.
id_equivalent('Ã', ['A']) :-	!.
id_equivalent('Ä', ['A',e]) :-	!.
id_equivalent('Ç', ['C']) :-	!.
id_equivalent('È', ['E']) :-	!.
id_equivalent('É', ['E']) :-	!.
id_equivalent('Ñ', ['N']) :-	!.
id_equivalent('Ö', ['O',e]) :-	!.
id_equivalent('Ü', ['U',e]) :-	!.
id_equivalent('ß', [s,s]) :-	!.
id_equivalent('á', [a]) :-	!.
id_equivalent('ã', [a]) :-	!.
id_equivalent('ä', [a,e]) :-	!.
id_equivalent('ç', [c]) :-	!.
id_equivalent('è', [e]) :-	!.
id_equivalent('é', [e]) :-	!.
id_equivalent('ñ', [n]) :-	!.
id_equivalent('ö', [o,e]) :-	!.
id_equivalent('ü', [u,e]) :-	!.
id_equivalent('Ē', ['E']) :-	!.
id_equivalent('ē', [e]) :-	!.
id_equivalent('Ō', ['O']) :-	!.
id_equivalent('ō', [o]) :-	!.
id_equivalent(_, [' ']).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

esc_latex(X, Y) :-
	atom_chars(X, Cs),
	esc_latex_1(Cs, Cs1),
	atom_chars(Y, Cs1).

esc_latex_1([X|Xs], ['\\',X|Xs1]) :-
	latex_esc_char(X),
	!,
	esc_latex_1(Xs, Xs1).
esc_latex_1([X|Xs], Xs1) :-
	latex_special_char(X, Xs1, Xs2),
	!,
	esc_latex_1(Xs, Xs2).
esc_latex_1([X,Y|Xs], Xs1) :-
	latex_unicode_postfix(X, Y, Xs1, Xs2),
	!,
	esc_latex_1(Xs, Xs2).
esc_latex_1([X|Xs], [X|Xs1]) :-
	esc_latex_1(Xs, Xs1).
esc_latex_1([], []).
			 
esc_latex_pretty(X, Y) :-
	atom_chars(X, Cs),
	esc_latex_pretty_1(Cs, Cs1),
	atom_chars(Y, Cs1).
		  
esc_latex_pretty_1([X|Xs], ['\\',X|Xs1]) :-
	latex_esc_char(X),
	!,
	esc_latex_pretty_1(Xs, Xs1).
esc_latex_pretty_1([X|Xs], Xs1) :-
	latex_special_char_pretty(X, Xs1, Xs2),
	!,
	esc_latex_pretty_1(Xs, Xs2).
esc_latex_pretty_1([X,Y|Xs], Xs1) :-
	latex_unicode_postfix(X, Y, Xs1, Xs2),
	!,
	esc_latex_pretty_1(Xs, Xs2).
esc_latex_pretty_1([X|Xs], [X|Xs1]) :-
	esc_latex_pretty_1(Xs, Xs1).
esc_latex_pretty_1([], []).

latex_unicode_postfix(C, '̆', ['{', '\\', 'u', '{', C, '}', '}'|X], X).

latex_esc_char('%').
latex_esc_char('&').
latex_esc_char('|').

latex_special_char('\\', ['\\',t,e,x,t,b,a,c,k,s,l,a,s,h|X], X).

latex_special_char_pretty(X, Y, Z) :-
        latex_special_char(X, Y, Z),
        !.			    
latex_special_char_pretty('<', ['{','\\',t,e,x,t,l,a,n,g,l,e,'}'|X], X).
latex_special_char_pretty('>', ['{','\\',t,e,x,t,r,a,n,g,l,e,'}'|X], X).

			 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


			 
