%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(wikinames, [convert_wikinames_to_prolog/0,
		      install_wikinames/0,
		      localized_wiki_url/2]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_base(localization)).
:- use_module(kbset_base(filename)).

:- dynamic yago_literal_triple/3.
:- dynamic wiki_url_to_de/2.

localized_wiki_url(URL, URL1) :-
	user_language(de),
	wiki_url_to_de(URL, URL1),
	!.
localized_wiki_url(URL, URL).

install_wikinames :-
	filename(yago_prolog, F),
	consult(F).

true(_,_,_).

convert_wikinames_to_prolog :-
	retractall(wiki_url_to_de(_, _)),
	filename(yago_multilingual_instances, F),
	load_yago_literal_file(F, true),
	( yago_literal_triple(S, 'rdfs:label', deu(O)),
	  atom_concat('http://en.wikipedia.org/wiki/', S, WE),
	  replace_space_with_underscore(O, O1),
	  atom_concat('http://de.wikipedia.org/wiki/', O1, WO),
	  assert(wiki_url_to_de(WE, WO)),
	  fail
	; true
	),
	filename(yago_prolog, Out),
	atom_concat(Out, '.pl', Out1),
	info(10, 'Writing ~w', [Out1]),
	onto_file((listing(wiki_url_to_de/2)), Out1),
	abolish(wiki_url_to_de/2).
		 
replace_space_with_underscore(W, W1) :-
	atom_chars(W, W2),
	map_rswu(W2, W3),
	atom_chars(W1, W3).

rswu(' ', '_') :-
	!.
rswu(X, X).

map_rswu([X|Xs], [X1|Xs1]) :-
	rswu(X, X1),
	map_rswu(Xs, Xs1).
map_rswu([], []).

load_yago_literal_file(F, TestPred) :-
	flag(cnt, _, 0),
	flag(cnt1, _, 0),
	info(10, 'Reading yago literal source: ~w', [F]),
	( repeat,
	  read_line_to_codes(current_input, Line),
	  flag(cnt1, N1, N1+1),
	  info_progress(20, 1000, N1),
	  ( Line == end_of_file ->
	    !
	  ; ( convert_yago_literal_row(Line, S, P, O) ->
	      true
	    ; atom_codes(A, Line),
	      err('Failed to convert row ~q', [A])
	    ),
	    Call =.. [TestPred, S, P, O],
	    call(Call),
	    assert(yago_literal_triple(S,P,O)),
	    flag(cnt, N, N+1),
	    fail
	  )
	)
	from_file F,
	flag(cnt, N, N),
	info_progress_done,
	info(10, 'Asserted ~w yago triples', [N]).


load_yago_file(F, TestPred) :-
	flag(cnt, _, 0),
	flag(cnt1, _, 0),
	info(10, 'Reading yago literal source: ~w', [F]),
	( repeat,
	  read_line_to_codes(current_input, Line),
	  flag(cnt1, N1, N1+1),
	  info_progress(20, 10000, N1),
	  ( Line == end_of_file ->
	    !
	  ; ( convert_yago_row(Line, S, P, O) ->
	      true
	    ; atom_codes(A, Line),
	      err('Failed to convert row ~q', [A])
	    ),
	    Call =.. [TestPred, S, P, O],
	    call(Call),
	    assert(yago_triple(S,P,O)),
	    flag(cnt, N, N+1),
	    fail
	  )
	)
	from_file F,
	flag(cnt, N, N),
	info_progress_done,
	info(10, 'Asserted ~w yago triples', [N]).

convert_yago_row(Row, S, P, O) :-
 	atom_codes(ARow, Row),
 	atomic_list_concat([_, S1, P1, O1|_], '\t', ARow),
	convert_yago_item(S1, S),
	convert_yago_item(P1, P),
	convert_yago_item(O1, O).

convert_yago_literal_row(Row, S, P, O) :-
 	atom_codes(ARow, Row),
 	atomic_list_concat([_, S1, P1, O1|_], '\t', ARow),
	convert_yago_item(S1, S),
	convert_yago_item(P1, P),
	convert_yago_literal(O1, O).

convert_yago_item(A, A1) :-
	sub_atom(A, 0, 1, _, '<'),
	sub_atom(A, _, 1, 0, '>'),
	!,
	sub_atom(A, 1, _, 1, A1).
convert_yago_item(A, A).


convert_yago_literal(A, O) :-
	sub_atom(A, _, _, 0, 'xsd:date'),
	!,
	sub_atom(A, 1, 4, _, Y),
	sub_atom(A, 6, 2, _, M),
	sub_atom(A, 9, 2, _, D),
	( atom_number(Y, Y1) ->
	  ( atom_number(M, M1), M1 >= 1, M1 =< 12 ->
	    ( atom_number(D, D1), D1 >= 1, D1 =< 31 ->
	      O = ymd(Y1,M1,D1)
	    ; O = ym(Y1,M1)
	    )
	  ; O = y(Y1)
	  )
	; sub_atom(A, 1, _, 11, D1),
	  O = weird_date(D1)
	).
convert_yago_literal(A, g(O)) :-
	sub_atom(A, _, _, 0, '<g>'),
	!,
	sub_atom(A, 1, _, 6, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, m(O)) :-
	sub_atom(A, _, _, 0, '<m>'),
	!,
	sub_atom(A, 1, _, 6, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, s(O)) :-
	sub_atom(A, _, _, 0, '<s>'),
	!,
	sub_atom(A, 1, _, 6, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, sqm(O)) :-
	sub_atom(A, _, _, 0, '<m^2>'),
	!,
	sub_atom(A, 1, _, 8, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, dollar(O)) :-
	sub_atom(A, _, _, 0, '<dollar>'),
	!,
	sub_atom(A, 1, _, 11, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, sqkm(O)) :-
	sub_atom(A, _, _, 0, '</km^2>'), %% sic yago June 2014
	!,
	sub_atom(A, 1, _, 10, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, deg(O)) :-
	sub_atom(A, _, _, 0, '<degrees>'),
	!,
	sub_atom(A, 1, _, 12, A1),
	( atom_number(A1, O) ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, isbn(O)) :-
	sub_atom(A, _, _, 0, '<yagoISBN>'),
	!,
	sub_atom(A, 1, _, 13, O).
convert_yago_literal(A, n(O)) :-
	sub_atom(A, _, _, 0, 'xsd:nonNegativeInteger'),
	!,
	sub_atom(A, 1, _, 25, A1),
	( atom_number(A1, O), O >= 0 ->
	  true
	; err('Failed to convert value ~q', [A])
	).
convert_yago_literal(A, eng(O)) :-
	sub_atom(A, _, _, 0, '@eng'),
	!,
	sub_atom(A, 1, _, 5, O).
convert_yago_literal(A, deu(O)) :-
	sub_atom(A, _, _, 0, '@deu'),
	!,
	sub_atom(A, 1, _, 5, O).
convert_yago_literal(A, O) :-
	\+ sub_atom(A, 0, _, _, '^^'),
	!,
	sub_atom(A, 1, _, 1, O).
convert_yago_literal(A, _) :-
	err('Failed to convert literal: ~q', [A]).

