%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% Operates on results of:
%%%% $ biber --tool --output-format=biblatexml BIBFILE.bib
%%%% 

:- module(bibconvert,
	  [load_biber_xml_files/1,
	   generate_biber_xml_files/1,
	   print_biber_xml_conversion_script/1,
	   generate_bibrefdata/0,
	   html_bibentries/2]).

:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(kbmappings)).
:- use_module(kbset_letters(nlx)).
:- use_module(kbset_letters(textconvert)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(readwrite)).
:- use_module(library(url)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(onto_atom)).
:- use_module(swilib(fromonto)).

:- flag(bibconvert_warnings, _, 0).

:- dynamic gen_referenced_biblabel/1.

generate_bibrefdata :-
	filename( gendata('generated_bibrefdata.pl'), Outfile),
	findall(k, enum_bibid(Id), Ks),
	length(Ks, Len),
	info(10, 'Number of bibentries: ~w', [Len]),
	info(10, 'Writing ~w', [Outfile]),
	info_progress_start,
	onto_file(( enum_bibid(Id),
		    entry_refdata(Id,A,AS,Y,S),
		    writeq(bibrefdata(Id,A,AS,Y,S)),
		    writeln('.'),
		    info_progress(10, 50),
		    fail
		  ; true
		  ),
		  Outfile).


%%%% 
%%%% Note: It seems that the --configfile option (needed to preserve the
%%%% keywords attributes) is in CYGWIN only considered properly if the
%%%% configfile is in the "current" directory with the bib files. Thus we copy
%%%% the configuration file to that directory.
%%%% 
print_biber_xml_conversion_script(Fs) :-
	filename( bib(''), BibDir ),
	absolute_file_name( kbset_letters('extras_biber/biber-tool-kbset.conf'),
			    CFG ),
	format('if cd ~w; then ', [BibDir]),
	format('cp ~w generated-biber-tool-kbset.conf; ', [CFG]),
	( member(F, Fs),
	  atom_concat(F, '.bib', F1),
	  format('biber --tool --configfile=generated-biber-tool-kbset.conf --output-format=biblatexml ~w; ',
		 [F1]),
	  fail
	; format('fi')
	).

%%%% 
%%%% This may not work on Windows, hence the way via
%%%% print_biber_xml_conversion_script/1
%%%%
generate_biber_xml_files(Fs) :-
	filename( bib(''), BibDir ),
	absolute_file_name( kbset_letters('extras_biber/biber-tool-kbset.conf'),
			    CFG ),
	( member(F, Fs),
	  atom_concat(F, '.bib', F1),
	  info(10, 'Processing bibfile ~w n ~w', [F1, BibDir]),
	  format(atom(Cmd),
		 'if cd ~w; then biber --tool --configfile=~w --output-format=biblatexml ~w; fi',
		 [BibDir, CFG, F1]),
	  info(10, 'Calling: ~w', [Cmd]),
	  shell(Cmd, Result),
	  ( Result \= 0 ->
	    err('Shell invocation failed: ~w', [Cmd])
	  ; true
	  ),
	  fail
	; true
	).

:- dynamic bibsrc/1.

load_biber_xml_files(Fs) :-
	retractall( bibsrc(_) ),
	( member(F, Fs),
	  filename(bib(F), F1),
	  atom_concat(F1, '_bibertool.bltxml', F2),
	  ( exists_file(F2) ->
	    info(10, 'Loading biber XML file ~w', [F2]),
	    load_xml(F2, XML, [space(remove)]),
	    asserta( bibsrc(XML) )
	  ; err('Missing file ~w. Please generate it with biber --tool --output-format=biblatexml <bibfile>', [F2])
	  ),
	  fail
	; true
	).

enum_bibid(Id) :-
	bibsrc(Es),
	enum_elem('bltx:entry', id=Id, Es, _).

enum_arg(Tag, Att, Es, E) :-
	enum_elem(Tag, Att, Es, E1),
	E1 = element(_, _, E).

enum_elem(Tag, Att, Es, E) :-
	member( E1, Es ),
	E1 = element(Tag1, Att1, Es1),
	( Tag1 = Tag, (var(Att) ; memberchk(Att, Att1)) ->
	  E = E1
	; enum_elem(Tag, Att, Es1, E)
	).

entry_year(Id, Year) :-
	bibsrc(Es),
	enum_arg('bltx:entry', id=Id, Es, E),
	!,
	( enum_arg('bltx:date', _, E, [Year]) ->
	  true
	; enum_arg('bltx:crossref', _, E, [Crossref|_]) ->
	  entry_year(Crossref, Year)
	; Year = ''
	).
	
entry_shorttitle(Id, Title) :-
	bibsrc(Es),
	enum_arg('bltx:entry', id=Id, Es, E),
	!,
	( enum_arg('bltx:shorttitle', _, E, [Title]) ->
	  true
	; enum_arg('bltx:title', _, E, [Title]) ->
	  true
	; Title = ''
	).

entry_authors_familynames(Id, Authors) :-
	bibsrc(Es),
	enum_arg('bltx:entry', id=Id, Es, E),
	!,
	( enum_arg('bltx:names', type=author, E, Names) ->
	  Suffix = ''
	; enum_arg('bltx:names', type=editor, E, Names) ->
	  Suffix = ' (Hrsg.)'
	; Names = [],
	  Suffix = ''
	),
	findall(Name,
		enum_arg('bltx:namepart', type=family, Names, [Name]),
		Names1),
	combine_authors(Names1, Authors1),
	atom_concat(Authors1, Suffix, Authors).

entry_authors_shortnames(Id, Authors) :-
	bibsrc(Es),
	enum_arg('bltx:entry', id=Id, Es, E),
	!,
	( enum_arg('bltx:names', type=author, E, Names) ->
	  true
	; enum_arg('bltx:names', type=editor, E, Names) ->
	  true
	; Names = []
	),
	findall(Name,
		enum_arg('bltx:namepart', type=family, Names, [Name]),
		Names1),
	combine_authors(Names1, Authors).


combine_authors(Names1, Authors1) :-
	( Names1 = [Authors1] ->
	  true
	; Names1 = [A1, A2] ->
	  atomic_list_concat([A1, ' und ', A2], Authors1)
	; Names1 = [A1, A2, A3] ->
	  atomic_list_concat([A1, ', ', A2, ' und ', A3], Authors1)
	; Names1 = [A1|_] ->
	  atomic_list_concat([A1, ' et al.'], Authors1)
	; Authors1 = 'Ohne Autor'  
	).

% map_family_name([X|Xs], [X1|Xs1]) :-
% 	family_name(X, X1),
% 	map_family_name(Xs, Xs1).
% map_family_name([], []).

entry_refdata(Id, Authors, AuthorsShort, Year, Shorttitle) :-
	entry_authors_familynames(Id, Authors),
	entry_authors_shortnames(Id, AuthorsShort),
	entry_year(Id, Year),
	entry_shorttitle(Id, Shorttitle1),
	bibtex_to_h(Shorttitle1, Shorttitle).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% CONVERTING ENTRIES
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic current_id/1.

entry_kvs(Id, KVs) :-
	findall(K-V, entry_kv(Id, K, V), KVs).

entry_kv(Id, id, Id).
entry_kv(Id, entrytype, V) :-
	bibsrc(Es),
	once(enum_elem('bltx:entry', id=Id, Es,
		       element(_, Atts, _))),
	memberchk(entrytype=V, Atts).
entry_kv(Id, K, V) :-
	bibsrc(Es),
	enum_arg('bltx:entry', id=Id, Es, E),
	!,
	retractall( current_id(_) ),
	asserta( current_id(Id) ),
	( member(E1, E),
	  ( item_elem(E1, K, V) ->
	    true
	  ; item_simple(E1, K, V) ->
	    true
	  ; item_range(E1, K, V) ->
	    true
	  ; item_names(E1, K, V) ->
	    true
	  ; flag(bibconvert_warnings, W, W+1),
	    info(10, 'WARNING: Entry ~w: Unsupported  type: ~q', [Id, E1]) ->
	    fail
	  ; err('Unsupported entry type: ~q ~q', [Id, E1])
	  )
	; fail
	).

item_elem(E, A, V) :-
	E = element(A, _,
		    [element('bltx:list', _,
			     [element('bltx:item', _, [V])])]),
	atom(V).

item_elem(E, A, val_list(Atoms)) :-
	E = element(A, _,
		    [element('bltx:list', _, Items)]),
	items_list(Items, Atoms).

items_list([element('bltx:item', _, [V])|L], [V|Vs]) :-
	atom(V),
	items_list(L, Vs).
items_list([], []).
	
item_range(E, A, val_range(V1,V2)) :-
	E = element(A, _,
		    [element('bltx:list',_,
			     [element('bltx:item',_,
				      [element('bltx:start',_,[V1]),
				       element('bltx:end',_,[V2])])])]),
	memberchk(A, ['bltx:pages']),
	atom(V1),
	atom(V2).
item_range(E, A, val_ranges(Pairs)) :-
	E = element(A, _,
		    [element('bltx:list',_, StartEndPairs)]),
	memberchk(A, ['bltx:pages']),
	StartEndPairs = [_|_],
	start_end_pairs(StartEndPairs, Pairs).
item_range(E, A, val_range(V1,V2)) :-
	E = element(A, _,
		    [element('bltx:start',_,[V1]),
		     element('bltx:end',_,[V2])]),
	memberchk(A, ['bltx:date']),
	atom(V1),
	atom(V2).
item_range(E, A, val_range(V1,'')) :-
	E = element(A, _,
		    [element('bltx:start',_,[V1]),
		     element('bltx:end',_,[])]),
	memberchk(A, ['bltx:date']),
	atom(V1).

start_end_pairs([element('bltx:item',_,
			 [element('bltx:start',_,[V1]),
			  element('bltx:end',_,[V2])])| SEPairs],
		[V1-V2|Pairs]) :-
	atom(V1),
	atom(V2),
	!,
	start_end_pairs(SEPairs, Pairs).
start_end_pairs([], []).

item_simple(E, A, V) :-
	E = element(A, _, [V]),
	atom(V).

item_names(E, A, val_names(V)) :-
	E = element('bltx:names', Atts, Es),
	memberchk(type=A, Atts),
	map_convert_name(Es, V).

convert_name(X, val_name(N1,N2,N3)) :-
	X = element('bltx:name', _, E),
	( E = [element('bltx:namepart', Atts1, M1),
	       element('bltx:namepart', Atts2, M2),
	       element('bltx:namepart', Atts3, M3)],
	  memberchk(type=prefix, Atts1),
	  memberchk(type=family, Atts2),
	  memberchk(type=given, Atts3)
	; E = [element('bltx:namepart', Atts1, M1),
	       element('bltx:namepart', Atts2, M2)],
	  memberchk(type=prefix, Atts1),
	  memberchk(type=family, Atts2),
	  M3 = []
	; E = [element('bltx:namepart', Atts2, M2),
	       element('bltx:namepart', Atts3, M3)],
	  memberchk(type=family, Atts2),
	  memberchk(type=given, Atts3),
	  M1 = []
	; E = [element('bltx:namepart', Atts2, M2)],
	  memberchk(type=family, Atts2),
	  M1 = [], M3 = []
	),
	convert_namepart(M1,N1),
	convert_namepart(M2,N2),
	convert_namepart(M3,N3),
	!.
convert_name(X, val_name([],M2,[])) :-
	X = element('bltx:name', _, E),
	member(element('bltx:namepart', Atts2, M2), E),
	memberchk(type=family, Atts2),
	!,
	convert_namepart(M2,N2),
	( current_id(Id) -> true ; Id = none ),
	flag(bibconvert_warnings, W, W+1),
	info(10, 'WARNING: Entry ~w: Improper name conversion ~q for: ~q', [Id, N2, X]).
convert_name(X, Id, val_name([],[],[])) :-
	( current_id(Id) -> true ; Id = none ),
	flag(bibconvert_warnings, W, W+1),
	info(10, 'WARNING: Entry ~w: Failed name conversion for: ~q', [Id, X]).

convert_namepart([], []) :-
	!.
convert_namepart([N], [N]) :-
	atom(N),
	!.
convert_namepart(E, N) :-
	map_convert_inner_namepart(E, N).

convert_inner_namepart(E, N) :-
	E = element('bltx:namepart', _Atts, [N1]),
	atom(N1),
	!,
	N = N1.
%  	( memberchk(initial=I, _Atts) -> N = name_initial(N1, I)
% 	; N = N1
%  	).
convert_inner_namepart(E, xxx(E)).


map_convert_inner_namepart([X|Xs], [X1|Xs1]) :-
	convert_inner_namepart(X, X1),
	map_convert_inner_namepart(Xs, Xs1).
map_convert_inner_namepart([], []).

map_convert_name([X|Xs], [X1|Xs1]) :-
	convert_name(X, X1),
	map_convert_name(Xs, Xs1).
map_convert_name([], []).

map_convert_namepart([X|Xs], [X1|Xs1]) :-
	convert_namepart(X, X1),
	map_convert_namepart(Xs, Xs1).
map_convert_namepart([], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% SORTING ENTRIES
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

entries(Es) :-
	info_progress_start,
	findall(K-E,
		( enum_bibid(X),
		  entry_kvs(X, E),
		  kve_sortkey(E, K),
		  info_progress(10, 50)
		),
		KEs),
	keysort(KEs, KEs1),
	map_val(KEs1, Es).

kve_sortkey(KVE, sortkey(T,A,D)) :-
	memberchk(entrytype-periodical, KVE),
	!,
	kve_titlekey(KVE, T),
	kve_authorkey(KVE, A),
	kve_datekey(KVE, D).
kve_sortkey(KVE, sortkey(A,D,T)) :-
	kve_authorkey(KVE, A),
	kve_datekey(KVE, D),
	kve_titlekey(KVE, T).

kve_datekey(KVE, D) :-
	memberchk('bltx:date'-D1, KVE),
	!,
	bibconvert_date_sortkey(D1, D).
kve_datekey(_, '').

bibconvert_date_sortkey(val_range(D,_), D) :-
	!.
bibconvert_date_sortkey(D, D).

kve_titlekey(KVE, K) :-
	( memberchk('bltx:title'-K, KVE) ->
	  true
	; memberchk('bltx:booktitle'-K, KVE) ->
	  true
	; K=''
	).

kve_authorkey(KVE, K) :-
	( memberchk(author-val_names(N), KVE) ->
	  true
	; memberchk(editor-val_names(N), KVE) ->
	  true
	; N = []
	),
	map_name_to_key(N, K).

name_to_key(val_name(N1,N2,N3), k(M2,M1,M3)) :-
	!,
	map_atom_sortkey(N1, M1),
	map_atom_sortkey(N2, M2),
	map_atom_sortkey(N3, M3).
name_to_key(N, k([M],'','')) :-
	atom(N),
	!,
	atom_sortkey(N, M).
name_to_key(N, k(M,'','')) :-
	map_atom_sortkey(N, M).

map_name_to_key([X|Xs], [X1|Xs1]) :-
	name_to_key(X, X1),
	map_name_to_key(Xs, Xs1).
map_name_to_key([], []).

map_atom_sortkey([X|Xs], [X1|Xs1]) :-
	atom_sortkey(X, X1),
	map_atom_sortkey(Xs, Xs1).
map_atom_sortkey([], []).

atom_sortkey(A, A1) :-
	atom(A),
	!,
	atom_chars(A, A2),
	map_c_sortkey(A2, A3),
	atom_chars(A1, A3).
atom_sortkey(X, X) :-
	info(10, 'Expected an atom, found ~q', [X]).

map_c_sortkey([X|Xs], Xs1) :-
	( c_sortkey(X, X1) ->
	  true
	; X1 = [X]
	),
	append(X1, Xs2, Xs1),
	map_c_sortkey(Xs, Xs2).
map_c_sortkey([], []).

c_sortkey('[', []).
c_sortkey(']', []).
c_sortkey('Ä', ['A',e]).
c_sortkey('Ö', ['O',e]).
c_sortkey('Ü', ['U',e]).
c_sortkey('ä', [a,e]).
c_sortkey('ö', [o,e]).
c_sortkey('ü', [u,e]).
c_sortkey('ß', [s,s]).
c_sortkey('é', [e]).
c_sortkey('É', ['E']).
c_sortkey('ç', [c]).
c_sortkey('-', [' ']).

map_val([_-X|Xs], [X|Xs1]) :-
	map_val(Xs, Xs1).
map_val([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

html_bibentries(Options, H) :-
	( memberchk(bibentries=all, Options) ->
	  Mode = all
	; Mode = cited,
	  filename( gendata('generated_referenced_biblabels'), CitedFile ),
	  atom_concat(CitedFile, '.pl', CitedFilePL),
	  (exists_file(CitedFilePL) ->
	   info(10, 'Consulting ~w', [CitedFile]),
	   consult(CitedFile)
	  ; info(10, 'WARNING: Working without generated referenced biblabels file')	  )
	),
	flag(bibconvert_warnings, _, 0),
	refresh_access_data,
	info(10, 'Computing entries'),
	entries(Es0),
	length(Es0, LE0),
	info(10, 'Total number of entries: ~w', [LE0]),
	( Mode = cited ->
	  %% There is no automatic preservation of entries that are CROSS
	  %% REFERENCED in cited entries. They may be added manually with
	  %% xnocite.
	  findall(E, (member(E, Es0),
		      memberchk(id-L, E),
		      gen_referenced_biblabel(L)),
		  Es),
	  length(Es, LE),
	  info(10, 'Number of cited entries: ~w', [LE])
	; Es = Es0
	),
	flag(bibconvert_warnings, W, W),
	info(10, 'Number of bibconvert warnings: ~w', [W]),
	info(10, 'Generating page'),
	( memberchk(bibsections=Sections, Options), Sections \= [] ->
	  H = [Toc|H1],
	  map_bibsection_toc(Sections, Toc),
	  map_bibsection(Sections, Es, H1)
	; findall(H1, (member(E, Es), entry_to_html(E, H1)), H)
	).

map_bibsection_toc(Sections, E) :-
	E = e(ul, [], LIs),
	findall(LI,
		( member(sect(Id, Title, _), Sections),
		  atom_concat('#', Id, Anchor),
		  LI = e(li, [], [e(a, [href=Anchor], [Title])])
		),
		LIs).

map_bibsection([Section|Sections], Es, H) :-
	( Section = sect(Id, Title, Condition) -> true
	; err('Bad bibsection specifier: ~q', [Section])
	),
	info(10, 'Bibsection ~q: Processing', [Title]),
	H = [e(h2, [id=Id], [Title]) | H1],
	findall(H2, (member(E, Es),
		     meets_bibsection_condition(Condition, E),
		     entry_to_html(E, H2)),
		H3),
	length(H3, L3),
	info(10, 'Bibsection ~q: Number of entries: ~w', [Title, L3]),
	append(H3, H4, H1),
	map_bibsection(Sections, Es, H4).
map_bibsection([], _, []).

has_keyword(E, K) :-
	memberchk('bltx:keywords'-Kwds, E),
	atom(Kwds),
	atomic_list_concat(Kwds1, ',', Kwds),
	memberchk(K, Kwds1).

meets_bibsection_condition(C, E) :-
	member(+K, C),
	has_keyword(E, K),
	!,
	\+ ( member(-K1, C),
	     has_keyword(K1, E) ).

% meets_bibsection_condition([+K|C], E) :-
% 	!,
% 	has_keyword(E, K),
% 	meets_bibsection_condition(C, E).
% meets_bibsection_condition([-K|C], E) :-
% 	\+ has_keyword(E, K),
% 	!,
% 	meets_bibsection_condition(C, E).
% meets_bibsection_condition([], _).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lookup_work(E, W) :-
	findall(W1, lookup_work_enum(E, W1), Ws),
	sort(Ws, Ws1),
	( Ws1 = [] ->
	  fail
	; Ws1 = [W] ->
	  true
	; ( memberchk('bltx:shorttitle'-ST, E),
	    atom(ST),
	    downcase_atom(ST, ST1),
	    ( member(W, Ws1),
	      work_title(W, WTitle),
	      downcase_atom(WTitle, WTitle1),
	      ST1 = WTitle1 -> true
	    ) -> true
	  ; memberchk('bltx:title'-T, E),
	    atom(T),
	    downcase_atom(T, T1),
	    member(W, Ws1),
	    work_title(W, WTitle),
	    downcase_atom(WTitle, WTitle1),
	    sub_atom(T1, 0, _, _, WTitle1) -> true
	  ) -> true
	; Ws1 = [W|_],
	  memberchk(id-Id, E),
	  memberchk('bltx:title'-T, E),
	  findall(T2, (member(W2, Ws1), work_title(W2, T2)), Ts1),
	  sort(Ts1, Ts2),
	  ( Ts2 = [T2] ->
	    info(10, 'WARNING: Bibref ~q (~q) matches multiple works ~q with same title: ~q',
		 [Id, T, Ws1, T2])
	  ;  info(10, 'WARNING: Bibref ~q (~q) matches multiple works ~q with different titles: ~q',
		  [Id, T, Ws1, Ts1])
	  )
	).

lookup_work_enum(E, W) :-
	memberchk('bltx:title'-T, E),
	atom(T),
	lookup_author_enum(E, A),
	enum_work(W),
	work_author(W, A),
	once(work_title(W, WTitle)),
	txt_match_included(WTitle, T).

	     
lookup_author_enum(E, A) :-
	( memberchk(author-val_names(Ns), E) -> true
	; memberchk(editor-val_names(Ns), E)
	),
	member(val_name(_, LN, _FN), Ns),
	atomic_list_concat(LN, ' ', LN1),
	enum_xperson(A),
	xperson_prettyname(A, N1),
	sub_atom(N1, 0, _, _, LN1).


txt_match_included(A1, A2) :-
	atom_to_words(A1, Ws1),
	atom_to_words(A2, Ws2),
	list_match_included(Ws1, Ws2),
	!.

list_match_included([], _) :- !.
list_match_included([X|Xs], [X|Xs1]) :-
	list_match_included(Xs, Xs1).
list_match_included(Xs, [_|Xs1]) :-
	list_match_included(Xs, Xs1).


atom_to_words(A, Ws) :-
	downcase_atom(A, A1),
	atom_chars(A1, Cs),
	aw1(Cs, Ws).

aw([C|Cs], W, Ws) :-
	char_type(C, alnum),
	!,
	aw(Cs, [C|W], Ws).
aw([_|Cs], W, [W1|Ws]) :-
	reverse(W, W2),
	atom_chars(W1, W2),
	aw1(Cs, Ws).
aw([], W, [W1]) :-
	reverse(W, W2),
	atom_chars(W1, W2).

aw1([C|Cs], Ws) :-
	char_type(C, alnum),
	!,
	aw([C|Cs], [], Ws).
aw1([_|Cs], Ws) :-
	aw1(Cs, Ws).
aw1([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

entry_to_html(E, H) :-
	H = e(div, [class=bibentry, id=Name], H0),
	H0 = [e(a, [class=anchor, href=Anchor],
		[e(span, [class=bibauthor], Aut)]) | H1],
	memberchk(id-Id, E),
	id_anchor(bib, Id, Name),
	atom_concat('#', Name, Anchor),
	memberchk(entrytype-Type, E),
	( e_to_h(Type, E, Aut, H1) ->
	  true
	; err('Unsupported bibentry type: ~q ~q', [Id, Type])
	).
 
e_to_h(Type, E, Aut, H) :-
	memberchk(Type, [book, mvbook, collection, proceedings,
			 unpublished, misc]),
	!,
 	( memberchk(author-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v1(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Aut = [Aut2, ': '],
	  ( memberchk(editor-val_names(Eds), E) ->
	    onto_atom(fmt_names_v2(Eds), Eds1),
	    Ed = ['Hg. von ', Eds1, '. ']
	  ; true ->
	    Ed = []
	  )
	; memberchk(editor-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v1(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Aut = [Aut2, ' (Hg.): '],
	  Ed = []
	; Aut = [], Ed = []
	),
	( ( memberchk('bltx:title'-T, E)
	  ; memberchk('bltx:booktitle'-T, E) ) ->
	  bibtex_to_h(T, T1),
	  add_fullstop_blank(T1, T2),
	  H1 = [e(span, [class=bibtitle], T2) | H2]
	; H1 = H2
	),
	( memberchk('bltx:volume'-V, E) ->
	  H2 = [e(span, [class=bibvolume], ['Bd.\u00A0', V, '. ']) | H3]
	; H2 = H3
	),
	( memberchk('bltx:volumes'-W, E) ->
	  H3 = [e(span, [class=bibvolumes], [W, '\u00A0Bde. ']) | H4]
	; H3 = H4
	),
	( Ed = [] ->
	  H4 = H5
	; H4 = [e(span, [class=bibedited], Ed) | H5]
	),
	findall(X,
		( once(( memberchk('bltx:location'-Y, E)
		       ; memberchk('bltx:address'-Y, E) )),
		  fmted_location(Y, X)
		; memberchk('bltx:publisher'-Y, E),
		  atom_or_member(X, Y)
		; memberchk('bltx:date'-Y, E),
		  fmted_date(Y, X)
		),
		Xs),
	atomic_list_concat(Xs, ', ', Xs1),
	( Xs1 = '' ->
	  H5 = H6
	; bibtex_to_h(Xs1, Xs2),
	  add_fullstop(Xs2, Xs3),
	  H5 = [e(span, [class=bibdata], Xs3) | H6]
	),
	e_to_h_after(Type, E, H6),
	H = H1.


e_to_h(Type, E, _, H) :-
	memberchk(Type, [periodical]),
	!,
	( memberchk(editor-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v2(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Ed = ['Hg. von ', Aut2, '. ']
	; Ed = []
	),
	( memberchk('bltx:title'-T, E) ->
	  bibtex_to_h(T, T1),
	  add_fullstop_blank(T1, T2),
	  H1 = [e(span, [class=bibtitle], T2) | H2]
	; H1 = H2
	),		
	( Ed = [] ->
	  H2 = H3
	; H2 = [e(span, [class=bibedited], Ed) | H3]
	),
	( memberchk('bltx:volume'-V, E) ->
	  H3 = [e(span, [class=bibvolume], ['Bd.\u00A0', V, '. ']) | H4]
	; H3 = H4
	),
	( memberchk('bltx:volumes'-W, E) ->
	  H4 = [e(span, [class=bibvolumes], [W, '\u00A0Bde. ']) | H5]
	; H4 = H5
	),
	findall(X,
		( once(( memberchk('bltx:location'-Y, E)
		       ; memberchk('bltx:address'-Y, E) )),
		  fmted_location(Y, X)
		; memberchk('bltx:publisher'-Y, E),
		  atom_or_member(X, Y)
		; memberchk('bltx:date'-Y, E),
		  fmted_date(Y, X)
		),
		Xs),
	atomic_list_concat(Xs, ', ', Xs1),
	( Xs1 = '' ->
	  H5 = H6
	; bibtex_to_h(Xs1, Xs2),
	  add_fullstop(Xs2, Xs3),
	  H5 = [e(span, [class=bibdata], Xs3) | H6]
	),
	e_to_h_after(Type, E, H6),
	H = H1.

e_to_h(Type, E, Aut, H) :-
	memberchk(Type, [article, injournal]),
	!,
 	( memberchk(author-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v1(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Aut = [Aut2, ': ']
	; Aut = []
	),
	( memberchk('bltx:title'-T, E) ->
	  bibtex_to_h(T, T1),
	  H1 = [e(span, [class=bibtitle], T1), '. ' | H2]
	; H1 = H2
	),
	( memberchk('bltx:journaltitle'-J, E) ->
	  bibtex_to_h(J, J1),
	  H2 = ['In: ', e(span, [class=bibjournal], J1), '.' | H3]
	; H2 = H3
	),
	findall(X,
		( memberchk('bltx:volume'-Y, E),
		  format(atom(X), 'Bd.\u00A0~w', [Y])
		; memberchk('bltx:issue'-X, E)
		; memberchk('bltx:number'-Y, E),
		  format(atom(X), '(~w)', [Y])
		; memberchk('bltx:date'-Y, E),
		  fmted_date(Y, Y1),
		  format(atom(X), '(~w)', [Y1])
		; memberchk('bltx:pages'-Y, E),
		  fmted_pages(Y, X)
		),
		Xs),
	atomic_list_concat(Xs, ', ', Xs1),
	( Xs1 = '' ->
	  H3 = H4
	; bibtex_to_h(Xs1, Xs2),
	  add_fullstop(Xs2, Xs3),
	  H3 = [e(span, [class=bibdata], [' '| Xs3]) | H4]
	),
	e_to_h_after(Type, E, H4),
	H = H1.

e_to_h(Type, E, Aut, H) :-
	memberchk(Type, [review]),
	!,
 	( memberchk(author-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v1(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Aut = [Aut2, ': ']
	; Aut = []
	),
	( memberchk('bltx:title'-T, E) ->
	  bibtex_to_h(T, T1),
	  H1 = [e(span, [class=bibtitle], T1), '. ' | H2]
	; H1 = H2
	),
	( memberchk('bltx:xref'-C, E) ->
	  bib_reference(st, '', C, XRef),
	  H2 = ['Rez. zu ', XRef, '. ' | H2a]
	; H2 = H2a
	),
	( memberchk('bltx:journaltitle'-J, E) ->
	  bibtex_to_h(J, J1),
	  H2a = ['In: ', e(span, [class=bibjournal], J1), '.' | H3]
	; H2a = H3
	),
	findall(X,
		( memberchk('bltx:volume'-Y, E),
		  format(atom(X), 'Bd.\u00A0~w', [Y])
		; memberchk('bltx:issue'-X, E)
		; memberchk('bltx:number'-Y, E),
		  format(atom(X), '(~w)', [Y])
		; memberchk('bltx:date'-Y, E),
		  fmted_date(Y, Y1),
		  format(atom(X), '(~w)', [Y1])
		; memberchk('bltx:pages'-Y, E),
		  fmted_pages(Y, X)
		),
		Xs),
	atomic_list_concat(Xs, ', ', Xs1),
	( Xs1 = '' ->
	  H3 = H4
	; bibtex_to_h(Xs1, Xs2),
	  add_fullstop(Xs2, Xs3),
	  H3 = [e(span, [class=bibdata], [' '| Xs3]) | H4]
	),
	e_to_h_after(Type, E, H4),
	H = H1.

e_to_h(Type, E, Aut, H) :-
	memberchk(Type, [incollection, inbook, inproceedings]),
	!,
 	( memberchk(author-val_names(Auts), E) ->
 	  onto_atom(fmt_names_v1(Auts), Aut1),
	  wrap_if_pseudoauthor(Aut1, E, Aut2),
	  Aut = [Aut2, ': ']
	; Aut = []
	),
	( memberchk('bltx:title'-T, E) ->
	  bibtex_to_h(T, T1),
	  H1 = [e(span, [class=bibtitle], T1), '. ' | H2a]
	; H1 = H2a
	),
	( memberchk('bltx:crossref'-C, E) ->
	  bib_reference(st, '', C, CrossRef),
	  H2a = ['In: ', CrossRef, '. ' | H4a]
	; ( memberchk(editor-val_names(Eds), E) ->
	    onto_atom(fmt_names_v2(Eds), Eds1),
	    H2a = ['In: ', e(span, [class=bibedited], [Eds1]), ' (Hg.): ' | H3a]
	  ; true ->
	    H2a = ['In: ' | H3a]
	  ),
	  ( memberchk('bltx:booktitle'-B, E) ->
	    bibtex_to_h(B, B1),
	    append(B1, ['. '], B2),
	    H3a = [e(span, [class=bibtitle], B2) | H4a]
	  ; H3a = H4a
	  )
	),
	( memberchk('bltx:volume'-V, E) ->
	  H4a = [e(span, [class=bibvolume], ['Bd.\u00A0', V, '. ']) | H3]
	; H4a = H3
	),
	( memberchk('bltx:volumes'-W, E) ->
	  H3 = [e(span, [class=bibvolumes], [W, '\u00A0Bde. ']) | H4]
	; H3 = H4
	),
	H5 = H4,
	findall(X,
		( once(( memberchk('bltx:location'-Y, E)
		       ; memberchk('bltx:address'-Y, E) )),
		  fmted_location(Y, X)
		; memberchk('bltx:publisher'-Y, E),
		  atom_or_member(X, Y)
		; memberchk('bltx:date'-Y, E),
		  fmted_date(Y, X)
		; memberchk('bltx:pages'-Y, E),
		  fmted_pages(Y, X)
		),
		Xs),
	atomic_list_concat(Xs, ', ', Xs1),
	( Xs1 = '' ->
	  H5 = H6
	; bibtex_to_h(Xs1, Xs2),
	  add_fullstop(Xs2, Xs3),
	  H5 = [e(span, [class=bibdata], Xs3) | H6]
	),
	e_to_h_after(Type, E, H6),
	H = H1.

e_to_h_after(_, E, H) :-
	( memberchk('bltx:library'-L, E) ->
	  bibtex_to_h(L, L1),
	  add_fullstop(L1, L2),
	  H0 = [' ', e(span, [class=biblibrary], L2) | H1]
	; H0 = H1
	),
	( memberchk('bltx:note'-N, E) ->
	  bibtex_to_h(N, N1),
	  add_fullstop(N1, N2),
	  H1 = [' ', e(span, [class=bibnote], N2) | H2]
	; H1 = H2
	),
	( memberchk('bltx:url'-U, E) ->
	  url_iri(U1, U),
	  H2 = [' ', e(a, [href=U1, class=biburl], ['[Online]']) | H3]
	; H2 = H3
	),
	( lookup_work(E, W) ->
	  id_url(work, W, UW),
	  H3 = [' ', e(a, [href=UW, class=biburl], ['[Verweise]']) | H4]
	; H3 = H4
	),
	H4 = [],
	H = H0.

add_fullstop(H, H1) :-
	( append(_, [Last], H),
	  atom(Last),
	  sub_atom(Last, _, _, 0, '.') ->
	  H1 = H
	; append(H, ['.'], H1)
	).

add_fullstop_blank(H, H1) :-
	( append(_, [Last], H),
	  atom(Last),
	  sub_atom(Last, _, _, 0, '.') ->
	  append(H, [' '], H1)
	; append(H, ['. '], H1)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fmted_location(X, X) :-
	atom(X),
	!.
fmted_location(val_list(Xs), X) :-
	!,
	atomic_list_concat(Xs, ', ', X).
fmted_location(X, _) :-
	err('Unsupported location format: ~q', [X]).

atom_or_member(X, X) :-
	!,
	atom(X).
atom_or_member(X, Y) :-
	member(X, Y).

fmted_date(val_range(A,B), D) :-
	!,
	format(atom(D), '~w–~w', [A, B]).
fmted_date(A, D) :-
	format(atom(D), '~w', [A]).

fmted_pages(val_ranges([A-B|Xs]), D) :-
	!,
	onto_atom(( format('S.\u00A0~w–~w', [A, B]),
		    ( member(A1-B1, Xs),
		      format(', ~w–~w', [A1, B1]),
		      fail
		    ; true
		    )),
		  D ).
fmted_pages(val_range(A,B), D) :-
	!,
	format(atom(D), 'S.\u00A0~w–~w', [A, B]).
fmted_pages(A, D) :-
	format(atom(D), 'S.\u00A0~w', [A]).

wrap_if_pseudoauthor(N, E, N1) :-
	( memberchk('bltx:options'-Options, E),
	  atom(Options),
	  sub_atom(Options, _, _, _, 'pseudoauthor') ->
	  format(atom(N1), '[~w]', [N])
	; N1 = N
	).

fmt_names_v1([N|Ns]) :-
	fmt_author_v1(N),
	( member(N1, Ns),
	  format(' und '),
	  fmt_author_v2(N1),
	  fail
	; true
	).

fmt_names_v2([N|Ns]) :-
	fmt_author_v2(N),
	( member(N1, Ns),
	  format(' und '),
	  fmt_author_v2(N1),
	  fail
	; true
	).

fmt_author_v1(val_name(Prefix, Family, Given)) :-
	!,
	( member(X, Prefix), format('~w ', X), fail
	; true
	),
	( Family = [] -> true
	; Family = [F|Fs] ->
	  format('~w', F),
	  ( member(X, Fs), format(' ~w', X), fail ; true )
	),
	( Given = [] -> true
	; Given = [G1|Gs] ->
	  ( Family = [] -> true ; format(', ') ),
	  format('~w', G1),
	  ( member(X, Gs), format(' ~w', X), fail ; true )
	).
fmt_author_v1(X) :-
	write(X).

fmt_author_v2(val_name(Prefix, Family, Given)) :-
	!,
	( member(X, Prefix), format('~w ', X), fail
	; true
	),
	( Given = [] -> true
	; Given = [G1|Gs] ->
	  format('~w', G1),
	  ( member(X, Gs), format(' ~w', X), fail ; true )
	),	
	( Family = [] -> true
	; Family = [F|Fs] ->
	  ( Given = [] -> true ; format(' ') ),
	  format('~w', F),
	  ( member(X, Fs), format(' ~w', X), fail ; true )
	).
fmt_author_v2(X) :-
	write(X).	


bibtex_to_h('', []) :-
	!.
bibtex_to_h(Bibtex, H) :-
	atom_codes(Bibtex, O1),
	read_in_latex(O1, O2),
	lx_to_nlx(O2, O3),
	nlx_to_h_inpar(O3, H).
