%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(date_handling,
	  [parse_ddate/2,
	   ddate_sortkey/2,
	   check_ddate/2,
	   ddate_pretty/2,
	   ddate_pretty_short/2,
	   ddate_pretty_daymonth/2,
	   ddate_pretty_from/2,
	   normdate_pretty/2,
	   normdate_year/2]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_letters(match_atom)).

parse_ddate(S, D) :-
	parse_ddate_special(S, D),
	!.
parse_ddate(S, D) :-
	parse_proper_date(S, D).

parse_proper_date(S, D) :-
	atom_codes(S, S1),
	append(S1, [0' ], S2),
	findall(D1, de_enum_date([0' |S2], D1, _, _), Ds1),
	( Ds1 = [], soleyear(S, Y) ->
	  D = Y
	; Ds1 = [D] ->
	  true
	; D = multi(Ds1)
	).

parse_ddate_special(S, after(D)) :-
	match_atom(['nach dem ', S1], S),
	!,
	parse_proper_date(S1, D).
parse_ddate_special(S, after(D)) :-
	match_atom(['nach ', S1], S),
	!,
	parse_proper_date(S1, D).
parse_ddate_special(S, before(D)) :-
	match_atom(['vor dem ', S1], S),
	!,
	parse_proper_date(S1, D).
parse_ddate_special(S, before(D)) :-
	match_atom(['vor ', S1], S),
	!,
	parse_proper_date(S1, D).
parse_ddate_special(S, between(D1,D2)) :-
	match_atom(['zwischen dem ', S1, ' und ', S2], S),
	!,
	parse_proper_date_or_day(S1, D1),
	parse_proper_date(S2, D2).
parse_ddate_special(S, between(D1,D2)) :-
	match_atom(['zwischen ', S1, ' und ', S2], S),
	!,
	parse_proper_date_or_day(S1, D1),
	parse_proper_date(S2, D2).
parse_ddate_special(S, range(D1,D2)) :-
	match_atom([S1, ' -- ', S2], S),
	!,
	parse_proper_date_or_day(S1, D1),
	parse_proper_date(S2, D2).
parse_ddate_special(S, range(D1,D2)) :-
	match_atom([S1, '--', S2], S),
	!,
	parse_proper_day(S1, D1),
	parse_proper_date(S2, D2).


parse_proper_date_or_day(S, D) :-
	parse_proper_date(S, D),
	D \= multi(_),
	!.
parse_proper_date_or_day(S, D) :-
	parse_proper_day(S, D).

parse_proper_day(S, D) :-
	sub_atom(S, _ , _, 0, '.'),
	sub_atom(S, 0, _, 1, N1),
	atom_number(N1, N),
	N > 0,
	N =< 31,
	!,
	D = d(N).



%%%% 
%%%% de_enum date discovers a sole year only if introduced by
%%%% some context
%%%%
soleyear(S, y(Y)) :-
	atom_length(S, 4),
	atom_number(S, Y),
	Y > 1200,
	Y < 2500.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


check_ddate(ymd(_,_,_), _) :-
	!.
check_ddate([], Ctx) :-
	!,
	info(10, 'Warning: Unspecified date in context: ~q', [Ctx]).
check_ddate(multi(Ds), Ctx) :-
	!,
	info(10, 'Warning: Ambiguous date ~q in context: ~q', [Ds, Ctx]).
check_ddate(D, Ctx) :-
	memberchk(D, [y(_), ym(_,_)]),
	!,
	info(10, 'Warning: Incomplete date: ~q in context ~q', [D, Ctx]).
check_ddate(D, Ctx) :-
	memberchk(D, [md(_,_), d(_)]),
	!,
	info(10, 'Warning: Date without year: ~q in context ~q', [D, Ctx]).
check_ddate(multi(D), Ctx) :-
	!,
	info(10, 'Warning: Ambigous date: ~q in context ~q', [multi(D), Ctx]).
check_ddate(D, Ctx) :-
	memberchk(D, [after(_), before(_), between(_,_), range(_,_)]),
	!,
	info(10, 'Warning: Special date: ~q in context ~q', [D, Ctx]).
check_ddate(X, Ctx) :-
	err('Bad date: ~q in context ~q', [X, Ctx]).

% ddate_sortkey(unknown, '????-??-??') :-
%         !.
% ddate_sortkey(D, K) :-
% 	date_normal(D, DN),
% 	atom_length(DN, L),
% 	( L = 10 ->
% 	  K = DN
% 	; L = 7 ->
% 	  format(atom(K), '~w-??', [DN])
% 	; L = 4 ->
% 	  format(atom(K), '~w-??-??', [DN])
% 	; K = '????-??-??'
% 	).

ddate_sortkey(ymd(Y,M,D), k(Y, M, D)) :- !.
ddate_sortkey(ym(Y,M),    k(Y, M, 99)) :- !.
ddate_sortkey(y(Y),       k(Y, 99, 99)) :- !.
ddate_sortkey(after(D),   K) :-
	!,
	ddate_sortkey(D, K1),
	inc_ddate_sortkey(K1, K).
ddate_sortkey(before(D),   K) :-
	!,
	ddate_sortkey(D, K1),
	dec_ddate_sortkey(K1, K).
ddate_sortkey(range(_,D),   K) :-
	!,
	ddate_sortkey(D, K).
ddate_sortkey(between(_,D),   K) :-
	!,
	ddate_sortkey(D, K).
ddate_sortkey(D, k(9999, 99, D)).

inc_ddate_sortkey(k(Y,M,D), k(Y,M,D1)) :- D1 is D+1.
dec_ddate_sortkey(k(Y,M,D), k(Y,M,D1)) :- D1 is D-1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% Date: Pretty Printing
%%%% 
ddate_pretty(D, P) :-
	with_output_to(atom(P), pp_date(D)).

pp_date(ymd(Y,M,D)) :-
	!,
	pretty_month(M, M1),
	format('~d. ~w ~d', [D, M1, Y]).
pp_date(before(D)) :-
	D = ymd(_, _, _),
	!,
	format('vor dem '),
	pp_date(D).
pp_date(after(D)) :-
	D = ymd(_, _, _),
	!,
	format('nach dem '),
	pp_date(D).
pp_date(between(D1, D2)) :-
	!,
	format('zwischen ~@ und ~@', [pp_date_first(D1,D2), pp_date(D2)]). 
pp_date(range(D1, D2)) :-
	!,
	format('~@ – ~@', [pp_date_first(D1,D2), pp_date(D2)]).
pp_date(y(Y)) :-
	!,
	format('~d', [Y]).
pp_date(ym(Y,M)) :-
	!,
	pretty_month(M, M1),
	format('~w ~d', [M1, Y]).
pp_date(m(M)) :-
	!,
	pretty_month(M, M1),
	format('~w [Jahr unbekannt]', [M1]).
pp_date(md(M,D)) :-
	!,
	pretty_month(M, M1),
	format('~d. ~w [Jahr unbekannt]', [D, M1]).
pp_date(d(D)) :-
	!,
	format('~d. [Jahr und Monat unbekannt]', [D]).
pp_date(X) :-
	err('Failed to print date: ~q', [X]).



%%%% 
%%%% Date: Pretty Printing, Short Version
%%%% 
ddate_pretty_short(D, P) :-
	with_output_to(atom(P), pp_date_short(D)).

pp_date_short(ymd(Y,M,D)) :-
	!,
	pretty_month_short(M, M1),
	format('~d.~w.~d', [D, M1, Y]).
pp_date_short(before(D)) :-
	D = ymd(_, _, _),
	!,
	format('vor dem '),
	pp_date_short(D).
pp_date_short(after(D)) :-
	D = ymd(_, _, _),
	!,
	format('nach dem '),
	pp_date_short(D).
pp_date_short(between(D1, D2)) :-
	!,
	format('zwischen ~@ und ~@',
	       [pp_date_short_first(D1,D2), pp_date_short(D2)]). 
pp_date_short(range(D1, D2)) :-
	!,
	format('~@–~@',
	       [pp_date_short_first(D1,D2), pp_date_short(D2)]).
pp_date_short(y(Y)) :-
	!,
	format('~d', [Y]).
pp_date_short(ym(Y,M)) :-
	!,
	pretty_month_short(M, M1),
	format('~w.~d', [M1, Y]).
pp_date_short(m(M)) :-
	!,
	pretty_month_short(M, M1),
	format('~w. [Jahr unbekannt]', [M1]).
pp_date_short(md(M,D)) :-
	!,
	pretty_month_short(M, M1),
	format('~d.~w. [Jahr unbekannt]', [D, M1]).
pp_date_short(d(D)) :-
	!,
	format('~d. [Jahr und Monat unbekannt]', [D]).
pp_date_short(X) :-
	err('Failed to print date: ~q', [X]).


pp_date_first(D1, D2) :-
	date_rangefirst(D1, D2, D3),
	!,
	pp_date_1(D3).
pp_date_first(D1, _) :-
	pp_date(D1).

pp_date_1(d(D)) :-
	!,
	format('~d.', [D]).
pp_date_1(md(M, D)) :-
	pretty_month(M, M1),
	format('~d. ~w', [D, M1]).
pp_date_1(D) :-
	pp_date(D).

pp_date_short_first(D1, D2) :-
	date_rangefirst(D1, D2, D3),
	!,
	pp_date_short_1(D3).
pp_date_short_first(D1, _) :-
	pp_date(D1).

pp_date_short_1(d(D)) :-
	!,
	format('~d.', [D]).
pp_date_short_1(md(M, D)) :-
	pretty_month_short(M, M1),
	format('~d. ~w', [D, M1]).
pp_date_short_1(D) :-
	pp_date_short(D).

date_rangefirst(ymd(Y,M,D), ymd(Y,M,_), d(D)).
date_rangefirst(md(M,D),    ymd(_,M,_), d(D)).
date_rangefirst(d(D),       _,          d(D)).
date_rangefirst(ymd(Y,M,D), ymd(Y,_,_), md(M,D)).
date_rangefirst(md(M,D),    _,          md(M,D)).

pretty_month_short(N, M) :-
	atom_number(M, N).

pretty_month(1, 'Januar').
pretty_month(2, 'Februar').
pretty_month(3, 'März').
pretty_month(4, 'April').
pretty_month(5, 'Mai').
pretty_month(6, 'Juni').
pretty_month(7, 'Juli').
pretty_month(8, 'August').
pretty_month(9, 'September').
pretty_month(10, 'Oktober').
pretty_month(11, 'November').
pretty_month(12, 'Dezember').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ddate_pretty_daymonth(ymd(_,M,D), PrettyDaymonth) :-
	!,
	ddate_pretty_daymonth(md(M,D), PrettyDaymonth).
ddate_pretty_daymonth(md(M,D), PrettyDaymonth) :-
	!,
	format(atom(PrettyDaymonth), '~|~`0t~d~2+.~|~`0t~d~2+.', [D,M]).
ddate_pretty_daymonth(ym(_,M), PrettyDaymonth) :-
	!,
	ddate_pretty_daymonth(m(M), PrettyDaymonth).
ddate_pretty_daymonth(m(M), PrettyDaymonth) :-
	!,
	format(atom(PrettyDaymonth), '??.~|~`0t~d~2+.', [M]).
ddate_pretty_daymonth(after(D), PD) :-
	!,
	ddate_pretty_daymonth(D, PD1),
	format(atom(PD), 'nach ~w', [PD1]).
ddate_pretty_daymonth(before(D), PD) :-
	!,
	ddate_pretty_daymonth(D, PD1),
	format(atom(PD), 'vor ~w', [PD1]).
ddate_pretty_daymonth(between(D1,D2), PD) :-
	!,
	ddate_pretty_daymonth(D2, PD2),
	( date_rangefirst(D1, D2, D3) ->
	  true
	; D3 = D1
	),
	( D3 = d(D4) ->
	  format(atom(PD1), '~|~`0t~d~2+.', [D4])
	; ddate_pretty_daymonth(D3, PD1)
	),
	format(atom(PD), 'zwischen ~w und ~w', [PD1, PD2]).
ddate_pretty_daymonth(range(D1,D2), PD) :-
	!,
	ddate_pretty_daymonth(D2, PD2),
	( date_rangefirst(D1, D2, D3) ->
	  true
	; D3 = D1
	),
	( D3 = d(D4) ->
	  format(atom(PD1), '~|~`0t~d~2+.', [D4])
	; ddate_pretty_daymonth(D3, PD1)
	),
	format(atom(PD), '~w–~w', [PD1, PD2]).
ddate_pretty_daymonth(_, '??.??.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ddate_pretty_from(ymd(_,_,_), 'vom ') :- !.
ddate_pretty_from(y(_), 'von ') :- !.
ddate_pretty_from(ym(_,_), 'von ') :- !.
ddate_pretty_from(_, '').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Normdate, for date of birth and death of persons
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

normdate_year(Date, Year) :-
	atom_chars(Date, [Y1,Y2,Y3,Y4,'-'|_]),
	!,
	atom_chars(Year, [Y1,Y2,Y3,Y4]).
normdate_year(Date, Date) :-
	atom_chars(Date, [_,_,_,_]),
	!.
normdate_year(_, '?').

normdate_pretty(Date, Pretty) :-
	atom_chars(Date, [Y1,Y2,Y3,Y4,'-',M1,M2,'-',D1,D2]),
	!,
	atom_chars(Pretty, [D1,D2,'.',M1,M2,'.',Y1,Y2,Y3,Y4]).
normdate_pretty(Date, Pretty) :-
	atom_chars(Date, [Y1,Y2,Y3,Y4]),
	!,
	atom_chars(Pretty, [Y1,Y2,Y3,Y4]).
normdate_pretty(Date, Pretty) :-
	atom_chars(Date, [Y1,Y2,Y3,Y4,'/',Z1,Z2,Z3,Z4]),
	!,
	atom_chars(Pretty1, [Y1,Y2,Y3,Y4]),
	atom_chars(Pretty2, [Z1,Z2,Z3,Z4]),
	atomic_list_concat([Pretty1,' - ', Pretty2], Pretty).
normdate_pretty(_, '?').

