%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% This module illustrates how Prolog fact bases, either present in the
%%%% KBSET system as result of e.g. extract_letters/0 (or extract_letters/2),
%%%% or by loading Prolog files with facts, can be converted to RDF and XML
%%%% formats and generate graph representations with Graphviz.
%%%%

:- module(export_demos, []).

:- use_module(library('semweb/rdf_db')).
:- use_module(library(rdf_write)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(xml_writer)).
:- use_module(swilib(dotgraph)).

:- use_module(kbset_base(dates), [date_w3c/2]).
:- use_module(kbset_letters(extract_letters), [letter/3, letter_revlink/3]).
:- use_module(kbset_letters(kbmappings), [id_url/3]).
:- use_module(kbset_letters(data_interface), [xperson_prettyname/2, geo_ename/2]).

:- rdf_register_prefix(let, 'http://www.sulzer-digital.de/sb/').
:- rdf_register_prefix(gnd, 'http://d-nb.info/gnd/').
:- rdf_register_prefix(xxx, 'http://my-domain/').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Example: Use of the Graphviz program to generate graph representations of
%%%% data. Invokes the "dot" program via SWI's shell predicate. The "dot"
%%%% program is part of the Graphviz software package.
%%%%
%%%% The example writes a file '/tmp/demo_graph.png'.
%%%%
%%%% See module swilib(dotgraph) for documentation of the Prolog syntax
%%%% version of dot's graph format.
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

graph_example :-
	graph_example('b:bs:1745-04-16').

graph_example(LetterId) :-
	findall(P, letter_revlink(P, person, LetterId), PersonIds),
	sort(PersonIds, PersonIds1),
	length(PersonIds1, N),
	info(10, 'Number of references persons: ~w', [N]),
	LetterNode = node(LetterId, []),
	findall(node(PersonId, [label=Name,
				shape=box,
				fontcolor=blue,
				color='#ccccff',
				style=filled]),
		( member(PersonId, PersonIds1),
		  xperson_prettyname(PersonId, Name) ),
		PersonNodes),
	findall(edge(LetterId, PersonId, [dir=forward, arrowsize=0.7]),
		member(PersonId, PersonIds1),
		Edges),
	append([LetterNode|PersonNodes], Edges, Graph),
	DotGraph = dotgraph(d1, [graph_attributes([])|Graph]),
	File = '/tmp/demo_graph.png',
	info(10, 'Writing ~w', [File]),
	write_png_dotgraph(DotGraph, File).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Example: Conversion of metainformation to XML/TEI
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% This example is operates on some predicates that become available after
%%%% calling extract_letters/0 (or extract_letters/2).
%%%% 

%%%% 
%%%% Prints a correspDesc stanza for each letter. The solutions are
%%%% enumerated, i.e. type in a semicolon to get the next printed. Or invoke
%%%% the example as "?- xml_example, fail." to print all the solutions for all
%%%% letters.
%%%%
xml_example :-
	enum_letter(X),
	corresp_desc_stanza(X,E),
	write_structure([E], [encoding(utf8)]),
	nl.

enum_letter(L) :-
	findall(L1, letter(L1, _, _), Ls),
	sort(Ls, Ls1),
	member(L, Ls1).
		 
corresp_desc_stanza(LetterId, E) :-
	E = e(correspDesc, [ref=LetterRef],
	      [e(correspAction,
		 [type=sent],
		 [e(persName, [ref=FromRef], [FromName]),
		  e(placeName, [], [Place])
		 | Date]),
	       e(correspAction,
		 [type=received],
		 [e(persName, [ref=ToRef], [ToName])])]),
	%% For an actual application, incomplete metainformation (e.g. missing
	%% place, from, or to) should also considered here (as for date
	%% below).
	letter(LetterId, from, From),
	letter(LetterId, to, To),
	letter(LetterId, place, Geo),
	xperson_prettyname(From, FromName),
	xperson_prettyname(To, ToName),
	geo_prettyname(Geo, Place),
	id_absolute_url(letter, LetterId, LetterRef),
	id_absolute_url(person, From, FromRef),
	id_absolute_url(person, To, ToRef),
	( letter(LetterId, date, Date1),
	  Date1 = ymd(_, _, _),
	  letter(LetterId, date_raw, RawDate) ->
	  date_w3c(Date1, Date2),
	  Date = [e(date, [when=Date2], [RawDate])]
	; letter(LetterId, date_raw, RawDate) ->
	  Date = [e(date, [], [RawDate])]
	; Date = []
	).

id_absolute_url(Type, Id, URL) :-
	id_url(Type, Id, URL1),
	atom_concat('http://www.sulzer-digital.de/sb/', URL1, URL).

geo_prettyname(geo(X), X) :- !.
geo_prettyname(X, Y) :- term_to_atom(X, Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Example: Conversion to RDF
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% Operates on letter/3, which is available after calling extract_letters/0
%%%% (or extract_letters/2).
%%%% 
letter_metainfo(S, P, O) :-
 	member(P, [from, to, date]),
 	letter(S, P, O).
% %%%% 
% %%%% Based on a loaded fact base with predicate my_letter/3:
% %%%% 
% letter_metainfo(S, P, O) :-
%  	member(P, [from, to, date]),
%  	my_letter(S, P, O).
		
rdf_example :-
	letter_metainfo_to_rdf_file('/tmp/tmp_metainfo.rdf').	

letter_metainfo_to_rdf_file(File) :-
	Map = map(entity(let, letter),
		  ns_add(xxx),
		  [from-entity(let, person),
		   to-entity(let, person),
		   date-literal(date)]),
	predicate_to_global_rdf(letter_metainfo, Map, RDF),
	write_rdf_triples(RDF, File).

write_rdf_triples(RDF, File) :-
	length(RDF, N),
	format('% Writing ~w triples~n', [N]),
	open(File, write, Out, [encoding(utf8)]),
        call_cleanup(rdf_write_xml(Out, RDF),
                     close(Out)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% predicate_to_global_rdf(+Pred, +Mapping, +RDF)
%%%% 
%%%% Convert the extension of a ternary Prolog predicate Pred to a list of triples
%%%% suitable for the SWI library predicate rdf_write_xml/2. The Mapping argument
%%%% controls the conversion of Prolog triples to RDF
%%%% specifiers:
%%%% 
%%%% Mapping := map(ItemMapping, ItemMapping, ItemMapping)
%%%% ItemMapping := IMSpec | list_of(IMSpec)
%%%% IMSpec := IMKey | Predicate-IMKey
%%%% IMKey :=    literal | literal(string) | ...
%%%%           | entity(NameSpace, KbLetType) | ns_add (NameSpace)
%%%%           | ns_prefix | keep
%%%% 
%%%%           See implementation of value_to_rdf/2 
%%%% 	  
%%%% The three item mappings are applied to Subject, Predicate and Object,
%%%% respectively, of each triple. If ItemMapping is a list, the first IMSpec
%%%% that matches is taken. That is, is of the form Predicate-IMKey where
%%%% Predicate unifies with the "predicate" (i.e. second argument) of the
%%%% input triple, or is just of the form IMKey.
%%%% 
%%%% Used namespaces should be registered with the SWI library predicate
%%%% rdf_register_prefix.
%%%% 		  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

predicate_to_global_rdf(Pred, Map, RDF) :-
	findall( RDFTriple,
		 ( Call =.. [Pred,S,P,O],
		   call(Call),
		   triple_to_global_rdf( triple(S,P,O), Map, RDFTriple )
		 ),
		 RDF1),
	sort(RDF1, RDF).

triple_to_global_rdf(Triple, Map, rdf(S, P, O)) :-
	triple_to_rdf(Triple, Map, rdf(S1, P1, O1)),
	rdf_global_id(S1, S),
	rdf_global_id(P1, P),
	rdf_global_object(O1, O).

triple_to_rdf(Triple, Map, rdf(S, P, O)) :-
	Triple = triple(S1, P1, O1),
	Map = map(MapS, MapP, MapO),
	( apply_mapping(MapS, P1, S1, S),
	  apply_mapping(MapP, P1, P1, P),
	  apply_mapping(MapO, P1, O1, O) ->
	  true
	; info(10, 'RDF conversion failed for: ~q', [Triple]),
	  fail
	).

apply_mapping(M, P, X, V) :-
	M \= [_|_],
	!,
	apply_mapping([M], P, X, V).
apply_mapping(M, P, X, V) :-
	( member(M1, M),
	  ( M1 = P1-M2, \+ \+ P=P1 ->
	    true
	  ; M1 \= _-_ ->
	    M2 = M1
	  ) -> true
 	; err('Mapping failed to match: ~q ~q', [M, P])
	),
	( value_to_rdf(M2, X, V) ->
	  true
	; % info(10, 'Mapping failure (value ignored): ~q ~q ~q', [M, P, X]),
	  fail
	).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

value_to_rdf(literal, X, literal(X)) :-
	%% Untyped literal
	!.
value_to_rdf(literal(string), X, literal(type(xsd:string, X))) :-
	!.
value_to_rdf(literal(integer), X, literal(type(xsd:integer, V))) :-
	!,
	( integer(X) ->
	  format(atom(V), '~d', X)
	; atom(X), atom_number(X, Y), integer(Y) ->
	  V = X
	).
value_to_rdf(literal(decimal), X, literal(type(xsd:decimal, V))) :-
	!,
	( number(X) ->
	  format(atom(V), '~d', X)
	; atom(X), atom_number(X, _) ->
	  V = X
	).
value_to_rdf(literal(date), X, literal(type(xsd:Type, V))) :-
	!,
	( ( X = ymd(_ ,_ ,_) -> Type = date
	  ; X = ym(_, _) -> Type = gYearMonth
	  ; X = y(_) -> Type = gYear
	  ; X = md(_) -> Type = gMonthDay
	  ; X = m(_) -> Type = gMonth
	  ; X = d(_) -> Type = gDay
	  ) ->
	  date_w3c(X, V)
	; Type = string,
	  format(atom(V), '~q', [X]),
	  info(10, 'Representing date as string: ~q', [X])
	).
value_to_rdf(entity(NS, Type), X, NS:Id) :-
	%% To convert "local" entities from the KBSET/KBLET system
	!,
	id_url(Type, X, Id).
value_to_rdf(ns_add(NS), X, NS:X) :-
	!.
value_to_rdf(ns_prefix, X, NS:V) :-
	%% e.g. for atoms of the form 'gnd:GNDID' used in KBSET
	!,
	sub_atom(X, N, _, _, ':'),
	!,
	sub_atom(X, 0, N, _, NS),
	N1 is N+1,
	sub_atom(X, N1, _, 0, V).
value_to_rdf(keep, X, X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
