%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(extract_letters,
	  [extract_letters/2,
	   letter/3,
	   letter_xl/3,
	   letter_revlink/3,
	   letter_kitem_id/2,
	   main_info_annot/3]).

:- use_module(kbset_base(readwrite)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_base(filename)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(fromonto)).

:- use_module(kbset_letters(legacy)).
:- use_module(kbset_letters(date_handling)).
:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(textutilities)).
:- use_module(kbset_letters(nlx)).

:- dynamic letter/3.
:- dynamic letter_revlink/3.
:- dynamic letter_xl/3.
:- dynamic main_info_annot/3.
:- dynamic letter_kitem_id/2.

latex_command(begin,[c,c,c,c,c,c],'').
latex_command(kitem,[c],'').
latex_command(vitem,[c],'').
latex_command(ksection,[t],'').
latex_command(noindent,[],'').
latex_command(sout,[t],'').
latex_command(textit,[t],'').
latex_command(textsf,[t],'').
latex_command(multicolumn,[c,c,t],'').
latex_command(xciteAY,[c],'').
latex_command(xciteAY,[o,c],'').
latex_command(xciteST,[c],'').
latex_command(xciteST,[o,c],'').
latex_command(xnocite,[c],'').
latex_command(xcon,[t],'').
latex_command(xcorp,[c,t],'').
latex_command(xcorpmain,[c,t],'').
latex_command(xdate,[t],'').
latex_command(xdateEnd,[t],'').
latex_command(xenglish,[t],'').
latex_command(xevent,[c,t],'').
latex_command(xeventmain,[c,t],'').
latex_command(xfrac,[t,t],'').
latex_command(xfrench,[t],'').
latex_command(xgeo,[c,t],'').
latex_command(xgeomain,[c,t],'').
latex_command(xgreek,[t],'').
latex_command(xhoch,[t],'').
latex_command(xmeta,[c,t],'').
latex_command(xins,[t],'').
latex_command(xinl,[t],'').
latex_command(xjournal,[c,t],'').
latex_command(xjournalmain,[c,t],'').
latex_command(xjournalq,[c,t],'').
latex_command(xl,[c,t],'').
latex_command(xlatin,[t],'').
latex_command(xletterRef,[c],'').
latex_command(xletterTitleRef,[c],'').
latex_command(xll,[c,c],'').
latex_command(xls,[t],'').
latex_command(xmar,[t],'').
latex_command(xmissing,[],'').
latex_command(xpar,[],'\n\n').
latex_command(xperson,[c,t],'').
latex_command(xpersonmain,[c,t],'').
latex_command(xquote,[t],'').
latex_command(xqwork,[t],'').
latex_command(xsalute,[t],'').
latex_command(xsout,[t],'').
latex_command(xsubject,[c,t],'').
latex_command(xsubjectmain,[c,t],'').
latex_command(xul,[t],'').
latex_command(xulul,[t],'').
latex_command(xus,[t],'').
latex_command(xwork,[c,t],'').
latex_command(xworkmain,[c,t],'').
latex_command(xworkq,[c,t],'').
latex_command(xxx,[],'XXX').
latex_command(xGulden,[],'Gulden').
latex_command(xSchilling,[],'Schilling').
latex_command(xPfennige,[],'Pfg').
latex_command(xPfund,[],'£').
latex_command(xReichstaler,[],'Rthlr').
latex_command(xGuldenPeriod,[],'Gulden. ').
latex_command(xPfennigePeriod,[],'Pfg. ').
latex_command(xPfundPeriod,[],'£. ').
latex_command(xReichstalerPeriod,[],'Rthl. ').
latex_command(xhref,[c,t],'').
latex_command(xurl,[c],'').

:- initialization( ( latex_command(C, A, T),
		     register_latex_command(C, A, T),
		     ( legacy_alias(C1, C) ->
		       register_latex_command(C1, A, T)
		     ; true
		     ),
		     fail
		   ; true
		   )).

extract_letters(LetterFileSpec, AnnotFileSpec) :-
	refresh_access_data,
	filename(AnnotFileSpec, AnnotFile),
	filename(LetterFileSpec, LetterFile),
	extract_annot_data(AnnotFile),
	set_letter_kitem_id,
	extract_letter_data(LetterFile),
	set_letter_revlink,
	set_letter_xl,
	set_letter_main_info_annot,
	set_letter_xmeta_properties.

extract_letter_data(NLX) :-
	findall(Att, letter(_,Att,_), Atts),
	sort(Atts, Atts1),
	( member(Att1, Atts1), Att1 \= annot,
	  retractall( letter(_,Att1,_) ),
	  fail
	; true
	),
	( atom(NLX) ->
	  info(10, 'Reading in ~w', [NLX]),
	  read_in_latex(NLX, NLX1),
	  info(10, 'Converting to NLX format'),
	  lx_to_nlx(NLX1, NLX2)
	; NLX2 =  NLX
	),
	info(10, 'Extracting letters'),
	( nlx_enum(NLX2,
		   nenv(LetterEnv, _, [Id, From, To, Place, Date|_], Text)),
	  ( LetterEnv = letter ; LetterEnv = brief ),
	  raw_person_to_pid(From, From1),
	  raw_person_to_pid(To, To1),
	  raw_place_to_gid(Place, Place1),
	  parse_ddate(Date, Date1),
	  assert(letter(Id, id, Id)),
	  assert(letter(Id, from, From1)),
	  assert(letter(Id, to, To1)),
	  assert(letter(Id, place, Place1)),
	  assert(letter(Id, date, Date1)),
	  assert(letter(Id, date_raw, Date)),
	  assert(letter(Id, text, Text)),
	  fail
	; true
	),
	findall(K, letter(K, _, _), Ks),
	sort(Ks, Ks1),
	length(Ks1, Len),
	info(10, 'Found ~w letters', [Len]).

extract_annot_data(NLX) :-
	retractall( letter(_,annot,_) ),
	( atom(NLX) ->
	  info(10, 'Reading in ~w', [NLX]),
	  read_in_latex(NLX, NLX1),
	  info(10, 'Converting to NLX format'),
	  lx_to_nlx(NLX1, NLX2)
	; NLX2 =  NLX
	),
	info(10, 'Extracting annots'),
	( ( nlx_enum(NLX2, nenv(kommentar,_,[Id|_], Text))
	  ; nlx_enum(NLX2, nenv(annotation,_,[Id|_], Text))
	  ),
	  assert(letter(Id, annot, Text)),
	  fail
	; true
	),
	findall(K, letter(K, annot, _), Ks),
	sort(Ks, Ks1),
	length(Ks1, Len),
	info(10, 'Found annots for ~w letters', [Len]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

raw_person_to_pid(Id, Id) :-
	enum_xperson(Id),
	!.
raw_person_to_pid(Name, Id) :-
	%% Try a reverse lookup in defheadername:
	headername(Id, Name),
	enum_xperson(Id),
	!.
raw_person_to_pid(X, _) :-
	err('Not registered as person id: ~q', [X]).

% raw_place_to_gid(Place, Id) :-
% 	atom_to_ascii(Place, Id).

raw_place_to_gid(Id, Id) :-
	enum_geo(Id),
	!.
raw_place_to_gid(Place, geo(Place)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Links Extracted from Text
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

text_prop(text).
text_prop(annot).

set_letter_revlink :-
	info(10, 'Extracting reverse links'),
	retractall( letter_revlink(_, _, _) ),
	( letter(Id, P, V),
	  text_prop(P),
	  text_get_links(V, Ls),
	  ( member(link(L, Type), Ls),
	    \+ letter_revlink(L, Type, Id),
	    assert( letter_revlink(L, Type, Id) ),
	    fail
	  ; true
	  ),
	  fail
	; true
	).

text_get_links(NLX, Links) :-
	findall(link(X,Type),
		( nlx_enum(NLX, ncmd(CMD, _, [X], _)),
		  link_command_and_type(CMD, Type) ),
	        Xs),
	sort(Xs, Links).

link_command_and_type(A, B) :-
	legacy_alias(A, A1),
	!, 
	link_command_and_type(A1, B).
link_command_and_type(xperson, person).
link_command_and_type(xpersonmain, person).
link_command_and_type(xwork, work).
link_command_and_type(xworkmain, work).
link_command_and_type(xworkq, work).
link_command_and_type(xgeo, geo).
link_command_and_type(xgeomain, geo).
link_command_and_type(xcorp, corp).
link_command_and_type(xcorpmain, corp).
link_command_and_type(xevent, event).
link_command_and_type(xeventmain, event).
link_command_and_type(xsubject, subject).
link_command_and_type(xsubjectmain, subject).
link_command_and_type(xjournal, journal).
link_command_and_type(xjournalmain, journal).
link_command_and_type(xjournalq, journal).

link_command_and_type(xpersonmain, main).
link_command_and_type(xworkmain, main).
link_command_and_type(xgeomain, main).
link_command_and_type(xcorpmain, main).
link_command_and_type(xeventmain, main).
link_command_and_type(xsubjectmain, main).
link_command_and_type(xjournalmain, main).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Annot-Labeling Information Extracted from Letters
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_letter_xl :-
	info(10, 'Extracting xl information'),
	retractall( letter_xl(_, _, _) ),
	( letter(Id, text, V),
	  text_get_xl(V, Label, Value),
	  assert( letter_xl(Id, Label, Value) ),
	  fail
	; true
	).

text_get_xl(NLX, Label, Value) :-
	nlx_enum(NLX, ncmd(xl, _, [Label|_], [Value|_])).
text_get_xl(NLX, Label, Value) :-
	nlx_enum(NLX, ncmd(xll, _, [Label,Value1|_], _)),
	atom_codes(Value1, Codes),
	( Codes = [] -> Value = []
	; read_in_latex(Codes, Value)
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Information from Annot:  Main Entries, Kitems
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic tmp_kitem/1.

set_letter_main_info_annot :-
	info(10, 'Extracting information on main entity references'),	
	retractall( main_info_annot(_, _, _) ),
	( letter(LID, annot, Annot),
	  retractall( tmp_kitem(_) ),
	  nlx_enum(Annot, X),
	  ( X = ncmd(kitem, _, [L], _) ->
	    assertz( tmp_kitem(L) )
	  ; X = ncmd(vitem, _, [L], _) ->
	    assertz( tmp_kitem(L) )
	  ; true
	  ),
	  ( X = ncmd(XMain, _, [U], _),
	    link_command_and_type(XMain, main),
	    tmp_kitem(L1) ->
	    assertz( main_info_annot(U, LID, L1) )
	  ; true
	  ),
	  fail
	; true
	).

set_letter_kitem_id :-
	info(10, 'Extracting identifiers of annotated text positions'),
	retractall( letter_kitem_id(_, _) ),
	( letter(LID, annot, Annot),
	  nlx_enum(Annot, X),
	  ( X = ncmd(kitem, _, [KItemId], _) -> true
	  ; X = ncmd(vitem, _, [KItemId], _)
	  ),
	  assertz( letter_kitem_id(LID, KItemId) ),
	  fail
	; true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_letter_xmeta_properties :-
	info(10, 'Extracting xmeta declarations'),
 	( ( letter(Id, text, E)
	  ; letter(Id, annot, E)
	  ),
	  member(ncmd(xmeta, _, [Prop|_], [Txt|_]), E),
	  xmeta_prop(Prop, Prop1),
	  assertz( letter(Id, Prop1, Txt) ),
	  fail
	; true
	).

xmeta_prop(transliteration, transliteration).
xmeta_prop(annotationAuthor, annotation_author).
xmeta_prop(status, status).

