%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2020 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% For now mainly to illustrate the possibilities of an export of meta data
%%%% in Prolog syntax. Should mature into a specified set of predicates
%%%% suitable to represent the meta data for correspondences.
%%%% 

:- module(factbase_export,
	  [list_metadata/0]).

:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(extract_letters)).
:- use_module(kbset_letters(date_handling)).
:- use_module(kbset_base(readwrite)).
:- use_module(kbset_base(dates)).

%%%% 
%%%% Expects that extract_letters/2 has been called before
%%%% to populate letter data predicates
%%%% 
list_metadata :-
	list_letter_metadata,
	list_letter_refers,
	list_data.
	
list_letter_metadata :-
	findall(letter(A, B, C),
		( member(B, [date,from,date_raw,to]),
		  letter(A, B, C)
		; member(B, [transliteration, annotation_author]),
		  letter(A, B, D),
		  lxelems_plaintxt_codes(D, D1),
		  atom_codes(C, D1)
		; letter(A, date, D),
		  ( B = date_year, ddate_year(D, C)
		  ; ddate_w3c(D, B, C)
		  )
		; letter(A, place, G),
		  ( G = geo(G1),
		    data_interface:defgeo(C, G1) -> B=place
		  ; data_interface:defsubgeo(C, _, G1) -> B=place
		  ; C = G1, B=place_raw
		  )
		),
		FsLetters1),
	sort(FsLetters1, FsLetters),
	( member(F, FsLetters),
	  portray_clause(F),
	  fail
	; true
	).

list_letter_refers :-
	findall(letter_refers(A,B,C),
		( letter_revlink(C,B,A), B\=main ),
		Links1),
	sort(Links1, Links2),
	( member(F, Links2),
	  portray_clause(F),
	  fail
	; true
	).

data_predicate(data_interface:defcorp/2).
data_predicate(data_interface:defcorpSee/2).
data_predicate(data_interface:defevent/2).
data_predicate(data_interface:defeventSee/2).
data_predicate(data_interface:defgeo/2).
data_predicate(data_interface:defgeoSee/2).
data_predicate(data_interface:defjournal/2).
data_predicate(data_interface:defjournalSee/2).
data_predicate(data_interface:defperson/2).
data_predicate(data_interface:defpersonSee/2).
data_predicate(data_interface:defsubevent/3).
data_predicate(data_interface:defsubgeo/3).
data_predicate(data_interface:defsubject/2).
data_predicate(data_interface:defsubjectSee/2).
data_predicate(data_interface:defwork/3).
data_predicate(data_interface:defworkThree/5).
data_predicate(data_interface:defworkTwo/4).
data_predicate(person_external_id/3).

person_external_id(P, X, Y) :-
	enum_xperson(P), 
	xperson_links(P, Links),
	( member(gnd-Y, Links),
	  X = gnd
	; member(wiki_de-Y1, Links),
	  strip_wiki_de_prefix(Y1, Y),
	  X = wiki_de
	).

strip_wiki_de_prefix(X, Y) :-
	sub_atom(X, 0, N, _, 'wikiDE:'),
	!,
	sub_atom(X, N, _, 0, Y).
ensure_wiki_de_prefix(X, X).


list_data :-
	( data_predicate(P),
	  sorted_extension(P, FsData),
	  ( member(F, FsData),
	    portray_clause(F),
	    fail
	  ; true
	  ),
	  fail
	; true
	).

sorted_extension(PSpec, E) :-
	( PSpec = Module:P/N ->
	  functor(Template, P, N),
	  Call = Module:Template
	; PSpec = P/N ->
	  functor(Template, P, N),
	  Call = Template
	),
	findall(Template, Call, E1),
	sort(E1, E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ddate_year(y(Y), Y).
ddate_year(ym(Y,_), Y).
ddate_year(ymd(Y,_,_), Y).
ddate_year(range(_,D), Y) :- date_year(D, Y).
ddate_year(between(_,D), Y) :- date_year(D, Y).
ddate_year(after(D), Y) :- date_year(D, Y).
ddate_year(before(D), Y) :- date_year(D, Y).

ddate_w3c(range(X,Y), A, V) :-
	!,
	( A = date_range_from_w3c,
	  date_w3c(X, V)
	; A = date_range_to_w3c,
	  date_w3c(Y, V)
	).
ddate_w3c(between(X,Y), A, V) :-
	!,
	( A = date_between_from_w3c,
	  date_w3c(X, V)
	; A = date_between_to_w3c,
	  date_w3c(Y, V)
	).
ddate_w3c(after(X), date_after_w3c, V) :-
	!,
	date_w3c(X, V).
ddate_w3c(before(X), date_before_w3c, V) :-
	!,
	date_w3c(X, V).
ddate_w3c(X, date_w3c, V) :-
	date_w3c(X, V).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
