%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(htmlexport_letters,
	  [ gen_html/1,
	    gen_html_bibpage/1] ).
	    
:- use_module(swilib(fromonto)).
:- use_module(swilib(xml_writer)).
:- use_module(swilib(pretty)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_letters(date_handling)).
:- use_module(kbset_letters(geo_handling)).
:- use_module(kbset_letters(kbmappings)).
:- use_module(kbset_letters(textutilities)).
:- use_module(kbset_letters(extract_letters)).
:- use_module(kbset_letters(textconvert)).
:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(bibconvert)).
:- use_module(kbset_letters(name_conversions)).

:- flag(offline, _, false).
% :- flag(offline, _, true).

:- dynamic html_options/1.
:- dynamic html_maintitle/1.
:- dynamic html_correspondence_navi/2.

:- dynamic pid_person_cache/2.

clear_caches :-
	retractall(pid_person_cache(_, _)).

% clear(e(br, [clear=all], [])).
clear(e(div, [style='clear:both; overflow:hidden'], [])).

set_html_options(Options) :-
	retractall( html_options(_) ),
	asserta( html_options(Options) ),
	( memberchk(maintitle=Maintitle, Options) ->
	  set_maintitle(Maintitle)
	; true
	).

maintitle(MainTitle) :-
	( html_maintitle(Title) ->
	  MainTitle = xml(Title)
	; MainTitle = 'MAINTITLE'
	).

set_maintitle(Atom) :-
	retractall( html_maintitle(_) ),
	asserta( html_maintitle(Atom) ).

html_correspondence_navi(A, B) :-
	( hcn(A, B) ->
	  true
	; hcn(B, A)
	).

hcn(A, B) :-
	( letter(L, from, A), letter(L, to, B) ->
	  ( letter(L1, from, B), letter(L1, to, A) ->
	    true
	  ; letter(L1, from, A), L1 \= L, letter(L1, to, B) ->
	    true
	  )
	).

%%%% 
%%%% Expects that extract_letters/2 has been called before
%%%% to populate letter data predicates
%%%% 
gen_html(Options) :-
	set_html_options(Options),
	refresh_access_data,
	clear_caches,
	( memberchk(skip_letters=true, Options) -> true
	; write_letters
	),
 	write_geos,
	write_letter_registers,
	write_letters_page,
 	write_persons_page,
	write_persons_by_refs_page,
 	write_geos_page,
	write_geos_by_refs_page,
	write_genregs,
%	write_events_page,
	write_persons,
	write_chain_pages,
	generate_referenced_biblabels.
% 	write_drawings_page.


write_letters :-
	lids(Bs),
	( member(B, Bs),
	  ( write_letter_page(B) ->
	    fail
	  ; err('write_letter_page failed: ~q', [B])
	  )
	; true
	).

write_persons :-
	persons(Ps),
	length(Ps, L),
	info(10, 'Writing ~w person pages', [L]),
	( member(P, Ps),
	  write_person_page(P),
	  fail
	; true
	).

write_geos :-
	geos(Os),
	( member(O, Os),
	  write_geo_page(O),
	  fail
	; true
	).

pid_person(Id, P) :-
	pid_person_cache(Id, P),
	!.
pid_person(Id, P) :-
	xperson(P),
	get_val(P, id, Id1),
	Id1 = Id,
	!,
	assert(pid_person_cache(Id, P)).

pid_shortname(Id, N) :-
	xperson_prettyname(Id, N1),
	short_person_name(N1, N),
	!.
pid_shortname(_, 'FOO').
	
map_initial([X|Xs], [X1,'. '|Xs1]) :-
	sub_atom(X, 0, 1, _, X1),
	map_initial(Xs, Xs1).
map_initial([], []).

pid_ref(Id, E) :-
	pid_ref(Id, [], E).
	
pid_ref(Id, Options, E) :-
	pid_person(Id, P),
	person_ref(P, Options, E),
	!.
pid_ref(Id, _, _) :-
	err('Undefinierte Person ~w', [Id]).

pid_ref_nolink(Id, E) :-
	pid_person(Id, P),
	person_ref_nolink(P, E),
	!.
pid_ref_nolink(Id, _) :-
	err('Undefinierte Person ~w', [Id]).

pid_ref_nolink_short(Id, E) :-
	pid_person(Id, P),
	person_ref_nolink_short(P, E),
	!.
pid_ref_nolink_short(Id, _) :-
	err('Undefinierte Person ~w', [Id]).

void_dates(Ds) :-
	\+ (member(X, Ds), X \= '?').

person_ref_nolink(P, E) :-
	get_val(P, prettyname, Name),
	get_val(P, birthdate, DG1),
	normdate_year(DG1, DG),
	get_val(P, deathdate, DT1),
	normdate_year(DT1, DT),
	( void_dates([DG,DT]) ->  D = []
	; D = [' (', DG, '–', DT, ')']
	),
	atomic_list_concat([Name|D], E).


person_ref_nolink_short(P, E) :-
	get_val(P, prettyname, Name),
	short_person_name(Name, Name1),
	get_val(P, birthdate, DG1),
	normdate_year(DG1, DG),
	get_val(P, deathdate, DT1),
	normdate_year(DT1, DT),
	( void_dates([DG,DT]) ->  D = []
	; D = [' (', DG, '–', DT, ')']
	),
	atomic_list_concat([Name1|D], E).


person_ref_with_id(Id, E) :-
	atom(Id),
	!,
	pid_person(Id, P),
	person_ref_with_id(P, E).
person_ref_with_id(P, E) :-
	get_val(P, id, Id),
	get_val(P, prettyname, Name),
	get_val(P, birthdate, DG1),
	id_url(person, Id, PersonURL),
	normdate_year(DG1, DG),
	get_val(P, deathdate, DT1),
	normdate_year(DT1, DT),
	id_anchor(person, Id, Id1),
	( void_dates([DG,DT]) ->  D = []
	; D = [' (', DG, xml('&ndash;'), DT, ')']
	),
	E = e(a,
	      [class=enav, id=Id1, href=PersonURL],
	      [Name | D]).

person_ref(P, E) :-
	person_ref(P, [], E).

person_ref(Id, Options, E) :-
	atom(Id),
	!,
	pid_person(Id, P),
	person_ref(P, Options, E).
person_ref(P, Options, E) :-
	get_val(P, id, Id),
	get_val(P, prettyname, Name),
	get_val(P, birthdate, DG1),
	id_url(person, Id, PersonURL),
	normdate_year(DG1, DG),
	get_val(P, deathdate, DT1),
	normdate_year(DT1, DT),
	( memberchk(extras=Extras0, Options) ->
	  true
	; Extras0 = []
	),
	( memberchk(number_of_letters, Options) ->
	  pid_number_of_letters(Id, Extras1)
	; Extras1 = []
	),
	( memberchk(stats, Options) ->
	  pid_stats(Id, Extras2)
	; Extras2 = []
	),
	append(Extras1, Extras2, Extras3),
	( Extras3 = [] ->
	  Extras4 = []
	; Extras4 = [e(span, [class=plinfo], Extras3)]
	),
	append(Extras0, Extras4, Extras),
	( memberchk(class=Class, Options) -> true ; Class=pinline ),
	( memberchk(element=Element, Options) -> true ; Element=span ),
	( void_dates([DG,DT]) ->  D = []
	; D = [' (', DG, xml('&ndash;'), DT, ')']
	),

	E = e(Element, [class=Class],
	      [e(a,
		 [class=enav, href=PersonURL],
		 [Name | D])
	      | Extras]).

person_ref_nodates(Id, E) :-
	atom(Id),
	!,
	pid_person(Id, P),
	person_ref_nodates(P, E).
person_ref_nodates(P, E) :-
	get_val(P, id, Id),
	get_val(P, prettyname, Name),
	id_url(person, Id, PersonURL),
	id_anchor(person, Id, Id1),
	E = e(a, [class=enav, id=Id1, href=PersonURL], [Name]).


entity_links(P, Class, Links) :-
	get_val(P, links, Links1),
	get_val(P, id, PID),
	( main_info_annot(PID, A, B) ->
	  main_info_url(A, B, KItem),
	  Links2 = [kitem-KItem|Links1]
	; Links2 = Links1
	),
	map_entity_link_add_sortkey(Links2, L3),
	sort(L3, L4),
	map_val(L4, Links4),
	map_pretty_link(Links4, Class, Links).

map_entity_link_add_sortkey([X|Xs], [X1|Xs1]) :-
	entity_link_add_sortkey(X, X1),
	map_entity_link_add_sortkey(Xs, Xs1).
map_entity_link_add_sortkey([], []).

entity_link_add_sortkey(kitem-X, 5-(kitem-X)) :- !.
entity_link_add_sortkey(wiki_de-X, 10-(wiki_de-X)) :- !.
entity_link_add_sortkey(wiki_en-X, 15-(wiki_en-X)) :- !.
entity_link_add_sortkey(wiki_fr-X, 17-(wiki_fr-X)) :- !.
entity_link_add_sortkey(gnd-X,  20-(gnd-X)) :- !.
entity_link_add_sortkey(ndb-X,  30-(ndb-X)) :- !.
entity_link_add_sortkey(wwp-X,  40-(wwp-X)) :- !.
entity_link_add_sortkey(geohack_de-X,  50-(geohack_de-X)) :- !.
entity_link_add_sortkey(K-X,   100-(K-X)).

pretty_link(kitem-URL, Class,
	    e(a, [class=Class, href=URL1], ['Kommentar'])) :-
	!,
	kitem_link(URL, URL1).
pretty_link(wiki_de-URL, Class,
	    e(a, [class=Class, href=URL1], ['Wikipedia'])) :-
	!,
	wiki_de_link(URL, URL1).

pretty_link(wiki_en-URL, Class,
	    e(a, [class=Class, href=URL1], ['Wikipedia (Englisch)'])) :-
	!,
	wiki_en_link(URL, URL1).

pretty_link(wiki_fr-URL, Class,
	    e(a, [class=Class, href=URL1], ['Wikipedia (Französisch)'])) :-
	!,
	wiki_fr_link(URL, URL1).

pretty_link(ndb-URL, Class,
	    e(a, [class=Class, href=URL1], ['Deutsche Biographie'])) :-
	!,
	ndb_link(URL, URL1).
pretty_link(gnd-URL, Class,
	    e(a, [class=Class, href=URL1], ['Gemeinsame Normdatei'])) :-
	!,
	gnd_link(URL, URL1).
pretty_link(wwp-URL, Class,
	    e(a, [class=Class, href=URL1], ['WW-Person'])) :-
	!,
	wwp_link(URL, URL1).
pretty_link(geohack_de-URL, Class,
	    e(a, [class=Class, href=URL1], ['GeoHack'])) :-
	!,
	geohack_de_link(URL, URL1).
pretty_link(link(Pretty)-URL, Class,
	    e(a, [class=Class, href=URL], [Pretty])) :-
	!.
pretty_link(_-URL, Class, e(a, [class=Class, href=URL], [URL])).

map_pretty_link([X], C, [X1]) :-
	pretty_link(X, C, X1),
	!.
map_pretty_link([X|Xs], C, [X1, ' | ' |Xs1]) :-
	pretty_link(X, C, X1),
	map_pretty_link(Xs, C, Xs1).
map_pretty_link([], _, []).

write_persons_page :-
	write_persons_page([]).

write_persons_by_refs_page :-
	write_persons_page([file='persons_by_refs.html',
			    self='persons_by_refs',
			    sort_by_number_of_references]).

write_persons_page(Options) :-
	letterpersons(BPs0),
	nonletterpersons(NPs0),
	( memberchk(self=Self, Options) -> true ; Self=persons ),
	( memberchk(sort_by_number_of_references, Options) ->
	  sort_by_number_of_letters(BPs0, BPs),
	  sort_by_number_of_references(NPs0, person, NPs)
	; BPs = BPs0,
	  NPs = NPs0
	),
	map_person_ref(BPs, [number_of_letters,
			     element=div,
			     class=pline], BLines),
	map_person_ref(NPs, [stats,
			     element=div,
			     class=pline], NLines),
	furthernames_refs(WLines),
	Es = [NaviTop,
	      e(hr, [class=navitop], []), CB,
	      e(h1, [], ['Briefverfasser und Briefempfänger']),
	      e(div, [class=xlines], BLines),
	      e(h1, [], ['Weitere Personen']),
	      e(div, [class=xlines], NLines),
	      e(h1, [], ['Weitere Namen der aufgeführten Personen']),
	      e(div, [class=xlines], WLines),
	      CE,
	      e(hr, [class=navibot], []),
	      NaviBot],
	navi(Self, top, NaviTop, CB),
	navi(Self, bot, NaviBot, CE),
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle, ': ', 'Personen'])]),
		   Body1])],
	mk_body(Es, Body1),
	( memberchk(file=File, Options) -> true ; File='persons.html' ),
	write_page(Page, File).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

furthernames_refs(Es) :-
	findall(N-P, ( enum_xperson(P),
		       id_is_referenced(person, P),
		       xperson_furthernames(P, Ns),
		       member(N, Ns)
		       ),
		NPs),
	keysort(NPs, NPs1),
	map_furthernames_ref(NPs1, Es).

% furthernames_refs(Es) :-
% 	findall(N-P, ( enum_xperson(P),
% 		       xperson_furthernames(P, Ns),
% 		       member(K-N, Ns),
% 		       \+ memberchk(K, [pseudonym, pseudonym_of])
% 		     ),
% 		NPs),
% 	keysort(NPs, NPs1),
% 	map_furthernames_ref(NPs1, Es).

map_furthernames_ref([N-P], [X1]) :-
	!,
	furthernames_ref(N, P, X1).
map_furthernames_ref([N-P|Xs], [X1|Xs1]) :-
	furthernames_ref(N, P, X1),
	map_furthernames_ref(Xs, Xs1).
map_furthernames_ref([], []).

furthernames_ref(N, P, E) :-
	pid_ref(P, [], PR),
	E = e(div, [class=pline],
	      [N, e(span, [class=plsee], [', siehe ']), PR]).

%%%% 
%%%% keysort a list with person ids as keys according to the persons, i.e.
%%%% prettyname, birthdate
%%%%
keysort_with_pidkeys(PIDVALs, PIDVALs1) :-
	findall(K-(PID-V),
		( member(PID-V, PIDVALs),
		  pid_person(PID, P),
		  add_keyvals(P, [person_sortname, birthdate], K-_) ),
		KPs),
	keysort(KPs, KPs1),
	map_val(KPs1, PIDVALs1).


persons(Ps) :-
 	findall(P, (xperson(P), is_referenced(person, P)), Ps1),
 	sort_plist(Ps1, [person_sortname, birthdate], Ps).

letterpersons(Ps) :-
 	findall(P, (xperson(P), is_letterperson(P)), Ps1),
 	sort_plist(Ps1, [person_sortname, birthdate], Ps).

nonletterpersons(Ps) :-
 	findall(P, ( xperson(P),
		     \+ is_letterperson(P),
		     is_referenced(person, P)
		   ),
		Ps1),
 	sort_plist(Ps1, [person_sortname, birthdate], Ps).


is_letterperson(P) :-
	get_val(P, id, Id),
	id_is_letterperson(Id).

id_is_letterperson(Id) :-
	( letter(_, from, Id) ->
	  true
	; letter(_, to, Id) ->
	  true
	).

id_is_referenced(Type, Id) :-
	letter_revlink(Id, Type, _),
	!.
id_is_referenced(person, Id) :-
	id_is_letterperson(Id),
	!.
id_is_referenced(person, Id) :-
	work_author(Id1, Id),
	id_is_referenced(work, Id1),
	!.
id_is_referenced(geo, Id) :-
	geo_subgeos(Id, Subs),
	member(Id1, Subs),
	id_is_referenced(geo,Id1),
	!.
id_is_referenced(Type, Id) :-
	genreg_subs(Type, Id, Subs),
	member(Id1, Subs),
	id_is_referenced(Type, Id1).

is_referenced(Type, Item) :-
 	get_val(Item, id, Id),
	id_is_referenced(Type, Id),
	!.

xperson([id-ID,
	 prettyname-NN,
	 furthernames-WN,
	 birthdate-DG,
	 deathdate-DT,
	 links-LS]) :-
	enum_xperson(ID),
	xperson_prettyname(ID, NN),
	xperson_furthernames(ID, WN),
	xperson_birth_normdate(ID, DG),
	xperson_death_normdate(ID, DT),
	xperson_links(ID, LS).

person_sortname(X, N) :-
	memberchk(prettyname-N1, X),
	atom_sortkey(N1, N).

atom_sortkey(X, Y) :-
	atom_to_ascii(X, X1),
	downcase_atom(X1, Y).

sort_plist(L, Ks, L1) :-
	map_add_keyvals(L, Ks, L2),
	keysort(L2, L3),
	map_val(L3, L1).

add_keyvals(L, Ks, Vs-L) :-
	get_vals_or_call(L, Ks, z(_,_,_,_), Vs).

map_add_keyvals([X|Xs], Ks, [X1|Xs1]) :-
	add_keyvals(X, Ks, X1),
	map_add_keyvals(Xs, Ks, Xs1).
map_add_keyvals([], _, []).

get_val(L, K, V) :-
	( memberchk(K-V, L) ->
	  true
	; err('Unspecified value: ~q in ~q', [K, L])
	).

get_val(L, K, D, V) :-
	( memberchk(K-V, L) ->
	  true
	; V = D
	).
			
get_vals_or_call(L, [K|Ks], D, [V|Vs]) :-
	!,
	( memberchk(K-V, L) ->
	  true
	; Call =.. [K,L,V], call(Call) ->
	  true
	; V = D
	),
	get_vals_or_call(L, Ks, D, Vs).
get_vals_or_call(_, [], _, []).
		  
map_val([_-X|Xs], [X|Xs1]) :-
	map_val(Xs, Xs1).
map_val([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

navi(Self, Type, Navi, ContentMark) :-
	maintitle(Maintitle),
	( html_options(Options) -> true ; Options = [] ),
	EMaintitle = e(b, [], [Maintitle, ': ']),
	navi_item(Self, haupseite, 'Hauptseite',
		  'index.html', NMain),
	navi_item(Self, letters, 'Briefe', 'letters.html', NLet),
	navi_item(Self, persons, 'Personen/Werke', 'persons.html', NPer),
	navi_item(Self, persons_by_refs, 'Personen/Werke nach Verweisen',
		  'persons_by_refs.html', NRPer),
	navi_item(Self, geos, 'Orte', 'geos.html', NGeo),
	navi_item(Self, geos_by_refs, 'Orte nach Verweisen',
		  'geos_by_refs.html', NRGeo),
	navi_item(Self, events, 'Ereignisse', 'events.html', NEve),
	navi_item(Self, journals, 'Periodika', 'journals.html', NJou),
	navi_item(Self, subjects, 'Sachregister', 'subjects.html', NSub),
	navi_item(Self, biblio, 'Bibliographie', 'biblio.html', NBib),
	navi_item(Self, corpors, 'Körperschaften', 'corpors.html', NCor),
%	navi_item(Self, drawings, 'Zeichnungen', 'drawings.html', N6),
	( Type = top, memberchk(search_action=Action, Options) ->
	  SearchBar=searchbarTop,
	  Search = e(div, [class=SearchBar],
		    [e(form,
		       [class=SearchBar,
			method=get,
			action=Action],
		       [e(input, [class=SearchBar,
				  type=text, size='15', name=q, value=''], []),
			e(input, [class=SearchBar,
				  type=submit, value='Suche'], [])])]),
	  Navi1 = [Search|Navi2]
	; Navi1 = Navi2
	),
 	Navi = e(div, [], Navi1),
	Navi2 =	[e(div, [class=naviline1],
		   [EMaintitle, NMain,
		    ' | ', NLet, ' | ', NPer]),	   
		 e(div, [class=naviline2],
		   [NRPer, ' | ', NCor, ' | ', NEve, ' | ', NGeo]),
		 e(div, [class=naviline3],
		   [NRGeo, ' | ', NSub, ' | ', NJou, ' | ',  NBib])],
	( Type = top ->
	  ContentMark = xml('\n<!--BEGIN-CONTENT-->\n')
	; Type = bot ->
	  ContentMark = xml('\n<!--END-CONTENT-->\n')
	; ContentMark = ''
	).
	
% map_navi_item([X|Xs], Y1, [X1|Xs1]) :-
% 	navi_item(X, Y1, X1),
% 	map_navi_item(Xs, Y1, Xs1).
% map_navi_item([], _, []).

navi_item(Self, Self, Pretty, _, Pretty) :-
	!.
navi_item(_, _, Pretty, Link, e(a, [class=nav, href=Link], [Pretty])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_person_page(P) :-
	get_val(P, id, Id),
	navi(person, top, NaviTop, CB),
	navi(person, bot, NaviBot, CE),
	get_val(P, prettyname, Name),
	get_val(P, birthdate, DG1),
	normdate_year(DG1, YG),
	get_val(P, deathdate, DT1),
	normdate_year(DT1, YT),
	( void_dates([YG,YT]) ->  D = []
	; D = [' (', YG, xml('&ndash;'), YT, ')']
	),
	E1 = e(h1, [], [Name | D]),
	Body = [NaviTop,
		e(hr, [class=navitop], []), CB, E1 | Rest0],
	( id_topimage(Id, TopImage) ->
	  Rest0 = [TopImage | Rest01]
	; Rest0 = Rest01
	),
	person_page_info(P, PersonInfo),
	xperson_page_persontext(Id, PersonText),
	( PersonText = '' ->
	  Rest01 = [PersonInfo|Rest3]
	; Rest01 = [PersonInfo, e(div, [class=ptext], PersonText) |Rest3]
	),
	letters_to(Id, BriefeAn),
	letters_recvd(Id, BriefeEmpf),
	append(BriefeAn, BriefeEmpf, BriefeAE),
	referring_letters_section(Id, person, Verweise),
	clear(CLEAR),
	append([CLEAR|BriefeAE],
	       Verweise, Weiteres),
	append(Weiteres, Rest4, Rest3),
	person_works_section(Id, Rest4, Rest4a),		
	id_images(Id, Images),
	append(Images, Rest5, Rest4a),
	Rest5 = [CE, e(hr, [class=navibot], []), NaviBot],
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [],
			[Maintitle, ': ', Name | D ])]),
		   Body1])],
	mk_body(Body, Body1),
	id_url(person, Id, File),
	write_page(Page, File).

xperson_page_persontext(P, PersonText) :-
	xperson_text(P, PersonText),
	!.
xperson_page_persontext(_, '').

person_page_info(P, E) :-
	E = e(div, [class=pinfo], Es),
	get_val(P, birthdate, DG1),
	normdate_pretty(DG1, DG),
	get_val(P, deathdate, DT1),
	normdate_pretty(DT1, DT),
	( void_dates([DG,DT]) ->  LD = ['unbekannt']
	; LD = [DG, xml('&ndash;'), DT]
	),
	Es = [e(div, [class=pinfoline], ['Lebensdaten: ' | LD]) | Es1],
	xperson_furthernames_pretty(P, WeitereNamen),
	( WeitereNamen = [] ->
	  Es1 = Es2
	; Es1 = [e(div, [class=pinfoline],
		   ['Weitere Namen: ' | WeitereNamen])
		| Es2]
	),
        entity_links(P, enav, Links),
	( Links = [] ->
	  Es2 = Es3 
	; Es2 = [e(div, [class=pinfoline], ['Links: ' | Links])
		| Es3]
	),
	Es3 = [].

%%%% 
%%%% The chainlink should open a new window or re-use the existing window
%%%% "kbsetchainlist". To avoid opening in a new tab it seems that
%%%% "menubar=no" is required for Chromium. Firefox seems not to depend on
%%%% this.
chainlink_window_params('left=0,top=0,width=800,height=400,menubar=no,toolbar=yes,location=yes,personalbar=yes,status=yes,resizable=yes,scrollbars=yes').
%%%%
chainlink(Id, Type, E) :-
	chain_url(Type, Id, URL),
	atomic_list_concat([URL, '#i1'], URL1),
	%% AATTS = [href=URL1, target='_blank'],
	chainlink_window_params(WindowParams),
	format(atom(Open),
	       'var w=window.open(\'~w\',\'kbsetchainlist\',\'~w\'); w.focus(); return false;',
	       [URL1, WindowParams]),
	AATTS = [class=chainlink, onclick=Open],
	E = e(p, [class=chainlink],
	      [e(a, AATTS, [xml('&rarr;'),
			    ' Verweise als Kette'])]).

referring_letters_section(Id, Type, H) :-
	referring_letters(Id, Type, [], H1),
	( H1 = []  -> H = H1
	; ( chain_page_is_useful(Id, Type) ->
	    chainlink(Id, Type, E),
	    H2 = [E|H1]
	  ; H2 = H1
	  ),
	  H =  [e(h2, [], ['Verweise']) | H2]
	).

pid_number_of_letters(Id, E) :-
	findall(V, letter(V, from, Id), Vs),
	sort(Vs, Vs1),
	length(Vs1, NV),
	findall(A, letter(A, to, Id), As),
	sort(As, As1),
	length(As1, NA),
	( xperson_gender(Id, w) ->
	  Verfasser = 'Verfasserin',
	  Empfaenger = 'Empfängerin'
	; % xperson_gender(Id, m) ->
	  Verfasser = 'Verfasser',
	  Empfaenger = 'Empfänger'
	),
	( NV = 0 ->
	  TV = ''
	; NV = 1 ->
	  format(atom(TV), '~w von 1 Brief', [Verfasser])
	; format(atom(TV), '~w von ~d Briefen', [Verfasser, NV])
	),
	( NA = 0 ->
	  TA = ''
	; NA = 1 ->
	  format(atom(TA), '~w von 1 Brief', [Empfaenger])
	; format(atom(TA), '~w von ~d Briefen', [Empfaenger, NA])
	),
	( TV = '' ->
	  ( TA = '' -> E = []
	  ; E = [TA]
	  )
	; ( TA = '' -> E = [TV]
	  ; E = [TV, ', ', TA]
	  )
	).


sort_by_number_of_letters(Ps, Ps1) :-
	reverse(Ps, Ps2),
	map_add_nol(Ps2, Ps3),
	keysort(Ps3, Ps4),
	map_val(Ps4, Ps5),
	reverse(Ps5, Ps1).

map_add_nol([X|Xs], [X1|Xs1]) :-
	add_nol(X, X1),
	map_add_nol(Xs, Xs1).
map_add_nol([], []).

add_nol(P, N-P) :-
	get_val(P, id, P1),
	findall(L, (letter(L, from, P1) ; letter(L, to, P1)), Ls),
	sort(Ls, Ls1),
	length(Ls1, N).

sort_by_number_of_references(Ps, Type, Ps1) :-
	reverse(Ps, Ps2),
	map_add_nor(Ps2, Type, Ps3),
	keysort(Ps3, Ps4),
	map_val(Ps4, Ps5),
	reverse(Ps5, Ps1).

map_add_nor([X|Xs], Type, [X1|Xs1]) :-
	add_nor(X, Type, X1),
	map_add_nor(Xs, Type, Xs1).
map_add_nor([], _, []).

add_nor(P, Type, N-P) :-
	get_val(P, id, Id),
	nor_ids(Id, Type, Ids),
	findall(B, ( member(Id1, Ids),
		     letter_revlink(Id1, Type, B) ),
		Ls),
	sort(Ls, Ls1),
	length(Ls1, N).

nor_ids(Id, geo, [Id|SubIds]) :-
 	!,
 	geo_subgeos(Id, SubIds).
nor_ids(Id, _, [Id]).

id_stats(Id, Type, [Text]) :-
	nor_ids(Id, Type, Ids),
	findall(B, ( member(Id1, Ids),
		     letter_revlink(Id1, Type, B)
		   ),
		Ls),
	sort(Ls, Ls1),
	length(Ls1, N),
	( N = 0 ->
	  Text = xml('&mdash;')
	; N = 1 ->
	  Text = ' 1 Verweis'
	; format(atom(Text), ' ~w Verweise', [N])
	).

pid_stats(Id, E) :- id_stats(Id, person, E).
gid_stats(Id, E) :- id_stats(Id, geo, E).

referring_letters(Id, Type, _, []) :-
	\+ letter_revlink(Id, Type, _),
	!.
referring_letters(Id, Type, Options, Vs) :-
	findall((D-V-A)-(B-Options1),
		( letter_revlink(Id, Type, B),
		  letter(B, date, D1),
		  letter(B, from, V),
		  letter(B, to, A),
		  ddate_sortkey(D1, D),
		  ( letter_revlink(Id, main, B) ->
		    Options1 = [main|Options]
		  ; Options1 = Options
		  )),
		Bs),
	sort(Bs, Bs1),
	map_val(Bs1, Bs2),
	( memberchk(just_id_options_pairs, Options) ->
	  Vs = Bs2
	; map_letterlink_global(Bs2, Vs)
	).

map_letterlink_global([X], [X1]) :-
	!,
	letterlink_global(X, X1).
map_letterlink_global([X|Xs], [X1, e(br, [], []) |Xs1]) :-
	letterlink_global(X, X1),
	map_letterlink_global(Xs, Xs1).
map_letterlink_global([], []).

letters_to(PID, []) :-
	\+ letter(_, from, PID),
	!.
letters_to(PID, [H|BriefeAn]) :-
	H = e(h2, [], ['Briefe an']),
	findall(A1-J1, ( letter(B1, from, PID),
			 letter(B1, to, A1),
			 letter(B1, date, D1),
			 ddate_pretty_year(D1, J1)),
		AJs1),
	sort(AJs1, AJs2),
	setof(A3-B3, setof(B4, (member(A3-B4, AJs2)), B3), AJs3),
	sort_br_ans(AJs3, AJs4),
	map_br_an(AJs4, PID, BriefeAn).

letters_recvd(PID, []) :-
	\+ letter(_, to, PID),
	!.
letters_recvd(PID, [H|BriefeAn]) :-
	H = e(h2, [], ['Empfangene Briefe von']),
	findall(A1-J1, ( letter(B1, to, PID),
			 letter(B1, from, A1),
			 letter(B1, date, D1),
			 ddate_pretty_year(D1, J1)),
		AJs1),
	sort(AJs1, AJs2),
	setof(A3-B3, setof(B4, (member(A3-B4, AJs2)), B3), AJs3),
	sort_br_ans(AJs3, AJs4),
	map_br_empf(AJs4, PID, BriefeAn).


sort_br_ans(AJs, AJs1) :-
	map_add_kv_pidkeys(AJs, AJs2),
	keysort(AJs2, AJs3),
	map_val(AJs3, AJs1).

add_kv_pidkeys(PID-V, K-(PID-V)) :-
	xperson_prettyname(PID, NN),
	xperson_birth_normdate(PID, DD),
	K = k(NN,DD).

map_add_kv_pidkeys([X|Xs], [X1|Xs1]) :-
	add_kv_pidkeys(X, X1),
	map_add_kv_pidkeys(Xs, Xs1).
map_add_kv_pidkeys([], []).

br_an(PID-Jahre, VPID, E) :-
	pid_ref(PID, P),
	E = [e(h3, [], [P]) | Jahre1],
	map_br_an_2(Jahre, PID, VPID, Jahre1).

map_br_an([X|Xs], Y1, Xs1) :-
	br_an(X, Y1, X2),
	append(X2, Xs2, Xs1),  
	map_br_an(Xs, Y1, Xs2).
map_br_an([], _, []).

br_an_2(J, A, V, e(a, [class=enav, href=URL], [J])) :-
	reg_url(V, A, J, URL).

map_br_an_2([X], Y1, Y2, [X1]) :-
	!,
	br_an_2(X, Y1, Y2, X1).
map_br_an_2([X|Xs], Y1, Y2, [X1, ' | ' |Xs1]) :-
	br_an_2(X, Y1, Y2, X1),
	map_br_an_2(Xs, Y1, Y2, Xs1).
map_br_an_2([], _, _, []).

br_empf(PID-Jahre, VPID, E) :-
	pid_ref(PID, P),
	E = [e(h3, [], [P]) | Jahre1],
	map_br_empf_2(Jahre, PID, VPID, Jahre1).

map_br_empf([X|Xs], Y1, Xs1) :-
	br_empf(X, Y1, X2),
	append(X2, Xs2, Xs1),  
	map_br_empf(Xs, Y1, Xs2).
map_br_empf([], _, []).

br_empf_2(J, A, V, e(a, [class=enav, href=URL], [J])) :-
	reg_url(A, V, J, URL).

map_br_empf_2([X], Y1, Y2, [X1]) :-
	!,
	br_empf_2(X, Y1, Y2, X1).
map_br_empf_2([X|Xs], Y1, Y2, [X1, ' | ' |Xs1]) :-
	br_empf_2(X, Y1, Y2, X1),
	map_br_empf_2(Xs, Y1, Y2, Xs1).
map_br_empf_2([], _, _, []).

xperson_furthernames_pretty(P, Names) :-
	get_val(P, furthernames, Names1),
	map_insert_separator(Names1, '; ', Names).

map_insert_separator([X], _, [X]) :- !.
map_insert_separator([X|Xs], S, [X,S|Xs1]) :-
	map_insert_separator(Xs, S, Xs1).
map_insert_separator([], _, []).


% 	map_pretty_name(Namen1, namen, Namen).
% 
% pretty_name(birthname-Name, _Class, 'geb. ', Name) :-
% 	!.
% pretty_name(altName-Name, _Class, 'geb. ', Name) :-
% 	!.
% pretty_name(pseudonym-P, _Class, 'pseud. ', Name) :-
% 	!,
% 	person_ref_nodates(P, Name).
% pretty_name(pseudonym_of-P, _Class, 'eigentl. ', Name) :-
% 	!,
% 	person_ref(P, Name).
% pretty_name(_-Name, _Class, 'auch: ', Name).
% % pretty_name(eigentlich-Name, _Class, 'eigentlich: ', Name) :-
% % 	!.
% % pretty_name(auch-Name, _Class, 'auch: ', Name) :-
% % 	!.
% % pretty_name(spitz-Name, _Class, 'Kosename ', Name) :-
% % 	!.
% 
% 
% omit_alt_name_key(kuerzel).
% omit_alt_name_key(index).
% 
% map_pretty_name([K-_|Xs], Y1, Xs1) :-
% 	omit_alt_name_key(K),
% 	!,
% 	map_pretty_name(Xs, Y1, Xs1).
% map_pretty_name([X], Y1, [A, B]) :-
% 	!,
% 	pretty_name(X, Y1, A, B).
% map_pretty_name([X|Xs], Y1, [A, B, e(br,[],[]) |Xs1]) :-
% 	pretty_name(X, Y1, A, B),
% 	map_pretty_name(Xs, Y1, Xs1).
% map_pretty_name([], _, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

docstring('<!DOCTYPE html>').
% docstring('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">').

write_page(Page, File) :-
	filename(html_output(File), File1),
	info(10, 'Writing: ~w', [File1]),
	date(date(Y,M,D)),
	date_w3c(ymd(Y,M,D), Date),
	format(atom(Note), '<!-- Generated with KBSET, ~w -->', [Date]),
	docstring(Docstring),
 	( writeln(Docstring),
	  writeln(Note),
 	  write_structure(Page, [space(html), style(html4), encoding(ascii)]))
	onto_file File1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

map_person_ref([X], Options, [X1]) :-
	!,
	person_ref(X, Options, X1).
map_person_ref([X|Xs], Options, Xs1) :-
	( memberchk(br, Options) ->
	  Xs1 = [X1, e(br,[],[]) |Xs2]
	; Xs1 = [X1|Xs2]
	),
	person_ref(X, Options, X1),
	map_person_ref(Xs, Options, Xs2).
map_person_ref([], _, []).

map_person_ref_commasep([X], [X1]) :-
	!,
	person_ref(X, [class=pinline], X1).
map_person_ref_commasep([X|Xs], [X1, ', ' | Xs1]) :-
	person_ref(X, [class=pinline], X1),
	map_person_ref_commasep(Xs, Xs1).
map_person_ref_commasep([], []).

% map_person_ref_index([X], [X1]) :-
% 	!,
% 	person_ref_index(X, X1).
% map_person_ref_index([X|Xs], [X1, e(br,[],[]) |Xs1]) :-
% 	person_ref_index(X, X1),
% 	map_person_ref_index(Xs, Xs1).
% map_person_ref_index([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

kitem_link(URL, URL).

wiki_en_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/wiki_en/', URL, '.htm'], URL1).
wiki_en_link(URL, URL1) :-
 	atomic_list_concat(['http://en.wikipedia.org/wiki/', URL], URL1).

wiki_fr_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/wiki_fr/', URL, '.htm'], URL1).
wiki_fr_link(URL, URL1) :-
 	atomic_list_concat(['http://fr.wikipedia.org/wiki/', URL], URL1).

wiki_de_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/wiki_de/', URL, '.htm'], URL1).
wiki_de_link(URL, URL1) :-
	( sub_atom(URL, 0, L, _, 'wikiDE:') ->
	  sub_atom(URL, L, _, 0, URL2)
	; URL2 = URL
	),
 	atomic_list_concat(['http://de.wikipedia.org/wiki/', URL2], URL1).

geohack_de_link(URL, URL1) :-
 	atomic_list_concat(['https://tools.wmflabs.org/geohack/geohack.php?language=de&',
		     URL],
		    URL1).

ndb_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/', URL, '.html'], URL1).
ndb_link(URL, URL1) :-
 	atomic_list_concat(['http://www.deutsche-biographie.de/', URL, '.html'], URL1).

gnd_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/', URL, '.html'], URL1).
gnd_link(URL, URL1) :-
	( sub_atom(URL, 0, L, _, 'gnd:') ->
	  sub_atom(URL, L, _, 0, URL2)
	; URL2 = URL1
	),
 	atomic_list_concat(['http://d-nb.info/gnd/', URL2], URL1).

wwp_link(URL, URL1) :-
	flag(offline, true, true),
	!,
	atomic_list_concat(['../webcache/P=', URL, '.html'], URL1).
wwp_link(URL, URL1) :-
	atomic_list_concat(['http://wwperson.informatik.uni-erlangen.de/cgi-bin/llmult/LANG=germ/P=', URL], URL1).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lid_title(LID, Title) :-
	letter(LID, id, Title).

lids(LIDs) :-
	findall(LID, letter(LID, _, _), LIDs1),
	sort(LIDs1, LIDs).

write_letter_page(LID) :-
	info(10, 'Processing letter ~q', [LID]),
	set_global_context_lid(LID),
	findall(P-V, letter(LID, P, V), B),
	get_val(B, from, Von),
	get_val(B, to, An),
	get_val(B, place, 'Unbekannt', Place0),
	get_val(B, geo_hinweis, '', PlaceHinweis),
	get_val(B, date, Datum),
	get_val(B, date_hinweis, '', DatumHinweis),
	get_val(B, text, '', Text),
	get_val(B, annot, '', Annot),
	ddate_pretty_from(Datum, VOM),
	xperson_prettyname(Von, VonPrettyName),
	xperson_prettyname(An, AnPrettyName),
	short_person_name(VonPrettyName, VonShortName),
	short_person_name(AnPrettyName, AnShortName),
%	ddate_pretty_year(Datum, Jahr),
	Es = [NaviTop,
	      e(hr, [class=navitop], []),
	      LetterNavi,
	      CB,
	      e(h1, [], ['Brief ', VOM, Datum1, ', von ', VonShortName,
			 ' an ', AnShortName]),
	      e(div, [class=annotContainer],
		[e(div, [class=annotItemMain],
		   [ e(div, [class=annotmeta], ['Von: ', Von1]),
		     e(div, [class=annotmeta], ['An: ', An1]),
		     e(div, [class=annotmeta], ['Ort: ', Place1]),
		     e(div, [class=annotmeta], ['Datum: ', Datum1])
		   | EsBrief ]),
		 e(div, [class=annotItemSpace], []),
		 e(div, [class=annotItemAnnot], EsAnnot)])
	     | Rest1],
	letternavi(LID, LetterNavi, []),
	( Text = '' ->
	  EsBrief = []
	; nlx_to_h(Text, HText),
	  DIVText = e(div, [class=xdivmetaletter], []),
	  EsBrief = [DIVText|HText]
	),
	letter_xmeta_after(LID, XMeta),
	( Annot == '' ->
	  EsAnnot = XMeta
	; nlx_to_h(Annot, Annot1),
	  %% HAnnot = e(h2, [class=xhannotation], ['Kommentar']),
	  %% append([HAnnot | Annot1], Rest2, Rest1)
	  append(Annot1, XMeta, EsAnnot)
	),
	Rest1 = Rest2,
	letter_images(LID, Abb),
	( Abb = [] ->
	  Rest2 = Rest3
	; Abb = [_] ->
	  append([e(h2, [], ['Autograph']) | Abb], Rest3, Rest2)
	; Abb = [_, _ |_] ->
	  append([e(h2, [], ['Autographen']) | Abb], Rest3, Rest2)
	),
	Rest3 =	[CE,
		 e(hr, [class=navibot], []),
		 LetterNavi,
		 e(hr, [class=navibot], []),
		 NaviBot],
	pid_ref(Von, Von1),
	pid_ref(An, An1),

	geo_sortkey(Place0, Place),
	gid_ref(Place, Place0, Place2),
	add_hinweis(Place2, PlaceHinweis, Place1),

	ddate_pretty(Datum, Datum2),
	add_hinweis(Datum2, DatumHinweis, Datum1),
	navi(letter, top, NaviTop, CB),
	navi(letter, bot, NaviBot, CE),
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle, ': ', 'Brief ', VOM, Datum1,
				    ', von ', VonShortName,
				    ' an ', AnShortName])]),
		   Body1])],
        mk_body(Es, Body1),
	id_url(letter, LID, URL),
	%% lid_title(LID, Title),
	write_page(Page, URL).

add_hinweis(X, '', X) :-
	!.
add_hinweis(X, H, e(span, [], [X, ' [', xml(H1), ']'])) :-
	hinweis_to_html(H, H1).

letter_images(B, Abbs) :-
	findall(Prio-File, aut(B, _Typ, Prio, File, _), Abbs1),
	keysort(Abbs1, Abbs2),
	map_val(Abbs2, Abbs3),
	map_aut(Abbs3, Abbs).

map_aut([A|As], [E|As1]) :-
	E = e(table, [border=0, class=autograph],
	      [e(colgroup, [span=1], []),
	       e(tr,
		 [class=autograph],
		 [e(td, [class=autograph, align=center],
		    [e(a, [class=enav, href=HiURL],
		       [e(img, [class=autograph,
				src=LoURL, width='100%',
				alt='Autograph'], [])])])]),
	       e(tr,
		 [class=license],
		 [e(td, [class=license, align=right], [LE])])]),
	aut_license(A, License),
	license_elem(License, LE),
	aut_url(A, LoURL, HiURL),
	map_aut(As, As1).
map_aut([], []).

letternavi(LID, LetterNavi, _Options) :-
	once((letter(LID, from, V),
	      letter(LID, to, A),
	      letter(LID, date, Datum))),
	ddate_pretty_year(Datum, Year),
	reg_url(V, A, Year, BVYear),
	letter_from_sequence(V, Seq),
	letter_dialog_sequence(V, A, DSeq),
	( prev_memberchk(LID, Seq, Prev) ->
	  id_url(letter, Prev, PrevURL),
	  NPrev = e(a, [class=nav, href=PrevURL], ['Voriger Brief'])
	; NPrev = 'Voriger Brief'
	),
	( next_memberchk(LID, Seq, Next) ->
	  id_url(letter, Next, NextURL),
	  NNext = e(a, [class=nav, href=NextURL], ['Nächster Brief'])
	; NNext = 'Nächster Brief'
	),
	( html_correspondence_navi(V, A) ->
	  ( prev_memberchk(LID, DSeq, DPrev) ->
	    id_url(letter, DPrev, DPrevURL),
	    DNPrev = e(a, [class=nav, href=DPrevURL],
		       ['Voriger Brief'])
	  ; DNPrev = 'Voriger Brief'
	  ),
	  ( next_memberchk(LID, DSeq, DNext) ->
	    id_url(letter, DNext, DNextURL),
	    DNNext = e(a, [class=nav, href=DNextURL],
		       ['Nächster Brief'])
	  ; DNNext = 'Nächster Brief'
	  ),
	  CorrNavi = [' ', xml('&bull;'),
		      ' Korrespondenz: ', DNPrev, ' | ', DNNext,
		      ' | ',  e(a, [class=nav, href=BVYear], [Year])]
	; CorrNavi = []
	),
	LetterNavi = e(p, [class=navi],
		      ['Autor: ', NPrev, ' | ',
		       NNext | CorrNavi]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

xmeta_prop(transliteration, 'Transkription: ').
xmeta_prop(annotation_author, 'Kommentar: ').
xmeta_prop(status, 'Status: ').

letter_xmeta_after(Id, E) :-
	findall(e(div, [class=xmetaentry], [K|V]),
		( xmeta_prop(Prop, K),
		  letter(Id, Prop, V1),
		  nlx_to_h_inpar(V1, V) ),
		E1),
	( E1 = [] ->
	  E = []
	; E = [e(h3, [class=xksection], ['Bearbeitung'])|E1]
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

letterlink_global(LID-Options, Pretty) :-
	!,
	letterlink_global_1(LID, Options, Pretty).
letterlink_global(LID, Pretty) :-
	letterlink_global_1(LID, [], Pretty).

letterlink_global_1(LID, Options, Pretty) :-
	( memberchk(main, Options) ->
	  Class=gllinkmain
	; Class=gllink
	),
	%% a possibility to export just the values:
	( memberchk(values(URL, LinkE), Options) -> true ; true ),
	LinkE = ['Brief ', Vom, DatumKurz,
		 ' von ', Von1, ' an ', An1],
	( memberchk(element, Options) -> true ; Element=span ),
	Pretty = e(Element, [class=Class],
		   [e(a, [class=enav, href=URL], LinkE)]),
        id_url(letter, LID, URL),
	once(( letter(LID, date, Datum),
	       letter(LID, from, Von),
	       letter(LID, to, An) )),
	pid_ref_nolink_short(Von, Von1),
	pid_ref_nolink_short(An, An1),
% 	( letter(LID, abstract, Kurztext) ->
% 	  Pretty = [DatumKurz, ' ', xml(Kurztext)]
% 	; Pretty = [DatumKurz]
% 	),
	ddate_pretty_from(Datum, Vom),
	ddate_pretty_short(Datum, DatumKurz).
%	date_kurz(Datum, DatumKurz).

letterlink_for_year(LID, Y, E) :-
	id_url(letter, LID, URL),
	letter(LID, date, Date),
	ddate_pretty_daymonth(Date, DayMonth),
	Pretty = [DayMonth,Y],
	( letter(LID, abstract, Kurztext) ->
	  E = e(span, [], [e(a, [class=enav, href=URL], Pretty),
			   ' ', xml(Kurztext)])
	; E = e(a, [class=enav, href=URL], Pretty)
	).

% Kurztext (abstract) may contain links!

map_letterlink_for_year([X], Y, [X1]) :-
	!,
	letterlink_for_year(X, Y, X1).
map_letterlink_for_year([X|Xs], Y, [X1, e(br, [], []) |Xs1]) :-
	letterlink_for_year(X, Y, X1),
	map_letterlink_for_year(Xs, Y, Xs1).
map_letterlink_for_year([], _, []).

write_letter_registers :-
	findall(PV-PA-J,
		( letter(B, from, PV),
		  letter(B, to, PA),
		  letter(B, date, D),
		  ddate_pretty_year(D, J)
		),
		VAJs),
	sort(VAJs, VAJs1),
	( member(V-A-J, VAJs1),
	  reg_prev_year(V, A, J, VAJs1, PJ),
	  reg_next_year(V, A, J, VAJs1, NJ),
	  write_letter_register_page(V, A, J, PJ, NJ),
	  fail
	; true
	).

reg_next_year(V, A, J, VAJs, J1) :-
	member(V-A-J1, VAJs),
	J @< J1,
	!.
reg_next_year(_, _, _, _, none).

reg_prev_year(V, A, J, VAJs, J1) :-
	reverse(VAJs, VAJs1),
	member(V-A-J1, VAJs1),
	J @> J1,
	!.
reg_prev_year(_, _, _, _, none).

letter_fromto_sequence(PIDVon, PIDAn, Seq) :-
	findall(D-B, ( letter(B, from, PIDVon),
		       letter(B, to, PIDAn),
		       letter(B, date, D1),
		       ddate_sortkey(D1, D)),
		DBs),
	keysort(DBs, DBs1),
	map_val(DBs1, Seq).


letter_from_sequence(PIDVon, Seq) :-
	findall(D-B, ( letter(B, from, PIDVon),
		       letter(B, date, D1),
		       ddate_sortkey(D1, D)),
		DBs),
	keysort(DBs, DBs1),
	map_val(DBs1, Seq).

letter_dialog_sequence(PIDVon, PIDAn, Seq) :-
	( PIDVon @< PIDAn ->
	  PID1 = PIDVon, PID2 = PIDAn
	; PID1 = PIDAn, PID2 = PIDVon
	),
	findall(D-B, ( letter(B, from, PID1),
		       letter(B, to, PID2),
		       letter(B, date, D1),
		       ddate_sortkey(D1, D)),
		DBs),
	findall(D1-B1, ( letter(B1, from, PID2),
			 letter(B1, to, PID1),
			 letter(B1, date, D2),
			 ddate_sortkey(D2, D1)),
		DBs1),
	append(DBs, DBs1, DBs2),
	keysort(DBs2, DBs3),
	map_val(DBs3, Seq).

prev_memberchk(Y, [X,Y|_], X) :-
	!.
prev_memberchk(Y, [_|Xs], Z) :-
	prev_memberchk(Y, Xs, Z).

next_memberchk(X, [X,Y|_], Y) :-
	!.
next_memberchk(Y, [_|Xs], Z) :-
	next_memberchk(Y, Xs, Z).

write_letter_register_page(PIDVon, PIDAn, Year, PrevYear, NextYear) :-
	navi(bv, top, NaviTop, CB),
	navi(bv, bot, NaviBot, CE),
	findall(Datum-LID, ( letter(LID, from, PIDVon),
			     letter(LID, to, PIDAn),
			     letter(LID, date, Datum1),
			     ddate_pretty_year(Datum1, Year),
			     ddate_sortkey(Datum1, Datum) ),
		LIDs),
	keysort(LIDs, LIDs1),
	map_val(LIDs1, LIDs2),
	map_letterlink_for_year(LIDs2, Year, EBriefe),
	BR = e(br, [], []),
	pid_shortname(PIDVon, FromShort),
	pid_shortname(PIDAn, ToShort),
	Title = ['Briefverzeichnis ', FromShort, ' an ', ToShort, ', ', Year],
	E1 = e(h1, [], Title),
	pid_ref(PIDVon, PVon),
	pid_ref(PIDAn, PAn),
	E2 = e(p, [], ['Von ', PVon, BR, 'An ', PAn]), 
	( PrevYear \= none ->
	  reg_url(PIDVon, PIDAn, PrevYear, PrevURL),
	  PREV = e(a, [class=nav, href=PrevURL], [PrevYear])
	; PREV = ''
	),
	( NextYear \= none ->
	  reg_url(PIDVon, PIDAn, NextYear, NextURL),
	  NEXT = e(a, [class=nav, href=NextURL], [NextYear])
	; NEXT = ''
	),
	( PREV = '', NEXT = '' ->
	  E3 = e(p, [class=navi], [])
	; E3 = e(p, [class=navi], [PREV, ' | ', NEXT])
	),
	append([NaviTop,
		e(hr, [class=navitop], []),
		CB,
		E3,
		E1,
	        E2,
	        e(p, [], EBriefe)], Rest1, Body),
	Rest1 = [CE, e(hr, [class=navibot], []), NaviBot],
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [],
			[Maintitle, ': '| Title])]),
		   Body1])],
	mk_body(Body, Body1),
	reg_url(PIDVon, PIDAn, Year, File),
	write_page(Page, File).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

geos(Os) :-
 	findall(O, (geo(O), is_referenced(geo,O)), Os1),
 	sort_plist(Os1, [name], Os).

geo([id-ID | Props]) :-
	enum_geo(ID),
	\+ geo_supergeo(ID, _),
	geo_props(ID, Props).

geo_props(ID, [name-NN, furthernames-WN, links-LS | Rest]) :-
	geo_ename(ID, NN),
	geo_furthernames(ID, WN),
	geo_links(ID, LS),
	( geo_supergeo(ID, _) ->
	  Rest = []
	; geo_subgeos(ID, Subs),
	  findall(ID1, (member(ID1, Subs), id_is_referenced(geo,ID1)), Subs0),
	  map_geo_props(Subs0, Subs1),
	  sort_plist(Subs1, [name], Subs2),
	  Rest = [subs-Subs2]
	).
	
map_geo_props([X|Xs], [[id-X|X1]|Xs1]) :-
	geo_props(X, X1),
	map_geo_props(Xs, Xs1).
map_geo_props([], []).

write_geo_page(O) :-
	get_val(O, id, Id),
	navi(geo, top, NaviTop, CB),
	navi(geo, bot, NaviBot, CE),
	get_val(O, name, Name),
	BR = e(br, [], []),
	E1 = e(h1, [], [Name]),
	Body = [NaviTop,
		e(hr, [class=navitop], []), CB,
		E1 | Rest01 ],
	( id_topimage(Id, TopImage) ->
	  Rest01 = [TopImage | Rest1]
	; Rest01 = Rest1
	),
	geo_furthernames_pretty(O, WeitereNamen),
	( WeitereNamen = [] ->
	  Rest1 = Rest2
	; append([BR | WeitereNamen], Rest2, Rest1)
	),
        entity_links(O, enav, Links),
	( Links = [] ->
	  Rest2 = Rest3
	; Rest2 = [e(p, [], ['Links: ' | Links]) | Rest3]
	),
	id_images(Id, Images),
	append(Images, Rest4, Rest3),
	referring_letters_section(Id, geo, Verweise),
	clear(CLEAR),
	geo_subgeos_section(Id, Subgeos, Rest5),
	append([CLEAR|Verweise], Subgeos, Rest4),
	Rest5 = [CE, e(hr, [class=navibot], []), NaviBot],
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [],
			[Maintitle, ': ', Name])]),
		   Body1])],
	mk_body(Body, Body1),
	id_url(geo, Id, File),
	write_page(Page, File).

geo_subgeos_section(G, H, HRest) :-
	geo_subgeos(G, SubGeos),
	findall(N-G1, ( member(G1, SubGeos),
			geo_ename(G1, N1),
			atom_sortkey(N1, N)
		      ),
		SubGeos1),
	sort(SubGeos1, SubGeos2),
	map_val(SubGeos2, SubGeos3),
	( SubGeos3 = [] ->
	  H = HRest
	; mk_hsubgeos(SubGeos3, G, HSubGeos),
	  H = [ e(h2, [], ['Untergeordnete Orte']),
		e(dl, [class=subgeos], HSubGeos) | HRest]
	).

mk_hsubgeos(SubGeos, G, H) :-
	findall(H1,
		( member(G1, SubGeos),
		  id_is_referenced(geo, G1),
		  id_anchor(geo, G1, Anchor),
		  geo_ename(G1, Name),
		  referring_letters(G1, geo, [], Ls),
		  ( Ls = [] ->
		    info(20, 'WARNING: Geo not referenced in any letter: ~q ~q ~w', [G1, G, Name])
		  ; true
		  ),
		  ( chain_page_is_useful(G1, geo) ->
		    chainlink(G1, geo, E),
		    Ls1 = [E|Ls]
		  ; Ls1 = Ls
		  ),
		  atom_concat('#', Anchor, TgtAnchor),
		  ( H1 = e(dt, [class=geosubs, id=Anchor],
			   [e(a, [class=anchor, href=TgtAnchor], [Name])])
		  ; H1 = e(dd, [class=geosubs], Ls1)
		  )
		),
		H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


id_topimage(Id, E) :-
	img(Id, top, File, Alt, License),
	!,
	img_url(File, URL),
	license_elem(License, LE),
	E = e(div, [class=topimg],
	      [e(img, [class=topimgpic, src=URL, alt=Alt, title=Alt], []),
	       e(span, [class=topimglic], [LE])]).

id_images(Id, Es) :-
	findall(Sortkey-img(File, Alt, License),
		( img(Id, Sortkey, File, Alt, License), Sortkey \= top ),
		Imgs),
	keysort(Imgs, Imgs1),
	map_val(Imgs1, Imgs2),
	map_img_elem(Imgs2, Es1),
	( Es1 = [] ->
	  Es = []
	; clear(CLEAR),
	  Es = [CLEAR,
		e(h2, [], ['Abbildungen']) | Es1]
	).

map_img_elem([X|Xs], [X1|Xs1]) :-
	img_elem(X, X1),
	map_img_elem(Xs, Xs1).
map_img_elem([], []).

img_elem(img(File, Caption, License), E) :-
	E = e(table, [border=0, class=abbildung],
	      [e(colgroup, [span=1], []),
	       e(tr,
		 [class=abbildung],
		 [e(td, [class=abbildung, align=center],
		    [e(img, [class=abbildung,
			     src=URL, alt='Abbildung', title=''], [])])]),
	       e(tr,
		 [class=license],
		 [e(td, [class=license, align=right], [LE])]),
	       e(tr,
		 [class=caption],
		 [e(td, [class=caption, align=center], [Caption])])]),
	img_url(File, URL),
	license_elem(License, LE).

license_elem(L, E) :-
	( L = pd ->
	  E = 'Bild: Public domain'
	; L = jk ->
	  E = 'Bild: J. Kittelmann'
	; L = zbz ->
	  E = 'Scan: Zentralbibliothek Zürich'
	; L = pd(URL) ->
	  E = e(a, [href=URL, class=license], ['Bild: Public domain'])
	; atomic_list_concat(['Bild: ', L], E)
	).

geo_furthernames_pretty(O, Namen) :-
	get_val(O, furthernames, Namen1),
	map_pretty_geoname(Namen1, namen, Namen).

pretty_geoname(pol-Name, _Class, 'heutiger polnischer Name: ', Name) :-
	!.
pretty_geoname(_-Name, _Class, 'weiterer Name: ', Name).

map_pretty_geoname([X], Y1, [A, B]) :-
	!,
	pretty_geoname(X, Y1, A, B).
map_pretty_geoname([X|Xs], Y1, [A, B, e(br,[],[]) |Xs1]) :-
	pretty_geoname(X, Y1, A, B),
	map_pretty_geoname(Xs, Y1, Xs1).
map_pretty_geoname([], _, []).

gid_ref(z(ohne_ort), _, 'ohne Ort') :-
	!.
gid_ref(ohne_ort, _, 'ohne Ort') :-
	!.
gid_ref(Id, _, E) :-
	gid_geo(Id, O),
	!,
	geo_ref(O, E).
gid_ref(_, geo(Name), Name) :-
	!.
gid_ref(Id, _, Id).
%	gid_ref(X, e(a, [href='ort'], [X])),

geo_ref(O, E) :-
	geo_ref(O, [], E).
	
geo_ref(O, Options, E) :-
	get_val(O, id, Id),
	get_val(O, name, Name),
	id_url(geo, Id, URL),
	id_anchor(geo, Id, Id1),
	( memberchk(stats, Options) ->
	  gid_stats(Id, Extras1)
	; Extras1 = []
	),
	( Extras1 = [] ->
	  Extras = []
	; Extras = [e(span, [class=glinfo], Extras1)]
	),
	( memberchk(class=Class, Options) -> true ; Class=ginline ),
	( memberchk(element=Element, Options) -> true ; Element=span ),
	E = e(Element, [class=Class],
	      [ e(a, [class=enav, id=Id1, href=URL], [Name])
	      | Extras ]).

gid_geo(Id, O) :-
	geo(O),
	get_val(O, id, Id1),
	Id1 = Id,
	!.

write_geos_page :-
	write_geos_page([]).

write_geos_by_refs_page :-
	write_geos_page([file='geos_by_refs.html',
			 self='geos_by_refs',
			 sort_by_number_of_references]).

write_geos_page(Options) :-
	geos(Gs0),
	( memberchk(sort_by_number_of_references, Options) ->
	  sort_by_number_of_references(Gs0, geo, Gs)
	; Gs = Gs0
	),
	( memberchk(self=Self, Options) -> true ; Self=geos ),
	map_geo_ref(Gs, [stats,
			 class=gline,
			 with_subs,
			 element=div|Options], Lines),
	Es = [NaviTop,
	      e(hr, [class=navitop], []), CB,
	      e(h1, [], ['Orte']),
	      e(div, [class=xlines], Lines), CE,
	      e(hr, [class=navibot], []),
	      NaviBot],
	navi(Self, top, NaviTop, CB),
	navi(Self, bot, NaviBot, CE),
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle, ': ', 'Orte'])]),
		   Body1])],
	mk_body(Es, Body1),
	( memberchk(file=File, Options) -> true ; File='geos.html' ),
	write_page(Page, File).

		      
map_geo_ref([X|Xs], Options, Xs1) :-
	geo_ref(X, Options, X1),
	( memberchk(with_subs, Options),
	  get_val(X, subs, [], Subs) ->
	  ( memberchk(sort_by_number_of_references, Options) ->
	    sort_by_number_of_references(Subs, geo, Subs1)
	  ; Subs1 = Subs
	  ),
	  map_geo_ref(Subs1, [class=glinesub|Options], SubRefs)
	; SubRefs=[]
	),
	append([X1|SubRefs], Xs2, Xs1),
	map_geo_ref(Xs, Options, Xs2).
map_geo_ref([], _, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% --------------------------------------------------
% von adelheid
% 
%   an lucie
%     1899 | 1990 |
%     
%   an heinrich
% 
%   an pueckler

write_letters_page :-
	findall(PV-PA-J,
		( letter(B, from, PV),
		  letter(B, to, PA),
		  letter(B, date, D),
		  ddate_pretty_year(D, J)
		),
		VAJs),
	sort(VAJs, VAJs1),
	setof(SV-SAJs,
	      setof(SA-SJs,
		    setof(SJ,
			  member(SV-SA-SJ, VAJs1), SJs), SAJs), Grps1),
	keysort_with_pidkeys(Grps1, Grps2),
	map_letters_grp(Grps2, Grps),
	append([NaviTop,
		e(hr, [class=navitop], []), CB,
		e(h1, [], ['Briefe']) | Grps],
	       [CE, e(hr, [class=navibot], []),
		NaviBot],
	       Es),
	navi(letters, top, NaviTop, CB),
	navi(letters, bot, NaviBot, CE),
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle, ': ', 'Briefe'])]),
		   Body1])],
	mk_body(Es, Body1),
	write_page(Page, 'letters.html').

letters_grp(VON-ANs, E) :-
	pid_ref(VON, Person),
	E = e(div, [], [e(h2, [], ['Von ', Person]) | ANs1]),
	keysort_with_pidkeys(ANs, ANs2),
	map_letters_to_grp(ANs2, VON, ANs1).

letters_to_grp(AN-Js, VON, E) :-
	pid_ref(AN, Person),
	E = e(div, [], [e(h3, [], ['An ', Person]) | Js1]),
	sort(Js, Js2),
	map_letters_reg_url(Js2, VON, AN, Js1).

letters_reg_url(X, Y1, Y2, E) :-
	reg_url(Y1, Y2, X, URL),
	E = e(a, [class=enav, href=URL], [X]).

map_letters_reg_url([X], Y1, Y2, [E]) :-
	!,
	letters_reg_url(X, Y1, Y2, E).
map_letters_reg_url([X|Xs], Y1, Y2, [E, ' | ' |Xs1]) :-
	letters_reg_url(X, Y1, Y2, E),
	map_letters_reg_url(Xs, Y1, Y2, Xs1).
map_letters_reg_url([], _, _, []).

map_letters_to_grp([X|Xs], Y1, [X1|Xs1]) :-
	letters_to_grp(X, Y1, X1),
	map_letters_to_grp(Xs, Y1, Xs1).
map_letters_to_grp([], _, []).

map_letters_grp([X|Xs], [X1|Xs1]) :-
	letters_grp(X, X1),
	map_letters_grp(Xs, Xs1).
map_letters_grp([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hinweis_to_html(X, X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

wzp_1(z(VonP,VonG,Datum)-LID)  :-
	aut(LID, z, _, _, _),
	once(( letter(LID, from, Von),
	       xperson_prettyname(Von, VonP),
	       xperson_birth_normdate(Von, VonG),
	       letter(LID, date, Datum1),
	       ddate_sortkey(Datum1, Datum) )).

	       
write_drawings_page :-
	navi(drawings, top, NaviTop, CB),
	navi(drawings, bot, NaviBot, CE),
	findall(Z, wzp_1(Z), Zs),
	sort(Zs, Zs1),
	map_val(Zs1, Bs),
	map_letterlink_global(Bs, EBriefe),
	E1 = e(h1, [], ['Zeichnungen']),
	E2 = e(p, [], ['Folgende Briefe enthalten Zeichnungen, die ',
		       'digitalisiert vorliegen:']),
	append([NaviTop,
		e(hr, [class=navitop], []), CB,
		E1,
		E2,
	        e(p, [], EBriefe)], Rest1, Body),
	Rest1 = [CE, e(hr, [class=navibot], []), NaviBot],
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [],
			[Maintitle, ': ', 'Zeichnungen'])]),
		   Body1])],
	mk_body(Body, Body1),
	write_page(Page, 'drawings.html').


mk_body(Body, Body1) :-
	Body1 = e(body, [], [e(div, [id=centerContainer], Body)]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% Date: Special Pretty Version, Used by HTML Export
%%%% 
ddate_pretty_year(Date, PrettyYear) :-
	date_year(Date, Y),
	!,
	date_normal(y(Y), PrettyYear).
ddate_pretty_year(D, PrettyYear) :-
	special_date_year_date(D, D1),
	!,
	ddate_pretty_year(D1, PrettyYear).
ddate_pretty_year(_, 'Undatiert').

special_date_year_date(before(D), D).
special_date_year_date(after(D), D).
special_date_year_date(range(_, D), D).
special_date_year_date(between(_, D), D).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
<!--                  <input type="hidden" name="ul"
                  value="http://www.pueckler-digital.de/npch/%">
wird allerdings nicht an die Ergebnisseite "vererbt"
 -->
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

person_works(P, Works) :-
	findall(T-W, ( work_author(W, P),
		       id_is_referenced(work, W),
		       work_title(W, T1),
		       atom_sortkey(T1, T) ), Ws),
	sort(Ws, Works1),
	map_val(Works1, Works).

person_works_section(P, H, HRest) :-
	person_works(P, Works),
	( Works = [] -> H = HRest
	; mk_hworks(Works, P, HWorks),
	  H = [ e(h2, [], ['Erwähnte Werke']),
		e(dl, [class=persworks], HWorks) | HRest]
	).

mk_hworks(Works, P, H) :-
	findall(H1,
		( member(W, Works),
		  id_anchor(work, W, Anchor),
		  work_title(W, Title),
		  referring_letters(W, work, [], Ls),
		  ( Ls = [] ->
		    info(20, 'WARNING: Work not referenced in any letter: ~q ~q ~w', [W, P, Title])
		  ; true
		  ),
		  ( chain_page_is_useful(W, work) ->
		    chainlink(W, work, E),
		    Ls1 = [E|Ls]
		  ; Ls1 = Ls
		  ),
		  %% Ls \= [],
		  findall(A, work_author(W, A), As),
		  ( select(P, As, As1), As1 \= [] ->
		    map_person_ref_commasep(As1, Ps1),
		    append(['Mit ' | Ps1], [e(br,[],[]) | Ls1], Ls2)
		  ; Ls2 = Ls1
		  ),
		  atom_concat('#', Anchor, TgtAnchor),
		  ( H1 = e(dt, [class=persworks, id=Anchor],
			   [e(a, [class=anchor, href=TgtAnchor], [Title])])
		  ; H1 = e(dd, [class=persworks], Ls2)
		  )
		),
		H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generic Register Pages
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_genregs :-
	write_genreg_page(events, event, 'Ereignisse'),
	write_genreg_page(subjects, subject, 'Sachregister'),
	write_genreg_page(corpors, corp, 'Körperschaften und Institutionen'),
	write_genreg_page(journals, journal, 'Periodika').

genreg_enum(corp, Id) :- enum_corp(Id).
genreg_enum(event, Id) :- enum_event(Id).
genreg_enum(journal, Id) :- enum_journal(Id).
genreg_enum(subject, Id) :- enum_subject(Id).
genreg_enum(person, Id) :- enum_xperson(Id).
genreg_enum(geo, Id) :- enum_geo(Id).
genreg_enum(work, Id) :-
	findall(W, (enum_xperson(P), person_works(P,W1), member(W,W1)), Ws),
	sort(Ws, Ws1),
	member(Id, Ws1).
genreg_is_subitem(corp, _) :- false.
genreg_is_subitem(event, Id) :- event_superevent(Id, _).
genreg_is_subitem(journal, _) :- false.
genreg_is_subitem(subject, _) :- false.
genreg_links(corp, _, []).
genreg_links(event, Id, Links) :- event_links(Id, Links).
genreg_links(journal, _, []).
genreg_links(subject, _, []).
genreg_name(corp, Id, Name) :- corp_ename(Id, Name).
genreg_name(event, Id, Name) :- event_ename(Id, Name).
genreg_name(journal, Id, Name) :- journal_ename(Id, Name).
genreg_name(subject, Id, Name) :- subject_ename(Id, Name).
genreg_name(person, Id, Name) :- pid_ref_nolink(Id, Name).
genreg_name(geo, Id, Name) :- geo_ename(Id, Name).
genreg_name(work, Id, Ref) :- once(work_title(Id, Ref)).
genreg_subs(corp, _, []).
genreg_subs(event, Id, Subs) :- event_subevents(Id, Subs).
genreg_subs(journal, _, []).
genreg_subs(subject, _, []).
genreg_supers(corp, _, []).
genreg_supers(event, Id, Supers) :- event_superevent(Id, Supers).
genreg_supers(journal, _, []).
genreg_supers(subject, _, []).

write_genreg_page(PageId, Type, Title) :-
	genreg_refs(Type, Lines),
	Es = [NaviTop,
	      e(hr, [class=navitop], []), CB,
	      e(h1, [], [Title]),
	      e(div, [class=xlines], Lines), CE,
	      e(hr, [class=navibot], []),
	      NaviBot],
	navi(PageId, top, NaviTop, CB),
	navi(PageId, bot, NaviBot, CE),
	maintitle(Maintitle),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle, ': ', Title])]),
		   Body1])],
	mk_body(Es, Body1),
	atomic_list_concat([PageId, '.html'], PageFile),
	write_page(Page, PageFile).

genreg_refs(Type, Rs) :-
	findall(E, ( genreg_enum(Type, Id),
		     id_is_referenced(Type, Id),
		     \+ genreg_is_subitem(Type, Id),
		     genreg_id_to_plist(Type, Id, E)
		   ),
		Es),
	sort_plist(Es, [sortkey], Es1),
	map_genreg_ref(Es1, Type, false, Rs).

map_genreg_ref([E|Es], Type, IsSub, Rs) :-
	genreg_ref(Type, E, IsSub, R1),
	append(R1, Rs1, Rs),
	map_genreg_ref(Es, Type, IsSub, Rs1).
map_genreg_ref([], _, _, []).

genreg_id_to_plist(Type, ID, PList) :-
	PList = [id-ID,
		 name-Name,
		 sortkey-SortKey,
		 subs-Sub,
		 supers-Super,
		 links-LS],
	genreg_name(Type, ID, Name),
	name_to_idlike_sortkey(Name, SortKey),
	genreg_subs(Type, ID, Sub1),
	findall(ID1, (member(ID1, Sub1), id_is_referenced(Type, ID1)), Sub),
	genreg_links(Type, ID, LS),
	(genreg_supers(Type, ID, Super) -> true ; Super = none).

genreg_ref(Type, X, IsSub, R) :-
	get_val(X, name, Name),
	get_val(X, subs, Sub),
	get_val(X, id, Id),
	id_anchor(Type, Id, Id1),
	atom_concat('#', Id1, Anchor),
	referring_letters(Id, Type, [], LetterRefs),
	( chain_page_is_useful(Id, Type) ->
	  chainlink(Id, Type, E),
	  LetterRefs1 = [E|LetterRefs]
	; LetterRefs1 = LetterRefs
	),
	( IsSub == false ->
	  R = [e(div, [class=genregh, id=Id1],
		 [e(a, [class=anchor, href=Anchor], [Name])]),
	       e(div, [class=genregref], LetterRefs1) | RSub],
	  findall(X1, ( member(I, Sub),
			genreg_id_to_plist(Type, I, X1) ),
		  XsSub),
	  sort_plist(XsSub, [sortkey], XsSub1),
	  map_genreg_ref(XsSub1, Type, true, RSub)
	; R = [e(div, [class=genregsubh, id=Id1],
		 [e(a, [class=anchor, href=Anchor], [Name])]),
	       e(div, [class=genregsubrefs], LetterRefs1)]
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_html_bibpage(Options) :-
	set_html_options(Options),
	html_bibentries(Options, Entries),
	append(Entries, Bottom, Rest),
	Es = [NaviTop,
	      e(hr, [class=navitop], []), CB,
	      e(h1, [], ['Bibliographie']) |
	     Rest],
	navi(biblio, top, NaviTop, CB),
	navi(biblio, bot, NaviBot, CE),
	Bottom = [CE, e(hr, [class=navibot], []),
		  NaviBot],
	maintitle(Maintitle1),
 	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], [Maintitle1, ': ', 'Bibliographie'])]),
		   Body1])],
	mk_body(Es, Body1),
	filename(html_output(''), Out),
	cd(Out),
	write_page(Page, 'biblio.html').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_referenced_biblabels :-
	filename( gendata('generated_referenced_biblabels.pl'), Outfile ),
	referenced_biblabels(Ls),
	length(Ls, Len),
	info(10, 'Number of referenced biblabels: ~w', [Len]),
	info(10, 'Writing ~w', [Outfile]),
	onto_file(( member(L, Ls),
		    writeq(gen_referenced_biblabel(L)),
		    writeln('.'),
		    fail
		  ; true),
		  Outfile).

referenced_biblabels(Ls) :-
	findall(L, ((letter(_, text, T) ; letter(_, annot, T)),
		    nlx_biblabel(T, L)),
		Ls1),
	sort(Ls1, Ls).

citation_command('xciteAY').
citation_command('xciteST').
citation_command('xnocite').
citation_command('cite').
citation_command('nocite').

nlx_biblabel(NLX, L) :-
	sub_term(ncmd(Cite, _, [L], _), NLX),
	citation_command(Cite).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_chain_pages :-
	( member(Type, [person, geo, event, subject, journal, corp, work]),
	  genreg_enum(Type, Id),
	  id_is_referenced(Type, Id),
	  chain_page_is_useful(Id, Type),
	  write_chain_page(Id, Type),
	  fail
	; true
	).

chain_page_is_useful(Id, Type) :-
	referring_letters(Id, Type, [just_id_options_pairs], Ls),
	length(Ls, N1),
	N1 > 1.

write_chain_page(Id, Type) :-
	referring_letters(Id, Type, [just_id_options_pairs], Ls),
	chain_list(Ls, 1, ChainBody),
	length(Ls, NResults),
	findall(X, member(X-_, Ls), Ls1),
	chain_next_prev_js(Ls1, JS1),
	js_code_chainresult(JS0),
	js_code_fixprevcolor(JS2),
	genreg_name(Type, Id, NameE),
	NVoid is NResults+1,
	chain_position_id(NVoid, IdNVoid),
	chain_position_id(0, IdPVoid),
	ChainHead =
	[e(hr, [class=chainhr], []),
	 e(p, [class=chaininfo],
	   ['Kette: Verweise auf ', NameE]),
	 e(p, [class=chainnavi],
	   [e(a, [id=IdPVoid, class=chainprev, onclick='chainPrev();'],
	      ['Voriges Ergebnis']), ' | ',
	    e(a, [id=IdNVoid, class=chainnext, onclick='chainNext();'],
	      ['Nächstes Ergebnis']),
	    ' (', NResults, ' Ergebnisse', ')'])],
	ChainVspace = e(div, [id=chainvspace], []),
	append(ChainBody, [ChainVspace], ChainBody1),
	genreg_name(Type, Id, Name),
	Page = [e(html, [],
		  [e(head, [],		  
		     [e(meta, [http-equiv='content-type',
			       content='text/html; charset=utf-8'], []),
		      e(link, [href='style.css',
			       rel='styleSheet',
			       type='text/css'], []),
		      e(title, [], ['Kette: Verweise auf ', Name]),
		      e(script, [type='text/javascript'],
			[JS0, JS1, JS2])]),
		   e(body, [class=chainbody],
		     [e(div, [id=chainfix0],
			[e(div, [id=chainfix1],
			   [e(div, [id=chainfix2], ChainHead)])]),
		      e(div, [id=chainlist],
			ChainBody1)])])],
	chain_url(Type, Id, File),
	write_page(Page, File).

chain_position_id(N, CPID) :-
	atom_concat(i, N, CPID).

chain_list([L-Options|Ls], N, [E|Es]) :-
	( Ls = [] -> N1 = N ; N1 is N+1 ),
	chain_position_id(N, CPID),
	chain_position_id(N1, CPID1),
	letterlink_global_1(L, [values(URL,E1)], _),
	format(atom(OnClick),
	       'chainResult(\'~w\',\'#~w\')', [URL, CPID1]),
	( memberchk(main, Options) ->
	  Class = chainlinemain
	; Class = chainline
	),
	E = e(div, [id=CPID, class=Class],
	      [e(a, [class=chain, onclick=OnClick],
		 [e(span, [class=chainnum], [N, '. ']), ' ' | E1])]),
	chain_list(Ls, N1, Es).
chain_list([], _, []).

chain_next_prev_js([], '') :- !.
chain_next_prev_js(Ls, JS) :-
	reverse(Ls, [LL|_]),
	append(Ls, [LL], Ls1),
	Ls = [L|_],
	chain_js_cases(Ls1, 1, L, L, 1, NX0, PX0),
	once(append(NX, [_], NX0)),
	PX0 = [_|PX],
	map_js_chain_case(NX, NX1),
	map_js_chain_case(PX, PX1),
	atomic_list_concat(NX1, NX2),
	atomic_list_concat(PX1, PX2),
	format(atom(ChainNext),
	       'function chainNext() {~nvar h = location.hash;~n\c
	      switch (h) {~n ~wdefault : break;  }~n}~n', [NX2]),
	format(atom(ChainPrev),
	       'function chainPrev() {~nvar h = location.hash;~n\c
	      switch (h) {~n ~wdefault :  break;  }~n}~n', [PX2]),
	atomic_list_concat([ChainNext,ChainPrev], JS).

js_chain_case(val(URL,IN1,IN2), Case) :-
	format(atom(Case),
	       'case "#~w" : chainResult("~w","#~w"); break;~n',
	       [IN1, URL, IN2]).

map_js_chain_case([X|Xs], [X1|Xs1]) :-
	js_chain_case(X, X1),
	map_js_chain_case(Xs, Xs1).
map_js_chain_case([], []).

chain_js_cases([L|Ls], N, LP, LP1, NP,
	       [val(URLN,IN1,IN2)|NX],
	       [val(URLP,IN1,IP2)|PX]) :-
	( Ls = [] ->
	  N1 = N,
	  LP2 = LP,
	  NP2 = NP
	; N1 is N+1,
	  LP2 = LP,
	  NP2 = NP
	),
	chain_position_id(N, IN1),
	chain_position_id(N1, IN2),
	chain_position_id(NP2, IP2),
	letterlink_global_1(L, [values(URLN,_)], _),
	letterlink_global_1(LP2, [values(URLP,_)], _),
	chain_js_cases(Ls, N1, LP1, L, N, NX, PX).
chain_js_cases([], _, _, _, _, [], []).

js_code_chainresult(X) :-
	X = 'function chainResult(x,y) {
    var strWindowFeatures = "menubar=yes,toolbar=yes,location=yes,personalbar=yes,status=yes,resizable=yes,scrollbars=yes";
    var w = window.open(x, "kbsetchainresult", strWindowFeatures);
    w.focus();
    window.open(y, "_self");
}\n'.

js_code_fixprevcolor(X) :-
	X = 'function fixPrevColor() {
      var h = location.hash;
      var d = document.getElementById("i0");
       if (h == "#i1") { d.style.color="black"; d.style.background="none";
      } else { d.style.color=""; d.style.background="";};}
      window.onhashchange = function() { fixPrevColor(); }
      window.onload = function() { fixPrevColor(); }\n'.

