%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(kbmappings,
	  [ id_url/3,
	    id_anchor/3,
	    aut_url/3,
	    img_url/2,
	    reg_url/4,
	    chain_url/3,
	    % bibrefdata/5,
	    main_info_url/3]).

:- use_module(swilib(err)).
:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(extract_letters)).

id_url(bib, Id, URL) :- !, bid_url(Id, URL).
id_url(corp, Id, URL) :- !, cid_url(Id, URL).
id_url(event, Id, URL) :- !, eid_url(Id, URL).
id_url(geo, Id, URL) :- !, gid_url(Id, URL).
id_url(journal, Id, URL) :- !, jid_url(Id, URL).
id_url(letter, Id, URL) :- !, lid_url(Id, URL).
id_url(person, Id, URL) :- !, pid_url(Id, URL).
id_url(subject, Id, URL) :- !, sid_url(Id, URL).
id_url(work, Id, URL) :- !, wid_url(Id, URL).
id_url(X, _, _) :- err('No id to url conversion for type ~q', [X]).

id_anchor(bib, Id, URL) :- !, bid_anchor(Id, URL).
id_anchor(corp, Id, URL) :- !, cid_anchor(Id, URL).
id_anchor(event, Id, URL) :- !, eid_anchor(Id, URL).
id_anchor(geo, Id, URL) :- !, gid_anchor(Id, URL).
id_anchor(journal, Id, URL) :- !, jid_anchor(Id, URL).
% id_anchor(letter, Id, URL) :- !, lid_anchor(Id, URL).
id_anchor(person, Id, URL) :- !, pid_anchor(Id, URL).
id_anchor(subject, Id, URL) :- !, sid_anchor(Id, URL).
id_anchor(work, Id, URL) :- !, wid_anchor(Id, URL).
id_anchor(X, _, _) :- err('No id to anchor conversion for type ~q', [X]).

% bibrefdata(A, B, C, D, E) :-
% 	bibrefdata1(A, B, C, D, E).

% id_url(person, X, Y) :- pid_url(X, Y).
% id_url(letter, X, Y) :- lid_url(X, Y).

lid_url(LID, URL) :-
	( letter(LID, id, LID1) ->
	  true
	; LID1 = LID
	),
	!,
	( sub_atom(LID1, 0, _, _, 'b:') ->
	  sub_atom(LID1, 2, _, 0, LID2)
	; LID2 = LID1
	),
	stdize_id(LID2, LID3),
	atomic_list_concat(['letter-', LID3, '.html'], URL).

pid_url(ID, URL) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['person-', ID1, '.html'], URL).

chain_url(Type, ID, URL) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['chain-', Type, '-', ID1, '.html'], URL).

pid_anchor(BibLabel, URL) :-
	stdize_id(BibLabel, BibLabel1),
	atomic_list_concat(['person-', BibLabel1], URL).

gid_url(ID, URL) :-
	geo_supergeo(ID, SuperID),
	!,
	stdize_id(SuperID, SuperID1),
	gid_anchor(ID, ID1),
	atomic_list_concat(['geo-', SuperID1, '.html#', ID1], URL).
gid_url(ID, URL) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['geo-', ID1, '.html'], URL).

gid_anchor(ID, URL) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['geo-', ID1], URL).

wid_anchor(ID, FRAG) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['work-', ID1], FRAG).

wid_url(ID, URL) :-
	( work_author(ID, PID) ->
	  true
	; PID = unknown
	),
	pid_url(PID, UPID),
	wid_anchor(ID, FRAG),
	atomic_list_concat([UPID, '#', FRAG], URL).

eid_anchor(ID, FRAG) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['event-', ID1], FRAG).

eid_url(ID, URL) :-
	eid_anchor(ID, FRAG),
	atomic_list_concat(['events.html', '#', FRAG], URL).

sid_anchor(ID, FRAG) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['subject-', ID1], FRAG).

sid_url(ID, URL) :-
	sid_anchor(ID, FRAG),
	atomic_list_concat(['subjects.html', '#', FRAG], URL).

cid_anchor(ID, FRAG) :-
	stdize_id(ID, ID1),
	atomic_list_concat(['corp-', ID1], FRAG).

cid_url(ID, URL) :-
	cid_anchor(ID, FRAG),
	atomic_list_concat(['corpors.html', '#', FRAG], URL).

jid_url(ID, URL) :-
	jid_anchor(ID, FRAG),
	atomic_list_concat(['journals.html', '#', FRAG], URL).

jid_anchor(BibLabel, URL) :-
	stdize_id(BibLabel, BibLabel1),
	atomic_list_concat(['journal-', BibLabel1], URL).

main_info_url(LID, TARGET, URL) :-
	lid_url(LID, URL1),
	atomic_list_concat([URL1, '#', TARGET], URL).

reg_url(PIDVon, PIDAn, Jahr, URL) :-
	stdize_id(PIDVon, PIDVon1),
	stdize_id(PIDAn, PIDAn1),
	atomic_list_concat(['reg-to-', PIDVon1, '-', PIDAn1, '-', Jahr, '.html'], URL).

img_url(File, URL) :-
	img_prefix(Prefix),
	atomic_list_concat([Prefix, File], URL).

img_prefix('img/').
aut_prefix('aut/').

aut_url(Hi, LoURL, HiURL) :-
	aut_prefix(Prefix),
	( sub_atom(Hi, _, B, 0,'_full.jpg')
	; sub_atom(Hi, _, B, 0,'_full.png')
	; sub_atom(Hi, _, B, 0,'_full.gif')
	),
	!,
	sub_atom(Hi, 0, _, B, Base),
	atomic_list_concat([Prefix, Base, '_800.jpg'], LoURL),
	atomic_list_concat([Prefix, Hi], HiURL).
aut_url(Hi, URL, URL) :-
	aut_prefix(Prefix),
	atomic_list_concat([Prefix, Hi], URL).

bid_url(BibLabel, URL) :-
	stdize_id(BibLabel, Id),
	atomic_list_concat(['biblio.html#bib-', Id], URL).

bid_anchor(BibLabel, URL) :-
	stdize_id(BibLabel, BibLabel1),
	atomic_list_concat(['bib-', BibLabel1], URL).

