%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(letters_config, [get_kbset_config/2,
			   get_kbset_config/3,
			   letters_initialize/0,
			   ensure_letters_initialized/0]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_base(filename)).

:- dynamic user:kbset_config/2.

get_kbset_config(Key, Value) :-
	user:kbset_config(Key, Value1),
	!,
	Value1 = Value.
get_kbset_config(Key, _) :-
	err('Missing kbset_config entry for ~q', [Key]).

get_kbset_config(Key, Value, _) :-
	user:kbset_config(Key, Value1),
	!,
	Value1 = Value.
get_kbset_config(_, Default, Default).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Initialization
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic inittime/0.

%%%% 
%%%% Requires the user:kbset_config predicate, e.g. by loading a configuration
%%%% file.
%%%% 
letters_initialize :-
	retractall( inittime ),
	set_docdir,
	register_required_dirs,
	register_fixed_files,
	asserta( inittime ).

ensure_letters_initialized :-
	inittime,
	predicate_property(user:kbset_config(_,_),
			   last_modified_generation(T1)),
	predicate_property(inittime,
			   last_modified_generation(T2)),
	T1  =< T2,
	!.
ensure_letters_initialized :-
	letters_initialize.

required_dir(appdata_dir, appdata(_)).
required_dir(bib_dir, bib(_)).
required_dir(gendata_dir, gendata(_)).
required_dir(html_output_dir, html_output(_)).

register_required_dirs :-
	docdir(LettersRoot),
	register_file_path(letters(_), LettersRoot),
	( required_dir(ConfKey, Template),
	  get_kbset_config(ConfKey, DirSpec),
	  restricted_expand_filename(DirSpec, DirName),
	  ensure_trailing_dir_sep(DirName, DirName1),
	  info(10, 'Setting ~w to ~w', [ConfKey, DirName1]),
	  register_file_path(Template, DirName1),
	  fail
	; true
	),
	once(absolute_file_name(kbset_data(''), KBSetDataDir)),
	register_file_path(kbset_data(_), KBSetDataDir).

fixed_file(generated_reordered, letters('generated_reordered.tex')).
fixed_file(generated_defperson, letters('generated_defperson.tex')).
fixed_file(generated_annotations, letters('generated_annotations.tex')).
fixed_file(generated_gnd_person_data, gendata('generated_gnd_person_data.pl')).

register_fixed_files :-
	( fixed_file(Key, File),
	  register_file(Key, File),
	  fail
	; true
	).

ensure_trailing_dir_sep(X, X1) :-
	( sub_atom(X, _, 1, 0, '/') ->
	  X1 = X
	; atom_concat(X, '/', X1)
	).
 
