%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(letters_interface,
	  [letters_initialize/0,
	   batch/1,
	   gen_reordered/0,
	   gen_annotations/0,
	   gen_dbs/0,
	   print_biber_xml_conversion_script/0,
	   print_copy_html_extras_script/0,
	   gen_html_bib/0,
	   gen_html/0,
	   extract_letters/0,
	   gen_metadata_export/0
	   ]).

:- use_module(swilib(info)).
:- use_module(swilib(err)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_letters(reorder_letters)).
:- use_module(kbset_letters(extract_letters)).
:- use_module(kbset_letters(make_data)).
:- use_module(kbset_letters(bibconvert)).
:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(htmlexport_letters)).
:- use_module(kbset_letters(letters_config)).
:- use_module(kbset_letters(factbase_export)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ensure_postfix(F, Postfix, F1) :-
	sub_atom(F, _, _, 0, Postfix),
	!,
	F1 = F.
ensure_postfix(F, Postfix, F1) :-
	atom_concat(F, Postfix, F1).

letter_files(F) :-
	get_kbset_config(letter_latex_sources, F1),
	findall(letters(F2),
		( member(F3, F1), ensure_postfix(F3, '.tex', F2) ),
		F).

annot_files(F) :-
	get_kbset_config(annotation_latex_sources, F1),
	findall(letters(F2),
		( member(F3, F1), ensure_postfix(F3, '.tex', F2) ),
		F).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% For wrappping -g arguments in shell scripts. Otherwise the message hook
%%%% is seemingly not obeyed during evaluating of them (SWI 7 and 8), leading
%%%% to "Unknown error term" messages instead of properly formatted messages
%%%% for err/1 and err/2 from swilib(err).
%%%%
%%%%
batch(Call) :-
	catch(Call, error(err(A,B),C),
	      (user:message_hook(error(err(A,B),C),error,_),halt)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generating LaTeX version with letters reordered (e.g. by date,
%%%% independently of author)
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_reordered :-
	ensure_letters_initialized,
	letter_files(F),
	generate_reordered(F, generated_reordered).

%
% Not maintained:
%
% generate_reordered(Separators) :-
% 	letter_files(F),
% 	generate_reordered_split(F, Separators).
%
% :- ( between(1, 20, I),
%      format(atom(Key), 'generate_reordered_~|~`0t~d~2+', [I]),
%      format(atom(File), '~w.tex', [Key]),
%      register_file(Key, letters(File)),
%      fail
%    ; true
%    ).
% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generating cleaned-up and ordered annotations
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_annotations :-
	ensure_letters_initialized,
	letter_files(F),
	generate_reordered(F, none),
	annot_files(K),
	generate_annots(K,
			[rm_empty_ksections, sort_ksections, reorder],
			generated_annotations).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Convert Latexdbs to Prolog
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_dbs :-
	ensure_letters_initialized,
	get_kbset_config(factbase_latex_sources, DBs),
	( member(DB, DBs),
	  format(atom(InFile), '~w.tex', DB),
	  format(atom(OutFile), 'generated_from_latexdb_~w.pl', DB),
	  filename(gendata(OutFile), OutFile1),
	  latexdb_to_pl(letters(InFile)) onto_file OutFile1,
	  fail
	; true
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generating HTML
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

extract_letters :-
	ensure_letters_initialized,
	filename(letters('generated_annotations.tex'), K),
	filename(letters('generated_reordered.tex'), G),
	extract_letters(G, K).

gen_html :-
	ensure_letters_initialized,
	get_kbset_config(html_options, Options),
 	gen_html(Options).

print_copy_html_extras_script :-
	ensure_letters_initialized,
	filename(html_output(''), Tgt),
	get_kbset_config(html_extras, Extras),
	( member(E, Extras),
	  restricted_expand_filename(E, E1),
	  format('cp -rv ~w ~w; ', [E1, Tgt]),
	  fail
	; true
	).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generating HTML bibliography
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_biber_xml_conversion_script :-
	ensure_letters_initialized,
	get_kbset_config(biblatex_sources, Bibs),
	print_biber_xml_conversion_script(Bibs).
	
gen_html_bib :-
	ensure_letters_initialized,
	get_kbset_config(biblatex_sources, Bibs),
	get_kbset_config(html_options, HTMLOptions),	
	load_biber_xml_files(Bibs),
	generate_bibrefdata,
	gen_html_bibpage(HTMLOptions).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_metadata_export :-
	extract_letters,
	OutFile = 'generated_metadata_export.pl',
	filename(gendata(OutFile), OutFile1),
	info(10, 'Writing ~w', [OutFile1]),
	get_kbset_config(html_options, HTMLOptions),
	( memberchk(maintitle=Title, HTMLOptions) ->
	  format(atom(ProjectInfo), '%%%% Project: ~w ', [Title])
	; ProjectInfo = ''
	),
	( date(date(Y,M,D)),
	  date_w3c(ymd(Y,M,D), Date),
	  format('%%%% Generated with KBSET by list_metadata/0 on ~w~n',
		 [Date]),
	  ( ProjectInfo = '' -> true
	  ; writeln(ProjectInfo)
	  ),
	  list_metadata
	)
	onto_file OutFile1.

	
