%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% This module need revision: Originally the idea was to merge data
%%%% specified in LaTeX DBs with data from other sources such as the GND to
%%%% "effective" data. Now the workflow is simpler: The data sources are the
%%%% LaTeX DBs and some DBs *previously generated* with sources like the GND
%%%% and stored in the appdata directory. The simpler workflow works but there
%%%% are still traces of the more complex flow in the code.
%%%%
:- module(make_data,
	  [ pl_to_latexdb/2,
	    generate_latex_seeindex/0,
	    latexdb_to_pl/1,
	    generate_gnd_person_data/0,
%	    generate_effective_person_data/0,
	    refresh_mk_data/0,
	    check_sorting/0,
	    check_person_ids_for_sorting/0,
	    check_person_gnd/0,
	    check_defperson/0,
	    check_works/0 ]).

:- use_module(swilib(err)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_base(readwrite)).
:- use_module(kbset_base(textutils)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(gnd)).
:- use_module(kbset_base(dates)).
:- use_module(swilib(info)).
:- use_module(kbset_letters(data_interface)).
:- use_module(kbset_letters(textutilities)).
:- use_module(kbset_letters(legacy)).
:- use_module(kbset_letters(nlx)).

:- multifile person/3.

:- multifile defperson/2.
:- multifile gnd_dewiki/2.
:- multifile gnd_person/7.

latex_command(defperson,[c,t],'').
latex_command(defwork,[c,c,t],'').
latex_command(defworkTwo,[c,c,c,t],'').
latex_command(defworkThree,[c,c,c,c,t],'').
latex_command(defgeo,[c,t],'').
latex_command(defsubgeo,[c,c,t],'').
latex_command(defcorp,[c,t],'').
latex_command(defjournal,[c,t],'').
latex_command(defevent,[c,t],'').
latex_command(defsubevent,[c,c,t],'').
%%
latex_command(defworkInfo,[c,c,t,t],'').
%%
latex_command(defpersonSee,[c,t],'').
latex_command(defsubjectSee,[c,t],'').
latex_command(defgeoSee,[c,t],'').
latex_command(defcorpSee,[c,t],'').
latex_command(defjournalSee,[c,t],'').
latex_command(defeventSee,[c,t],'').
%%
latex_command(defheadername,[c,t],'').

:- initialization( ( latex_command(C, A, T),
		     register_latex_command(C, A, T),
		     ( legacy_alias(C1, C) ->
		       register_latex_command(C1, A, T)
		     ; true
		     ),
		     fail
		   ; true
		   )).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

refresh_mk_data :-
	info(10, 'Refreshing preparation data'),
	Files = ['generated_gnd_person_data',
		 'gnd_to_dewiki_1780',
		 'person_gnd',
		 'person_data',
		 'defperson'],
	( member(F, Files),
	  filename( appdata(F), F1 ),
	  info(10, 'Consulting ~w', [F1]),
	  consult(F1),
	  fail
	; true
	).

refresh_defperson_data :-
	info(10, 'Refreshing defperson data'),
	Files = ['defperson'],
	( member(F, Files),
	  filename( appdata(F), F1 ),
	  info(10, 'Consulting ~w', [F1]),
	  consult(F1),
	  fail
	; true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Conversion of defperson etc. Prolog <-> LaTeX
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% 
%%%% 
pl_to_latexdb(Pattern, Cmd) :-
	refresh_access_data,
	( ( Pattern = _:Pattern1 -> true ; Pattern1 = Pattern ),
	  Pattern1 =.. [_|Args],
	  call(Pattern),
	  format('\\~w', [Cmd]),
	  ( member(Arg, Args),
	    esc_latex_pretty(Arg, Arg1),
	    format('{~w}', [Arg1]),
	    fail
	  ; true
	  ),
	  format('~n', []),
	  fail
	; true
	).

:- dynamic ldb_info/1.

%%%% 
%%%% 
%%%%
latexdb_to_pl(NLX) :-
	( NLX = [] ->
	  true
	; NLX = [_|_] ->
	  latexdb_to_pl_1(NLX)
	; filename(NLX, File) ,
	  info(10, 'Reading in ~w', [NLX]),
	  read_in_latex(File, NLX1),
	  info(10, 'Converting to NLX format'),
	  lx_to_nlx(NLX1, NLX2),
	  latexdb_to_pl(NLX2)
	).

latexdb_to_pl_1(NLX) :-	
	findall( def(CMD, Symbols, Value),
		 ( % nlx_enum(NLX2, E),
	           member(E, NLX),		 
		   ( E = ncmd(CMD, _, Symbols, [Value|_]),
		     defcmd(CMD) ->
		     retractall(ldb_info(_)),
		     assert(ldb_info(CMD-Symbols-Value))
		   ; E = com(_) -> fail
		   ; E = spc(_) -> fail
		   ; ( ldb_info(LCMD) -> true ; LCMD='none' ),
		     err('Warning: Unexpected latex code in data file: ~q, last defcmd: ~q', [E, LCMD])
		   )
		 ),
		 Defs
	       ),
	findall(D, ( member( def(T1, S1, V1), Defs ),
		     lx_to_plain(V1, V2),
		     append(S1, [V2], Args),
		     effective_functor(T1, T2),
		     D =.. [T2|Args]
		   ),
		Ds),
	sort(Ds, Ds1),
	( member(D, Ds1),
	  format('~q.~n', D),
	  fail
	; true
	),
	check_defs_for_duplicates(Ds).

member_and_tail(X, [X|Y], Y).
member_and_tail(X, [_|Y], Z) :-
	member_and_tail(X, Y, Z).

check_defs_for_duplicates(Ds) :-
	( member_and_tail(D1, Ds, Ds1),
	  D1 =.. [DefCmd, Label1 | Value1],
	  ( member(D2, Ds1),
	    D2 =.. [DefCmd1, Label2 | Value2],
	    \+ seecmd(DefCmd),
	    \+ seecmd(DefCmd1),
	    ( Label1 = Label2 ->
	      ( Value1 = Value2 ->
		info(10, 'WARNING: duplicate entry: ~q ~q', [Label1, Value1])
	      ; info(10, 'WARNING: duplicate label: ~q', [Label1])
	      )
	    ; Value1 = Value2 ->
	      info(10, 'WARNING: duplicate value: ~q', [Value1])
	    ; true
	    ),
	    fail
	  ; true
	  ),
	  fail
	; true
	).

lx_to_plain(X, Y) :-
	findall(Z, (member(Z1, X), lxelem_enum_plaintxt(Z1, Z)), Z2),
 	atomic_list_concat(Z2, Y).

defcmd(defperson).
defcmd(defwerk).
defcmd(defwerkTwo).
defcmd(defwerkThree).
defcmd(defwork).
defcmd(defworkTwo).
defcmd(defworkThree).
defcmd(defgeo).
defcmd(defsubgeo).
defcmd(defcorp).
defcmd(defjournal).
defcmd(defevent).
defcmd(defsubevent).
defcmd(defsubject).
defcmd(defsubsubject).
defcmd(defwerkInfo).
defcmd(defworkInfo).
%%
defcmd(defheadername).
%%
defcmd(defpersonSee).
defcmd(defsubjectSee).
defcmd(defgeoSee).
defcmd(defcorpSee).
defcmd(defjournalSee).
defcmd(defeventSee).
%%
seecmd(defpersonSee).
seecmd(defsubjectSee).
seecmd(defgeoSee).
seecmd(defcorpSee).
seecmd(defjournalSee).
seecmd(defeventSee).
seecmd(defheadername).

effective_functor(defwerk, defwork) :- !.
effective_functor(defwerkTwo, defworkTwo) :- !.
effective_functor(defwerkThree, defworkThree) :- !.
effective_functor(defwerkInfo, defwork) :- !.
effective_functor(defwerkTwoInfo, defworkTwo) :- !.
effective_functor(defwerkThreeInfo, defworkThree) :- !.
effective_functor(X, X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Generate GND Persondb, extracted from full GND set up with KBSET
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% 
%%%% 
generate_gnd_person_data :-
	info(10, 'Assuming KBSET is set up with full data loaded'),
	%%
	%% Requires kbset with assistance_sb and load_data
	%%
	refresh_defperson_data,
	Row = gnd_person(P, N,Ns,G,YB,YD,Ls),
	( defperson(P, _),
	  ( person(P, gnd, GPID),
	    person_preferred_name(GPID, N) ->
	    ( person_gender(GPID, G) -> true
	    ; G = unknown
	    ),
	    ( person_year_of_birth(GPID, YB1) ->
	      date_normal(y(YB1), YB)
	    ; YB = unknown
	    ),
	    ( person_year_of_death(GPID, YD1) ->
	      date_normal(y(YD1), YD)
	    ; YD = unknown
	    ),
	    ( gnd_to_wiki(GPID, de, DEWiki) ->
	      Ls = [wiki_de-DEWiki|Ls1]
	    ; Ls = Ls1
	    ),
	    ( gnd_to_wiki(GPID, en, ENWiki) ->
	      Ls1 = [wiki_en-ENWiki|Ls2]
	    ; Ls1 = Ls2
	    ),
	    Ls2 = [gnd-GPID],
	    Ns = [],
	    format('~q.~n', [Row]),
	    fail
	  ; fail
	  ),
	  fail
	; true
	).

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%% 
% %%%% Generate Effective Person Data
% %%%% 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% %%%% 
% %%%% 
% %%%% 
% generate_effective_person_data :-
% 	refresh_mk_data,
% 	Row = person(P,N,Ns,G,YB,YD,Ls),
% 	( defperson(P, DefPersonInfo),
% %	  \+ p_is_pseudonym_id(P),
% 	  ( gnd_person(P,N,Ns,G,YB,YD,Ls1) ->
% 	    ( \+ memberchk(wiki_de-_, Ls1),
% 	      memberchk(gnd-GND, Ls1),
% 	      gnd_dewiki(GND, Wiki) ->
% 	      Ls = [wiki_de-Wiki|Ls1]
% 	    ; Ls = Ls1
% 	    )
% 	  ; extract_defperson_info(DefPersonInfo, N, YB, YD),
% 	    Ns = [],
% 	    G = unknown,
% 	    ( person(P, gnd, GND) ->
% 	      Ls = [gnd-GND]
% 	    ; Ls = []
% 	    )
% 	  ),
% 	  fix_person_info(Row, Row1),
% 	  format('~q.~n', [Row1]),
% 	  fail	
% 	; true
% 	).
% 
% extract_defperson_info(X, N, YB, YD) :-
% 	( atomic_list_concat([N, D1], ' (', X) ->
% 	  ( sub_atom(D1, _, _, 0, ')'),
% 	    sub_atom(D1, 0, _, 1, D2),
% 	    atomic_list_concat([YB1, YD1], '--', D2),
% 	    atom_number(YB1, YB2),
% 	    atom_number(YD1, YD2),
% 	    date_normal(y(YB2), YB),
% 	    date_normal(y(YD2), YD) ->
% 	    true
% 	  ; YB = unknown,
% 	    YD = unknown
% 	  )
% 	; N = X,
% 	  YB = unknown,
% 	  YD = unknown
% 	).
% 
% 	
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% fix_person_info(person(P,N,_Ns,G,YB,YD,Ls),
% 		person(P,N1,Ns1,G1,YB1,YD1,Ls1)) :-
% 	( person(P, preferredName, N1) ->
% 	  true
% 	; N1 = N
% 	),
% 	( person(P, dateOfBirth, YB2), atom_number(YB2, YB3) ->
% 	  date_normal(y(YB3), YB1)
% 	; YB1 = YB
% 	),
% 	( person(P, dateOfDeath, YD2), atom_number(YD2, YD3) ->
% 	  date_normal(y(YD3), YD1)
% 	; YD1 = YD
% 	),
% 	( person(P, gender, G2), memberchk(G2, [m,f]) ->
% 	  G1 = G2
% 	; G1 = G
% 	),
% 	p_further_names(P, Ns1),
% 	findall(LK-L, person(P, link, LK-L), Links1),
% 	append(Links1, Ls, Ls1).
% 
% p_further_names(P, Ns1) :-
% 	findall( X,
% 		 ( X = 1-(pseudonym_of-P1), person(P, pseudonymOf, P1)
% 		 ; X = 2-(pseudonym-P1), person(P1, pseudonymOf, P)
% 	         ; X = 3-(altname-N), person(P, altName, N)
% 	         ; X = 4-(birthname-N), person(P, birthName, N)
% 		 ),
% 		 Ns2
% 	       ),
% 	sort(Ns2, Ns3),
% 	map_val(Ns3, Ns1).
% 
% p_is_pseudonym_id(P) :-
% 	person(P, pseudonymOf, _),
% 	!.
% 
% map_val([_-X|Xs], [X|Xs1]) :-
% 	map_val(Xs, Xs1).
% map_val([], []).
% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% CHECKS
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_sorting :-
	refresh_mk_data,
	findall(X-N, ( enum_xperson(X),
		       xperson_prettyname(X,N1),
		       xperson_birth_normdate(X,B),
		       xperson_death_normdate(X,D),
		       format(atom(N), '~w (~w-~w)', [N1, B, D] )),
		Ps),
	sort(Ps, Ps1),
	show_sort_inconsistencies(Ps1).

show_sort_inconsistencies([K1-N1,K2-N2|KNs]) :-
	atom_to_ascii(N1, M1),
	atom_to_ascii(N2, M2),
	(\+ (M1 @< M2) ->
	  info(0, 'Predence error: ~q < ~q but not ~w < ~w', [K1, K2, N1, N2])
	; true
	),
	show_sort_inconsistencies([K2-N2|KNs]).
show_sort_inconsistencies([_]).
show_sort_inconsistencies([]).
	  
			   
check_person_ids_for_sorting :-
	refresh_mk_data,
	( enum_xperson(X),
	  xperson_prettyname(X, N),
	  sub_atom(X, 0, 1, _, A1),
	  sub_atom(N, 0, 1, _, A2),
	  downcase_atom(A2, A3),
	  A1 \= A3,
	  info(0, 'Different first letter: ~q ~q', [X, N]),
	  fail
	; true
	).

:- flag(count, _, 0).

check_person_gnd :-
	refresh_mk_data,
	flag(count, _, 0),
	( enum_xperson(X),
	  ( person(X,gnd,A), A \= 'gnd:missing' ->
	    ( person(X,gnd,B), B \= 'gnd:missing', B \= A ->
	      info(0, 'Multiple GND mappings ~q: ~q and ~q', [X, A, B])
	    ; true
	    )
	  ; flag(count, C, C+1),
	    ( defperson(X, Info) ->
	      info(0, '~d, Not mapped to GND: ~q: NAME ~w', [C, X, Info])
	    ; info(0, '~d, Not mapped to GND: ~q:', [C, X])
	    )
	  ),
	  fail
	; true
	).

check_works :-
	refresh_mk_data,
	( enum_work(W),
	  findall(X, work_title(W,X), T),
	  ( T = [_] ->
	    true
	  ; info(0, 'Work ~q has multiple titles: ~q', [W, T])
	  ),
	  fail
	; true
	).

check_defperson :-
	refresh_mk_data,
	( defperson(A,B),
	  defperson(A,C),
	  B @< C,
	  format('Multiple definition for ~q: ~w and ~w~n', [A, B, C]),
	  fail
	; true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_latex_seeindex :-
	refresh_access_data,
	( enum_xperson(P),
	  xperson_furthernames(P, Ns),
	  ( member(K-N, Ns),
	    prt_seeindex(K, P, N),
	    fail
	  ; true
	  ),
	  fail
	; true
	).

furthername_name_type(birthname).
furthername_name_type(altname).
furthername_id_type(pseudonym).
furthername_id_type(pseudonym_of).

prt_seeindex(Type, P, N) :-
	furthername_name_type(Type),
	!,
	name_to_idlike_sortkey(N, Sortname),
	format('\\index[per]{\\pgfkeysvalueof {person/~w}@~w|see{\\pgfkeysvalueof {person/~w}}}%~n',
	       [Sortname, N, P]).
prt_seeindex(Type, P, P1) :-
	furthername_id_type(Type),
	!,
	format('\\index[per]{\\pgfkeysvalueof {person/~w}|seealso{\\pgfkeysvalueof {person/~w}}}%~n',
	       [P1, P]).


%     \index{Entry Y@\hypertarget{Entry Y}{Entry Y}}
%     \index{Entry X|see{\hyperlink{Entry Y}{Entry Y}}}
