%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(nlx,
	  [nlx_enum/2,
	   nlxelem_enum_plaintxt/2,
	   nlx_used_cmds/2,
	   lx_to_nlx/2]).

:- use_module(swilib(err)).
:- use_module(kbset_base(readwrite)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% nlx format
%%%%
%%%% Like the lx format (see kbset/readwrite.pl) with exception that commands
%%%% and environments are represented in a nested fashion.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% For environments {letter}[5] and {enclosure}[2]
%%%% 
:- initialization(( register_latex_command(begin,[c,c,c,c,c,c],''),
		    register_latex_command(begin,[c,c,c],'') )).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% 
%%%% 
nlx_enum(Es, E) :-
	member(E1, Es),
	( E = E1
	; nlx_enum_1(E1, E)
	).

nlx_enum_1(nenv(_,_,_,B), E) :-
	nlx_enum(B, E).
nlx_enum_1(ncmd(_,_,_,Bs), E) :-
	member(B, Bs),
	nlx_enum(B, E).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% 
%%%% 
nlx_used_cmds(NLX, Cmds) :-
	findall(C, nlx_enum_cmd(NLX, C), Cmds1),
	sort(Cmds1, Cmds).

nlx_enum_cmd(NLX, C) :-
	member(E, NLX),
	nlx_enum_cmd_1(E, C).

nlx_enum_cmd_1(nenv(Env,_,_,_), env(Env)).
nlx_enum_cmd_1(nenv(_,_,_,B), C) :-
	nlx_enum_cmd(B, C).
nlx_enum_cmd_1(ncmd(C,_,_,_), C).
nlx_enum_cmd_1(ncmd(C,_,_,Bs), C) :-
	member(B, Bs),
	nlx_enum_cmd(B, C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% 
%%%% 
nlxelem_enum_plaintxt(nenv(_, _, _, X), Y) :-
	!,
	member(Z, X),
	nlxelem_enum_plaintxt(Z, Y).
nlxelem_enum_plaintxt(ncmd(_, _, _, X), Y) :-
	!,
	member(Z1, X),
	member(Z, Z1),
	nlxelem_enum_plaintxt(Z, Y).
nlxelem_enum_plaintxt(X, Y) :-
	lxelem_enum_plaintxt(X, Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% 
%%%% 
lx_to_nlx([], []) :-
	!.
lx_to_nlx(Es, [E|Es1]) :-
	Es \= [],
	lx_to_nlx_1(Es, E, Es2),
	lx_to_nlx(Es2, Es1).

lx_to_nlx_1([cmd(begin,A)|Es], nenv(Env, Opts, Args, Body), Es2) :-
	!,
 	eparse_args(A, Es, Opts, Args1, _, Es3),
	Args1 = [Env|Args],
	lx_to_nlx_2(Es3, Body, Es2).
lx_to_nlx_1([cmd(Cmd,A)|Es], ncmd(Cmd, Opts, Args, Sub), Es2) :-
	!,
	eparse_args(A, Es, Opts, Args, Sub, Es2).
lx_to_nlx_1([E|Es], E, Es).

lx_to_nlx_2([end|Es], [], Es) :-
	!.
lx_to_nlx_2([cmd(end,A)|Es], [], Es1) :-
 	!,
 	eparse_args(A, Es, _, _, _, Es1).
lx_to_nlx_2(Es, [E|Es1], Es2) :-
	lx_to_nlx_1(Es, E, Es3),
	lx_to_nlx_2(Es3, Es1, Es2).

eparse_args([], Es, [], [], [], Es) :-
	!.
eparse_args([o|A], [opt(X)|Es], [X|Xs], Ys, Zs, Es1) :-
	!,
	eparse_args(A, Es, Xs, Ys, Zs, Es1).
eparse_args([o|A], [ctr(X)|Es], Xs, Ys, Zs, Es1) :-
	!,
	eparse_args(A, [ctr(X)|Es], Xs, Ys, Zs, Es1).
eparse_args([o|A], [beg|Es], Xs, Ys, Zs, Es1) :-
	!,
	eparse_args(A, [beg|Es], Xs, Ys, Zs, Es1).
eparse_args([c|A], [ctr(Y)|Es], Xs, [Y|Ys], Zs, Es1) :-
	!,
	eparse_args(A, Es, Xs, Ys, Zs, Es1).
eparse_args([t|A], [beg|Es], Xs, Ys, [Z|Zs], Es1) :-
	!,
	lx_to_nlx_2(Es, Z, Es3),
	eparse_args(A, Es3, Xs, Ys, Zs, Es1).
eparse_args(A, [spc(_)|Es], Xs, Ys, Zs, Es1) :-
	!,
	eparse_args(A, Es, Xs, Ys, Zs, Es1).
eparse_args(A, [com(_)|Es], Xs, Ys, Zs, Es1) :-
	!,
	eparse_args(A, Es, Xs, Ys, Zs, Es1).
eparse_args(A, [E|_], _, _, _, _) :-
	err('Failed to parse next argument in ~w from ~q', [A, E]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%