%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(reorder_letters,
	  [generate_reordered/2,
	   % generate_reordered_split/2,
	   generate_annots/3
	   ]).

:- use_module(kbset_base(filename)).
:- use_module(kbset_base(readwrite)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_letters(date_handling)).
:- use_module(kbset_letters(letters_config)).

:- dynamic(letterid_sortkey/2).

generate_reordered(InFileSpecs, OutFileSpec) :-
	findall(F, (member(F1, InFileSpecs), filename(F1, F)), Files),
	filename(OutFileSpec, Outfile),
	read_letter_sources(Files, Ls),
	length(Ls, Len),
	info(10, 'Total number of letters ~w', [Len]),
	map_add_letter_sortkey(Ls, Ls1),
	set_letterid_sortkey(Ls1),
	( OutFileSpec = none ->
	  true
	; keysort(Ls1, Ls2),
	  map_rm_key(Ls2, Ls3),
	  info(10, 'Writing ~w', [Outfile]),
	  onto_file(( member(L, Ls3),
		      lxelems_write_src(L),
		      nl,
		      nl,
		      fail
		    ; true
		    ),
		    Outfile)
	).

set_letterid_sortkey(KLs) :-
	retractall(letterid_sortkey(_, _)),
	( member(Key-_, KLs),
	  Key = k(_Date, _From, _To, Id),
	  assert(letterid_sortkey(Id, Key)),
	  fail
	; true
	).

id_letter(Id, [cmd(begin, _), ctr(letter), ctr(Id)|_]).
id_letter(Id, [cmd(begin, _), ctr(brief), ctr(Id)|_]).

split_letters(Bs, [], [Bs]) :-
	!.
split_letters([B|Bs], Ids, [[B|Bs1]|Bs2]) :-
	!,
	split_letters(Bs, Ids, Bs1, Bs2).
split_letters([], _, []).

split_letters([B|Bs], Ids, [], [[B|Bs1]|Bs2]) :-
	member(Id, Ids),
	id_letter(Id, B),
	!,
	split_letters(Bs, Ids, Bs1, Bs2).
split_letters([B|Bs], Ids, [B|Bs1], Bs2) :-
	split_letters(Bs, Ids, Bs1, Bs2).
split_letters([], _, [], []).

generate_reordered_split(InFileSpecs, Separators) :-
	findall(F, (member(F1, InFileSpecs), filename(F1, F)), Files),
	read_letter_sources(Files, Ls),
	length(Ls, Len),
	info(10, 'Total number of letters: ~w', [Len]),
	map_add_letter_sortkey(Ls, Ls1),
	keysort(Ls1, Ls2),
	map_rm_key(Ls2, Ls3),
	split_letters(Ls3, Separators, Lss4),
	length(Lss4, LenG),
	info(10, 'Number of letter groups: ~w', [LenG]),
	map_grs_1(Lss4, 1).

map_grs_1([Ls|Lss], I) :-
	length(Ls, Len),
	info(10, 'Number of letters in group ~w: ~w', [I, Len]),
	format(atom(OutFileSpec), 'generate_reordered_~|~`0t~d~2+', [I]),
	filename(OutFileSpec, Outfile),
	info(10, 'Writing ~w', [Outfile]),
	onto_file(( member(L, Ls),
		    lxelems_write_src(L),
		    nl,
		    nl,
		    fail
		  ; true
		  ),
		  Outfile),
	I1 is I+1,
	map_grs_1(Lss, I1).
map_grs_1([], _).


%%%% 
%%%% For environment {letter}[5] 
%%%%
:- initialization(( register_latex_command(begin,[c,c,c,c,c,c],'' ))).

read_letter_sources(Files, Ls) :-
	findall(L, ( member(File, Files),
		     read_letter_source(File, Ls1),
		     member(L, Ls1)
		   ),
		Ls).

read_letter_source(X, Ls) :-
	info(10, 'Reading letter source ~w', [X]),
	read_in_latex(X, Es),
	info(10, 'Extracting letters', []),
	eparse_letters(Es, Ls),
	length(Ls, N),
	info(10, 'Read ~w letters', [N]).

eparse_letters(Es, [L|Ls]) :-
	eparse_letter(Es, L, Es1),
	!,
	eparse_letters(Es1, Ls).
eparse_letters(_, []).

eparse_letter([cmd(begin,[c,c,c,c,c,c])|Es],
	      Meta-[cmd(begin,[c,c,c,c,c,c])|L], Es1) :-
	eparse_ctr(6, Es, [letter|Args], L-R, Es2),
	!,
	Meta =.. [letter|Args],
	to_env_end(Es2, letter, R, Es1).
eparse_letter([cmd(begin,[c,c,c,c,c,c])|Es],
	      Meta-[cmd(begin,[c,c,c,c,c,c])|L], Es1) :-
	eparse_ctr(6, Es, [brief|Args], L-R, Es2),
	!,
	Meta =.. [letter|Args],
	to_env_end(Es2, brief, R, Es1).
eparse_letter([E|Es], L, Es1) :-
	( E \= spc(_), E \= com(_) ->
	  up_to_6_elems(Es, Ctx),
	  info(10, 'Skipping between letters: ~q, before: ~q', [E, Ctx])
	; info(50, 'Skipping between letters: ~q', [E])
	),
	eparse_letter(Es, L, Es1).

up_to_6_elems([A,B,C,D,E,F|_], [A,B,C,D,E,F]) :- !.
up_to_6_elems(X, X).

to_env_end([cmd(end, [c])|Es], ENV, [cmd(end, [c])|L], Es1) :-
	eparse_ctr(1, Es, [ENV], L-[], Es1),
	!.
to_env_end([E|Es], ENV, [E|L], Es1) :-
	to_env_end(Es, ENV, L, Es1).

eparse_ctr(0, Es, [], R-R, Es) :-
	!.
eparse_ctr(N, [ctr(X)|Es], [X|Xs], [ctr(X)|Es1]-R, Es2) :-
	!,
	N1 is N - 1,
	eparse_ctr(N1, Es, Xs, Es1-R, Es2).
eparse_ctr(N, [spc(S)|Es], Xs, [spc(S)|Es1]-R, Es2) :-
	!,
	eparse_ctr(N, Es, Xs, Es1-R, Es2).
eparse_ctr(N, [com(S)|Es], Xs, [com(S)|Es1]-R, Es2) :-
	!,
	eparse_ctr(N, Es, Xs, Es1-R, Es2).
eparse_ctr(_, _, _, _, _) :-
	fail.

add_letter_sortkey(Meta-L, K-L) :-
	( Meta = letter(Id, A, B, _, D) ->
	  true
	; err('Bad letter meta format: ~q', [Meta])
	),
	parse_ddate(D, D1),
	check_ddate(D1, D-Id),
	ddate_sortkey(D1, DK),
	K = k(DK,A,B,Id).

map_add_letter_sortkey([X|Xs], [X1|Xs1]) :-
	add_letter_sortkey(X, X1),
	map_add_letter_sortkey(Xs, Xs1).
map_add_letter_sortkey([], []).

map_rm_key([_-X|Xs], [X|Xs1]) :-
	map_rm_key(Xs, Xs1).
map_rm_key([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% For Annotations (Kommentare)
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- initialization(( register_latex_command(begin,[c,c],'' ))).

generate_annots(InFileSpecs, Options, OutFileSpec) :-
	findall(F, (member(F1, InFileSpecs), filename(F1, F)), Files),
	filename(OutFileSpec, Outfile),
	read_annot_sources(Files, Ks),
	length(Ks, Len),
	info(10, 'Total number of annots ~w', [Len]),
	( memberchk(rm_empty_ksections, Options) ->
	  info(10, 'Removing empty ksections'),
	  map_remove_empty_ksections(Ks, Ks01)
	; Ks01 = Ks
	),
	( memberchk(sort_ksections, Options) ->
	  info(10, 'Initializing sorting of ksections'),
	  init_sort_ksections,
	  ( \+ ksection_title_sortkey(_, _ ) ->
	    Ks1 = Ks01
	  ; info(10, 'Sorting ksections'),
	    map_sort_ksections(Ks01, Ks1)
	  )
	; Ks1 = Ks01
	),
	( memberchk(reorder, Options) ->
	  ( \+ letterid_sortkey(_, _) ->
	    err('Tried to reorder annots without reordering letters before')
	  ; info(10, 'Reordering annots according to letters'),
	    map_add_annot_sortkey(Ks1, Ks2),
	    keysort(Ks2, Ks3),
	    ( Ks2 == Ks3 ->
	      info(10, 'Input ordering of annots was preserved')
	    ; info(10, 'WARNING: Input ordering of annots was not preserved'),
	      first_keydiff(Ks3, Ks2, KK3, KK2),
	      info(10, 'WARNING: First difference to input ordering: Expected: ~w Input: ~w',
		   [KK3, KK2])
	    )
	  )
	; Ks3 = Ks1
	),
	map_rm_key(Ks3, Ks4),
	info(10, 'Writing ~w', [Outfile]),
	onto_file(( member(K, Ks4),
		    lxelems_write_src(K),
		    nl,
		    nl,
		    fail
		  ; true
		  ),
		  Outfile).

first_keydiff([X-_|Xs], [Y-_|Ys], X1, Y1) :-
	X == Y,
	!,
	first_keydiff(Xs, Ys, X1, Y1).
first_keydiff([X-_|_], [Y-_|_], X, Y) :-
	!.
first_keydiff(_, _, none, none).

add_annot_sortkey(Meta-K, KEY-K) :-
	( Meta = annot(Id) ->
	  ( letterid_sortkey(Id, KEY) -> true
	  ; info(10, 'WARNING: No annot sortkey found for ~q', [Id]),
	    KEY = z(z,z,z,z,z,z,z,z,z,z)
	  )
	; err('Bad annot meta format: ~q', [Meta])
	).

map_add_annot_sortkey([X|Xs], [X1|Xs1]) :-
	add_annot_sortkey(X, X1),
	map_add_annot_sortkey(Xs, Xs1).
map_add_annot_sortkey([], []).

read_annot_sources(Files, Ks) :-
	findall(K, ( member(File, Files),
		     read_annot_source(File, Ks1),
		     member(K, Ks1)
		   ),
		Ks).

read_annot_source(X, Ks) :-
	info(10, 'Reading annot source ~w', [X]),
	read_in_latex(X, Es),
	info(10, 'Extracting annots', []),
	eparse_annots(Es, Ks),
	length(Ks, N),
	info(10, 'Read ~w annots', [N]).

eparse_annots(Es, [K|Ks]) :-
	eparse_annot(Es, K, Es1),
	!,
	eparse_annots(Es1, Ks).
eparse_annots(_, []).

eparse_annot([cmd(begin,[c,c])|Es],
	    Meta-[cmd(begin,[c,c])|K], Es1) :-
	eparse_ctr(2, Es, [annotation,Arg], K-R, Es2),
	!,
	Meta = annot(Arg),
	to_env_end(Es2, annotation, R, Es1).
eparse_annot([cmd(begin,[c,c])|Es],
	    Meta-[cmd(begin,[c,c])|K], Es1) :-
	eparse_ctr(2, Es, [kommentar,Arg], K-R, Es2),
	!,
	Meta = annot(Arg),
	to_env_end(Es2, kommentar, R, Es1).
eparse_annot([E|Es], K, Es1) :-
	( E \= spc(_), E \= com(_) ->
	  up_to_6_elems(Es, Ctx),
	  info(10, 'Skipping between annots: ~q, before: ~q', [E, Ctx])
	; info(50, 'Skipping between annots: ~q', [E])
	),
	eparse_annot(Es, K, Es1).

remove_empty_ksections(M-Es, M-Es1) :-
	rmek(Es, Es1).

map_remove_empty_ksections([X|Xs], [X1|Xs1]) :-
	remove_empty_ksections(X, X1),
	map_remove_empty_ksections(Xs, Xs1).
map_remove_empty_ksections([], []).

rmek([E|Es], Es1) :-
	E = cmd(ksection, _),
	!,
	rmek1(Es, [E], Es1).
rmek([E|Es], [E|Es1]) :-
	rmek(Es, Es1).
rmek([], []).

rmek1([E|Es], REs, Es1) :-
	E = cmd(ksection, _),
	!,
	add_revksection(REs, Es2, Es1),
	rmek([E|Es], Es2).
rmek1(EndOfAnnot, REs, Es1) :-
	EndOfAnnot = [cmd(end, _), ctr(annotation)],
	!,
	add_revksection(REs, EndOfAnnot, Es1).
rmek1(EndOfAnnot, REs, Es1) :-
	EndOfAnnot = [cmd(end, _), ctr(kommentar)],
	!,
	add_revksection(REs, EndOfAnnot, Es1).
rmek1([E|Es], REs, Es1) :-
	!,
	rmek1(Es, [E|REs], Es1).
rmek1(Es, REs, _) :-
	err('Unexpected annot parts: ~q ~q', [Es, REs]).

add_revksection(REs, Rest, Es) :-
	reverse(REs, Es1),
	( is_nonempty_ksection(Es1) ->
	  append(Es1, Rest, Es)
	; Es = Rest
	).

is_nonempty_ksection([_|Es]) :-
	eskip_t(Es, Es1),
	member(E, Es1),
	E \= spc(_),
	!.

eskip_t(E, E1) :-
	eskip_t(E, 0, E1).

eskip_t([beg|E], N, E1) :-
	!,
	N1 is N+1,
	eskip_t(E, N1, E1).
eskip_t([end|E], 1, E) :-
	!.
eskip_t([end|E], N, E1) :-
	!,
	N1 is N-1,
	eskip_t(E, N1, E1).
eskip_t([_|E], N, E1) :-
	eskip_t(E, N, E1).
eskip_t([], _, _) :-
	err('Unable to skip LaTeX argument').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic ksection_title_sortkey/2.
	
init_sort_ksections :-
	ensure_letters_initialized,
	get_kbset_config(annotation_section_order, Titles, []),
	retractall( ksection_title_sortkey(_, _ ) ),
	map_set_ksection_title_sortkey(Titles, 0).

map_set_ksection_title_sortkey([X|Xs], N) :-
	assertz(ksection_title_sortkey(X, N)),
	N1 is N+1,
	map_set_ksection_title_sortkey(Xs, N1).
map_set_ksection_title_sortkey([], _).
	
map_sort_ksections([X|Xs], [X1|Xs1]) :-
	sort_ksections(X, X1),
	map_sort_ksections(Xs, Xs1).
map_sort_ksections([], []).

map_add_ksection_sortkey([X|Xs], [K-X|Xs1]) :-
	X = section(Title, _),
	( ksection_title_sortkey(Title, K) ->
	  true
	; info(10, 'WARNING: Unspecified annotation section order for ~q',
	       [Title]),
	  K = z(Title)
	),
	map_add_ksection_sortkey(Xs, Xs1).
map_add_ksection_sortkey([], []).

sort_ksections(K-X, K-X1) :-
	( phrase(lx_annot(H,S,F), X, []) ->
	  true
	; err('Failed to parse annotation for ~q', [K])
	),
	map_add_ksection_sortkey(S, S1),
	keysort(S1, S2),
	map_val(S2, S3),
	findall(E, ( member(E, H)
		   ; member(section(_, Es), S3),
		     member(E, Es)
		   ; member(E, F)
		   ),
		X1).

lx_annot(Header, Sections, Footer) -->
	lx_annot_header(Header),
	!,
	lx_annot_sections(Sections),
 	lx_annot_footer(Footer),
	!.

lx_annot_section_intro(Title, Xs) -->
	[X1, X2],
	{ X1 = cmd(ksection, [t]),
	  X2 = beg
	},
	lx_items(Xs3),
	[X4],
	{ X4 = end },
	{ append([X1,X2|Xs3], [X4], Xs),
	  findall(W, member(wrd(W), Xs3), Ws),
	  atomic_list_concat(Ws, ' ', Title)
	}.

lx_items([]) -->
	[].
lx_items([X|Xs]) -->
	[X],
	lx_items(Xs).

lx_annot_section_main([X|Xs]) -->
	[X],
	{ X \=  cmd(ksection, _) },
	lx_annot_section_main(Xs).
lx_annot_section_main([]) -->
	[].

lx_annot_footer([X1,X2]) -->
	[X1,X2],
	{ X1 = cmd(end,[c]),
	  ( X2 = ctr(annotation)
	  ; X2 = ctr(kommentar)
	  ) }.

lx_annot_sections([S|Ss]) -->
	lx_annot_section(S),
	lx_annot_sections(Ss).
lx_annot_sections([]) -->
	[].

lx_annot_section(section(Title, Xs)) -->
	lx_annot_section_intro(Title, Xs1),
	!,
	lx_annot_section_main(Xs2),
	{ append(Xs1, Xs2, Xs) }.

lx_annot_header([X|Xs]) -->
	[X],
	{ X \= cmd(ksection, _) },
	!,
	lx_annot_header(Xs).
lx_annot_header([]) -->
	[].

map_val([_-X|Xs], [X|Xs1]) :-
	map_val(Xs, Xs1).
map_val([], []).


