%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(textconvert,
	  [nlx_to_h/2,
	   nlx_to_h_inpar/2,
	   bib_reference/4,
	   global_context_lid/1,
	   set_global_context_lid/1,
	   register_html_conversion/2
	  ]).

:- use_module(kbset_base(readwrite)).
:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_letters(nlx)).
:- use_module(kbset_letters(kbmappings)).
:- use_module(kbset_letters(data_interface), [bibrefdata/5,stdize_id/2]).
:- use_module(kbset_letters(legacy)).
:- use_module(kbset_letters(extract_letters),
	      [letter_kitem_id/2, letter_xl/3]).

:- dynamic global_context_lid/1.

set_global_context_lid(Lid) :-
	retractall( global_context_lid(_) ),
	assert( global_context_lid(Lid) ).

is_par_breaking_spc(spc(S)) :-
	atom_chars(S, S1),
	findall(k, member('\n', S1), K),
	length(K, LK),
	LK >= 2.

split_into_pars([], []).
split_into_pars([X|Xs], Xs1) :-
	(( is_at_par_level(X) ; is_par_breaking_spc(X)) ->
	  split_into_pars_1([X|Xs], _, Xs1)
	; Xs1 = [par(Ys)|Xs2],
	  split_into_pars_1([X|Xs], Ys, Xs2)
	).

split_into_pars_1([X|Xs], [], [top([X])| Xs1]) :-
	is_at_par_level(X),
	!,
	split_into_pars(Xs, Xs1).

split_into_pars_1([X|Xs], [], Xs1) :-
	is_par_breaking_spc(X),
	!,
	split_into_pars(Xs, Xs1).
split_into_pars_1([X|Xs], [X|Ys], Xs1) :-
	!,
	split_into_pars_1(Xs, Ys, Xs1).
split_into_pars_1([], [], []).

remove_empty_pars(Ps, Ps1) :-
	findall(P, (member(P, Ps), par_is_nonempty(P)), Ps1).

remove_headers_of_empty_sections([top([X]),top([Y])|Ps], Ps1) :-
	is_section_header(X),
	is_section_header(Y),
	!,
	remove_headers_of_empty_sections([top([Y])|Ps], Ps1).
remove_headers_of_empty_sections([P|Ps], [P|Ps1]) :-
	remove_headers_of_empty_sections(Ps, Ps1).
remove_headers_of_empty_sections([], []).	

par_is_nonempty(top(_)) :-
	!.
par_is_nonempty(par(X)) :-
	member(X1, X),
	\+ memberchk(X1, [spc(_), com(_)]),
	!.

nlx_to_h(NLX, Hs) :-
	split_into_pars(NLX, Pars),
	remove_empty_pars(Pars, Pars1),
	remove_headers_of_empty_sections(Pars1, Pars2),
	findall(H, (member(P, Pars2), npx_to_h(P, H)), Hs).

npx_to_h(par(NLX), H) :-
	!,
	H = e(p, [class=xstd], Hs),
	nlx_to_h_inpar(NLX, Hs).
npx_to_h(top(NLX), H) :-
	nlx_to_h_inpar_1(NLX, H).

nlx_to_h_inpar(NLX, Hs) :-
	findall(H, nlx_to_h_inpar_1(NLX, H), Hs1),
	map_swallow_double_periods(Hs1, Hs).

nlx_to_h_inpar_1(NLX, H) :-
	member(N, NLX),
	nlx_to_h_2(N, H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_item_separator(ncmd(kitem,_,_,_)) :- !.
is_item_separator(ncmd(vitem,_,_,_)) :- !.
is_item_separator(ncmd(item,_,_,_)) :- !.

split_into_items([X|Xs], Xs1) :-
	is_item_separator(X),
	!,
	split_into_items_1([X|Xs], _, Xs1).
split_into_items([_|Xs], Xs1) :-
	split_into_items(Xs, Xs1).
split_into_items([], []).

split_into_items_1([X|Xs], [], [item(X,Xs1)|Xs2]) :-
	is_item_separator(X),
	!,
	split_into_items_1(Xs, Xs1, Xs2).
split_into_items_1([X|Xs], [X|Xs1], Xs2) :-
	split_into_items_1(Xs, Xs1, Xs2).
split_into_items_1([], [], []).

process_klist(NLX, H) :-
	findall(H1, process_klist_1(NLX, H1), H).

process_klist_1(NLX, H) :-
	split_into_items(NLX, Items),
	( member(item(ncmd(kitem, _, [L], _), NLX1), Items)
	; member(item(ncmd(vitem, _, [L], _), NLX1), Items)
	),
	nlx_to_h(NLX1, H1),
	( H1 = [e(p, _, H2)|H3] ->
	  append(H2, H3, H4)
	; H4 = H1
	),
	( global_context_lid(Lid), letter_xl(Lid, L, DT1) ->
	  nlx_to_h_inpar(DT1, DT2)
	; DT2 = ['Undefined: ', L]
	),
	item_fragment_name(L, LName),
	text_fragment_ref(L, LText),
	fix_nested_a(e(a, [class=kitem, id=LName, href=LText], DT2), DT3),
	( H = e(dt, [class=kitem], DT3)
	; H = e(dd, [class=kitem], H4)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nlx_is_nonempty(NLX) :-
	nlx_enum(NLX, X),
	\+ memberchk(X, [wrd(_), pun(_)]),
	!.

is_at_par_level(ncmd(CMD, _, _, _)) :-
	par_level_cmd(CMD).
is_at_par_level(nenv(ENV, _, _, _)) :-
	par_level_env(ENV).

par_level_cmd(C) :- legacy_alias(C, C1), !, par_level_cmd(C1).

par_level_cmd(ksection).
par_level_cmd(xdate).
par_level_cmd(xdateEnd).
par_level_cmd(xsalute).
par_level_cmd(xcloser).
par_level_cmd(xcenter).

par_level_env(E) :- legacy_alias(E, E1), !, par_level_env(E1).

par_level_env(xverse).
par_level_env(klist).
par_level_env(enclosure).
par_level_env(inclosure).
par_level_env(center).

is_section_header(ncmd(ksection, _, _, _)).

foreign_language_cmd(xlatin).
foreign_language_cmd(xfrench).
foreign_language_cmd(xenglish).
foreign_language_cmd(xgreek).

tex_unicode_map('&','&').	       
tex_unicode_map(ss,'ß').
tex_unicode_map(l,'ł').
tex_unicode_map(aa,'å').	       
tex_unicode_map(ae,'æ').
tex_unicode_map(oe,'œ').
tex_unicode_map(o,'ø').	       	       
tex_unicode_map('S','§').	       
tex_unicode_map('L','Ł').
tex_unicode_map('AA','Å').	       
tex_unicode_map('AE','Æ').
tex_unicode_map('OE','Œ').
tex_unicode_map('O','Ø').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

translate_pun(',', ',') :- !.
translate_pun('.', '.') :- !.
translate_pun('?', '?') :- !.
translate_pun('!', '!') :- !.
translate_pun('--', '–') :- !. % ndash
translate_pun('---', '—') :- !. % mdash
translate_pun(A, A1) :-
	atom_chars(A, A2),
	translate_pun_1(A2, A3),
	atom_chars(A1, A3).

translate_pun_1(['"','`'|A], ['„'|A1]) :-
	!,
	translate_pun_1(A, A1).
translate_pun_1(['"','\''|A], ['”'|A1]) :-
	!,
	translate_pun_1(A, A1).
translate_pun_1(['~'|A], ['\u00A0'|A1]) :-
	!,
	translate_pun_1(A, A1).
translate_pun_1([C|A], [C|A1]) :-
	translate_pun_1(A, A1).
translate_pun_1([], []).
	
item_fragment_ref(X, Y) :-
	stdize_id(X, X1),
	atomic_list_concat(['#item-', X1], Y).
item_fragment_name(X, Y) :-
	stdize_id(X, X1),
	atomic_list_concat(['item-', X1], Y).

text_fragment_ref(X, Y) :-
	stdize_id(X, X1),
	atomic_list_concat(['#txt-', X1], Y).
text_fragment_name(X, Y) :-
	stdize_id(X, X1),
	atomic_list_concat(['txt-', X1], Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% DEBUG
% fix_nested_a(X, Y) :-
% 	assert(fna(X)),
% 	fail.

fix_nested_a_illegal(X, Y) :-
	fix_nested_a(X, Y),
	( [X] \== Y ->
	  info(10, 'WARNING: Nested "a": ~q ~q', [X,Y])
	; true
	).

fix_nested_a(e(a, A, [e(span, AS, H)]), E) :-
	contains_a(H),
	!,
	E = [e(span, AS, H1)],
	fix_nested_a(e(a, A, H), H1).
fix_nested_a(e(a, A, E), [e(a, A, E1) | E2]) :-
	split_off_span_with_a(E, E1, E2),
	!.
fix_nested_a(e(a, A, H), H1) :-
	contains_a(H),
	!,
	( select(id=_, A, AN) -> true
	; AN = A
	),
	fix_nested_a_1(H, A, AN, H1).
fix_nested_a(E, [E]).

split_off_span_with_a([E|E1], [E|E2], E3) :-
	soswa(E1, E2, E3).

soswa([e(span,S,E)|E1], E2, E3) :-
	contains_a(E),
	!,
	E2 = [],
	E3 = [e(span,S,E)|E1].
soswa([E|E1], [E|E2], E3) :-
	soswa(E1, E2, E3).

contains_a(E) :-
	member(e(a,_,_), E),
	!.
contails_a(E) :-
	member(e(_,_,E1), E),
	contains_a(E1).

fix_nested_a_1(H, A1, AN, H1) :-
	split_a(H, H2),
	map_wrap_fix_nested_a(H2, A1, AN, H1).

map_wrap_fix_nested_a([aelem(E)|Hs], A1, AN, [E|Hs1]) :-
	!,
	map_wrap_fix_nested_a(Hs, A1, AN, Hs1).
map_wrap_fix_nested_a([elems(H)|Hs], A1, AN, [e(a, A1, H)|Hs1]) :-
	map_wrap_fix_nested_a_1(Hs, AN, Hs1).
map_wrap_fix_nested_a([], _, _, []).

map_wrap_fix_nested_a_1([aelem(E)|Hs], AN, [E|Hs1]) :-
	!,
	map_wrap_fix_nested_a_1(Hs, AN, Hs1).
map_wrap_fix_nested_a_1([elems(H)|Hs], AN, [e(a, AN, H)|Hs1]) :-
	map_wrap_fix_nested_a_1(Hs, AN, Hs1).
map_wrap_fix_nested_a_1([], _, []).	

split_a([e(a,X,Y)|Hs], [aelem(e(a,X,Y))|Hs1]) :-
	!,
	split_a(Hs, Hs1).
split_a([X|Hs], [elems([X|Hs1])|Hs2]) :-
	split_a_1(Hs, Hs1, Hs2).
split_a([], []).

split_a_1([e(a,X,Y)|Hs], [], Hs1) :-
	!,
	split_a([e(a,X,Y)|Hs], Hs1).
split_a_1([H|Hs], [H|Hs1], Hs2) :-
	split_a_1(Hs, Hs1, Hs2).
split_a_1([], [], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

quote_work(Es, [e(i, [], Es)]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic special_html_conversion/2.

register_html_conversion(X, Y) :-
	retractall( special_html_conversion(X, _) ),
	asserta( special_html_conversion(X, Y) ).

% special_cmd(xReichstaler, E) :-
% 	E = e(img, [alt='Rthlr.', style='height:2.5ex',
% 		    src='symbols/symbol_reichstaler.svg'], []).


special_cmd(xGulden, E) :-
	E = e(img, [alt='Gulden', style='height:2.5ex;vertical-align:-25%;',
 		    src='glyphs/florin64.png'], []).
% special_cmd(xGulden, 'Gulden').
special_cmd(xGuldenPeriod, 'Gulden.').
special_cmd(xGuldenPunkt, 'Gulden.').
special_cmd(xPfennige, 'Pfg.').
special_cmd(xPfennigePeriod, 'Pfg.').
special_cmd(xPfennigePunkt, 'Pfg.').
special_cmd(xPfund, E) :-
	E = e(img, [alt='Pfund', style='height:1.8ex;vertical-align:-2%;',
 		    src='glyphs/pfund64.png'], []).
% special_cmd(xPfund, '£').
special_cmd(xPfundPeriod, '£.').
special_cmd(xPfundPunkt, '£.').
% special_cmd(xReichstaler, 'Rthlr.').
% special_cmd(xReichstalerPeriod, 'Rthlr.').
% special_cmd(xReichstalerPunkt, 'Rthlr.').
special_cmd(xReichstaler, E) :-
	E = e(img, [alt='Rthlr.', style='height:2.5ex;vertical-align:-22%;',
 		    src='glyphs/reichstaler64.png'], []).
special_cmd(xReichstalerPeriod, E) :-
	E = e(span, [], [e(img, [alt='Rthlr.',
				 style='height:2.5ex;vertical-align:-22%;',
				 src='glyphs/reichstaler64.png'], []),
			 '.']).
special_cmd(xReichstalerPunkt, E) :-
	E = e(span, [], [e(img, [alt='Rthlr.',
				 style='height:2.5ex;vertical-align:-22%;',
				 src='glyphs/reichstaler64.png'], []),
			 '.']).
special_cmd(xSchilling, E) :-
	E = e(img, [alt='Schilling', style='height:2.5ex;vertical-align:-25%;',
 		    src='glyphs/schilling64.png'], []).

bib_reference(st, O, L, H) :-	
	bibrefdata(L, Authors, _AuthorsShort, Year, Shorttitle),
	!,
	id_url(bib, L, URL),
	Pretty = [Authors, ' ', e(i, [], Shorttitle), ' ', Year],
	bib_append_option(Pretty, O, Pretty1),
	H = e(a, [class=bibref, href=URL], Pretty1).
bib_reference(ay, O, L, H) :-	
	bibrefdata(L, _Authors, AuthorsShort, Year, _Shorttitle),
	!,
	id_url(bib, L, URL),
	Pretty = [AuthorsShort, ' ', Year],
	bib_append_option(Pretty, O, Pretty1),
	H = e(a, [class=bibref, href=URL], Pretty1).
bib_reference(_, O, L, H) :-	
	id_url(bib, L, URL),
	Pretty = [L],
	bib_append_option(Pretty, O, Pretty1),
	H = e(a, [class=bibref, href=URL], Pretty1).

bib_append_option(Pretty, O, Pretty1) :-
	( O = [O1], O1 \= '' ->
	  atom_codes(O1, O2), read_in_latex(O2, O3), nlx_to_h_inpar(O3, O4),
	  append(Pretty, [', '|O4], Pretty1)
	; Pretty1 = Pretty
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

href_cmd(A, B, C) :- legacy_alias(A, A1), !, href_cmd(A1, B, C).
href_cmd(xcorp, xcorp, corp).
href_cmd(xcorpmain, xjournal, corp).
href_cmd(xjournal, xjournal, journal).
href_cmd(xjournalmain, xjournal, journal).
href_cmd(xperson, xperson, person).
href_cmd(xpersonmain, xperson, person).
href_cmd(xwork, xwork, work).
href_cmd(xworkmain, xwork, work).
href_cmd(xgeo, xgeo, geo).
href_cmd(xgeomain, xgeo, geo).
href_cmd(xevent, xevent, event).
href_cmd(xeventmain, xevent, event).
href_cmd(xsubject, xsubject, subject).
href_cmd(xsubjectmain, xsubject, subject).

href_cmd_q(A, B, C) :- legacy_alias(A, A1), !, href_cmd_q(A1, B, C).
href_cmd_q(xworkq, xwork, work).
href_cmd_q(xjournalq, xjournal, work).

:- flag(in_quote, _, 0).

nlx_to_h_2(ncmd(C, [], [], []), H) :-
	special_html_conversion(C, H),
	!.
nlx_to_h_2(ncmd(xfrac, [], [], [A,B]), H) :-
	!,
	nlx_to_h_inpar(A, A1),
	nlx_to_h_inpar(B, B1),
	H = e(span, [], [e(sup, [], A1), xml('&frasl;'), e(sub, [], B1)]).
nlx_to_h_2(ncmd(xhoch, [], [], [A]), H) :-
	!,
	nlx_to_h_inpar(A, A1),
	H = e(sup, [], A1).

nlx_to_h_2(nenv(xverse, _, _, X), H) :-
	!,
	nlx_to_h_inpar(X, Hs),
	H = e(p, [class=xpoem], Hs).
nlx_to_h_2(nenv(center, _, _, X), H) :-
	!,
	nlx_to_h_inpar(X, Hs),
	H = e(div, [class=xcenter], Hs).
%%%%
%%%%
nlx_to_h_2(ncmd(Cmd, _, [U], [T]), H) :-
	href_cmd(Cmd, Class, UrlType),
	!,
	nlx_to_h_inpar(T, Hs),
	id_url(UrlType, U, U1),
	fix_nested_a_illegal(e(a, [class=Class, href=U1], Hs), Hs1),
	member(H, Hs1).
nlx_to_h_2(ncmd(Cmd, _, [U], [T]), H) :-
	href_cmd_q(Cmd, Class, UrlType),
	!,
	nlx_to_h_inpar(T, Hs),
	id_url(UrlType, U, U1),
	quote_work(Hs, Hs1),
	fix_nested_a_illegal(e(a, [class=Class, href=U1], Hs1), Hs2),
	member(H, Hs2).
%%%%
nlx_to_h_2(ncmd(xl, _, [L], [T]), H) :-
	!,
	nlx_to_h_inpar(T, Hs),
	( global_context_lid(Lid), letter_kitem_id(Lid, L) ->
	  item_fragment_ref(L, LR),
	  text_fragment_name(L, LN),
	  fix_nested_a(e(a, [class=xl, id=LN, href=LR], Hs), Hs1)
	; Hs1 = Hs
	),
	member(H, Hs1).
nlx_to_h_2(ncmd(xll, _, [L|_], _), H) :-
	!,
	( global_context_lid(Lid), letter_kitem_id(Lid, L) ->
	  item_fragment_ref(L, LR),
	  text_fragment_name(L, LN),
	  fix_nested_a(e(a, [class=xl, id=LN, href=LR],
			 ['[',xml('&rarr;'),']']), Hs1)
	; Hs1 = []
	),
	member(H, Hs1).
%%%%
nlx_to_h_2(ncmd(CMD, _, _, [T]), H) :-
	convert_inpar(CMD, Hs, Hs1),
	!,
	nlx_to_h_inpar(T, Hs),
	member(H, Hs1).
nlx_to_h_2(ncmd(xxx, _, _, []), H) :-
	!,
	H = e(span, [class=xxx], ['XXX']).
%%%%
nlx_to_h_2(ncmd(xciteAY, O, [L], _), H) :-
	!,
	bib_reference(ay, O, L, H).
nlx_to_h_2(ncmd(xciteST, O, [L], _), H) :-
	!,
	bib_reference(st, O, L, H).
%%%%
nlx_to_h_2(nenv(klist, _, _, T), H) :-
	!,
	process_klist(T, Hs),
	H = e(dl, [class=xklist], Hs).
nlx_to_h_2(ncmd(ksection, _, _, [T]), H) :-
	!,
	nlx_to_h_inpar(T, Hs),
	H = e(h3, [class=xksection], Hs).
%%%%
nlx_to_h_2(nenv(inclosure, _, _, X), H) :-
	!,
	( H = e(div, [class=xhinclosure], ['Einschluß'])
	; nlx_to_h(X, Hs),
	  member(H, Hs)
	).
nlx_to_h_2(nenv(enclosure, _, [Id, B|_], X), H) :-
	!,
	stdize_id(Id, Id1),
	( H = e(div, [id=Id1, class=xhenclosure], ['Beilage: ', B])
	; nlx_to_h(X, Hs),
	  member(H, Hs)
	).
%%%%
%%%%
nlx_to_h_2(ncmd('\\', _, _, _), H) :-
	!,
	H = e(br, [], []).
nlx_to_h_2(ncmd(X, _, _, _), Y) :-
	tex_unicode_map(X, Y),
	!.
nlx_to_h_2(ncmd(xquote, _, _, [T]), H) :-
	!,
	flag(in_quote, Q, Q+1),
	nlx_to_h_inpar(T, Hs),
	flag(in_quote, _, Q),
	( Q > 0 ->
	  append(['›'|Hs], ['‹'], Hs1)
	; append(['»'|Hs], ['«'], Hs1)
	),
	H = e(span, [class=xquote], Hs1).
nlx_to_h_2(ncmd(xmeta, _, _, _), _) :-
	!,
	fail.
nlx_to_h_2(ncmd(xhref, _, [U|_], [T|_]), H) :-
	!,
	nlx_to_h_inpar(T, Hs),
	H = e(a, [href=U, class=txthref], Hs).
nlx_to_h_2(ncmd(xurl, _, [U|_], _), H) :-
	!,
	H = e(a, [href=U, class=txthref], [U]).
nlx_to_h_2(ncmd(xletterRef, _, [U|_], _), H) :-
	!,
	id_url(letter, U, URL),
	H = e(a, [href=URL, class=txthref], [URL]).
nlx_to_h_2(ncmd(C, X1, X2, X3), H) :-
	legacy_alias(C, C1),
	!,
	nlx_to_h_2(ncmd(C1, X1, X2, X3), H).
nlx_to_h_2(nenv(C, X1, X2, X3), H) :-
	legacy_alias(C, C1),
	!,
	nlx_to_h_2(nenv(C1, X1, X2, X3), H).
nlx_to_h_2(ncmd(C, [], [], []), H) :-
	special_cmd(C, H),
	!.
%%%%
%%%%
nlx_to_h_2(nenv(tabular, _, [Tabspec|_], X), H) :-
	!,
	( convert_tabular(Tabspec, X, H) ->
	  true
	; info(10, 'WARNING: could not convert a tabular with spec ~w',
	       [Tabspec]),
	  nlx_to_h_inpar_1(X, H)
	).
nlx_to_h_2(nenv(_, _, _, X), H) :-
	!,
	nlx_to_h_inpar_1(X, H).
nlx_to_h_2(ncmd(_, _, _, X), H) :-
	!,
	member(X1, X),
	nlx_to_h_inpar_1(X1, H).
nlx_to_h_2(spc(_), ' ') :-
	!.
nlx_to_h_2(pun(A), H) :-
	!,
	translate_pun(A, H).	
nlx_to_h_2(X, H) :-
	lxelem_enum_plaintxt(X, H),
	H \= ''.

convert_inpar(Type, Hs, Hs1) :-
	legacy_alias(Type, Type1),
	!,
	convert_inpar(Type1, Hs, Hs1).
convert_inpar(C, Hs,         [e(span, [class=xforeign], Hs)]) :-
	foreign_language_cmd(C),
	!.
convert_inpar(xsout, Hs,     [e(del, [class=xsout], Hs)]).
convert_inpar(xul, Hs,       [e(span, [class=xul], Hs)]).
convert_inpar(xulul, Hs,     [e(span, [class=xulul], Hs)]).
convert_inpar(xdate, Hs,     [e(p, [class=xdate], Hs)]).
convert_inpar(xdateEnd, Hs,  [e(p, [class=xdateend], Hs)]).
convert_inpar(xsalute, Hs,   [e(p, [class=xsalute], Hs)]).
convert_inpar(xcloser, Hs,   [e(p, [class=xcloser], Hs)]).
convert_inpar(xcenter, Hs,   [e(div, [class=xcenter], Hs)]).
convert_inpar(xls, Hs,       [e(span, [class=xls], Hs)]).
convert_inpar(textit, Hs,    [e(i, [], Hs)]).
convert_inpar(xqwork, Hs,    [e(i, [], Hs)]).
convert_inpar(xqjournal, Hs, [e(i, [], Hs)]).
convert_inpar(xus, Hs,       ['⟨', e(span, [class=xus], Hs), '⟩']).
convert_inpar(xins, Hs,      [xml('&lceil;'), e(span, [class=xins], Hs), xml('&rceil;')]).
convert_inpar(xinl, Hs,      ['|', e(span, [class=inl], Hs), '|']).
convert_inpar(xmar, Hs,      [xml('&uarr;'), e(span, [class=xmar], Hs), xml('&uarr;')]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

swallow_next_period(e(a, A, E)) :-
	memberchk(class=bibref, A),
	append(_, [Last], E),
	atom(Last),
	sub_atom(Last, _, 1, 0, '.'),
	!.

swallow_double_periods(e(T, A, E), e(T, A, E1)) :-
	!,
	map_swallow_double_periods(E, E1).
swallow_double_periods(X, X).

map_swallow_double_periods([E|E1], [E2|E3]) :-
	swallow_double_periods(E, E2),
	( E1 = ['.'|E4],
	  swallow_next_period(E2) ->
	  true
	; E4 = E1
	),
	map_swallow_double_periods(E4, E3).
map_swallow_double_periods([], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Rudimentary Support for Tables
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

convert_tabular(Tabspec, NLX, H) :-
	( parse_tabspec(Tabspec, ColSpecs), map_to_tcg(ColSpecs, Styles) ->
	  true
	; info(10, 'Failed to parse tabular spec: ~w', [Tabspec])
	),
	split_list_by_separator(NLX, ncmd('\\',_,_,_), Rows),
	map_to_hrow(Rows, Styles, HRows),
	%% Unfortunately it seems not possible to center a whole table in CSS
	%% by the surrounding div. So we need to center the table itself.
	H = e(table, [class=convtable], HRows).

map_to_tcg([vline,X|Xs], [X1|Xs1]) :-
	!,
	to_tcg(X, 'border-left-style:solid;', X1),
	map_to_tcg(Xs, Xs1).
map_to_tcg([X,vline|Xs], [X1|Xs1]) :-
	!,
	to_tcg(X, 'border-right-style:solid;', X1),
	map_to_tcg(Xs, Xs1).
map_to_tcg([X|Xs], [X1|Xs1]) :-
	to_tcg(X, '', X1),
	map_to_tcg(Xs, Xs1).
map_to_tcg([], []).

to_tcg(c, B, Style) :-
	format(atom(Style), 'text-align:center;~w', [B]).
to_tcg(l, B, Style) :-
	format(atom(Style), 'text-align:left;~w', [B]).
to_tcg(r, B, Style) :-
	format(atom(Style), 'text-align:right;~w', [B]).
to_tcg(c(W), B, Style) :-
	format(atom(Style), 'text-align:center;width:~wem;~w', [W, B]).
to_tcg(l(W), B, Style) :-
	format(atom(Style), 'text-align:left;width:~wem;~w', [W, B]).
to_tcg(r(W), B, Style) :-
	format(atom(Style), 'text-align:right;width:~wem;~w', [W, B]).

to_hrow(Row, Styles, H) :-
	split_list_by_separator(Row, pun('&'), Row1),
	map_to_hcell(Row1, Styles, HRow),
	H =  e(tr, [], HRow).

to_hcell(NLX, Style, H) :-
	nlx_to_h_inpar(NLX, H1),
	( var(Style) -> Atts = [class=convtd]
	; Atts = [class=convtd, style=Style]
	),
	H = e(td, Atts, H1).

map_to_hrow([X|Xs], S, [X1|Xs1]) :-
	to_hrow(X, S, X1),
	map_to_hrow(Xs, S, Xs1).
map_to_hrow([], _, []).

map_to_hcell([X|Xs], [S|Ss], [X1|Xs1]) :-
	to_hcell(X, S, X1),
	map_to_hcell(Xs, Ss, Xs1).
map_to_hcell([], _, []).

split_list_by_separator(A, S, [B|U]) :-
	slbs(A, S, B, U).

slbs([E|A], S, [], [B|C]) :-
	\+ \+ E = S,
	!,
	slbs(A, S, B, C).
slbs([E|A], S, [E|B], C) :-
	slbs(A, S, B, C).
slbs([], _, [], []).


parse_tabspec(Atom, Tabspec) :-
	atom_codes(Atom, Codes),
	phrase(tabspec(Tabspec), Codes, []),
	!.

tabspec([C|Cs]) -->
	tabcol_spec(C),
	tabspec(Cs).
tabspec([]) -->
	[].

tabcol_spec(l(W)) --> `L{`, num(W), `em}`.
tabcol_spec(c(W)) --> `C{`, num(W), `em}`.
tabcol_spec(r(W)) --> `R{`, num(W), `em}`.   
tabcol_spec(l) --> `l`.
tabcol_spec(c) --> `c`.
tabcol_spec(r) --> `r`.
tabcol_spec(vline) --> `|`.
	  
num(D) -->
	floatnum(D).
num(D) -->
	natnum(D).

natnum(D) -->
        digit(D0),
	digits(T),
        { number_codes(D, [D0|T])
        }.
floatnum(D) -->
        digit(D0),
	digits(T),
	`.`,
	digit(D1),
	digits(T1),
	{ append([D0|T], [0'.,D1|T1], D2),
	  number_codes(D, D2)
	}.

	
digit(D) --> [D], { code_type(D, digit) }.
digits([D|T]) --> digit(D), !, digits(T).
digits([]) --> [].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
