%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(date_ner,
	  [show_date_vwr/1,
	   e_to_entity_date/4,
	   date_ner_init_ctx/2,
	   date_ner_update_ctx/3,

	   specifies_day/4,
	   day_of_week/4]).

:- use_module(kbset_ner(layers)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_ner(ctx)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_base(localization)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(ner_utils)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





e_to_entity_date(E, CTX, _, _) :-
	e_content(E, wrd(W)),
	assistance_entry(no_entity(date, CReq)),
	memberchk(word=W, CReq),
	context_requirement_matches(CReq, CTX, _MatchReason),
	!,
	fail.
e_to_entity_date(E, CTX, Date, Reason) :-
	e_content(E, wrd(W)),
	assistance_entry(entity(date, DateSpec, CReq)),
	memberchk(word=W, CReq),
	context_requirement_matches(CReq, CTX, MatchReason),
	!,
	Reason = [explicit_date([word=W|MatchReason])],
	normal_to_date(DateSpec, Date1),
	Date = begin_end(Date1).
e_to_entity_date(E, CTX, begin(ignore), [suppressed_date]) :-
	e_get(E, parsed_date, begin(_)),
	assistance_entry(no_entity(date, CReq)),
	ctx_get(CTX, elems_after, Es),
	once(( split_list(Es, Es0, [E1|Es1]),
	       e_get(E1, parsed_date, end)) ),
	ctx_set(CTX, elems_after, Es1, CTX1),
	member(E0, Es0),
	e_content(E0, wrd(W)),
	memberchk(word=W, CReq),
	context_requirement_matches(CReq, CTX1, _MatchReason),
	!.
e_to_entity_date(E, CTX, Date, Reason) :-
	e_get(E, parsed_date, begin(FoundDate)),
	assistance_entry(entity(date, DateSpec, CReq)),
	memberchk(date=CDate, CReq),
	normal_to_date(CDate, FoundDate),
	ctx_get(CTX, elems_after, Es),
	once((mem([E1|Es1], Es), e_get(E1, parsed_date, end))),
	ctx_set(CTX, elems_after, Es1, CTX1),
	context_requirement_matches(CReq, CTX1, MatchReason),
	!,
	Reason = [explicitly_defaulted_parsed_date(FoundDate,
						   [found_date=FoundDate|MatchReason])],
	normal_to_date(DateSpec, Date1),
	Date = begin(Date1).
e_to_entity_date(E, CTX, Date, [complete_parsed_date]) :-
	e_get(E, parsed_date, begin(D)),
	is_complete_date(D, CTX),
	!,
	Date = begin(D).
e_to_entity_date(E, CTX, Date,
		 [defaulted_parsed_date(FoundDate,TemplateDate)]) :-
	e_get(E, parsed_date, begin(FoundDate)),
	ctx_get(CTX, date_ner_info, DNI),
	dni_default_datespec(FoundDate, DNI, Date1, TemplateDate),
	!,
	Date = begin(Date1).
e_to_entity_date(E, _CTX, Date, [incomplete_parsed_date]) :-
	e_get(E, parsed_date, begin(Date1)),
	!,
	Date = begin(Date1).
e_to_entity_date(E, _CTX, Date, []) :-
	e_get(E, parsed_date, end),
	!,
	Date = end.

dni_default_datespec(FoundDate, DNI, Date, TemplateDate) :-
	member_upto(TemplateDate, DNI, 30),
	dni_matches(FoundDate, TemplateDate, Date),
	!.

dni_matches(d(D), ym(Y,M), ymd(Y,M,D)) :-
	!.
dni_matches(d(D), ymd(Y,M,_), ymd(Y,M,D)) :-
	!.
dni_matches(m(M), y(Y), ym(Y,M)) :-
	!.
dni_matches(m(M), ym(Y,_), ym(Y,M)) :-
	!.
dni_matches(m(M), ymd(Y,_,_), ym(Y,M)) :-
	!.
dni_matches(md(M,D), ym(Y,M), ymd(Y,M,D)) :-
	!.
dni_matches(md(M,D), ymd(Y,M1,D1), ymd(Y2,M,D)) :-
	findall(AbsDiff-Y0,
		( ( Y0 = Y ; M1 > 10, Y0 is Y+1 ; M1 < 3, Y0 is Y-1 ),
		  date_diff(ymd(Y,M1,D1), ymd(Y0,M,D), NDays),
		  AbsDiff is abs(NDays) ),
		DiffYs),
	keysort(DiffYs, [AbsDiff1-Y2|_]),
	AbsDiff1 < 50,
	!.
dni_matches(ymd(Y,M,D), ymd(Y1,_,_), ymd(Y1,M,D)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(ymd(Y,M,D), ym(Y1,_), ymd(Y1,M,D)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(ymd(Y,M,D), y(Y1), ymd(Y1,M,D)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(ym(Y,M), ymd(Y1,_,_), ym(Y1,M)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(ym(Y,M), ym(Y1,_), ym(Y1,M)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(ym(Y,M), y(Y1), ym(Y1,M)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(y(Y), ymd(Y1,_,_), y(Y1)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(y(Y), ym(Y1,_), y(Y1)) :-
	is_short_and_full_year(Y, Y1).
dni_matches(y(Y), y(Y1), y(Y1)) :-
	is_short_and_full_year(Y, Y1).
%%%%
dni_matches(ymd(Y,M,D), ymd(Y1,_,_), ymd(Y2,M,D)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(ymd(Y,M,D), ym(Y1,_), ymd(Y2,M,D)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(ymd(Y,M,D), y(Y1), ymd(Y2,M,D)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(ym(Y,M), ymd(Y1,_,_), ym(Y2,M)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(ym(Y,M), ym(Y1,_), ym(Y2,M)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(ym(Y,M), y(Y1), ym(Y2,M)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(y(Y), ymd(Y1,_,_), y(Y2)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(y(Y), ym(Y1,_), y(Y2)) :-
	short_to_full_year_like(Y, Y1, Y2).
dni_matches(y(Y), y(Y1), y(Y2)) :-
	short_to_full_year_like(Y, Y1, Y2).



date_diff(ymd(Y1,M1,D1), ymd(Y2,M2,D2), NDays) :-
	date_time_stamp(date(Y1,M1,D1,0,0,0,0,-,-), S1),
	date_time_stamp(date(Y2,M2,D2,0,0,0,0,-,-), S2),
	NDays is round((S1 - S2) / 86400).
	
is_short_and_full_year(Y, Y1) :-
	format(atom(Y2), '~|~`0t~d~2+', [Y]),
	atom_chars(Y2, [C1, C2]),
	format(atom(Y3), '~|~`?t~d~4+', [Y1]),
	atom_chars(Y3, [C3, C4, C1, C2]),
	char_type(C3, digit),
	char_type(C4, digit).

short_to_full_year_like(Y, Y1, Y2) :-
	Y < 100,
	format(atom(Y3), '~|~`0t~d~2+', [Y]),
	atom_chars(Y3, [C1, C2]),
	format(atom(Y4), '~|~`?t~d~4+', [Y1]),
	atom_chars(Y4, [C3, C4, _, _]),
	char_type(C3, digit),
	char_type(C4, digit),
	number_chars(Y5, [C3, C4, C1, C2]),
	DIFF_THRESHOLD = 10,
	( abs(Y5 - Y1) < DIFF_THRESHOLD ->
	  Y2 = Y5
	; abs((Y5+100) - Y1) < DIFF_THRESHOLD ->
	  Y2 is Y5+100
	; abs((Y5-100) - Y1) < DIFF_THRESHOLD ->
	  Y2 is Y5-100
	).
	  
specifies_day(ymd(Y,M,D), Y, M, D).
%% specifies_day(ya(Y,A), Y, M, D).


is_complete_date(ymd(Y,_,_), _CTX) :- Y >= 1000.
is_complete_date(ym(Y,_), _CTX) :- Y >= 1000.
is_complete_date(y(Y), _CTX) :- Y >= 1000.
is_complete_date(ya(Y,_), _CTX) :- Y >= 1000.

%% those to default:
%% a, md


mk_date(D, date(D)).

date_ner_init_ctx(CTX, CTX1) :-
	ctx_set(CTX, date_ner_info, [], CTX1).
date_ner_update_ctx(CTX, D, CTX1) :-
	ctx_get(CTX, date_ner_info, DI),
	ctx_set(CTX, date_ner_info, [D|DI], CTX1).

day_of_week(Y, M, D, DOW) :-
        user_language(Lang),
	day_of_the_week(date(Y,M,D), W),
	dow(Lang, W, DOW).
	    
dow(de, W, DOW) :-
	!,
	( W=1 -> DOW = 'Montag'
	; W=2 -> DOW = 'Dienstag'
	; W=3 -> DOW = 'Mittwoch'
	; W=4 -> DOW = 'Donnerstag'
	; W=5 -> DOW = 'Freitag'
	; W=6 -> DOW = 'Samstag'
	; W=7 -> DOW = 'Sonntag'
	).

dow(_, W, DOW) :-
	!,
	( W=1 -> DOW = 'Monday'
	; W=2 -> DOW = 'Tuesday'
	; W=3 -> DOW = 'Wednesday'
	; W=4 -> DOW = 'Thursday'
	; W=5 -> DOW = 'Friday'
	; W=6 -> DOW = 'Saturday'
	; W=7 -> DOW = 'Sunday'
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_date_vwr(D) :-
        user_language(Lang),
	show_date_vwr(Lang, D).


show_date_vwr(de, ignore-_) :-
	!,
	writeln('Datum ist zu ignorieren').
show_date_vwr(_, ignore-_) :-
	!,
	writeln('Date to ignore').
show_date_vwr(de, D-R) :-
	!,
	date_normal(D, DN),
	( specifies_day(D, Y0, M0, D0) ->
	  day_of_week(Y0, M0, D0, DOW),
	  atomic_list_concat([', ',DOW], DOW1)
	; DOW1 = ''
	),
	( R = pro(_, R1),
	  memberchk(complete_parsed_date, R1) ->
	  format('~w~w~n        geparst~n', [DN, DOW1])
	; R = pro(_, R1),
	  memberchk(defaulted_parsed_date(D1, D2), R1) ->
	  date_normal(D1, D1N),
	  date_normal(D2, D2N),
	  format('~w~w~n        ~w geparst~n        ~w verknüpft~n',
		 [DN,DOW1,D1N,D2N])
	; R = pro(_, R1),
	  memberchk(explicit_date(RDP), R1) ->
	  format('~w~w~n        Explizit spezifiziert im Kontext:~n',
		 [DN,DOW1]),
	  show_context_match_rationale(8, RDP)

	; R = pro(_, R1),
	  memberchk(explicitly_defaulted_parsed_date(D1, RDP), R1) ->
	  date_normal(D1, D1N),
	  format('~w~w~n        ~w geparst~n        Explizit spezifiziert im Kontext:~n',
		 [DN,DOW1,D1N]),
	  show_context_match_rationale(8, RDP)
	; R = pro(_, R1),
	  memberchk(incomplete_parsed_date, R1) ->
	  format('~w~n         geparst, konnte nicht vervollständigt werden~n',
		 [DN])
	; write(DN-R)
	).
show_date_vwr(_, D-R) :-
	!,
	date_normal(D, DN),
	( specifies_day(D, Y0, M0, D0) ->
	  day_of_week(Y0, M0, D0, DOW),
	  atomic_list_concat([', ',DOW], DOW1)
	; DOW1 = ''
	),
	( R = pro(_, R1),
	  memberchk(complete_parsed_date, R1) ->
	  format('~w~w~n        parsed~n', [DN, DOW1])
	; R = pro(_, R1),
	  memberchk(defaulted_parsed_date(D1, D2), R1) ->
	  date_normal(D1, D1N),
	  date_normal(D2, D2N),
	  format('~w~w~n        ~w parsed~n        ~w combined~n',
		 [DN,DOW1,D1N,D2N])
	; R = pro(_, R1),
	  memberchk(explicit_date(RDP), R1) ->
	  format('~w~w~n        Explicitly specified in context:~n',
		 [DN,DOW1]),
	  show_context_match_rationale(8, RDP)

	; R = pro(_, R1),
	  memberchk(explicitly_defaulted_parsed_date(D1, RDP), R1) ->
	  date_normal(D1, D1N),
	  format('~w~w~n        ~w parsed~n        Explicitly specified in context:~n',
		 [DN,DOW1,D1N]),
	  show_context_match_rationale(8, RDP)
	; R = pro(_, R1),
	  memberchk(incomplete_parsed_date, R1) ->
	  format('~w~n         parsed, could not be completed~n',
		 [DN])
	; write(DN-R)
	).
show_date_vwr(L, D) :-
	show_date_vwr(L, D-[]).

% show_date_vwr(D) :-
% 	writeq(D),
% 	nl.
% 	date_normal(D, ND),
% 	format(atom(Info), 'Date: ~w', [ND]).


normal_to_date(A, D) :-
	atom_chars(A, A1),
	normal_to_date_1(A1, D).

normal_to_date_1([Y1,Y2,Y3,Y4], y(Y)) :-
	!,
	atomic_list_concat([Y1,Y2,Y3,Y4], YA),
	atom_number(YA, Y).
normal_to_date_1([Y1,Y2,Y3,Y4,'-',M1,M2], y(Y,M)) :-
	!,
	atomic_list_concat([Y1,Y2,Y3,Y4], YA),
	atom_number(YA, Y),
	atomic_list_concat([M1,M2], MA),
	atom_number(MA, M).
normal_to_date_1([M1,M2,'-',D1,D2], Date) :-
	!,
	normal_to_date_1(['?','?','?','?','-',M1,M2,'-',D1,D2], Date).
normal_to_date_1([Y1,Y2,Y3,Y4,'-',M1,M2,'-',D1,D2], Date) :-
	!,
	atomic_list_concat([Y1,Y2,Y3,Y4], YA),
	( YA = '????' ->
	  Date = md(M, D)
	; atom_number(YA, Y),
	  Date = ymd(Y,M,D)
	),
	atomic_list_concat([M1,M2], MA),
	atom_number(MA, M),
	atomic_list_concat([D1,D2], DA),
	atom_number(DA, D).
normal_to_date_1([Y1,Y2,Y3,Y4,'-'|A1], ya(Y,A)) :-
	!,
	atomic_list_concat([Y1,Y2,Y3,Y4], YA),
	atom_number(YA, Y),
	atomic_list_concat(A1, A).
normal_to_date_1(A, a(A1)) :-
	atomic_list_concat(A, A1).


