%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(detect, [es_set_quotes/2]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_ner(layers)).


es_set_quotes(Es, Es1) :-
	info(10, 'Processing quotation marks'),
	es_set_quotes(Es, [], Stack1, Es1),
	( Stack1 \= [] ->
	  err('Open quotes at end of document: ~q', [Stack1])
	; true
	).
	

es_set_quotes([E|Es], Stack, Stack1, [E1|Es1]) :-
	e_content(E, pun(Pun)),
	!,
	pun_stack_quotes(Pun, Stack, [E|Es], Quotes, Stack2),
	( Quotes = [] ->
	  e_del(E, quotes, E1)
	; e_set(E, quotes, Quotes, E1)
	),
	es_set_quotes(Es, Stack2, Stack1, Es1).
es_set_quotes([E,EA,EB,EC|Es], Stack, Stack1, [E1,EA,EB,EC|Es1]) :-
	e_content(E, cmd(xsicm, [t])),
	e_content(EA, beg),
	e_content(EB, pun(',,')),
	e_content(EC, end),
	!,
	pun_stack_quotes(',,', Stack, [E,EA,EB,EC|Es], Quotes, Stack2),
	( Quotes = [] ->
	  e_del(E, quotes, E1)
	; e_set(E, quotes, Quotes, E1)
	),
	es_set_quotes(Es, Stack2, Stack1, Es1).
es_set_quotes([E,EA,EB,EC|Es], Stack, Stack1, [E,EA,EB,EC1|Es1]) :-
	e_content(E, cmd(xsicm, [t])),
	e_content(EA, beg),
	e_content(EB, pun(PUN)),
	once((PUN='``' ; PUN='\'\'')),
	e_content(EC, end),
	!,
	pun_stack_quotes(PUN, Stack, [E,EA,EB,EC|Es], Quotes, Stack2),
	( Quotes = [] ->
	  e_del(EC, quotes, EC1)
	; e_set(EC, quotes, Quotes, EC1)
	),
	es_set_quotes(Es, Stack2, Stack1, Es1).
es_set_quotes([E|Es], Stack, Stack1, [E|Es1]) :-
	es_set_quotes(Es, Stack, Stack1, Es1).
es_set_quotes([], Stack, Stack, []).

pun_stack_quotes(W, Stack, RestEs, Quotes, Stack1) :-
 	atom_chars(W, Cs),
 	pun_stack_quotes_1(Cs, Stack, RestEs, Quotes, Stack1).

pun_stack_quotes_1([',',','|Cs], S, RestEs, [begin(S2)|Qs], S1) :-
	!,
	( memberchk(',,', S) ->
	  es_error('Nested use of quote: ~w', [RestEs])
	; S2 = [',,'|S]
	),
	pun_stack_quotes_1(Cs, S2, RestEs, Qs, S1).
pun_stack_quotes_1(['\'','\''|Cs], S, RestEs, [end(S)|Qs], S1) :-
	!,
	( S = [',,'|S2] ->
	  true
	; S2 = S,
	  es_error('Closing quote without opening match: ~w', [RestEs])
	),
	pun_stack_quotes_1(Cs, S2, RestEs, Qs, S1).
pun_stack_quotes_1(['`','`'|Cs], S, RestEs, [end(S)|Qs], S1) :-
	!,
	( S = [',,'|S2] ->
	  true
	; S2 = S,
	  es_error('Closing quote without opening match: ~w', [RestEs])
	),
	pun_stack_quotes_1(Cs, S2, RestEs, Qs, S1).
pun_stack_quotes_1([_|Cs], S, RestEs, Qs, S1) :-
	pun_stack_quotes_1(Cs, S, RestEs, Qs, S1).
pun_stack_quotes_1([], S, _, [], S).

