%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(esops, [es_words_by_frequency/3,
		  es_key_ngrams/3,
		  es_partition_into_sentences/3,
		  es_split_at_srcpos/4,
		  es_structure_context_at_end/2,
		  show_structure_context/1,
		  es_error/2,
		  con_es_after_env_seq/4,
		  env_open_command/2,
		  es_parse_cmd_args/3,
		  es_word_prefix/2,
		  es_word_prefix/3,
		  es_first_word/2,
		  es_first_punct_before_word/2,
		  es_enum_words_upto/3,
		  es_split_at_commands/3]).

:- use_module(swilib(onto_atom)).
:- use_module(swilib(err)).
:- use_module(kbset_data(stoplists)).
:- use_module(kbset_base(textutils)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_base(utils)).
:- use_module(swilib(err)).
:- use_module(kbset_ner(fmtutils)).
:- use_module(kbset_base(hash_kbset)).
:- use_module(kbset_ner(stemming)).
:- use_module(kbset_ner(libstuff)).
:- use_module(kbset_ner(keywords)).

:- module_transparent con_es_after_env_seq/4.
:- module_transparent con_es_after_env_seq_1/5.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Miscellaneous
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% 
%%%% 
es_words_by_frequency(Es, Options, Ws) :-
	( memberchk(nostop=true, Options),
	  memberchk(text_language=Lang, Options) ->
	  aux_map_es_word_nostop(Lang, Es, Ws0)
	; aux_map_es_word(Es, Ws0)
	),
	( memberchk(stems=true, Options),
	  memberchk(stemming_language=SLang, Options) ->
	  length(Ws0, Len),
	  HTSize is round(Len/8),
	  mk_htm(HTSize, Unstem),
	  map_make_stem(Ws0, SLang, Unstem, Ws00)
	; Ws00 = Ws0,
	  Unstem = nil
	),
	msort(Ws00, Ws1),
	reverse(Ws1, Ws2),
	acc_freq(Ws2, FWs3),
	( Ws2 = [] ->
	  FWs6 = []
	; memberchk(relate_to_global_frequencies=GFs, Options),
	  length(Ws2, Total),
	  map_div_freq_related_to_global(FWs3, Total, GFs, FWs4),
	  keysort(FWs4, FWs5),
	  reverse(FWs5, FWs6)
	; length(Ws2, Total),
	  map_div_freq(FWs3, Total, FWs4),
	  keysort(FWs4, FWs5),
	  reverse(FWs5, FWs6)
	),
	( memberchk(minfreq=Freq, Options) ->
	  rkv_list_prefix_greaterq(FWs6, Freq, FWs7)
	; FWs7 = FWs6
	),
	( memberchk(output_frequency=true, Options) ->
	  Ws8 = FWs7
	; map_val(FWs7, Ws8)
	),
	( Unstem \= nil ->
	  map_add_unstems(Ws8, Unstem, Ws)
	; Ws = Ws8
	).

map_add_unstems([X|Xs], Y1, [X1|Xs1]) :-
	add_unstems(X, Y1, X1),
	map_add_unstems(Xs, Y1, Xs1).
map_add_unstems([], _, []).

add_unstems(Freq-Y, U, Freq-Z) :-
	!,
	add_unstems(Y, U, Z).
add_unstems(X, U, X-Y) :-
	findall(Z, htm_get(U, X, Z), Z1),
	sort(Z1, Y).

map_make_stem([X|Xs], Y1, Y2, [X1|Xs1]) :-
	make_stem(X, Y1, Y2, X1),
	map_make_stem(Xs, Y1, Y2, Xs1).
map_make_stem([], _, _, []).

make_stem(W, VLang, UnstemTable, W1) :-
	downcase(W, W0),
	( VLang = en(v1) ->
	  stem_english_20(W0, W1)
	; VLang = de(20) ->
	  stem_de_20(W0, W1)
	; VLang = de(191) ->
	  stem_de_191(W0, W1)
	; VLang = de(19) ->
	  stem_de_19(W0, W1)
	; err('Unknown lang for stemming: ~q', [VLang])
	),
	( htm_get(UnstemTable, W1, W) -> true
	; htm_add(UnstemTable, W1, W)
	).

acc_freq([X|Xs], Ys) :-
	acc_freq(Xs, X, 1, Ys).

acc_freq([X|Xs], Y, N, Ys) :-
	X == Y,
	!,
	N1 is N+1,
	acc_freq(Xs, Y, N1, Ys).
acc_freq([X|Xs], Y, N, [N-Y|Ys]) :-
	X \== Y,
	acc_freq(Xs, X, 1, Ys).
acc_freq([], Y, N, [N-Y]).

map_div_freq_related_to_global([X|Xs], Y1, Z1, [X1|Xs1]) :-
	div_freq_related_to_global(X, Y1, Z1, X1),
	map_div_freq_related_to_global(Xs, Y1, Z1, Xs1).
map_div_freq_related_to_global([], _, _, []).

div_freq_related_to_global(N-X, Total, GlobalFrequencies, F-X) :-
	( memberchk(stem_frequency(X, GF, _), GlobalFrequencies) ->
	  true
	; length(GlobalFrequencies, L),
	  GF is 0.5 / L
	),
	_Threshold = 0.01,
	( N =< 2 ->
	  F = 0
	; % F is (N/Total)^2 / GF
	  F is N/Total / GF  
	).

map_div_freq([X|Xs], Y1, [X1|Xs1]) :-
	div_freq(X, Y1, X1),
	map_div_freq(Xs, Y1, Xs1).
map_div_freq([], _, []).

div_freq(N-X, Total, F-X) :-
	F is N / Total.

aux_map_es_word([E|Es], [W|Ws]) :-
	e_content(E, wrd(W)),
	!,
	aux_map_es_word(Es, Ws).
aux_map_es_word([_|Es], Ws) :-
	aux_map_es_word(Es, Ws).
aux_map_es_word([], []).

aux_map_es_word_nostop(Lang, Es, Ws) :-
	aux_map_es_word_nostop_1(Es, Lang, Ws).
	
aux_map_es_word_nostop_1([E|Es], L, [W|Ws]) :-
	e_content(E, wrd(W)),
	\+ is_stopword(L, W),
	!,
	aux_map_es_word_nostop_1(Es, L, Ws).
aux_map_es_word_nostop_1([_|Es], L, Ws) :-
	aux_map_es_word_nostop_1(Es, L, Ws).
aux_map_es_word_nostop_1([], _, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%%  Partition input into a list of list of Es, where each sublist
%%%%  corresponds to a sentence.
%%%% 
es_partition_into_sentences(Es, Options, Ess) :-
	( memberchk(text_language=Lang, Options) ->
	  true
	; Lang = de
	),
	es_sent_2_skipvoid(Es, Lang, S, S, Ess).

es_sent_2([E,EP|Es], Lang, S, D, Ss) :-
	e_content(E, wrd(_)),
	e_content(EP, pun(P)),
	is_sentence_delimiter(P),
	next_word_is_lowercase(Es),
	!,
	D = [E,EP|D1],
	es_sent_2(Es, Lang, S, D1, Ss).
es_sent_2([E,EP|Es], Lang, S, D, Ss) :-
	e_content(E, wrd(W)),
	e_content(EP, pun('.')),
	abbreviation_word(Lang, W),
	!,
	D = [E,EP|D1],
	es_sent_2(Es, Lang, S, D1, Ss).
es_sent_2([E,E1|Es], Lang, S, D, [S|Ss1]) :-
	e_content(E, pun(P)),
	is_sentence_delimiter(P),
	e_content(E1, spc(_)),
	!,
	D = [E],
	es_sent_2_skipvoid([E1|Es], Lang, S1, S1, Ss1).
es_sent_2([E|Es], Lang, S, D, Ss) :-
	D = [E|D1],
	es_sent_2(Es, Lang, S, D1, Ss).
es_sent_2([], _, S, D, [S]) :-
	D = [].

es_sent_2_skipvoid([E|Es], Lang, S, D, Ss) :-
	e_content(E, wrd(_)),
	!,
	( S == D ->
	  es_sent_2([E|Es], Lang, S, D, Ss)
	; D = [],
	  Ss = [S|Ss1],
	  es_sent_2([E|Es], Lang, S1, S1, Ss1)
	).
es_sent_2_skipvoid([E|Es], Lang, S, D, Ss) :-
	D = [E|D1],
	es_sent_2_skipvoid(Es, Lang, S, D1, Ss).
es_sent_2_skipvoid([], _, S, D, Ss) :-
	( S == D ->
	  Ss = []
	; D = [],
	  Ss = [S]
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

abbreviation_word(_, W) :-
	atom_length(W, L),
	L = 1,
	!.
abbreviation_word(_, W) :-
	atom_number(W, N),
	N > 0,
	N =< 31,
	!.
abbreviation_word(_, W) :-
	is_roman_number(W),
	!.
abbreviation_word(Lang, W) :-
	!,
	is_period_delimited_abbrv(Lang, W).

next_word_is_lowercase(Es) :-
	member(E, Es),
	e_content(E, wrd(W)),
	!,
	sub_atom(W, 0, 1, _, C),
	\+ is_upper_char(C).

is_sentence_delimiter(X) :-
	member(C, ['.', '!', '?', ':']),
	sub_atom(X, _, _, _, C),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_key_ngrams(Es, Options, NGrams) :-
	es_partition_into_sentences(Es, Options, Ess),
	map_es_wordlist_noempty(Ess, Ss),
	key_ngrams(Ss, Options, NGrams).

map_es_wordlist_noempty([X|Xs], Xs1) :-
	es_wordlist(X, X1),
	( X1 = [] ->
	  Xs1 = Xs2
	; Xs1 = [X1|Xs2]
	),
	map_es_wordlist_noempty(Xs, Xs2).
map_es_wordlist_noempty([], []).

es_wordlist([E|Es], [W|Ws]) :-
	e_content(E, wrd(W)),
	!,
	es_wordlist(Es, Ws).
es_wordlist([_|Es], Ws) :-
	es_wordlist(Es, Ws).
es_wordlist([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_split_at_srcpos(Es, Pos, RevBefore, After) :-
	es_split_at_srcpos(Es, Pos, 0, [], RevBefore, After).
es_split_at_srcpos([E|Es], Pos, CurrPos, RB, RB1, After) :-
	CurrPos < Pos,
	!,
	e_srctxt_size(E, Size),
	CurrPos1 is CurrPos + Size,
	es_split_at_srcpos(Es, Pos, CurrPos1, [E|RB], RB1, After).
es_split_at_srcpos(Es, _, _, RB, RB, Es).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
es_structure_context_at_end(Es, S) :-
	es_sae(Es, [], [], [], [], SC1),
	mk_structure_context(SC1, Es, S).

mk_structure_context(struct_ctx_1(DomDown, DomEnv, DomWrap, NUnProcessed), Es,
		      SC) :-
	( NUnProcessed > 0 ->
	  reverse(Es, REs),
	  list_cdr_upto(NUnProcessed, REs, REs1),
	  reverse(REs1, Es1),
	  es_srctxt_size(Es1, Position),
	  SC = [unprocessed_from_position=Position|SC1]
	; SC = SC1
	),
	reverse(DomEnv, Envs),
	SC1 = [environments=Envs|SC2],
	reverse(DomWrap, Wrap),
	SC2 = [commands=Wrap|SC3],
	cleanup_down_cmds(DomDown, Levelled, Unlevelled),
	SC3 = [levelled_structure=Levelled,
	       unlevelled_structure=Unlevelled].

show_structure_context(SCTX) :-
	format_header_1([en-'Structure Context',
			 de-'Struktureller Kontext']),
	( memberchk(unprocessed_from_position=Pos, SCTX), Pos > 0 ->
	  format('Unprocessed from position: ~w~n', [Pos])
	; true
	),
	( memberchk(levelled_structure=Structs, SCTX),
	  Structs \= [] ->
	  format_header_2([en-'Structure',
			   de-'Struktur']),
	  ( member(I-S, Structs),
	    sc_level(I, Level),
	    Indent is Level+1,
	    indent(Indent, '*'),
	    write(' '),
	    write_arginfo(S),
	    nl,
	    fail
	  ; true
	  )
	; true
	),
	( memberchk(unlevelled_structure=UStructs, SCTX),
	  UStructs \= [] ->
	  format_header_2([en-'Partitions',
			   de-'Partitionen']),
	  ( member(PType-PArgs, UStructs),
	    write(PType), write(' '), write_arginfo(PArgs), nl,
	    fail
	  ; true
	  )
	; true
	),
	( memberchk(environments=Envs, SCTX),
	  Envs \= [] ->
	  format_header_2([en-'Environments',
			   de-'Umgebungen']),
	  ( member(begin(E)-EArgs, Envs),
	    writeq(E), write(' '), write_arginfo(EArgs), nl,
	    fail
	  ; true
	  )
	; true
	),
	( memberchk(commands=Commands, SCTX),
	  Commands \= [] ->
	  format_header_2([en-'Commands',
			   de-'Kommandos']),
	  ( member(wrap(_N,Cmd,CArgs), Commands),
	    write(Cmd), % write(' '), write(N), write(' '),
	    write(' '),
	    write_arginfo(CArgs), nl,
	    fail
	  ; true
	  )
	; true
	).

write_arginfo(Xs) :-
	( member(A, Xs),
	  ( A = arg(X) ->
	    ( member(E, X),
	      e_enum_plaintxt(E, T),
	      write(T),
	      fail
	    ; true
	    )
	  ; A = ctr(X) ->
	    format('{~w}', [X])
	  ; A = opt(X) ->
	    format('[~w]', [X])
	  ),
	  fail
	; true
	).
	

cleanup_down_cmds(CXs, Levelled, Unlevelled) :-
	csc_l(CXs, 1000, Levelled1),
	reverse(Levelled1, Levelled),
	csc_ul(CXs, Unlevelled1),
	reverse(Unlevelled1, Unlevelled).

csc_ul([C-_|CXs], CXs1) :-
	sc_level(C, _),
	!,
	csc_ul(CXs, CXs1).
csc_ul([C-X|CXs], [C-X|CXs1]) :-
	csc_ul(CXs, CXs1).
csc_ul([], []).

csc_l([C-X|CXs], L, [C-X|CXs1]) :-
	sc_level(C, L1),
	L1 < L,
	!,
	csc_l(CXs, L1, CXs1).
csc_l([_|CXs], L, CXs1) :-
	csc_l(CXs, L, CXs1).
csc_l([], _, []).

indent(N, C) :-
	N > 0,
	!,
	write(C),
	N1 is N - 1,
	indent(N1, C).
indent(0, _).

sc_level(C, Level) :-
	structuring_commands(SCs),
	sc_level(SCs, C, 0, Level).

sc_level([Cs|_], C, N, N) :-
	memberchk(C-_, Cs),
	!.
sc_level([_|Css], C, N, N1) :-
	N2 is N+1,
	sc_level(Css, C, N2, N1).

structuring_commands([[xtitle-[t]],
		      [xchap-[t],xchapnc-[t],xchaph-[t,c]],
		      [xsec-[t],xsecn-[t],xsecnc-[t],xseca-[t,c]],
		      [xsubsec-[t],xsubseca-[t],xsubsecn-[t]],
		      [xpar-[t]]]).

is_downward_command(Cmd, Pattern) :-
	structuring_commands(SCs),
	member(SCClass, SCs),
	member(Cmd-Pattern, SCClass),
	!.
is_downward_command('xpage', [t]).
is_downward_command('xpagep', [t]).

%% *** xpagem xpagep


add_downward_command(Cmd, Args, Dom, Dom1) :-
	( select(Cmd-_, Dom, Dom2) ->
	  true
	; Dom2 = Dom
	),
	Dom1 = [Cmd-Args|Dom2].
	
add_env_command(begin, [ctr(Env)|Args], Dom, [begin(Env)-Args|Dom]).

rm_env_command(end, [ctr(Env)|_], [begin(Env)-_| Dom], Dom).

wrap_push(WrapSpec, Wrap, [WrapSpec|Wrap]).
wrap_pop([_|Wrap], Wrap).

toprocess_push(Cmd, Patt, Args, ToProcess, ToProcess1) :-
	args_cos(Patt, Args, Args1),
	map_at(Patt, 1, Cmd, Args1, ToProcess1, ToProcess).
toprocess_pop([X|Xs], X, Xs).

map_at([t|Patt], N, Cmd, Args, [wrap(N,Cmd,Args)|ToProcess1], ToProcess2) :-
	!,
	N1 is N+1,
	map_at(Patt, N1, Cmd, Args, ToProcess1, ToProcess2).
map_at([c|Patt], N, Cmd, Args, ToProcess, ToProcess1) :-
	!,
	N1 is N+1,
	map_at(Patt, N1, Cmd, Args, ToProcess, ToProcess1).
map_at([_|Patt], N, Cmd, Args, ToProcess, ToProcess1) :-
	map_at(Patt, N, Cmd, Args, ToProcess, ToProcess1).
map_at([], _, _, _, ToProcess, ToProcess).

args_cos([c|Patt], [A|Args], [A|Args1]) :-
	!,
	args_cos(Patt, Args, Args1).
args_cos([o|Patt], [A|Args], [A|Args1]) :-
	!,
	args_cos(Patt, Args, Args1).
args_cos([_|Patt], [_|Args], Args1) :-
	args_cos(Patt, Args, Args1).
args_cos([], [], []).	      

es_sae([E|Es], ToProcess, DomDown, DomEnv, DomWrap, Result) :-
	e_content(E, beg),
	!,
	toprocess_pop(ToProcess, WrapSpec, ToProcess1),
	wrap_push(WrapSpec, DomWrap, DomWrap1),
	es_sae(Es, ToProcess1, DomDown, DomEnv, DomWrap1, Result).
es_sae([E|Es], ToProcess, DomDown, DomEnv, DomWrap, Result) :-
	e_content(E, end),
	!,
	wrap_pop(DomWrap, DomWrap1),
	es_sae(Es, ToProcess, DomDown, DomEnv, DomWrap1, Result).
es_sae([E|Es], ToProcess, DomDown, DomEnv, DomWrap, Result) :-
	e_content(E, cmd(Cmd, Patt)),
	!,
	( es_parse_cmd_args(Es, Patt, Args) ->
	  ( is_downward_command(Cmd, Patt) ->
	    add_downward_command(Cmd, Args, DomDown, DomDown1),
	    DomEnv1 = DomEnv,
	    DomWrap1 = DomWrap
	  ; Cmd = begin ->
	    add_env_command(Cmd, Args, DomEnv, DomEnv1),
	    DomDown1 = DomDown,
	    DomWrap1 = DomWrap
	  ; Cmd = end ->
	    rm_env_command(Cmd, Args, DomEnv, DomEnv1),
	    DomDown1 = DomDown,
	    DomWrap1 = DomWrap	    
	  ; DomDown1 = DomDown,
	    DomEnv1 = DomEnv,
	    DomWrap1 = DomWrap
	  ),
	  toprocess_push(Cmd, Patt, Args, ToProcess, ToProcess1),
	  es_sae(Es, ToProcess1, DomDown1, DomEnv1, DomWrap1, Result)
	; length([E|Es], Len),
	  Result = struct_ctx_1(DomDown, DomEnv, DomWrap, Len)
	).
es_sae([_|Es], ToProcess, DomDown, DomEnv, DomWrap, Result) :-
	!,
	es_sae(Es, ToProcess, DomDown, DomEnv, DomWrap, Result).
es_sae([], _, DomDown, DomEnv, DomWrap, Result) :-
	Result = struct_ctx_1(DomDown, DomEnv, DomWrap, 0).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_parse_cmd_args(Es, Specs, Args) :-
	es_parse_cmd_args(Es, Specs, Args, _).
es_parse_cmd_args(Es, Specs, Args, Es1) :-
	parse_args_1(Specs, Es, Args, [], Es1).

parse_args_1([S|Ss], Es, Args, Args1, Es1) :-
	parse_arg(S, Es, Args, Args2, Es2),
	parse_args_1(Ss, Es2, Args2, Args1, Es1).
parse_args_1([], Es, Args, Args, Es).

parse_arg(t, [E|Es], [arg(Arg)|Args], Args, Es1) :-
	e_content(E, beg),
	!,
	( parse_be(Es, [beg], Arg, Es1) -> true
	; Arg=[]
	  % list_prefix_upto(20, Es, Es1),
 	  % err('Failed to parse LaTex beg command argument at: ~q', [[E|Es1]]),
	  % fail
	).
parse_arg(t, [E|Es], Args, Args1, Es1) :-
	e_is_blank(E),
	!,
	parse_arg(t, Es, Args, Args1, Es1).
parse_arg(c, [E|Es], [ctr(Arg)|Args], Args, Es) :-
	e_content(E, ctr(Arg)),
	!.
parse_arg(c, [E|Es], Args, Args1, Es1) :-
	e_is_blank(E),
	!,
	parse_arg(c, Es, Args, Args1, Es1).
parse_arg(c, _Es, _, _, _) :-
	!,
%	list_prefix_upto(20, Es, Es1),
%	err('Failed to parse LaTeX control argument at: ~q', [Es1]),
	fail.

parse_arg(o, [E|Es], [opt(Arg)|Args], Args, Es) :-
	e_content(E, opt(Arg)),
	!.
parse_arg(o, [E|Es], Args, Args1, Es1) :-
	e_is_blank(E),
	!,
	parse_arg(o, Es, Args, Args1, Es1).
parse_arg(o, Es, Args, Args, Es).

parse_be([E|Es], I, [E|Es1], Es2) :-
	e_content(E, beg),
	!,
	parse_be(Es, [beg|I], Es1, Es2).
parse_be([E|Es], [beg], [], Es) :-
	e_content(E, end),
	!.
parse_be([E|Es], [beg|I], [E|Es1], Es2) :-
	e_content(E, end),
	!,
	parse_be(Es, I, Es1, Es2).
parse_be([E|Es], I, [E|Es1], Es2) :-
	!,
	parse_be(Es, I, Es1, Es2).	
% parse_be(Es, I, U, V) :-
% 	list_prefix_upto(20, Es, Es1),
% 	err('Parse be failed: ~q ~q', [Es1, I]).


es_error(F, Arglist) :-
	append(FmtArgs, [InfoEs], Arglist),
	list_prefix_upto(40, InfoEs, Es1),
	onto_atom((member(E, Es1),
		   e_enum_srctxt(E, Txt),
		   write(Txt),
		   fail
		  ;true),
		  Info),
	append(FmtArgs, [Info], FmtArgs1),
	err(F, FmtArgs1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

con_es_after_env_seq(Es, Pred, Arg1, Arg2) :-
	mk_env(Env),
	con_es_after_env_seq_1(Es, Pred, Arg1, Arg2, Env).
	
con_es_after_env_seq_1([E|Es], P, X, Y, Env) :-
	Call =.. [P,E,X,Z,Es,Env],
	Call,
	update_env(E, Env, Es, Env1),
	con_es_after_env_seq_1(Es, P, Z, Y, Env1).
con_es_after_env_seq_1([], _, X, X, _).


map_es_after_env_seq(Es, Pred, FixedArg, Es1) :-
	mk_env(Env),
	map_es_after_env_seq_1(Es, Pred, FixedArg, Env, Es1, []).
	
map_es_after_env_seq_1([E|Es], P, X, Env, Es1, Es2) :-
	Call =.. [P,E,X,Es,Env,Es1,Es3],
	Call,
	update_env(E, Env, Es, Env1),
	map_es_after_env_seq_1(Es, P, X, Env1, Es3, Es2).
map_es_after_env_seq_1([], _, _, _, Es, Es).

env_open_command(Env, Cmd) :-
 	member(X, Env),
 	( X = arg(Cmd)
	; X = env(Cmd)
	; X = ctr(Cmd)
	).

mk_env([]).			      

update_env(E, Env, Es, Env1) :-
	( e_content(E, cmd(begin, [c|Patt])) ->
	  ( es_parse_cmd_args(Es, [c], [ctr(Cmd)]) ->
	    env_add_patt_args([c|Patt], Cmd, [env(Cmd)|Env], Env1)
	  ; es_error('Bad structure (failed to parse begin arg) at: ~w', [[E|Es]])
	  )
	; e_content(E, cmd(end, [c|Patt])) ->
	  ( es_parse_cmd_args(Es, [c], [ctr(Cmd)]),
	    Env=[env(Cmd)|Env2] ->
	    env_add_patt_args([c|Patt], Cmd, Env2, Env1)
	  ; es_error('Bad structure (failed to parse end arg) at: ~w', [[E|Es]])
	  )
	; e_content(E, end) ->
	  ( Env=[arg(_)|Env1] ->
	    true
	  ; es_error('Bad structure (unexpected text argument end) at: ~w', [[E|Es]])
	  )
	; e_content(E, ctr(_)) ->
	  ( Env=[ctr(_)|Env1] ->
	    true
	  ; es_error('Bad structure (unexpected control argument) at: ~w', [[E|Es]])
	  )
	; e_content(E, cmd(Cmd, Patt)) ->
	  env_add_patt_args(Patt, Cmd, Env, Env1)
	; Env1 = Env
	).

env_add_patt_args([c|Patt], Cmd, Env, [ctr(Cmd)|Env1]) :-
	!,
	env_add_patt_args(Patt, Cmd, Env, Env1).
env_add_patt_args([t|Patt], Cmd, Env, [arg(Cmd)|Env1]) :-
	!,
	env_add_patt_args(Patt, Cmd, Env, Env1).
env_add_patt_args([_|Patt], Cmd, Env, Env1) :-
	env_add_patt_args(Patt, Cmd, Env, Env1).
env_add_patt_args([], _, Env, Env).
	

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_word_prefix([], _) :-
	!.
es_word_prefix([W|Ws], [E|Es]) :-
	e_content(E, wrd(W1)),
	!,
	W1 = W,
	es_word_prefix(Ws, Es).
es_word_prefix(Ws, [_|Es]) :-
	es_word_prefix(Ws, Es).

es_word_prefix([], Rest, Rest) :-
	!.
es_word_prefix([W|Ws], [E|Es], Rest) :-
	e_content(E, wrd(W1)),
	!,
	W1 = W,
	es_word_prefix(Ws, Es, Rest).
es_word_prefix(Ws, [_|Es], Rest) :-
	es_word_prefix(Ws, Es, Rest).

es_first_word([E|_], W) :-
	e_content(E, wrd(W1)),
	!,
	W = W1.
es_first_word([_|Es], W) :-
	es_first_word(Es, W).

es_enum_words_upto([E|Es], N, W) :-
	N > 0,
	( e_content(E, wrd(W1)) ->
	  N1 is N-1,
	  ( W = W1
	  ; es_enum_words_upto(Es, N1, W)
	  )
	; es_enum_words_upto(Es, N, W)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_split_at_commands([], _, []).
es_split_at_commands([E|Es], Cs, Ess) :-
	sacs(Es, Cs, [E|Es1], Es1, Ess).

sacs([E|Es], Cs, Es0, [], [Es0|Ess1]) :-
	e_content(E, cmd(Cmd, _)),
	memberchk(Cmd, Cs),
	!,
	sacs(Es, Cs, [E|Es1], Es1, Ess1).
sacs([E|Es], Cs, Es0, [E|Es1], Ess) :-
	sacs(Es, Cs, Es0, Es1, Ess).
sacs([], _, Es, [], [Es]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

es_first_punct_before_word([E|_], W) :-
	e_content(E, pun(W)),
	!.
es_first_punct_before_word([E|_], _) :-
	e_content(E, wrd(_)),
	!,
	fail.
es_first_punct_before_word([_|Es], W) :-
	es_first_punct_before_word(Es, W).


