%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(keywords, [key_ngrams/3,
		     kwdata_unstem/2,
		     kwdata_stem_frequency/2,
		     kwdata_word_frequency/2,
		     map_kwdata_unstem/2,
		     map_kwdata_stem_frequency/2,
		     map_kwdata_word_frequency/2,
		     last_key_ngrams_result/1,
		     kwdata_display_result/2,
		     kwdata_display_result/1,
		     unstem_is_preferred/3,
		     word_to_stem/3
		     ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Keyword Extraction
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% - After Matsuo and Ishizuka, IJAIT 2004:
%%%%   Keyword Extraction from a Single Document Using Word Co-Occurrence
%%%%   Statistical Information
%%%% 
%%%%   Does not perform clustering
%%%%
%%%% - Not really tested for quality of results
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(swilib(err)).
:- use_module(swilib(fromonto)).
:- use_module(swilib(info)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_data(stoplists)).
:- use_module(kbset_ner(stemming)).
:- use_module(kbset_ner(libstuff)).
:- use_module(kbset_ner(ngrams)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_data(wordlists_de)).
:- use_module(kbset_base(textutils)).

:- dynamic feature_sentences/2.
:- dynamic feature_sentences_size_sum/2.
:- dynamic feature_sentence/2.
:- dynamic feature/1.
:- dynamic splitted_feature/2.
:- dynamic document_size/1.
:- dynamic sentence_size/2.
:- dynamic sentence/3.
:- dynamic language/1.
:- dynamic unstem/2.
:- dynamic last_key_ngrams_result/1.

:- flag(id, _, 1).

gen_id(N) :- flag(id, N, N+1).
	

reset :-
	info(10, 'Resetting data structures'),
	retractall(feature_sentences(_, _)),
	retractall(feature_sentences_size_sum(_, _)),
	retractall(feature_sentence(_, _)),
	retractall(feature(_)),
	retractall(splitted_feature(_, _)),
	retractall(document_size(_)),
	retractall(sentence_size(_, _)),
	retractall(sentence_size(_, _, _)),
	retractall(language(_)),
	retractall(unstem(_, _)).

get_option(Key, Options, Value) :-
	memberchk(Key=Value1, Options),	!, Value = Value1.
get_option(Key, _, Value) :-
	option_default(Key, Value1), !,	Value = Value1.
get_option(Key, _, _) :-
	err('Option value unspecified: ~k', [Key]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% key_ngrams(+Wordlists, +Options, -NGrams) :-
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Wordlists: List of list of words, each list of words representing a
%%%%            sentence
%%%%    NGrams: List of stemmed ngrams (each as an atom), ordered by
%%%%            relevance
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Options
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% max length of ngrams to construct
%%%%
option_default(ngram_max_n, 3).
%%%%
%%%% min frequency for ngrams to be considered
%%%%
option_default(ngram_min_f, 4).
%%%%
%%%% language, effects stemming and stopwords
%%%%
option_default(lang, de(19)).
%%%%
%%%% if true: create asserted datastructures anew
%%%% if false: work on previously asserted datastructures
%%%%
option_default(reset_before, true).
%%%%
%%%% if true: clean up asserted datastructures afterward
%%%%
option_default(reset_after, false).
%%%%
%%%% percentage of stems that are considered as "frequent words G"
%%%% (the algorithm handles these words differently from other words)
%%%%
option_default(freq_percent, 30).
%%%%
%%%% if true, output a list of key-value pairs, where the key inidcates
%%%% the relevance rating
%%%%
option_default(output_rating, false).
%%%%
%%%%
%%%%
%%%%
key_ngrams(Wordlists, Options, NGrams) :-
	( get_option(reset_before, Options, true) ->
	  key_ngrams_install_datastructures(Wordlists, Options)
	; info(10, 'Reusing existing datastructures')
	),
	get_option(freq_percent, Options, Percentage),
	get_option(output_rating, Options, OutputRating),
	info(10, 'Keyword extraction'),
	ordered_features(Percentage, OutputRating, NGrams),
	( get_option(reset_after, Options, true) ->
	  reset
	; info(10, 'Keeping datastructures')
	),
	retractall( last_key_ngrams_result(_) ),
	assert( last_key_ngrams_result(NGrams) ).
	
key_ngrams_install_datastructures(Wordlists, Options) :-
	reset,
	get_option(lang, Options, Lang),
	assert(language(Lang)),
	apply_append(Wordlists, Wordseq),
	info(10, 'Stemming'),
	words_to_stems(Wordseq, Lang, Stems),
	info(10, 'N-Grams'),
	get_option(ngram_max_n, Options, MaxN),
	get_option(ngram_min_f, Options, MaxF),
	generate_ngrams(Stems, MaxN, MaxF, NGrams),
	info(10, 'Creating data structures'),
	length(Stems, DocSize),
	assert(document_size(DocSize)),
	assert_features(NGrams),
	assert_sentences(Wordlists, Lang),
	assert_feature_sentences.

assert_sentences(Wordlists, Lang) :-
	( member(Words, Wordlists),
	  assert_sentence(Words, Lang),
	  fail
	; true
	).

assert_sentence(Words, Lang) :-
	gen_id(Id),
	words_to_stems(Words, Lang, Stems),
	length(Stems, Size),
	assert(sentence_size(Id, Size)),
	findall(F, stems_feature(Stems, F), Fs),
	sort(Fs, Fs1),
	( member(F1, Fs1),
	  assert(feature_sentence(F1, Id)),
	  fail
	; true
	),
	assert(sentence(Id, Words, Stems)).

assert_feature_sentences :-
	( feature(F),
	  findall(S, feature_sentence(F, S), Ss),
	  sort(Ss, Ss1),
	  assert(feature_sentences(F, Ss1)),
	  sum_sentence_size(Ss1, N),
	  N1 is N + 1, %% to prevent zero here
	  assert(feature_sentences_size_sum(F, N1)),
	  fail
	; true
	).

sum_sentence_size([], 0).
sum_sentence_size([A|As], N1) :-
	sum_sentence_size(As, N2),
	sentence_size(A, N),
	N1 is N + N2.

stems_feature([W|Ws], F) :-
	splitted_feature(W, B),
	headmatch(B, Ws),
	feature_id([W|B], F).
stems_feature([_|Ws], F) :-
	stems_feature(Ws, F).

headmatch([], _).
headmatch([W|Ws], [W|Ws1]) :-
	headmatch(Ws, Ws1).
	
feature_id(NGram, Atom) :-
	atomic_list_concat(NGram, ' ', Atom).
	
assert_features(NGrams) :-
	( member([A|B], NGrams),
	  assert(splitted_feature(A,B)),
	  feature_id([A|B], F),
	  assert(feature(F)),
	  fail
	; true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- flag(cnt, _, 0).

ordered_features(Percentage, OutputRating, Fs) :-
	frequent_features(Percentage, Gs),
	length(Gs, LG),
	findall(F, feature(F), Fs1),
	length(Fs1, LF),
	info(20, 'Number of ngrams - total: ~w frequent: ~w', [LF, LG]),
	flag(cnt, _, 0),
	map_x_prime_square(Fs1, Gs, NFs),
	info(10, 'Sorting'),
	keysort(NFs, NFs1),
	reverse(NFs1, Fs2),
	( OutputRating == true ->
	  Fs = Fs2
	; map_val(Fs2, Fs)
	).

frequent_features(Percentage, Fs) :-
	features_by_frequency(Fs1),
	list_prefix_by_percentage(Percentage, Fs1, Fs).

features_by_frequency(Fs) :-
	findall(FR-F, ( feature(F), feature_frequency(F,FR) ), FRFs),
	keysort(FRFs, FRFs1),
 	map_val(FRFs1, Fs1),
	reverse(Fs1, Fs).

feature_frequency(F, N) :-
	feature_sentences(F, Ss),
	length(Ss, N).

map_x_prime_square([X|Xs], Gs, [X1-X|Xs1]) :-
	flag(cnt, N, N+1),
	info_progress(20, 100, N),
	x_prime_square(X, Gs, X1),
	map_x_prime_square(Xs, Gs, Xs1).
map_x_prime_square([], _, []) :-
	info_progress_done.

x_prime_square(W, Gs, N) :-
	map_bias(Gs, W, Biases),
	sum(Biases, N1),
	max(Biases, N2),
	N is N1 - N2.

bias(G, W, Bias) :-
	p_g(G, PG),
	n_w(W, NW),
	freq(W, G, FREQ),
	NWPG is NW*PG,
	Bias is (FREQ - NWPG)^2 / NWPG.

map_bias([X|Xs], Y, [X1|Xs1]) :-
	bias(X, Y, X1),
	map_bias(Xs, Y, Xs1).
map_bias([], _, []).

p_g(G, N) :-
	feature_sentences_size_sum(G, GS),
	document_size(ND),
	N is GS / ND.

n_w(W, N) :-
	feature_sentences_size_sum(W, N).

freq(W, G, N) :-
	feature_sentences(W, As1),
	feature_sentences(G, As2),
	!,
	ord_intersection(As1, As2, As),
	length(As, N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% These operate on the data structures asserted by key_ngrams
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

kwdata_unstem(StemNGramAtom, WordsAtom) :-
	feature_id(Stems1, StemNGramAtom),
	map_unstem(Stems1, Words1),
	atomic_list_concat(Words1, ' ', WordsAtom).

kwdata_stem_frequency(StemNGramAtom, N) :-
	feature_frequency(StemNGramAtom, N).

kwdata_word_frequency(WordNGramAtom, N) :-
	atomic_list_concat(Words, ' ', WordNGramAtom),
	language(Lang),
	words_to_stems(Words, Lang, Stems),
	feature_id(Stems, StemNGramAtom),
	kwdata_stem_frequency(StemNGramAtom, N).	

map_kwdata_stem_frequency([X|Xs], [X1|Xs1]) :-
	kwdata_stem_frequency(X, X1),
	map_kwdata_stem_frequency(Xs, Xs1).
map_kwdata_stem_frequency([], []).

map_kwdata_word_frequency([X|Xs], [X1|Xs1]) :-
	kwdata_word_frequency(X, X1),
	map_kwdata_word_frequency(Xs, Xs1).
map_kwdata_word_frequency([], []).

map_kwdata_unstem([X|Xs], [X1|Xs1]) :-
	kwdata_unstem(X, X1),
	map_kwdata_unstem(Xs, Xs1).
map_kwdata_unstem([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

word_to_stem(W, VLang, S) :-
	downcase(W, W1),
	\+ is_stopword(VLang, W1),
	make_stem(VLang, W, S).
	
words_to_stems([W|Ws], VLang, [S|Ss]) :-
	word_to_stem(W, VLang, S),
	!,
	words_to_stems(Ws, VLang, Ss).
words_to_stems([_|Ws], VLang, Ss) :-
	words_to_stems(Ws, VLang, Ss).
words_to_stems([], _, []).

vlang_lang(de(_), de) :- !.
vlang_lang(en(_), en) :- !.
vlang_lang(X, _) :-  err('Unknown stemming language: ~q', [X]).

make_stem(VLang, W, W1) :-
	( VLang = en(20) ->
	  stem_english_20(W, W1)
	; VLang = de(20) ->
	  stem_de_20(W, W1)
	; VLang = de(191) ->
	  stem_de_191(W, W1)
	; VLang = de(19) ->
	  stem_de_19(W, W1)
	; VLang = de(18) ->
	  stem_de_18(W, W1)
	; err('Unknown lang for stemming: ~q', [VLang])
	),
	vlang_lang(VLang, Lang),
	register_unstem(Lang, W1, W).

register_unstem(Lang, X, Y) :-
	( unstem(X, Z) ->
	  ( unstem_is_preferred(Lang, Y, Z) ->
	    retract(unstem(X, Z)),
	    assert(unstem(X, Y))
	  ; true
	  )
	; assert(unstem(X,Y))
	).

unstem_is_preferred(_, W, W) :-	!, fail.
unstem_is_preferred_de(Lang, W1, W2) :-
	unstem_key(Lang, W1, K1),
	unstem_key(Lang, W2, K2),
	K1 @< K2.
	
unstem_key(de, W, k(Subst,Low,Len,W)) :-
	( substantive_de(W) -> Subst=0 ; Subst=1 ),
	( starts_with_upper_char(W) ->  Low=1 ; Low=0 ),
	atom_length(W, Len).

unstem_key(en, W, k(Low,Len,W)) :-
	( starts_with_upper_char(W) ->  Low=1 ; Low=0 ),
	atom_length(W, Len).

map_unstem([X|Xs], [X1|Xs1]) :-
	unstem(X, X1),
	map_unstem(Xs, Xs1).
map_unstem([], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

list_prefix_by_percentage(Percentage, Xs, Xs1) :-
	length(Xs, L),
	N is round((L * Percentage / 100)),
	list_prefix(N, Xs, Xs1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

test(X) :-
	N=500,
	list_cdr(N,X,X1),
	list_prefix(100,X1,X2),
	key_ngrams(X2, [lang=german(v3), freq_percent=30, output_rating=true], Y),
	kwdata_display_result(Y).

kwdata_display_result(X) :-
	kwdata_display_result(20, X).
	
kwdata_display_result(N, X) :-
	length(X, L),
	( L > N ->
	  list_prefix(N, X, Y)
	; Y = X
	),
	( member(E, Y),
	  kwdata_display_item(E),
	  fail
	; true
	).

kwdata_display_item(N-X) :-
	!,
	kwdata_stem_frequency(X, F),
	kwdata_unstem(X, Y),
	format('~w ~w ~w (~w)~n', [N, F, Y, X]).

kwdata_display_item(X) :-
	!,
	kwdata_stem_frequency(X, F),
	kwdata_unstem(X, Y),
	format('~w ~w (~w)~n', [F, Y, X]).

% test_wl(L) :-
% 	file_lines(src('tmp_test_04.txt'), L1),
% 	wss(L1, L).

wss(Xs, Xs1) :-
	wss(Xs, Z, Z, Xs1).

wss(['STOP'|Xs], Y, [], [Y|Xs1]) :-
	!,
	wss(Xs, Z, Z, Xs1).
wss([X|Xs], Y, [X|Z], Xs1) :-
	wss(Xs, Y, Z, Xs1).
wss([], _, [], []).


file_lines(FileSpec, Lines) :-
	filename(FileSpec, File),
	( set_stream(current_input, encoding(utf8)),
	  read_lines_1(Lines)
	)
	from_file File.

read_lines_1([]) :-
	at_end_of_stream,
	!.
read_lines_1([X|Xs]) :-
	read_line_to_codes(current_input, L),
	atom_codes(X, L),
	read_lines_1(Xs).
