%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(layers,
	  [file_to_es/2,
	   file_to_es/3,
	   es_to_file/2,
	   es_mergein/3,
	   mk_e/3,
	   e_content/2,
	   e_add_values_to_ordset/4,
	   e_get/3,
	   e_set/4,
	   e_del/3,
	   es_plain_codes/2,
	   e_plaintxt_size/2,
	   e_srctxt_size/2,
	   es_srctxt_size/2,
	   e_enum_plaintxt/2,
	   e_enum_srctxt/2,
	   es_get_with_srcpos/4,
	   es_get_with_srcpos_single_elem_prop/4,
	   e_is_blank/1]).
	   

:- use_module(swilib(info)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_base(readwrite)).
:- use_module(kbset_base(filename)).
:- use_module(swilib(fromonto)).

file_to_es(F, Es)  :-
	file_to_es(F, [], Es).
file_to_es(F, _Options, Es)  :-
	filename(F, F1),
	read_in_latex(F1, LEs),
	info(20, 'Converting'),
	map_mk_e(LEs, Es).

es_to_file(Es, F) :-
	info(10, 'Writing file ~w', [F]),
	onto_file(( member(E, Es),
		    e_enum_srctxt(E, T),
		    write(T),
		    fail
		  ; true
		  ),
		  F).

% es_plainpos_to_srcposs(PP, PPB, PPA) :-
% es_srcpos_to_srcposs(PP, PPB, PPA) :-

% es_insert_at_plainpos before after
% es_insert_before_plainpos before after
% es_insert_after_plainpos before after

% es_insert_at_srcpos before after


%%%% 
%%%% Xs is a list of elements of the form:
%%%%    before(Pos,E)
%%%%    behind(Pos,E)
%%%% 
%%%% The E components are merged into Es at the
%%%% plaintext position related to Pos:
%%%% - for before(Pos,E), before the first element whose
%%%%   textpos end exceeds or is equal to Pos
%%%% - for behind(Pos,E), after the first element whose
%%%%   textpos end exceeds or is equal to Pos
%%%%   
%%%% Notes: The Xs are processed from left to right, thus the specified
%%%% behavior is only guaranteed if they are ordered according to Pos.  The
%%%% plaintext position relevant for merging is always taken from the input Es
%%%% (changes due to the elements already processed are not counted).
%%%% 
es_mergein(Xs, Es, Es1) :-
	es_mergein(Xs, 0, Es, Es1).

es_mergein([before(P,E)|Xs], P0, [E1|Es1], Es2) :-
	!,
	e_plaintxt_size(E1, S),
	P1 is P0 + S,
	( P1 > P ->
	  Es2 = [E|Es3],
	  es_mergein(Xs, P0, [E1|Es1], Es3)
	; Es2 = [E1|Es3],
	  es_mergein([before(P,E)|Xs], P1, Es1, Es3)
	).
es_mergein([before(_,E)|Xs], _, [], [E|Es2]) :-
	!,
	es_mergein(Xs, _, [], Es2).
es_mergein([behind(P,E)|Xs], P0, [E1|Es1], Es2) :-
	!,
	( P0 >= P ->
	  Es2 = [E|Es3],
	  es_mergein(Xs, P0, [E1|Es1], Es3)
	; Es2 = [E1|Es3],
	  e_plaintxt_size(E1, S),
	  P1 is P0 + S,
	  es_mergein([behind(P,E)|Xs], P1, Es1, Es3)
	).
es_mergein([behind(_,E)|Xs], _, [], [E|Es2]) :-
	!,
	es_mergein(Xs, _, [], Es2).
es_mergein([], _, Es, Es).	

es_write_src(Es, F) :-
	filename(F, F1),
	onto_file((member(E, Es),
		   e_enum_srctxt(E, A),
		   write(A),
		   fail
		  ; true
		  ),
		  F1).

es_write_plain(Es, F) :-
	filename(F, F1),
	onto_file((member(E, Es),
		   e_enum_plaintxt(E, A),
		   write(A),
		   fail
		  ; true
		  ),
		  F1).

es_plain_codes(Es, Cs) :-
	findall(C, ( member(E, Es),
		     e_enum_plaintxt(E, A),
		     atom_codes(A, Cs1),
		     member(C, Cs1) ),
		Cs).

e_enum_plaintxt(e(LXE, _, _, _), T) :- lxelem_enum_plaintxt(LXE, T).
e_enum_srctxt(e(LXE, _, _, _), T) :- lxelem_enum_srctxt(LXE, T).
e_plaintxt_size(e(_, S, _, _), S).
e_srctxt_size(e(_, _, S, _), S).

es_srctxt_size(Es, S) :-
	es_srctxt_size_1(Es, 0, S).

es_srctxt_size_1([E|Es], S, S1) :-
	e_srctxt_size(E, S2),
	S3 is S + S2,
	es_srctxt_size_1(Es, S3, S1).
es_srctxt_size_1([], S, S).

es_get_with_rest(Es, K, V, Es1) :-
	mem(E, Es, Es1),
	e_get(E, K, V).

es_get_with_srcpos(Es, K, V, Pos) :-
	es_get_1(Es, K, V, 1, Pos).

es_get_with_srcpos_single_elem_prop(Es, K, V, Pos) :-
	es_get_2(Es, K, V1, 1, Pos1, E),
	( V = begin(V1),
	  Pos = Pos1
	; V = end,
	  e_srctxt_size(E, Size),
	  Pos is Pos1 + Size
	).

es_get_1([E|_], K, V, Pos, Pos1) :-
	e_get(E, K, V0),
	( V0 = begin_end(V1) ->
	  e_srctxt_size(E, S),
	  ( V = begin(V1), Pos1 = Pos
	  ; V = end, Pos1 = Pos + S
	  )
	; V = V0, Pos1 = Pos
	).
es_get_1([E|Es], K, V, Pos, Pos1) :-
	e_srctxt_size(E, S),
	Pos2 is Pos + S,
	es_get_1(Es, K, V, Pos2, Pos1).

% es_get_1([E|_], K, V, Pos, Pos) :-
% 	e_get(E, K, V).
% es_get_1([E|Es], K, V, Pos, Pos1) :-
% 	e_srctxt_size(E, S),
% 	Pos2 is Pos + S,
% 	es_get_1(Es, K, V, Pos2, Pos1).

es_get_2([E|_], K, V, Pos, Pos, E) :-
	e_get(E, K, V).
es_get_2([E|Es], K, V, Pos, Pos1, E1) :-
	e_srctxt_size(E, S),
	Pos2 is Pos + S,
	es_get_2(Es, K, V, Pos2, Pos1, E1).

e_get(e(_, _, _, KVs), K, V) :-
	memberchk(K=V1, KVs),
	V = V1.
e_set(e(A, B, C, KVs), K, V, e(A, B, C, KVs1)) :-
	( select(K=_, KVs, KVs2) ->
	  KVs1 = [K=V|KVs2]
	; KVs1 = [K=V|KVs]
	).

e_del(e(A, B, C, KVs), K, e(A, B, C, KVs1)) :-
	memberchk(K=_, KVs),
	!,
	( select(K=_, KVs, KVs2) ->
	  KVs1 = KVs2
	; KVs1 = KVs
	).
e_del(E, _, E).

e_add_values_to_ordset(e(A, B, C, KVs), K, V, e(A, B, C, KVs1)) :-
	( select(K=V1, KVs, KVs2) ->
	  ord_union(V, V1, V2),
	  KVs1 = [K=V2|KVs2]
	; KVs1 = [K=V|KVs]
	).

e_content(e(A, _, _, _), A).


mk_e(LXE, E) :-
	mk_e(LXE, [], E).

mk_e(LXE, KVs, e(LXE, SP, SS, KVs)) :-
	findall(LP, ( lxelem_enum_plaintxt(LXE, TP),
		      atom_length(TP, LP) ),
		LPs),
	sum(LPs, SP),
	findall(LS, ( lxelem_enum_srctxt(LXE, TS),
		      atom_length(TS, LS) ),
		LSs),
	sum(LSs, SS).

map_mk_e([X|Xs], [X1|Xs1]) :-
	mk_e(X, X1),
	map_mk_e(Xs, Xs1).
map_mk_e([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

e_is_blank(E) :-
	e_content(E, spc(_)),
	!.
e_is_blank(E) :-
	e_content(E, com(_)),
	!.
e_is_blank(E) :-
	e_plaintxt_size(E, 0).
