%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(location_ner,
	  [ deref_word_to_location_id/4,
	    not_blocked_word_location/2,
	    not_blocked_id_location/2,
	    pro_word_location_feature/3,
	    pro_id_location_feature/3,

	    not_exp_blocked_as_location_name_in_context/2,
	    
	    stdize_location/2,
	    show_location_vwr/1,
	    show_locations_vwr/1,
	    compile_location_assistance/0]).

:- use_module(swilib(info)).
:- use_module(kbset_ner(ctx)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_base(location)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(wordutils)).
:- use_module(kbset_base(textutils)).
:- use_module(kbset_data(stoplists)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(fmtutils)).
:- use_module(kbset_ner(germanutils)).
:- use_module(kbset_data(wordlists_de)).
:- use_module(swilib(err)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_ner(ner_utils)).
:- use_module(kbset_base(geo)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

deref_word_to_location_id(W, CTX, LID, W1) :-
	ctx_get_defaulted(CTX, text_language, Lang, de),
	ord_enum(LID-W1, ( location_possible_basename(Lang, W, W1),
			   word_to_location(W1, LID))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

not_blocked_word_location(W, CTX) :-
	word_is_capitalized(W),
	\+ ( ctx_get(CTX, element, E), e_identified_item(persons, E, _) ).

not_blocked_id_location(_,  _) :-
	true.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_word_location_feature(not_explicitly_blocked, W, CTX) :-
	not_exp_blocked_as_location_name_in_context(W, CTX).

% pro_word_location_feature(no_common_person_name, W, CTX) :-
% 	ctx_get(CTX, text_language, Lang),
% 	word_is_no_common_person_name(Lang, W).

pro_word_location_feature(no_stopword, W, CTX) :-
	pro_word_feature_no_stopword(W, CTX).

pro_word_location_feature(no_common_substantive, W, CTX) :-
	pro_word_feature_no_common_substantive(W, CTX).

pro_word_location_feature(no_common_downcase, W, CTX) :-
	pro_word_feature_no_common_downcase(W, CTX).

pro_word_location_feature(no_common_firstname, W, CTX) :-
	pro_word_feature_no_common_firstname(W, CTX).

pro_word_location_feature(no_common_lastname, W, CTX) :-
	pro_word_feature_no_common_lastname(W, CTX).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		
location_possible_basename(de, W, W1) :-
	!,
	remove_diacritics_used_in_handwriting(W, W0),
	location_possible_basename_de(W0, W2),
	enum_umlaut_variants(W2, W1).
location_possible_basename(en, W, W).
	
location_possible_basename_de(W, W).
location_possible_basename_de(N, N1) :-
	%% 'Dresdner' -> 'Dresden'
	sub_atom(N, _, L, 0, 'ner'),
	sub_atom(N, _, 1, 3, X),
	( X\=a, X\=e, X\=i, X\=o, X\=u, X\=ä, X \=ö, X\=ü),
	sub_atom(N, 0, _, L, Pre),
	atomic_list_concat([Pre, 'en'], N1).
location_possible_basename_de(W, W1) :- has_suffix(W, nsisches, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, nsischen, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, nsischer, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, nsische, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, nser, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, isches, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, ischen, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, ischer, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, ische, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, sches, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, schen, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, scher, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, sche, W1).
location_possible_basename_de(W, W1) :- has_suffix(W, er, W1).
location_possible_basename_de(W, W1) :-
	has_suffix(W, s, W1), not_has_suffix(W1, s).
		

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_id_location_feature(in_wikipedia, LID, _CTX) :-
	geo_localized_wiki_url(LID, _).

pro_id_location_feature(has_pid(LID), LID, CTX) :-
	ctx_get(CTX, location_pid, LID).

pro_id_location_feature(explicitly_specified_in_context(Arg), LID, CTX) :-
	exp_specified_location_in_context(LID, CTX, Arg).

pro_id_location_feature(explicitly_specified, LID, CTX) :-
	exp_specified_location_general(LID, CTX).

pro_id_location_feature(is_country, LID, CTX) :-
	location_feature_class(LID, Class),
	Class = 'A',
	ctx_get(CTX, word, W),
	country_name(W).
% 	location_feature_code(LID, Code),
% 	memberchk(Code, ['PCL', 'PCLF', 'PCLD',
% 			 'PCLH', 'PCLS', 'PCLI','ADM1','ADM1H']).


pro_id_location_feature(is_city_or_village, LID, _CTX) :-
	location_feature_class(LID, Class),
	Class = 'P'.

pro_id_location_feature(has_population_over_10000, LID, _CTX) :-
	location_population(LID, Population),
 	Population > 10000.

pro_id_location_feature(population(N), LID, _CTX) :-
	location_population(LID, N).

pro_id_location_feature(already_identified_in_context, LID, CTX) :-
	ID_ITEM_WINDOW = 10,
  	ctx_get(CTX, identified_items, Es),
	member_upto(LID1, Es, ID_ITEM_WINDOW),
	LID1 = LID,
	!.

pro_id_location_feature(near_to_already_identified_in_context(LID1),
			LID, CTX) :-
	ID_ITEM_WINDOW = 10,
	NEAR_THRESHOLD = 50,
  	ctx_get(CTX, identified_items, Es),
	member_upto(LID1, Es, ID_ITEM_WINDOW),
	LID1 \= LID,
	geo_distance(LID, LID1, D),
	D =< NEAR_THRESHOLD,
	!.

pro_id_location_feature(referenced_by_preferred_name, LID, CTX) :-
	ctx_get(CTX, element, E),
	e_content(E, wrd(W)),
	geo_localized_name(LID, W).

% 
% pro_id_location_feature(big_population_100000, LID, _CTX) :-
% 	location_population(LID, Population),
% 	Population > 100000.


% pro_id_location_feature(linked_to_location_in_context(LID1), LID, CTX) :-
% 	ctx_get(CTX, elems_rev_before, Es),
% 	member_upto(E, Es, 100),
% 	e_identified_location_id(E, LID1),
% 	location_linked_to_location(LID, LID1).

% pro_id_location_feature(already_identified_in_context, LID, CTX) :-
% 	ctx_get(CTX, elems_rev_before, Es),
% 	once(( member_upto(E, Es, 100),
% 	       e_identified_location_id(E, LID) )).
% 

% 
% 				%	ctx_get_multi(CTX, context_locations, LID),
% 	show_location(LID).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic exp_blocked_as_location_name/2.
:- dynamic exp_specified_location/2.

not_exp_blocked_as_location_name_in_context(W, CTX) :-
	\+ ( exp_blocked_as_location_name(W, CReq),
	     context_requirement_matches(CReq, CTX, _)
	   ).

exp_specified_location_general(Id, _CTX) :-
	exp_specified_location(Id, []),
	!.

exp_specified_location_in_context(Id, CTX, Arg) :-
	exp_specified_location(Id, CReq),
	CReq \= [],
	!,
	context_requirement_matches(CReq, CTX, Arg).

compile_location_assistance :-
	info(10, 'Compiling location assistance specification'),
	retractall( exp_blocked_as_location_name(_, _) ),
	retractall( exp_specified_location(_, _) ),
	retract_supplemented_locations,
	( assistance_entry(no_entity(location, CReq)),
	  ( select(word_in=Ws, CReq, CReq1),
	    member(W, Ws)
	  ; select(word=W, CReq, CReq1)
	  ),
	  assert( exp_blocked_as_location_name(W, CReq1) ),
	  fail
	; true
	),
	( assistance_entry(no_entity(location, CReq)),
	  ( select(prefix_in=Ws, CReq, CReq1),
	    member(W, Ws)
	  ; select(prefix=W, CReq, CReq1)
	  ),
	  assert( (exp_blocked_as_location_name(X, CReq1) :- sub_atom(X, 0, _, _, W)) ),
	  fail
	; true
	),
 	( assistance_entry(entity(location, Spec, CReq)),
 	  identify_location(Spec, Ids),
 	  ( Ids = [Id] ->
 	    assert( exp_specified_location(Id, CReq) )
	  ; Ids = [] ->
	    err('Void explicit location specification: ~q', [Spec])
	  ; err('Ambiguous explicit location specification: ~q (~q)',
		[Spec, Ids])
	  ),
	  fail
	; true
	),
 	( assistance_entry(supplement(location, Spec, Sup)),
 	  identify_location(Spec, Ids),
 	  ( Ids = [Id] ->
	    process_supplemented_location(Id, Sup)
	  ; Ids = [] ->
	    err('Void explicit location specification: ~q', [Spec])
	  ; err('Ambiguous explicit location specification: ~q (~q)',
		[Spec, Ids])
	  ),
	  fail
	; true
	).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_locations_vwr([P-pro(V0, R)|PRs]) :-
	!,
	show_locations_vwr([P-pro(V0, R)|PRs], V0).
show_locations_vwr([]) :-
	!.
show_locations_vwr(Ps) :-
	( member(P, Ps),
	  show_location_vw(P),
	  fail
	; true
	).
	
show_locations_vwr([P-pro(V1, R)|PRs], V0) :-
	( V1 @> V0 ->
	  write('..................................................'),
	  writeln('............................')
	; true
	),
	show_location_vwr(P-pro(V1, R)),
	show_locations_vwr(PRs, V1).
show_locations_vwr([], _).

show_location_vwr(P-pro(_, R)) :-
	!,
	show_location_vw(P),
	show_reasons(R).

show_location_vwr(P-_) :-
	!,
	show_location_vw(P).
show_location_vwr(P) :-
	show_location_vw(P).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stdize_location(P-_, P) :-
	!.
stdize_location(P, P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
