%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(ner_base,
	  [sort_entities/3,
	   sort_entities/4,
	   sort_features/3,
	   e_entities/4,
	   e_entities_ignore_threshold/5,
	   features_lock_until/3]).
	   

:- use_module(kbset_ner(layers)).	   
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(person_ner)).
:- use_module(kbset_ner(location_ner)).
:- use_module(kbset_ner(date_ner)).
:- use_module(kbset_ner(ctx)).
:- use_module(swilib(info)).
:- use_module(kbset_base(utils)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% DISPATCH
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_word_feature(person, F, W, CTX) :-
	!,
	pro_word_person_feature(F, W, CTX).
pro_word_feature(function_person, F, W, CTX) :-
	!,
	pro_word_function_person_feature(F, W, CTX).
pro_word_feature(location, F, W, CTX) :-
	!,
	pro_word_location_feature(F, W, CTX).

pro_id_feature(person, F, Id, CTX) :-
	!,
	pro_id_person_feature(F, Id, CTX).
pro_id_feature(function_person, F, Id, CTX) :-
	!,
	pro_id_function_person_feature(F, Id, CTX).
pro_id_feature(location, F, Id, CTX) :-
	!,
	pro_id_location_feature(F, Id, CTX).

deref_word_to_id(person, W, CTX, ID, W1) :-
	!,
	deref_word_to_person_id(W, CTX, ID, W1).
deref_word_to_id(function_person, W, CTX, ID, W1) :-
	!,
	deref_word_to_function_person_id(W, CTX, ID, W1).
deref_word_to_id(location, W, CTX, ID, W1) :-
	!,
	deref_word_to_location_id(W, CTX, ID, W1).

not_blocked_word(person, W, CTX) :-
	!,
	not_blocked_word_person(W, CTX).
not_blocked_word(function_person, W, CTX) :-
	!,
	not_blocked_word_function_person(W, CTX).
not_blocked_word(location, W, CTX) :-
	!,
	not_blocked_word_location(W, CTX).
not_blocked_word(date, _, _) :-
	!,
	fail.
not_blocked_word(_, _, _).

not_blocked_id(person, Id, CTX) :-
	!,
	not_blocked_id_person(Id, CTX).
not_blocked_id(function_person, Id, CTX) :-
	!,
	not_blocked_id_function_person(Id, CTX).
not_blocked_id(location, Id, CTX) :-
	!,
	not_blocked_id_location(Id, CTX).
not_blocked_id(_, _, _).

e_to_entity_direct(date, E, CTX, ID, Reasons) :-
	!,
	e_to_entity_date(E, CTX, ID, Reasons).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

features_lock_until(function_person, Features, Until) :-
	!,
	features_lock_until_for_function_person(Features, Until).
features_lock_until(person, Features, Until) :-
	!,
	features_lock_until_for_person(Features, Until).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Feature Properties
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

word_feature(not_explicitly_blocked).
word_feature(no_stopword).
word_feature(no_common_substantive).
word_feature(no_common_geoname).
word_feature(no_common_downcase).
word_feature(no_common_firstname).

nondeterministic_feature(linked_to_person_in_context(_)).

value_feature(linked_to_others(NLinks), Code, Keyval) :-
	Code = (Keyval is 1/(NLinks+1)).

value_feature(linked_to_many_others(NLinks), Code, Keyval) :-
	Code = (Keyval is 1/(NLinks+1)).

value_feature(population(NPeople), Code, Keyval) :-
	Code = (Keyval is 1/(NPeople+1)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Cached Local Predicates
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic threshold_feature/2.
% :- dynamic canonic_threshold_features/2.
% :- dynamic threshold_key/2.
:- dynamic word_features/2.

update_ner_control :-
	info(10, 'Updating ner control'),
%	retractall( canonic_threshold_features(_, _) ),
	retractall( threshold_feature(_, _) ),
%	retractall( threshold_key(_, _) ),
	retractall( word_features(_, _) ),
	( entity_features(Prop, Fs),
	  findall(WF, (member(WF, Fs), word_feature(WF)), WFs),
	  assert( word_features(Prop, WFs) ),
	  entity_threshold_features(Prop, TFs0),
	  sort_features(Prop, TFs0, TFs),
%	  assert( canonic_threshold_features(Prop, TFs) ),
%	  entity_features_key(Prop, TFs, TK),
%	  assert( threshold_key(Prop, TK) ),
	  ( member(F0, TFs),
	    assert( threshold_feature(Prop, F0) ),
	    fail
	  ; true
	  ),
	  fail
	; true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% e_entities(E, Prop, CTX, IRs) :-
% 	findall(I-pro(K, R), ( e_to_entity(Prop, E, CTX, I, R),
% 			       entity_features_key(Prop, R, K) ),
% 		IRs1),
% 	( IRs1 \= [] -> writeln(eee(IRs1)) ; true ),
% 	sort_entities(Prop, IRs1, default, IRs),
% 	( IRs1 \= [] -> writeln(fff(IRs1)) ; true ).

e_entities(E, Prop, CTX, IRs) :-
	findall(I-pro(K, R), ( e_to_entity(Prop, E, CTX, I, R),
			       entity_features_key(Prop, R, K) ),
		IRs1),
	sort_entities(Prop, IRs1, default, IRs).

e_to_entity(Prop, E, CTX, ID, Reasons) :-
	e_to_entity_direct(Prop, E, CTX, ID, Reasons).
e_to_entity(Prop, E, CTX, ID, Reasons) :-
	%% line break in word
	e_content(E, wrd(W)),
	ctx_get(CTX, elems_after, Es),
	word_break(Es, W1, Es1),
	!,
	atomic_list_concat([W, W1], W2),
	ctx_set(CTX, elems_after, Es1, CTX1),
        e_to_entity_1(Prop, W2, E, CTX1, ID, Reasons).	
e_to_entity(Prop, E, CTX, ID, Reasons) :-
	e_content(E, wrd(W)),
        e_to_entity_1(Prop, W, E, CTX, ID, Reasons).
	
e_to_entity_1(Prop, W, E, CTX, ID, Reasons) :-
	ctx_set(CTX, element, E, CTX1),
	not_blocked_word(Prop, W, CTX1),
	w_word_features(Prop, W, CTX1, WReasons),
	deref_word_to_id(Prop, W, CTX1, ID, W1),
	ctx_set(CTX1, word, W1, CTX2),
	not_blocked_id(Prop, ID, CTX2),
%	writeq(e_to_entity_1(Prop,W,W1)), nl,
	id_features(Prop, ID, CTX2, WReasons, Reasons).
%	writeq(success______(Prop,W,W1)), nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

word_break(Es, W, Es1) :-
	Es = [A1,A2,A3|Es2],
	e_content(A1, pun(Hyph)),
	is_word_breaking_hyphen(Hyph),
	( e_content(A2, cmd(_, _)) ->
	  B2 = A3,
	  Es2 = [B3|Es1]
	; B2 = A2,
	  B3 = A3,
	  Es2 = Es1
	),
	e_content(B2, spc(NL)),
	is_newline(NL),
	e_content(B3, wrd(W)),
	sub_atom(W, 0, 1, _, FirstChar),
	char_type(FirstChar, lower).
	  
is_word_breaking_hyphen('-').
is_word_breaking_hyphen('=').

is_newline('\n').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% For use in inspecting a single element
%%%%

e_entities_ignore_threshold(E, Prop, SortingMethod, CTX, IRs) :-
	findall(I-pro(K, R),
		( e_to_entity_ignore_threshold(Prop, E, CTX, I, R),
		  entity_features_key(Prop, R, K) ),
		IRs1),
	sort_entities(Prop, IRs1, SortingMethod, IRs).

e_to_entity_ignore_threshold(Prop, E, CTX, ID, Reasons) :-
	e_to_entity_direct(Prop, E, CTX, ID, Reasons).
e_to_entity_ignore_threshold(Prop, E, CTX, ID, Reasons) :-
	e_content(E, wrd(W)),
	ctx_set(CTX, element, E, CTX1),
	not_blocked_word(Prop, W, CTX1),
	w_word_features(Prop, W, CTX1, WReasons),
	deref_word_to_id(Prop, W, CTX1, ID, W1),
	ctx_set(CTX1, word, W1, CTX2),
	not_blocked_id(Prop, ID, CTX2),
	id_features_ignore_threshold(Prop, ID, CTX2, WReasons, Reasons).

id_features_ignore_threshold(Prop, ID, CTX, WReasons, Reasons) :-
	entity_features(Prop, Fs),
	findall(F, ( member(F, Fs),
		     \+ word_feature(F),
		     pro_id_feature(Prop, F, ID, CTX) ),
		Fs1),	
	append(WReasons, Fs1, Fs2),
	sort_features(Prop, Fs2, Reasons).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% e_to_entity(Prop, E, CTX, ID, Reasons) :-
% 	e_content(E, wrd(W)),
% 	ctx_set(CTX, element, E, CTX1),
% 	not_blocked_word(Prop, W, CTX1),
% 	writeq(	w_word_features(Prop, W, _CTX1, WReasons)), nl,
% 	w_word_features(Prop, W, CTX1, WReasons),
% 	writeln(done),
% 	deref_word_to_id(Prop, W, CTX1, ID),
% 	writeln(deref),
% 	not_blocked_id(Prop, ID, CTX1),
% 	writeln(nbi),
% 	writeq(id_features(Prop, ID, _CTX1, WReasons, Reasons)), nl,
% 	id_features(Prop, ID, CTX1, WReasons, Reasons),
% 	writeq(xxxx), nl.


w_word_features(Prop, W, CTX, Features) :-
	word_features(Prop, WFs),
	wwf_1(WFs, Prop, W, CTX, Features).

wwf_1([F|Fs], Prop, W, C, Fs1) :-
	( pro_word_feature(Prop, F, W, C) ->
	  Fs1 = [F|Fs2],
	  wwf_1(Fs, Prop, W, C, Fs2)
	; wwf_1(Fs, Prop, W, C, Fs1)
	).
wwf_1([], _, _, _, []).


id_features(Prop, ID, CTX, WReasons, Reasons) :-
	entity_features(Prop, Fs),
	idfs_1(Fs, Prop, ID, u, CTX, WReasons, Reasons).

% idfs_1([F|_], P, _, _, _, _, _) :-
%   	writeq(idfs(P,F)), nl, fail.

idfs_1([F|Fs], Prop, ID, Status, CTX, WReasons, Reasons1) :-
	threshold_feature(Prop, F),
	%% only allow atoms as threshold features
	!,
	( ( word_feature(F), memberchk(F, WReasons) ->
	    true
	  ; pro_id_feature(Prop, F, ID, CTX) ->
	    true
	  ) ->
	  Reasons1 = [F|Reasons2],
	  Status1 = Status
	; Status = t ->
	  Reasons1 = Reasons2,
	  Status1 = Status
	; fail
	),
	idfs_1(Fs, Prop, ID, Status1, CTX, WReasons, Reasons2).
idfs_1([F|Fs], Prop, ID, Status, CTX, WReasons, Reasons1) :-
	nondeterministic_feature(F),
	!,
	%% word features must be deterministic and
	%% thus do not appear here
	set_of(G, (pro_id_feature(Prop, F, ID, CTX), G=F), Fs1),
	( Fs1 = [] ->
	  Reasons1 = Reasons2,
	  Status1 = Status
	; Fs1 = [F1] ->
	  Reasons1 = [F1|Reasons2],
	  Status1 = t
	; append(Fs1, Reasons2, Reasons1),
	  Status1 = t
	),
	idfs_1(Fs, Prop, ID, Status1, CTX, WReasons, Reasons2).
idfs_1([F|Fs], Prop, ID, Status, CTX, WReasons, Reasons1) :-
	( ( word_feature(F), memberchk(F, WReasons) ->
	    true
	  ; pro_id_feature(Prop, F, ID, CTX) ->
	    true
	  ) ->
	  Reasons1 = [F|Reasons2],
	  Status1 = t
	; Reasons1 = Reasons2,
	  Status1 = Status
	),
	idfs_1(Fs, Prop, ID, Status1, CTX, WReasons, Reasons2).
idfs_1([], _, _, _, _, _, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

olist_prefix_upto_keygroup([], _, []).
olist_prefix_upto_keygroup([K-X|KXs], N, KXs1) :-
	olist_prefix_upto_keygroup([K-X|KXs], N, K, KXs1).

olist_prefix_upto_keygroup([K-X|KXs], N, K, [K-X|KXs1]) :-
	N > 0,
	!,
	olist_prefix_upto_keygroup(KXs, N, K, KXs1).
olist_prefix_upto_keygroup([_-X|KXs], N, K1, [K1-X|KXs1]) :-
	N1 is N-1,
	N1 > 0,
	!,
	olist_prefix_upto_keygroup(KXs, N1, K1, KXs1).
olist_prefix_upto_keygroup(_, _, _, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sort_features(Prop, Rs, Rs1) :-
	entity_features(Prop, Feat),
	map_add_feature_importance(Rs, Feat, IRs),
	sort(IRs, IRs1),
	map_val(IRs1, Rs1).

map_add_feature_importance([X|Xs], Feat, [I-X|Xs1]) :-
	position_of_element(Feat, X, I),
	map_add_feature_importance(Xs, Feat, Xs1).
map_add_feature_importance([], _, []).

position_of_element([X|_], X, 0) :-
	!.
position_of_element([_|Ys], X, N) :-
	position_of_element(Ys, X, N1),
	N is N1 + 1.
position_of_element([], _, 0).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sort_entities(Prop, PRs, PRs1) :-
	sort_entities(Prop, PRs, keep_all, PRs1).

sort_entities(_, [], _, []) :-
	!.
sort_entities(Prop, PRs, CutoffPolicy, PRs1) :-
	( CutoffPolicy = default ->
	  CutoffPolicy1 = keep_above(none, 3, 5)
	; CutoffPolicy = keep_best_group ->
	  CutoffPolicy1 = keep_above(none, 1, 5)
	; CutoffPolicy1 = CutoffPolicy
	),
	map_add_entity_key(PRs, KPRs),
	sort(KPRs, KPRs1),
	( CutoffPolicy1 = keep_all ->
	  KPRs2 = KPRs1
	; CutoffPolicy1 = keep_first ->
	  ( KPRs1 = [KPR|_] ->
	    KPRs2 = [KPR]
	  ; KPRs2 = []
	  )
	; CutoffPolicy1 = keep_above(Threshold, MaxGroups, MaxItems) ->
	  ( Threshold = none ->
	    KPRs3 = KPRs1
	  ; entity_features_key(Prop, Threshold, CutoffKey),
	    okv_list_prefix_lessq(KPRs1, CutoffKey, KPRs3)
	  ),
	  list_prefix_upto(MaxItems, KPRs3, KPRs4),
	  olist_prefix_upto_keygroup(KPRs4, MaxGroups, KPRs2)
	),
	!,
	map_val(KPRs2, PRs1).

map_add_entity_key([X-pro(K,R)|Xs], [K-(X-pro(K,R))|Xs1]) :-
	map_add_entity_key(Xs, Xs1).
map_add_entity_key([], []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

entity_features_key(Prop, ItemFeatures, Key) :-
	entity_features_key_1(Prop, ItemFeatures, Key). %% without caching 
%	entity_features_key_2(Prop, ItemFeatures, Key). %% with caching	

:- dynamic features_key_cache/3.

reset_features_key_cache :-
 	retractall( features_key_cache(_, _, _) ).
 
:- initialization(reset_features_key_cache).
 
entity_features_key_2(Prop, Features, Key) :-
	( member(X, Features), \+ atom(X) ->
	  entity_features_key_1(Prop, Features, Key)
	; entity_features_key_3(Prop, Features, Key)
	).

entity_features_key_3(Prop, Features, Key) :-
	features_key_cache(Prop, Features, Key),
	!.
entity_features_key_3(Prop, Features, Key) :-
	entity_features_key_1(Prop, Features, Key),
	assert( features_key_cache(Prop, Features, Key) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

entity_features_key_1(Prop, ItemFeatures, Key) :-
	entity_features(Prop, AllFeatures),
 	fk_2(AllFeatures, ItemFeatures, Args),
	%%	Key = Args.
	%%	atomic_list_concat(Args, Key).
 	Key =.. [k|Args].

fk_2([F|Fs], IFs, [N|Ns]) :-
	( atom(F) ->
	  ( memberchk(F, IFs) ->
	    N = 0.5
	  ; N = 1
	  )
	; value_feature(F, Code, N) ->
	  ( memberchk(F, IFs) ->
	    call(Code)
	  ; N = 1
	  )
	; nondeterministic_feature(F) ->
	  ( \+ \+ memberchk(F, IFs) ->
	    num_of_unifiable_members(IFs, F, NF),
	    N is 1/(NF+1)
	  ; N = 1
	  )
	; \+ \+ memberchk(F, IFs) ->
	  N = 0.5
	; N = 1
	),
	fk_2(Fs, IFs, Ns).
fk_2([], _, []).


% fk_2([F|Fs], IFs, [N|Ns]) :-
% 	( atom(F) ->
% 	  ( memberchk(F, IFs) ->
% 	    N = 8
% 	  ; N = 9
% 	  )
% 	; \+ \+ memberchk(F, IFs) ->
% 	  num_of_unifiable_members(IFs, F, NF),
% 	  N is 9 - min(9, NF)
% 	; N = 9
% 	),
% 	fk_2(Fs, IFs, Ns).
% fk_2([], _, []).


num_of_unifiable_members([X|Xs], Y, N) :-
	\+ \+ X = Y,
	!,
	num_of_unifiable_members(Xs, Y, N1),
	N is N1 + 1.
num_of_unifiable_members([_|Xs], Y, N) :-
	num_of_unifiable_members(Xs, Y, N).
num_of_unifiable_members([], _, 0).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- initialization(add_assistance_update_hook(ner_base:update_ner_control)).
