%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2016 , 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(ner_interface,
	  [show_person_summary/1,
	   show_person_summary_alphabetical/1,
	   show_person_summary_alphabetical_brief/1,
	   show_location_summary/1,
	   show_date_summary/1,
	   show_semantic_context_at_point/2,
	   show_structural_context_at_point/2,
	   inspect_item_at_srcpos/2,
	   process_source/1,
	   process_source/2,
	   set_verbosity/1]).

:- reexport(render,
	    [render_latex_annotated/1,
	     render_plain/1,
	     render_rxml_annotated/1,
	     render_ner_style/1,
	     render_entity_facts/1,
	     render_dariah_csv/1]).
	    
:- use_module(swilib(info)).
:- use_module(swilib(onto_atom)).
:- use_module(swilib(fromonto)).
:- use_module(swilib(err)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_base(gnd)).
:- use_module(kbset_ner(ner)).
:- use_module(kbset_ner(person_ner)).
:- use_module(kbset_ner(location_ner)).
:- use_module(kbset_ner(date_ner)).
:- use_module(kbset_ner(ner_base)).
:- use_module(kbset_ner(ner_utils)).
:- use_module(kbset_base(location)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(geo)). % show_location_wv
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(registry)).
:- use_module(kbset_ner(semctx)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_ner(latexannot)).
:- use_module(kbset_ner(esops)).

:- dynamic(last_es/1).

set_last_es(Es) :-
	retractall(last_es(_)),
	assert(last_es(Es)).

write_kbset_high_file(Src, Es) :-
	pure_file_base_name(Src, PureBase),
	filename(kbset_high_file_pattern, FilePattern),
	format(atom(File), FilePattern, [PureBase]),
	write_kbset_high_file(Src, Es, File).

write_kbset_high_file(_Src, Es, File) :-
	info(50, 'Writing highlight file: ~w', [File]),
	highlight_table(Es, Rows),
	onto_file( pp_lisp_highlight_table(Rows), File ),
	set_last_es(Es).

highlight_table(Es, Rows) :-
	dates_highlight_table(Es, Rows1),
	persons_highlight_table(Es, Rows2),
	locations_highlight_table(Es, Rows4),
	%%	function_names_highlight_table(Es, Rows3),
	%%	apply_append([Rows1, Rows2, Rows3, Rows4], Rows).
	apply_append([Rows1, Rows2, Rows4], Rows).	

dates_highlight_table(Es, Rows) :-
	findall(P-D, es_get_with_srcpos(Es, date, D, P), PDs),
	map_mk_highlight_table_row(PDs, date, Rows).

persons_highlight_table(Es, Rows) :-
	findall(P-D,
		es_get_with_srcpos_single_elem_prop(Es,
						    persons, D, P),
		PDs),
	map_mk_highlight_table_row(PDs, persons, Rows).

function_names_highlight_table(Es, Rows) :-
	findall(P-D,
		es_get_with_srcpos_single_elem_prop(Es,
						    function_name, D, P),
		PDs),
	map_mk_highlight_table_row(PDs, function_name, Rows).

locations_highlight_table(Es, Rows) :-
	findall(P-D,
		es_get_with_srcpos_single_elem_prop(Es,
						    locations, D, P),
		PDs),
	map_mk_highlight_table_row(PDs, locations, Rows).

pp_lisp_highlight_table(Rows) :-
	format('(setq kbset-high-table \'(~n'),
	( member(row(P1, P2, Kind, Info), Rows),
	  format('  (~d ~d ~w "~w")~n', [P1, P2, Kind, Info]),
	  fail
	; format('))')
	).

map_mk_highlight_table_row([_-begin(ignore-_),_-end|Xs], Kind, Xs1) :-
	!,
	map_mk_highlight_table_row(Xs, Kind, Xs1).
map_mk_highlight_table_row([P1-begin(D),P2-end|Xs], Kind,
			   [row(P1, P2, Kind, Info)|Xs1]) :-
	!,
	( highlight_table_info(Kind, D, Info) ->
	  true
	; err('highlight_table_info failed: ~q ~q', [Kind, D])
	),
	map_mk_highlight_table_row(Xs, Kind, Xs1).
map_mk_highlight_table_row([X|_], Kind, _) :-
	err('map_mk_highlight_table_row failed: ~q ~q', [Kind, X]).
map_mk_highlight_table_row([], _, []).

highlight_table_info(date, D, Info) :-
	!,
	onto_atom( (show_date_vwr(D), delim), Info ).
highlight_table_info(persons, PCs, Info) :-
	!,
	onto_atom( (show_persons(PCs), delim), Info ).
highlight_table_info(function_name, N, Info) :-
	!,
	format(atom(Info), 'Function: ~w', [N]).

highlight_table_info(locations, Ps, Info) :-
	!,
	onto_atom( (show_locations_vwr(Ps), delim), Info).

delim :-
	write('====================================='),
	writeln('=========================================').


show_persons(Ps) :-
	sort_entities(person, Ps, Ps1),
	show_persons_vwr(Ps1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

process_source(Src) :-
	process_source(Src, []).

process_source(Src, Options) :-
	file_to_es(Src, Options, Es),
	register_doc(Src, Options, Es),
	es_process(Es, Es1),
	register_processed_doc(Src, [], Es1),
	( memberchk(no_high_file, Options) ->
	  true
	; write_kbset_high_file(Src, Es1)
	).
	
es_process(Es, Es1) :-
	ner_processing_options(Options),
	es_ner(Es, Options, Es1),
	set_last_es(Es1).

inspect_item_at_srcpos(Src, SrcPos) :-
	doc_processed_es(Src, Es),
	!,
	ner_inspect_es_item_at_srcpos(Es, SrcPos).

show_person_summary(Src) :-
	doc_processed_es(Src, Es),
	!,
	ner_show_es_item_summary(person, Es,
				 [sort=frequency, show_rationales=true]).

show_person_summary_alphabetical(Src) :-
	doc_processed_es(Src, Es),
	!,
	ner_show_es_item_summary(person, Es,
				 [sort=alpha, show_rationales=true]).

show_person_summary_alphabetical_brief(Src) :-
	doc_processed_es(Src, Es),
	!,
	ner_show_es_item_summary(person, Es, [sort=alpha]).

show_location_summary(Src) :-
	doc_processed_es(Src, Es),
	!,
	ner_show_es_item_summary(location, Es,
				 [sort=frequency, show_rationales=true]).

show_date_summary(Src) :-
	doc_processed_es(Src, Es),
	!,
	ner_show_es_date_summary(Es).

show_words_by_frequency(Src) :-
	doc_processed_es(Src, Es),
	!,
	es_words_by_frequency(Es, [output_frequency=true], FWs),
	( member(F-W, FWs),
	  format('~w (~6f)~n', [W, F]),
	  fail
	; true
	).
	  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_semantic_context_at_point(Src, Point) :-
	doc_processed_es(Src, Es),
	!,
	make_semctx_at_point(Es, Point, SCTX),
	show_semctx(SCTX).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_structural_context_at_point(Src, Point) :-
	doc_processed_es(Src, Es),
	!,
	es_split_at_srcpos(Es, Point, REs1, _),
	reverse(REs1, Es1),
	es_structure_context_at_end(Es1, S),
	show_structure_context(S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% :- dynamic val/2.
% 
% set(Key, Value) :-
% 	retractall(val(Key, _)),
% 	assert(val(Key, Value)),
% 	info(10, 'Set ~w to ~q', [Key, Value]).
% 
% get(Key, Value) :-
% 	val(Key, Value).
% 
% set_interactively(Key) :-
% 	( val(Key, Oldval) ->
% 	  true
% 	; Oldval = []
% 	),
% 	format('Set ~w to: ~q', [Key, Oldval]),
% 	read(Val),
% 	set(Key, Val).
% 
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_verbosity(N) :-
	set_info_verbosity(N).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


