%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(ner_utils,
	  [context_requirement_matches/3,
	   show_context_match_rationale/2,
	   show_context_match_rationale/1,
	   e_identified_item/3,
	   e_identified_item_id/3,
	   ner_item_id/2,
	   ner_item_features/2,
	   pro_word_feature_no_stopword/2,
	   pro_word_feature_no_common_substantive/2,
	   pro_word_feature_no_common_downcase/2,
	   pro_word_feature_no_common_firstname/2,
	   pro_word_feature_no_common_lastname/2,
	   show_reasons/1]).

:- use_module(kbset_ner(layers)).
:- use_module(kbset_ner(fmtutils)).
:- use_module(kbset_data(stoplists)).
:- use_module(kbset_ner(wordutils)).
:- use_module(kbset_data(wordlists_de)).
:- use_module(kbset_base(textutils)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(ctx)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(geo)).
:- use_module(kbset_base(localization)).

context_requirement_matches(Conditions, CTX, Rationale) :-
	map_crm_cond(Conditions, CTX, Rationale).

map_crm_cond([X|Xs], Y1, [X1|Xs1]) :-
	crm_cond(X, Y1, X1),
	map_crm_cond(Xs, Y1, Xs1).
map_crm_cond([], _, []).

crm_cond(followed_by_words=Ws, CTX, followed_by_words=Ws) :-
	!,
	ctx_get(CTX, elems_after, After),
	es_word_prefix(Ws, After).
crm_cond(following_words=Ws, CTX, following_words=Ws) :-
	!,
	ctx_get(CTX, elems_rev_before, Before),
	reverse(Ws, Ws1),
	es_word_prefix(Ws1, Before).
crm_cond(near_word=W, CTX, near_word=W) :-
	!,
	WINDOWSIZE = 100,
	( ctx_get(CTX, elems_rev_before, RevBefore),
	  es_enum_words_upto(RevBefore, WINDOWSIZE, W) ->
	  true
	; ctx_get(CTX, elems_after, After),
	  es_enum_words_upto(After, WINDOWSIZE, W) ->
	  true
	).
crm_cond(near_word_in=Ws, CTX, R) :-
	!,
	member(W, Ws),
	crm_cond(near_word=W, CTX, R),
	!.
%% other conditions to be evaluated by special checks,
%% ignored here
crm_cond(_, _, true).


show_context_match_rationale(X) :-
	show_context_match_rationale(0, X).

show_context_match_rationale(_, []) :-
	!,
	true.
show_context_match_rationale(Indent, Rationale) :-
        user_language(L),
	map_show_context_match_rationale_item(Rationale, L, Indent).

map_show_context_match_rationale_item([RationaleItem|RIs], L, Indent) :-
	( RationaleItem == true ->
	  true
	; indent(Indent),
	  format('~q', [RationaleItem]),
	  ( RIs = [_|_] ->
	    format('~n', [])
	  ; true
	  )
	),
	map_show_context_match_rationale_item(RIs, L, Indent).
map_show_context_match_rationale_item([], _, _).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

e_identified_item(Prop, E, Item) :-
	e_get(E, Prop, [Item|Is1]),
	( Is1 = [] ->
	  true
	; Item = _-pro(K1, _),
	  Is1 = [_-pro(K2, _)|_],
	  K1 @< K2
	).

e_identified_item_id(Prop, E, Id) :-
	e_identified_item(Prop, E, Item),
	ner_item_id(Item, Id).

ner_item_id(Id-_, Id) :-
	!.
ner_item_id(Id, Id).

ner_item_features(_-pro(_,F), F) :-
	!.
ner_item_features(_, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


pro_word_feature_no_stopword(W, CTX) :-
	ctx_get(CTX, text_language, Lang),
	\+ is_stopword(Lang, W).

pro_word_feature_no_common_substantive(W, CTX) :-
	ctx_get(CTX, text_language_variant, VLang),
	\+ word_is_common_substantive(VLang, W).

pro_word_feature_no_common_downcase(W, CTX) :-
	ctx_get(CTX, text_language_variant, VLang),
	\+ word_is_common_downcase(VLang, W).

pro_word_feature_no_common_firstname(W, _CTX) :-
	\+ ( common_firstname(W, N), N > 0.002 ).

pro_word_feature_no_common_lastname(W, _CTX) :-
	findall(X, name_to_person(W, X), Xs),
	length(Xs, L),
	L =< 2.

% pro_word_feature_no_common_downcase_at_begin_of_sentence(W, CTX) :-
% 	\+ aux_common_downcase_at_begin_of_sentence(W, CTX).
% 
% aux_common_downcase_at_begin_of_sentence(W, CTX) :-
% 	ctx_get(CTX, elems_rev_before, Es),
% 	es_first_punct_before_word(Es, P),
% 	punct_stops_sentence(P),
% 	ctx_get(CTX, text_language_variant, VLang),
% 	word_is_common_downcase(VLang, W).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_pretty(de, Reason, Fmt) :-
	pro_pretty_de(Reason, Fmt),
	!.
pro_pretty(en, Reason, Fmt) :-
	pro_pretty_en(Reason, Fmt),
	!.
pro__pretty(_, Reason, Fmt) :-
	format(atom(Fmt), Reason).

show_reasons(Reasons) :-
	main_context(CTX),
	show_reasons(Reasons, CTX).
	
show_reasons(Reasons, _CTX) :-
	Reasons1 = Reasons,
	%% no need to sort, the new implem should already yields them sorted,
	%% if it should be used, the show_person stuff should go to another
	%% module since sort_entity_features in in module base_ner which
	%% imports person_ner *** sort_entity_features(person, Reasons,
	%% Reasons1),
	user_language(Lang),
	( member(Reason, Reasons1),
	  ( pro_pretty(Lang, Reason, Fmt) ->
	    pp_fmt(Fmt, 16)
	  ; indent(16),
	    writeln(Reason)
	  ),
	  fail
	; true
	).

pp_fmt(Atom-Args, Indent) :-
	!,
	indent(Indent),
	format(Atom, Args),
	nl.
pp_fmt(Atom, Indent) :-
	indent(Indent),
	format(Atom),
	nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_pretty_de(not_explicitly_blocked,
	      'Nicht explizit blockiert').
pro_pretty_de(no_stopword, 'Kein Stopwort').
pro_pretty_de(no_common_substantive, 'Kein geläufiges Substantiv').
pro_pretty_de(no_common_geoname, 'Kein geläufiger Ortsname').
pro_pretty_de(no_common_firstname, 'Kein geläufiger Vorname').
pro_pretty_de(no_common_lastname, 'Kein geläufiger Nachname').
pro_pretty_de(in_wikipedia_de,
	      'In der deutschen Wikipedia').
pro_pretty_de(in_wikipedia,
	      'In der Wikipedia').
pro_pretty_de(followed_by_matching_roman_number(RN),
	      'Vor von römischer Zahl: ~w'-[RN]).
pro_pretty_de(followed_by_matching_specifier(Ws),
	      'Funktion: ~@'-[write_members_separated_by_blanks(Ws)]).
pro_pretty_de(followed_by_matching_extension(Extension),
	      'Vor Zusatz: ~w'-[Extension]).
pro_pretty_de(preceded_by_matching_first_names_initials(As),
	      'Nach Initialen: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_de(followed_by_comma_and_matching_first_names_initials(As),
	      'Vor Komma und Initialen: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_de(preceded_by_matching_first_names(As),
	      'Nach Vornamen: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_de(followed_by_comma_and_matching_first_names(As),
	      'Vor Komma und Vornamen: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_de(followed_by_matching_extra_names(As),
	      'Vor Zusatz: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_de(has_pid(A),
	      'Identifikator ist ~w'-[A]).
pro_pretty_de(explicitly_specified,
	      'Explizit spezifiziert').
pro_pretty_de(explicitly_specified_in_context(Arg),
	      'Explizit spezifiziert im Kontext ~@'
	      -[show_context_match_rationale(Arg)]).
pro_pretty_de(born_in_span_before_year_in_context(YB, YC),
	      'Geboren ~q, passend zu Kontextjahr ~w'-[YB,YC]).
pro_pretty_de(occupation_mentioned(W),
	      'Beruf oder Beschäftigung erwähnt: ~w'-[W]).
pro_pretty_de(linked_to_person_in_context(PID),
	      'Verlinkt zu Person im Kontext: ~@'
	      -[show_person_nonl(PID)]).
pro_pretty_de(linked_to_others(N),
	      'Verlinkt mit ~w anderen'-[N]).
pro_pretty_de(linked_to_many_others(N),
	      'Verlinkt mit mehr als ~w anderen'-[N]).
pro_pretty_de(already_identified_in_context,
	      'Bereits identifiziert').
pro_pretty_de(near_to_already_identified_in_context(LID),
	      'Nah bei bereits identifiziertem Ort: ~@'-
	      [(geo_localized_name(LID, Name), write(Name))]).
pro_pretty_de(referenced_by_preferred_name,
	      'Im Text steht der bevorzugte Name').
pro_pretty_de(no_common_downcase,
	      'Kein kleinschreibbares Wort').
pro_pretty_de(population(N),
	      'Population: ~w'-[N]).
pro_pretty_de(has_population_over_10000,
	      'Population >= 10.000').
pro_pretty_de(has_population_over_100000,
	      'Population >= 100.000').
pro_pretty_de(is_city_or_village,
	      'Stadt oder Dorf').
pro_pretty_de(is_country,
	      'Land').


%% Q: associate with a name, from the assistance
pro_pretty_de(born_before(Y), 'Vor ~w geboren'-[Y]).
pro_pretty_de(year_of_birth(Y), '~w geboren'-[Y]).
pro_pretty_de(year_of_death(Y), '~w gestorben'-[Y]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_pretty_en(not_explicitly_blocked,
	      'Not explicitly blocked').
pro_pretty_en(no_stopword, 'No stop word').
pro_pretty_en(no_common_substantive, 'No common noun').
%%
%% note: "common noun" has a particular meaning in English, it is distinguised
%% against "proper noun".
%%
pro_pretty_en(no_common_geoname, 'No common location name').
pro_pretty_en(no_common_firstname, 'No common first name').
pro_pretty_en(no_common_lastname, 'No common last name').
pro_pretty_en(in_wikipedia_de,
	      'In the German Wikipedia').
pro_pretty_en(in_wikipedia,
	      'In the Wikipedia').
pro_pretty_en(followed_by_matching_roman_number(RN),
	      'Before Roman numerals: ~w'-[RN]).
pro_pretty_en(followed_by_matching_specifier(Ws),
	      'Function: ~@'-[write_members_separated_by_blanks(Ws)]).
pro_pretty_en(followed_by_matching_extension(Extension),
	      'Before addition: ~w'-[Extension]).
pro_pretty_en(preceded_by_matching_first_names_initials(As),
	      'After initials: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_en(followed_by_comma_and_matching_first_names_initials(As),
	      'Before comma and initials: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_en(preceded_by_matching_first_names(As),
	      'After first name(s): ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_en(followed_by_comma_and_matching_first_names(As),
	      'Before comma and first name(s): ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_en(followed_by_matching_extra_names(As),
	      'Before addition: ~@'
	      -[write_members_separated_by_blanks(As)]).
pro_pretty_en(has_pid(A),
	      'Identifier is ~w'-[A]).
pro_pretty_en(explicitly_specified,
	      'Explicitly specified').
pro_pretty_en(explicitly_specified_in_context(Arg),
	      'Explicitly specified in context ~@'
	      -[show_context_match_rationale(Arg)]).
pro_pretty_en(born_in_span_before_year_in_context(YB, YC),
	      'Born ~q, matching context year ~w'-[YB,YC]).
pro_pretty_en(occupation_mentioned(W),
	      'Profession or occupation mentioned: ~w'-[W]).
pro_pretty_en(linked_to_person_in_context(PID),
	      'Linked to person in context: ~@'
	      -[show_person_nonl(PID)]).
pro_pretty_en(linked_to_others(N),
	      'Linked to ~w others'-[N]).
pro_pretty_en(linked_to_many_others(N),
	      'Linked to more than ~w others'-[N]).
pro_pretty_en(already_identified_in_context,
	      'Already identified').
pro_pretty_en(near_to_already_identified_in_context(LID),
	      'Close to already identified location: ~@'-
	      [(geo_localized_name(LID, Name), write(Name))]).
pro_pretty_en(referenced_by_preferred_name,
	      'The preferred name is in the text').
pro_pretty_en(no_common_downcase,
	      'Not commonly used in lowercase').
pro_pretty_en(population(N),
	      'Population: ~w'-[N]).
pro_pretty_en(has_population_over_10000,
	      'Population >= 10.000').
pro_pretty_en(has_population_over_100000,
	      'Population >= 100.000').
pro_pretty_en(is_city_or_village,
	      'City or village').
pro_pretty_en(is_country,
	      'Country').

pro_pretty_en(born_before(Y), 'Born before ~w'-[Y]).
pro_pretty_en(year_of_birth(Y), 'Born in ~w'-[Y]).
pro_pretty_en(year_of_death(Y), 'Died in ~w'-[Y]).

