%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(ngrams, [generate_ngrams/4]).

:- use_module(kbset_base(hash_kbset)).
:- use_module(kbset_base(utils)).
:- use_module(swilib(pretty)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% N-Grams (Sequences of Words)
%%%% 
%%%% An implementation of the apriori algorithm shown in Fuernkranz:
%%%% A study using n-gram features for text categorization, 1998
%%%% (Remark: the original algorithm takes a set of documents, but seems to
%%%% consider only the overall frequency of n_grams)
%%%% 
%%%% generate_ngrams(+Documents, +MaxN, +MinF, -NGrams)
%%%% 
%%%%    Documents: List of words (atoms)
%%%%    MaxN: Maximal size of n_grams (word sequences) to detect
%%%%    MinF: Minimal frequency with which the n_grams must occur
%%%%    NGrams: List of n-grams, where an n-gram is a list of words (atoms)
%%%%    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ngram_feature_tab_hashsize(10000).
ngram_counter_tab_hashsize(10000).

generate_ngrams(Doc, MaxN, MinF, Features) :-
	ngram_feature_tab_hashsize(HSIZEF),
	ngram_counter_tab_hashsize(HSIZEC),
	mk_ht(HSIZEF, FeatureTab),
	ht_put(FeatureTab, [], true),
	( count_up(MaxN, N),
	  mk_ht(HSIZEC, CounterTab),
	  ( n_gram(Doc, N, NGram),
	    n_gram_butlast(NGram, INGram),
	    n_gram_butfirst(NGram, FNGram),
	    ( ht_get(FeatureTab, INGram, true),
	      ht_get(FeatureTab, FNGram, true) ->
	      ht_inc(CounterTab, NGram)
	    ; true
	    ),
	    fail
	  ; true
	  ),
	  ( ht_key(CounterTab, K),
	    ht_get(CounterTab, K, V),
	    ( V >= MinF ->
	      ht_put(FeatureTab, K, true)
	    ; true
	    ),
	    fail
	  ; true
	  ),
	  fail
	; true
	),
	findall(F, ht_key(FeatureTab, F), Features1),
	sort(Features1, [_|Features]).
	    
ht_inc(T, Key) :-
	( ht_get(T, Key, N) ->
	  N1 is N + 1,
	  ht_put(T, Key, N1)
	; ht_put(T, Key, 1)
	).
	
n_gram(E, N, NG) :-
	n_head(E, N, NG).
n_gram([_|Es], N, NG) :-
	n_gram(Es, N, NG).

n_head(_, 0, []).
n_head([E|Es], N, [E|Es1]) :-
	N > 0,
	N1 is N - 1,
	n_head(Es, N1, Es1).

n_gram_butfirst([_|Es], Es).
n_gram_butlast(Es, Es1) :- append(Es1, [_], Es), !.


% 
% pp_t(T) :-
% 	( ht_key(T, K),
% 	  ht_get(T, K, V),
% 	  writeq(kv(K, V)),
% 	  writeln('.'),
% 	  fail
% 	; true
% 	).