%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2016 , 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(person_ner,
	  [ deref_word_to_person_id/4,
	    not_blocked_word_person/2,
	    not_blocked_id_person/2,
	    pro_word_person_feature/3,
	    pro_id_person_feature/3,

	    deref_word_to_function_person_id/4,
	    not_blocked_word_function_person/2,
	    not_blocked_id_function_person/2,
	    pro_word_function_person_feature/3,
	    pro_id_function_person_feature/3,
	    
	    features_lock_until_for_function_person/2,
	    features_lock_until_for_person/2,
	    
	    stdize_person/2,
	    show_person_vwr/1,
	    show_persons_vwr/1,
	    
	    compile_person_assistance/0]).

:- use_module(kbset_ner(ctx)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(function)).
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(wordutils)).
:- use_module(kbset_base(textutils)).
:- use_module(swilib(info)).
:- use_module(kbset_data(stoplists)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(fmtutils)).
:- use_module(kbset_ner(germanutils)).
:- use_module(kbset_data(wordlists_de)).
:- use_module(swilib(err)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_ner(ner_utils)).
:- use_module(kbset_base(function)).

:- use_module(kbset_ner(location_ner), [not_exp_blocked_as_location_name_in_context/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

deref_word_to_person_id(W, CTX, PID, W1) :-
	ctx_get_defaulted(CTX, text_language, Lang, de),
	ord_enum(PID-W1, ( name_possible_basename(Lang, W, W1),
			   name_to_person(W1, PID) )).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

not_blocked_word_person(W, _) :-
	word_is_capitalized(W).

not_blocked_id_person(Id, CTX) :-
	\+ blocked_id_person(Id, CTX).

blocked_id_person(PID, CTX) :-	
	ctx_get(CTX, person_born_before, Y),
	person_year_of_birth(PID, Y1),
	Y1 > Y.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_word_person_feature(not_explicitly_blocked, W, CTX) :-
	not_exp_blocked_as_person_name_in_context(W, CTX).

pro_word_person_feature(no_stopword, W, CTX) :-
	pro_word_feature_no_stopword(W, CTX).

pro_word_person_feature(no_common_substantive, W, CTX) :-
	pro_word_feature_no_common_substantive(W, CTX),
	!.
pro_word_person_feature(no_common_substantive, _, CTX) :-
	ctx_preceded_by_person_intro(CTX).
	
pro_word_person_feature(no_common_firstname, W, CTX) :-
	pro_word_feature_no_common_firstname(W, CTX).

pro_word_person_feature(no_common_geoname, W, CTX) :-
	ctx_get(CTX, text_language, Lang),
	( word_is_common_geoname(Lang, W) ->
	  \+ not_exp_blocked_as_location_name_in_context(W, CTX)
	; true
	).

pro_word_person_feature(no_common_downcase, W, CTX) :-
	pro_word_feature_no_common_downcase(W, CTX),
	!.
pro_word_person_feature(no_common_downcase, _, CTX) :-
	ctx_preceded_by_person_intro(CTX).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ctx_preceded_by_person_intro(CTX) :-
	ctx_get(CTX, elems_rev_before, Es),
	es_first_word(Es, W),
	is_person_intro_word(W),
	!.

is_person_intro_word('D').
is_person_intro_word('Doctor').
is_person_intro_word('Doktor').
is_person_intro_word('Dr').
is_person_intro_word('Fr').
is_person_intro_word('Frau').
is_person_intro_word('Frl').
is_person_intro_word('Fräulein').
is_person_intro_word('H').
is_person_intro_word('Herr').
is_person_intro_word('Herren').
is_person_intro_word('Herrn').
is_person_intro_word('Hr').
is_person_intro_word('Hrn').
is_person_intro_word('M').
is_person_intro_word('MM').
is_person_intro_word('Madame').
is_person_intro_word('Mademoiselle').
is_person_intro_word('Mister').
is_person_intro_word('Mle').
is_person_intro_word('Mlle').
is_person_intro_word('Mme').
is_person_intro_word('Monsieur').
is_person_intro_word('Mr').
is_person_intro_word('Mrs').
is_person_intro_word('Ms').
is_person_intro_word('Mag').
is_person_intro_word('Magister').
is_person_intro_word('Prof').
is_person_intro_word('Professor').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pro_id_person_feature(in_wikipedia_de, PID, _CTX) :-
	person_is_in_dewiki(PID).

pro_id_person_feature(in_wikipedia, PID, _CTX) :-
	person_is_in_wiki(PID).

pro_id_person_feature(linked_to_others(N), PID, _CTX) :-
	findall(X, person_linked_to_person(PID, X), Xs),
	sort(Xs, Xs1),
	length(Xs1, N),
	N > 0.

pro_id_person_feature(linked_to_many_others(N), PID, CTX) :-
	findall(X, person_linked_to_person(PID, X), Xs),
	sort(Xs, Xs1),
	length(Xs1, N1),
	ctx_get_defaulted(CTX, person_ner_many_links_granularity, Gran, 50),
	N2 is N1 div Gran,
	N2 > 0,
	N is N2 * Gran.

pro_id_person_feature(occupation_mentioned(W), PID, CTX) :-
	person_occupation(PID, W),
	WINDOW = 6,
	( ctx_get(CTX, elems_rev_before, Es),
	  es_enum_words_upto(Es, WINDOW, W) ->
	  true
	; ctx_get(CTX, elems_after, Es),
	  es_enum_words_upto(Es, WINDOW, W) ->
	  true
	),
	!.

pro_id_person_feature(preceded_by_matching_first_names_initials(Ws), PID, CTX) :-
	ctx_get(CTX, elems_rev_before, Es),
	ctx_get(CTX, word, W),
	person_first_names_initials(PID, W, Ws),
	Ws \= [],
	reverse(Ws, Ws1),
	( es_word_prefix(Ws1, Es, Rest),
	  \+ ( es_first_word(Rest, PreceedingWord),
	       seems_initial(PreceedingWord)
	     ) ->
	  true	
	; Ws = [_,_|_],
	  es_first_word(Es, W2),
	  atomic_list_concat(Ws, W2)
	),
	!.

pro_id_person_feature(preceded_by_matching_first_names(Ws), PID, CTX) :-
	ctx_get(CTX, elems_rev_before, Es),
	ctx_get(CTX, word, W),
	person_first_names(PID, W, Ws),
	Ws \= [],
	reverse(Ws, Ws1),
	es_word_prefix(Ws1, Es),
	!.

pro_id_person_feature(followed_by_comma_and_matching_first_names(Ws), PID, CTX) :-
	ctx_get(CTX, elems_after, Es),
	Es = [E1|_],
	e_content(E1, pun(',')),
	ctx_get(CTX, word, W),
	person_first_names(PID, W, Ws),
	Ws \= [],
	es_word_prefix(Ws, Es),
	!.

pro_id_person_feature(followed_by_comma_and_matching_first_names_initials(Ws), PID, CTX) :-
	ctx_get(CTX, elems_after, Es),
	Es = [E1|_],
	e_content(E1, pun(',')),
	ctx_get(CTX, word, W),
	person_first_names_initials(PID, W, Ws),
	Ws \= [],
	es_word_prefix(Ws, Es, Rest),
	\+ ( es_first_word(Rest, PreceedingWord),
	     seems_initial(PreceedingWord)
	   ),
	!.

pro_id_person_feature(followed_by_matching_extra_names(Ws), PID, CTX) :-
	ctx_get(CTX, elems_after, Es),
	ctx_get(CTX, word, W),
	person_extra_names(PID, W, Ws),
	Ws \= [],
	es_word_prefix(Ws, Es),
	!.

% 2014 GND
%
% pro_id_person_feature(followed_by_matching_roman_number(RN), PID, CTX) :-
% 	ctx_get(CTX, elems_after, Es),
% 	ctx_get(CTX, word, Name),
% 	es_first_word(Es, RN),
% 	is_roman_number(RN),
% 	person_extra_name(PID, Name, W),
% 	starts_with_roman_number(W, RN),
% 	!.

pro_id_person_feature(followed_by_matching_roman_number(RN), PID, CTX) :-
	ctx_get(CTX, elems_after, Es),
	ctx_get(CTX, word, Name),
	es_first_word(Es, RN),
	is_roman_number(RN),
	( atomic_list_concat([Name, ' ', RN], Name1)
	; atomic_list_concat([Name, ' ', RN, '.'], Name1)
	),
	person_last_name(PID, Name1),
	!.

pro_id_person_feature(referenced_by_preferred_name, PID, CTX) :-
	ctx_get(CTX, element, E),
	e_content(E, wrd(W)),
	person_preferred_last_name(PID, PrefName),
	( W = PrefName ->
	  ( ctx_get(CTX, elems_after, Es),
	    es_first_word(Es, RN),
	    is_roman_number(RN) ->
	    atomic_list_concat([W, ' ', RN, '.'], PrefName)
	  ; true
	  )
	; sub_atom(W, 0, _, _, PrefName)
	).

pro_id_person_feature(born_before(Y), PID, CTX) :-
	ctx_get(CTX, person_born_before, Y),
	person_year_of_birth(PID, Y1),
	Y1 =< Y.

pro_id_person_feature(has_pid(PID), PID, CTX) :-
	ctx_get(CTX, person_pid, PID).

pro_id_person_feature(explicitly_specified_in_context(Arg), PID, CTX) :-
	exp_specified_person_in_context(PID, CTX, Arg).

pro_id_person_feature(explicitly_specified, PID, CTX) :-
	exp_specified_person_general(PID, CTX).

pro_id_person_feature(year_of_birth(Y), PID, CTX) :-
	ctx_get(CTX, person_year_of_birth, Y),
	person_year_of_birth(PID, Y).

pro_id_person_feature(year_of_death(Y), PID, CTX) :-
	ctx_get(CTX, person_year_of_death, Y),
	person_year_of_death(PID, Y).

pro_id_person_feature(born_in_span_before_year_in_context(X, Y), PID, CTX) :-
	ctx_get_multi(CTX, context_years, Y),
	person_year_of_birth(PID, X),
	D is Y - X,
	D =< 100,
	D >= 15.

pro_id_person_feature(linked_to_person_in_context(PID1), PID, CTX) :-
	ctx_get(CTX, identified_items, PIDs),
	ctx_get_defaulted(CTX, person_ner_linked_context_window_size, WIN, 20),
	member_upto(PID1, PIDs, WIN),
	person_linked_to_person(PID, PID1).

pro_id_person_feature(already_identified_in_context, PID, CTX) :-
  	ctx_get(CTX, identified_items, Es),
	memberchk(PID, Es).

% pro_id_person_feature(linked_to_person_in_context(PID1), PID, CTX) :-
% 	ctx_get(CTX, elems_rev_before, Es),
% 	member_upto(E, Es, 100),
% 	e_identified_item_id(persons, E, PID1),
% 	person_linked_to_person(PID, PID1).
% 
% 
% %% 
% % pro_id_person_feature(already_identified_in_context, PID, CTX) :-
% %  	ctx_get(CTX, elems_rev_before, Es),
% %  	once(( member_upto(E, Es, 100),
% %  	       e_identified_item_id(persons, E, PID) )).
% %  

% 
% 				%	ctx_get_multi(CTX, context_persons, PID),
% 	show_person(PID).


features_lock_until_for_person(Features, Until) :-
	memberchk(followed_by_matching_extra_names(Ws), Features),
	last(Ws, W),
	Until = wrd(W).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

seems_initial(X) :-
	atom_length(X, 1),
	!.
seems_initial('Ch').
seems_initial('Th').
seems_initial('Ph').
seems_initial('Ps').

name_possible_basename(de, W, W1) :-
	!,
	remove_diacritics_used_in_handwriting(W, W0),
	name_possible_basename_de(W0, W2),
	enum_umlaut_variants(W2, W1).
name_possible_basename(en, W, W).

name_possible_basename_de(W, W).
name_possible_basename_de(W, W1) :- has_suffix(W, isches, W1), has_suffix(W1, n).
name_possible_basename_de(W, W1) :- has_suffix(W, ischen, W1), has_suffix(W1, n).
name_possible_basename_de(W, W1) :- has_suffix(W, ischer, W1), has_suffix(W1, n).
name_possible_basename_de(W, W1) :- has_suffix(W, ische, W1), has_suffix(W1, n).
name_possible_basename_de(W, W1) :- has_suffix(W, isches, W1), has_suffix(W1, r).
name_possible_basename_de(W, W1) :- has_suffix(W, ischen, W1), has_suffix(W1, r).
name_possible_basename_de(W, W1) :- has_suffix(W, ischer, W1), has_suffix(W1, r).
name_possible_basename_de(W, W1) :- has_suffix(W, ische, W1), has_suffix(W1, r).
name_possible_basename_de(W, W1) :- has_suffix(W, n, W1), has_suffix(W1, e).
name_possible_basename_de(W, W1) :- has_suffix(W, ens, W1), has_suffix(W1, s).
name_possible_basename_de(W, W1) :- has_suffix(W, ens, W1), has_suffix(W1, z).
name_possible_basename_de(W, W1) :- has_suffix(W, ens, W1), has_suffix(W1, 'ß').
name_possible_basename_de(W, W1) :- has_suffix(W, sches, W1).
name_possible_basename_de(W, W1) :- has_suffix(W, schen, W1).
name_possible_basename_de(W, W1) :- has_suffix(W, scher, W1).
name_possible_basename_de(W, W1) :- has_suffix(W, sche, W1).
name_possible_basename_de(W, W1) :- has_suffix(W, en, W1).
name_possible_basename_de(W, W1) :- has_suffix(W, s, W1), not_has_suffix(W1, s).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic exp_blocked_as_person_name/2.
:- dynamic exp_specified_person/2.

not_exp_blocked_as_person_name_in_context(W, CTX) :-
	\+ ( exp_blocked_as_person_name(W, CReq),
	     context_requirement_matches(CReq, CTX, _)
	   ).

exp_specified_person_general(Id, _CTX) :-
	exp_specified_person(Id, []),
	!.

exp_specified_person_in_context(Id, CTX, Arg) :-
	exp_specified_person(Id, CReq),
	CReq \= [],
	!,
	context_requirement_matches(CReq, CTX, Arg).

compile_person_assistance :-
	info(10, 'Compiling person assistance specification'),
	retract_supplemented_persons,
	retractall( exp_blocked_as_person_name(_, _) ),
	retractall( exp_specified_person(_, _) ),
	( assistance_entry(no_entity(person, CReq)),
	  ( select(word_in=Ws, CReq, CReq1),
	    member(W, Ws)
	  ; select(word=W, CReq, CReq1)
	  ),
	  assert( exp_blocked_as_person_name(W, CReq1) ),
	  fail
	; true
	),
	( assistance_entry(no_entity(person, CReq)),
	  ( select(prefix_in=Ws, CReq, CReq1),
	    member(W, Ws)
	  ; select(prefix=W, CReq, CReq1)
	  ),
	  assert( (exp_blocked_as_person_name(X, CReq1) :- sub_atom(X, 0, _, _, W)) ),
	  fail
	; true
	),
 	( assistance_entry(entity(person, Spec, CReq)),
 	  identify_person(Spec, Ids),
 	  ( Ids = [Id] ->
 	    assert( exp_specified_person(Id, CReq) )
	  ; Ids = [] ->
	    err('Void explicit person specification: ~q', [Spec])
	  ; err('Ambiguous explicit person specification: ~q (~q)',
		[Spec, Ids])
	  ),
	  fail
	; true
	),
	( assistance_entry(supplement(person, Spec, Sup)),
 	  identify_person(Spec, Ids),
 	  ( Ids = [Id] ->
	    process_supplemented_person(Id, Sup),
	    install_function_info(Id)
	  ; Ids = [] ->
	    err('Void explicit person specification: ~q', [Spec])
	  ; err('Ambiguous explicit person specification: ~q (~q)',
		[Spec, Ids])
	  ),
	  fail
	; true
	),
	( ext_person(Id, _),
	  install_function_info(Id),
	  fail
	; true
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_persons_vwr([P-pro(V0, R)|PRs]) :-
	!,
	show_persons_vwr([P-pro(V0, R)|PRs], V0).
show_persons_vwr([]) :-
	!.
show_persons_vwr(Ps) :-
	( member(P, Ps),
	  show_person_vw(P),
	  fail
	; true
	).
	
show_persons_vwr([P-pro(V1, R)|PRs], V0) :-
	( V1 @> V0 ->
	  write('..................................................'),
	  writeln('............................')
	; true
	),
	show_person_vwr(P-pro(V1, R)),
	show_persons_vwr(PRs, V1).
show_persons_vwr([], _).

show_person_vwr(P-pro(_, R)) :-
	!,
	show_person_vw(P),
	show_reasons(R).

show_person_vwr(P-_) :-
	!,
	show_person_vw(P).
show_person_vwr(P) :-
	show_person_vw(P).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% This should be done after loading initial person data, since evaluating the
%% assistance file involves disambiguation of person specifications.
%%
% :- add_assistance_update_hook(person_ner:compile_person_assistance).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% FUNCTION_PERSON
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

deref_word_to_function_person_id(W, CTX, Id, W0) :-
	ctx_get_defaulted(CTX, text_language, Lang, de),
	standardize_function_name(Lang, W, Function),
	ctx_get(CTX, elems_after, EsAfter),
	( es_word_prefix([W1, W2, W3], EsAfter),
	  function_junction(Lang, W1, W2),
	  function_index(W3, Function, Id),
	  W0 = [Function, W1, W2, W3]
	; es_word_prefix([W1, W2], EsAfter),
	  function_junction(Lang, W1),
	  function_index(W2, Function, Id),
	  W0 = [Function, W1, W2]
	; es_word_prefix([W1], EsAfter),
	  function_index(W1, Function, Id),
	  W0 = [Function, W1]
	).
	
pro_word_function_person_feature(X, Y, Z) :-
	pro_word_person_feature(X, Y, Z).

pro_id_function_person_feature(followed_by_matching_specifier(W), _, CTX) :-
	%% *** for now: the test is actually done in the deref, this here is
	%% just to move the words to the rationale features
	ctx_get(CTX, word, W).
pro_id_function_person_feature(X, Y, Z) :-
	pro_id_person_feature(X, Y, Z).

standardize_function_name(Lang, F, F) :-
	function_name(Lang, F),
	!.
standardize_function_name(Lang, F, F1) :-
	function_name_variant(Lang, F, F1),
	!.

not_blocked_word_function_person(W, CTX) :-
	ctx_get_defaulted(CTX, text_language, Lang, de),
	standardize_function_name(Lang, W, _).

not_blocked_id_function_person(ID, CTX) :-
	not_blocked_id_person(ID, CTX).

features_lock_until_for_function_person(Features, Until) :-
	memberchk(followed_by_matching_specifier(Ws), Features),
	last(Ws, W),
	Until = wrd(W).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stdize_person(P-_, P) :-
	!.
stdize_person(P, P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
