%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(prepare, [dep_ensure/1]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).

:- use_module(kbset_base(filename)).
:- use_module(swilib(graphs)).

children(call(Children, _), Children) :- !.
children(shell(Children, _), Children) :- !.
children(items(Children), Children) :- !.
children(gunzip(Child), [Child]) :- !.
children(unzip(Child), [Child]) :- !.
children(cat(Children), Children) :- !.
children(qcompile(Child), [Child]) :- !.
children(_, []).


dep_ensure(X) :-
	dep_ensure_strongly(X, '').


dep_ensure_weakly(Item, I, Value) :-
	info(10, '~w Ensuring ~w', [I, Item]),
	dep_file(Item, File),
	exists_file(File),
	!,
	time_file(File, ThisTime),
	dep(Dep),
	( memberchk(Item-Method, Dep) ->
	  true
	; err('No method to build ~w', [Item])
	),
	children(Method, Items),
	atom_concat(' ', I, I1),
	map_dep_ensure_weakly(Items, I1, Values),
	( member(Value, Values), Value > ThisTime ->
	  select_missing_items(Items, Values, MissingItems),
	  map_dep_ensure_strongly(MissingItems, I1),
	  info(10, '~w Building ~w', [I, Item]),
	  dep_build(Method, Item),
	  time_file(File, Value)
	; Value = ThisTime
	).
dep_ensure_weakly(Item, I, 0) :-
	info(10, '~w Missing  ~w', [I, Item]).

dep_ensure_strongly(Item, I) :-
	info(10, '~w Strongly ~w', [I, Item]),
	dep(Dep),
	( memberchk(Item-Method, Dep) ->
	  true
	; err('No method to build ~w', [Item])
	),
	atom_concat(' ', I, I1),
	children(Method, Items),
	map_dep_ensure_weakly(Items, I1, Values),
	select_missing_items(Items, Values, MissingItems),
 	map_dep_ensure_strongly(MissingItems, I1),
	info(10, '~w Building ~w', [I, Item]),
	dep_build(Method, Item).

select_missing_items([I|Is], [0|Vs], [I|Is1]) :-
	!,
	select_missing_items(Is, Vs, Is1).
select_missing_items([_|Is], [_|Vs], Is1) :-
	select_missing_items(Is, Vs, Is1).
select_missing_items([], [], []).

map_dep_ensure_weakly([X|Xs], I, [X1|Xs1]) :-
	dep_ensure_weakly(X, I, X1),
	map_dep_ensure_weakly(Xs, I, Xs1).
map_dep_ensure_weakly([], _, []).

map_dep_ensure_strongly([X|Xs], I) :-
	dep_ensure_strongly(X, I),
	map_dep_ensure_strongly(Xs, I).
map_dep_ensure_strongly([], _).


dep_build(Method, Item) :-
	catch( dep_build_1(Method, Item),
	       E,
	       ( dep_file(Item, F1),
		 ( exists_file(F1) ->
		   atom_concat(Item, '.interrupted', Item1),
		   dep_file(Item1, F2),
		   info(10, 'Renaming ~w to ~w', [F1, F2]),
		   rename_file(F1, F2)
		 ; true
		 ),
		 throw(E)
	       )).
		 
dep_build_1(qcompile(Child), _) :-
	!,
	dep_file(Child, File),
	info(10, 'Qcompiling ~w', [File]),
	qcompile(File).
dep_build_1(call(_Children, Call), _) :-
	!,
	call(Call).
dep_build_1(shell(_Children, Cmd), _) :-
	!,
	dep_dir(Dir),
	format(atom(Cmd1), 'cd ~w ; ~w', [Dir, Cmd]),
	info(10, 'Executing ~w', [Cmd1]),
	shell_wrapped(Cmd1).
dep_build_1(items(_), Parent) :-
	!,
	dep_file(Parent, File),
	format(atom(Cmd), 'echo >~w dummy', [File]),
	shell_wrapped(Cmd).
dep_build_1(gunzip(Child), Parent) :-
	!,
	dep_file(Child, File),
	dep_file(Parent, PFile),
	format(atom(Cmd), 'gzip -d -k ~w ; touch ~w', [File, PFile]),
	info(10, 'Extracting: ~w', [Cmd]),
	shell_wrapped(Cmd).
dep_build_1(unzip(Child), _Parent) :-
	!,
	dep_dir(Dir),
	dep_file(Child, File),
	format(atom(Cmd), 'unzip -o -B -DD -d ~w ~w', [Dir, File]),
	info(10, 'Extracting: ~w', [Cmd]),
	shell_wrapped(Cmd).
dep_build_1(cat(Children), Parent) :-
	!,
	dep_dir(Dir),
	dep_file(Parent, File),
	atomic_list_concat(Children, ' ', Children1),
	format(atom(Cmd), 'cd ~w; cat >~w ~w', [Dir, File, Children1]),
	info(10, 'Concatenating: ~w', [File]),
	shell_wrapped(Cmd).
dep_build_1(download(URL), Item) :-
	!,
	dep_file(Item, File),
	format(atom(Cmd), 'wget "~w" -O ~w', [URL,File]),
	dep_file('download.log', Log),
	format(atom(LogCmd), 'date >>~w ; echo >>~w ~w', [Log, Log, Cmd]),
	shell_wrapped(LogCmd),
	info(10, 'Dowloading ~w from ~w', [File, URL]),
	shell_wrapped(Cmd),
	format(atom(LogCmd2), 'ls -lt ~w >>~w ; md5sum ~w >>~w',
	       [File, Log, File, Log]),
	shell_wrapped(LogCmd2).
dep_build_1(Method, _) :-
	err('No implementation to process: ~w', [Method]).

shell_wrapped(Cmd) :-
	shell(Cmd, R),
	( R = 0 -> true
	; err('Shell command failed: ~w', [Cmd])
	).

dep_file(Item, File) :-
	filename(bigdata(Item), File).

dep_dir(Dir) :-
	filename(bigdata(''), Dir).

