%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(registry, [register_doc/3,
		     register_processed_doc/3,
		     registered_doc/1,
		     doc_es/2,
		     doc_processed_es/2,
		     doc_property/3]).

:- use_module(swilib(err)).
:- use_module(swilib(info)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_base(utils)).
	      
:- dynamic current_es/2.
:- dynamic current_processed_es/2.
:- dynamic current_es_property/3.

%%%% 
%%%% Returns latest registered as first solution
%%%% 
registered_doc(X) :-
	current_es(X, _).

doc_es(Doc, Es) :-
	( current_es(Doc, Es) ->
	  true
	; current_es(Doc1, Es) ->
	  info(10, 'WARNING: Using document: ~w instead of: ~w', [Doc1, Doc])
	; err('No registered document')
	).
	     
register_processed_doc(Src, _, Es) :-
	info(10, 'Registering processed ~w', [Src]),
	retractall(current_processed_es(Src, _)),
	asserta(current_processed_es(Src, Es)).

doc_processed_es(Doc, Es) :-
	( current_processed_es(Doc, Es) ->
	  true
	; current_processed_es(Doc1, Es) ->
	  info(10, 'WARNING: Using document: ~w instead of: ~w', [Doc1, Doc])
	; err('No registered processed document')
	).

register_doc(Src, _Options, Es) :-
	info(10, 'Registering ~w', [Src]),
	retractall(current_es(Src, _)),
	asserta(current_es(Src, Es)),
	retractall(current_es_property(Src, _, _)),
	ctx_get_defaulted(stemming_language, SLANG, de(19)),
	ctx_get_defaulted(text_language, LANG, de),
	asserta(current_es_property(Src, stemming_language, SLANG)),
	asserta(current_es_property(Src, text_language, LANG)).


doc_property(Src, Prop, Value) :-
	current_es_property(Src, Prop, Value),
	!.
doc_property(Src, Prop, Value) :-
	compute_doc_property(Prop, Src, Value),
	asserta( current_es_property(Src, Prop, Value) ).

compute_doc_property(global_stem_frequencies, Src, Value) :-
	!,
	current_es(Src, Es),
	doc_property(Src, stemming_language, SLANG),
	doc_property(Src, text_language, LANG),
	es_words_by_frequency(Es,
			      [output_frequency=true,
			       nostop=true,
			       text_language=LANG,
			       stems=true,
			       stemming_language=SLANG],
			      FWs),
	map_fw_to_wf(FWs, Value).

compute_doc_property(global_key_ngrams, Src, Value) :-
	!,
	current_es(Src, Es),
	doc_property(Src, stemming_language, SLANG),
	doc_property(Src, text_language, LANG),
	es_key_ngrams(Es, [output_rating=false,
			   text_language=LANG,
			   stemming_language=SLANG],
		      NGrams),
	list_prefix_upto(25, NGrams, Value).

compute_doc_property(Prop, _, _) :-
	err('No method for document property ~q', [Prop]).



fw_to_wf(F-(W-Unstems),
	 stem_frequency(W,F,Unstems)) :-
	!.
fw_to_wf(F-W,
	 stem_frequency(W,F,[])).

map_fw_to_wf([X|Xs], [X1|Xs1]) :-
	fw_to_wf(X, X1),
	map_fw_to_wf(Xs, Xs1).
map_fw_to_wf([], []).
	       
			      
			       