%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(render,
	  [render_latex_annotated/1,
	   render_plain/1,
	   render_rxml_annotated/1,
	   render_ner_style/1,
	   render_entity_facts/1,
	   render_dariah_csv/1]).

:- use_module(swilib(info)).
:- use_module(swilib(onto_atom)).
:- use_module(swilib(fromonto)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_base(gnd)).
:- use_module(kbset_ner(ner)).
:- use_module(kbset_ner(person_ner)).
:- use_module(kbset_ner(location_ner)).
:- use_module(kbset_ner(date_ner)).
:- use_module(kbset_ner(ner_base)).
:- use_module(kbset_ner(ner_utils)).
:- use_module(kbset_base(location)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(dates)).
:- use_module(kbset_base(geo)). % show_location_wv
:- use_module(kbset_base(utils)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(registry)).
:- use_module(kbset_ner(semctx)).
:- use_module(kbset_base(filename)).
:- use_module(kbset_ner(latexannot)).
:- use_module(kbset_ner(rxmlannot)).
:- use_module(kbset_base(localization)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_latex_annotated(Src) :-
	latex_target_file(Src, Target),
	render_latex_annotated(Src, Target).

render_latex_annotated(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	findall(A, ( annotated_by(Src1, AnnotSrc),
		     filename(Src1, Src2),
		     Src2 = Src,
		     file_to_abouts(AnnotSrc, About),
		     member(A, About)),
		As),
	time(es_validate_abouts(Es, As)),
	time(es_latannot_abouts(Es, As, Es1)),
	time(es_latannot_persons(Es1, Es2)),
	time(es_latannot_locations(Es2, Es2a)),
	time(es_latannot_dates(Es2a, Es3)),
	time(es_latannot_quotes(Es3, Es4)),
	time(es_latannot_keywords(Es4, Es5)),
	es_to_file(Es5, Target).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_rxml_annotated(Src) :-
	xml_target_file(Src, Target),
	render_rxml_annotated(Src, Target).

render_rxml_annotated(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	es_rxmlannot_persons(Es, Es1),
	es_rxmlannot_locations(Es1, Es2),
	es_rxmlannot_dates(Es2, Es3),
	es_to_file(Es3, Target).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_plain(Src) :-
	plain_target_file(Src, Target),
	render_plain(Src, Target).

render_plain(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	info(10, 'Writing ~w', [Target]),
	onto_file(( member(E, Es),
		    e_enum_plaintxt(E, T),
		    write(T),
		    fail
		  ; true
		  ),
		  Target).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_ner_style(Src) :-
	plain_target_file(Src, Target),
	render_ner_style(Src, Target).

render_ner_style(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	info(10, 'Writing ~w', [Target]),
	onto_file(( member(E, Es),
		    ( e_content(E, spc(Spc)) ->
		      ( sub_atom(Spc, _, _, _, '\n\n') ->
			format('~n')
		      ; true
		      )
		    ; ( e_content(E, wrd(_)) ; e_content(E, pun(_))) ->
		      ( e_identified_item(persons, E, _) ->
			Class = 'I-PER'
		      ; e_identified_item(locations, E, _) ->
			Class = 'I-LOC'
		      ; e_get(E, persons, _) ->
			Class = 'I-PER'
		      ; e_get(E, locations, _) ->
			Class = 'I-LOC'
		      ; Class = 'O'
		      ),
		      e_enum_plaintxt(E, T),
		      format('~w/~w ', [T, Class])
		    ),
		    fail
		  ; true
		  ),
		  Target).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_entity_facts(Src) :-
	prolog_target_file(Src, Target),
	render_entity_facts(Src, Target).

render_entity_facts(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	info(10, 'Writing ~w', [Target]),
	onto_file(( entity_fact(Es, Fact),
		    format('~q.~n', [Fact]),
		    fail
		  ; true
		  ),
		  Target).

entity_fact(Es, Fact) :-
	    member(E, Es),
	    ( e_identified_item_id(persons, E, P),
	      person_pretty(P, PPretty),
	      Fact = entity(person, P, PPretty)
	    ; e_identified_item_id(locations, E, L),
	      location_pretty(L, LPretty),
	      Fact = entity(location, L, LPretty)
	    ; e_get(E, date, begin(Date-_Rationale)),
	      Date \= ignore,
	      date_normal(Date, DPretty),
	      Fact = entity(date, Date, DPretty)
	    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

render_dariah_csv(Src) :-
	csv_target_file(Src, Target),
	render_dariah_csv(Src, Target).

render_dariah_csv(Src, Target) :-
	doc_processed_es(Src, Es),
	!,
	findall(Fact, entity_fact(Es, Fact), Facts),
	map_dariah_fact(Facts, nil, nil, DFacts),
	info(10, 'Writing ~w', [Target]),
	( format('Name,Address,Description,Longitude,Latitude,TimeStamp~n'),
	  ( member(d(N,A,S,LO,LA,T), DFacts), 
	    format('"~w","~w","~w","~w","~w","~w"~n',
		   [N,A,S,LO,LA,T]),
	    fail
	  ; true
	  )
	) onto_file Target.

map_dariah_fact([entity(date,D,_)|EFs], _, P, DFs) :-
	( D = ymd(_,_,_) ; D = ym(_) ; D = y(_) ),
	!,
	map_dariah_fact(EFs, D, P, DFs).
map_dariah_fact([entity(person, P, _)|EFs], D, _, DFs) :-
	!,
	map_dariah_fact(EFs, D, P, DFs).
map_dariah_fact([entity(location, L, _)|EFs], D, P, [DF|DFs]) :-
	!,
	dariah_fact(L, D, P, DF),
	map_dariah_fact(EFs, D, P, DFs).
map_dariah_fact([_|EFs], D, P, DFs) :-
	map_dariah_fact(EFs, D, P, DFs).
map_dariah_fact([], _, _, []).

dariah_localized(de, mention, 'Erwähnung') :- !.
dariah_localized(_, mention, 'Mention') :- !.
dariah_localized(de, empty_description, 'Keine Information') :- !.
dariah_localized(_, empty_description, 'No information') :- !.
dariah_localized(de, fmt_person_description, 'Zuletzt erwähnte Person: ~w') :- !.
dariah_localized(_, fmt_person_description, 'Last mentioned person: ~w') :- !.

dariah_quote(A, B) :-
	A0 = A,
        text_replace(A0, '<', '&lt;', A1),
	text_replace(A1, '>', '&gt;', A2),
	text_replace(A2, '"', '', A3),
	B = A3.

text_replace(A, O, N, B) :-
	sub_atom(A, N1, N2, _, O),
	!,
	sub_atom(A, 0, N1, _, A1),
	N4 is N1 + N2,
	sub_atom(A, N4, _, 0, A2),
	text_replace(A2, O, N, A3),
	atomic_list_concat([A1, N, A3], B).
text_replace(A, _, _, A).

dariah_fact(Location, LastDate, LastPerson,
	    d(Name,Address,Description,Long,Lati,TimeStamp)) :-
	user_language(Lang),
	dariah_localized(Lang, mention, Name),
	( LastDate = nil ->
	  get_time(X),
	  stamp_date_time(X, D, 'UTC'),
	  date_time_value(year, D, TimeStamp)
	; date_normal(LastDate, TimeStamp)
	),
	( LastPerson = nil ->
	  dariah_localized(Lang, empty_description, Description)
	; dariah_localized(Lang, fmt_person_description, FmtDescription),
	  person_pretty(LastPerson, PPretty),
	  dariah_quote(PPretty, PPretty1),
	  format(atom(Description), FmtDescription, [PPretty1])
	),
	location_pretty(Location, Address1),
	dariah_quote(Address1, Address),
	geo_latlon(Location, Lati-Long).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

plain_target_file(Src, FileName) :-
	target_file(Src, '.txt', FileName).

latex_target_file(Src, FileName) :-
	target_file(Src, '.tex', FileName).

xml_target_file(Src, FileName) :-
	target_file(Src, '.xml', FileName).

prolog_target_file(Src, FileName) :-
	target_file(Src, '.pl', FileName).

csv_target_file(Src, FileName) :-
	target_file(Src, '.csv', FileName).

target_file(Src, Extension, FileName) :-
	pure_file_base_name(Src, SimpleBase),
	!,
	atomic_list_concat([out_, SimpleBase, Extension], OutBase),
	filename(tmp(OutBase), FileName).
