%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2016 , 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Merging in annotations as XML elements.
%%%% 
%%%% XML is used here not in form of parsed structures, but on a
%%%% character/word level, where the markup (everything in "<>") is just
%%%% handled as "opaque", that is, skipped for operations like entity
%%%% identification. To distignuish this from a possible more substantial
%%%% future xml support we call it "Rough XML" or "rxml". See also
%%%% rxml_elems/2 in readwrite.pl.
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(rxmlannot,
	  [ es_rxmlannot_persons/2,
	    es_rxmlannot_locations/2,
	    es_rxmlannot_dates/2 ]).

:- use_module(swilib(fromonto)).
:- use_module(swilib(err)).
:- use_module(swilib(term_handling)).
:- use_module(swilib(info)).

% :- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(layers)).
:- use_module(kbset_ner(ner_utils)).
:- use_module(kbset_base(person)).
:- use_module(kbset_base(location)).
:- use_module(kbset_base(dates)).
	
es_rxmlannot_persons(Es, Es1) :-
	info(10, 'Merging with person information'),
	es_rxmlannot(Es, person_teiwrap, Es1).

es_rxmlannot_locations(Es, Es1) :-
	info(10, 'Merging with location information'),
	es_rxmlannot(Es, location_teiwrap, Es1).

es_rxmlannot_dates(Es, Es1) :-
	info(10, 'Merging with date information'),
	es_rxmlannot(Es, date_teiwrap, Es1).

es_rxmlannot([E|Es], Prop, Es1) :-
	e_rxmlannot(Prop, E, Open, Close),
	!,
	mk_e(opq(Open), [], EOpen),
	mk_e(opq(Close), [], EClose),
	Es1 = [EOpen, E, EClose|Es2],
	es_rxmlannot(Es, Prop, Es2).
es_rxmlannot([E|Es], Prop, [E|Es1]) :-
	es_rxmlannot(Es, Prop, Es1).
es_rxmlannot([], _, []).

e_rxmlannot(person_teiwrap, E, O, C) :-
	e_person_teiwrap(E, O, C).
e_rxmlannot(location_teiwrap, E, O, C) :-
	e_location_teiwrap(E, O, C).
e_rxmlannot(date_teiwrap, E, O, C) :-		
	e_date_teiwrap(E, O, C).

e_person_teiwrap(E, Open, Close) :-
	e_identified_item_id(persons, E, P1),
	!,
	( person_url(P1, URL) ->
	  xml_quote_cdata(URL, URL1),
	  format(atom(Open),
		 '<name type="person" resp="auto" key="~w" ref="~w">',
		 [P1, URL1])
	; format(atom(Open),
		 '<name type="person" resp="auto" key="~w"">',
		 [P1])
	),
	Close = '</name>'.
e_location_teiwrap(E, Open, Close) :-
	e_identified_item_id(locations, E, P1),
	!,
	( location_url(P1, URL) ->
	  xml_quote_cdata(URL, URL1),
	  format(atom(Open),
		 '<name type="place" resp="auto" key="geo:~w" ref="~w">',
		 [P1, URL1])
	; format(atom(Open),
		 '<name type="place" resp="auto" key="geo:~w">',
		 [P1])
	),
	Close = '</name>'.
e_date_teiwrap(E, Open, Close) :-
	!,
	e_get(E, date, begin(Date)),
	once(( Date = Date1-_ ; Date1 = Date )),
	Date1 \= ignore,
	date_w3c(Date1, Normal),
	Normal \= '?',
	format(atom(Open),
	       '<date resp="auto" when="~w">',
	       [Normal]),
	Close = '</date>'.
	


