%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(semctx, [mk_semctx/4,
		   semctx_get/3,
		   semctx_get_defaulted/4,
		   show_semctx/1,
		   make_semctx_at_point/3]).

:- use_module(kbset_ner(ctx)).
:- use_module(kbset_ner(esops)).
:- use_module(kbset_ner(fmtutils)).
:- use_module(kbset_ner(registry)).
:- use_module(kbset_ner(assistance)).
:- use_module(kbset_ner(keywords)).
:- use_module(swilib(err)).
:- use_module(kbset_base(utils)).

mk_semctx(Es, Props, CTX, SemCtx) :-
	map_mk_semctx_prop(Props, Es, CTX, SemCtx).

map_mk_semctx_prop([X|Xs], Y1, Y2, Xs1) :-
	( mk_semctx_prop(X, Y1, Y2, X1) ->
	  Xs1 = [X=X1|Xs2]
	; err('Invalid semantic contex property: ~q', [X])
	),
	map_mk_semctx_prop(Xs, Y1, Y2, Xs2).
map_mk_semctx_prop([], _, _, []).

semctx_get(SCTX, P, V) :-
	ctx_get(SCTX, P, V).

semctx_get_defaulted(SCTX, P, V, D) :-
	ctx_get_defaulted(SCTX, P, V, D).

mk_semctx_prop(words, Es, CTX, Words) :-
	!,
	ctx_get_defaulted(CTX, text_language, Lang, de),
	ctx_get_defaulted(CTX, semctx_words_minfreq, Freq, 0.002),
	es_words_by_frequency(Es, [nostop=true,
				   text_language=Lang,
				   minfreq=Freq], Words).

mk_semctx_prop(stems, Es, CTX, Words) :-
	!,
	ctx_get_defaulted(CTX, text_language, Lang, de),
	ctx_get_defaulted(CTX, stemming_language, SLang, de(19)),
	ctx_get_defaulted(CTX, semctx_stems_minfreq, Freq, 0.002),
	es_words_by_frequency(Es, [nostop=true,
				   text_language=Lang,
				   stemming_language=SLang,
				   stems=true,
				   minfreq=Freq], Words).

mk_semctx_prop(weighed_stems, Es, CTX, Words) :-
	!,
	( registered_doc(Doc) ->
	  doc_property(Doc, global_stem_frequencies, GF),
	  ctx_get_defaulted(CTX, text_language, Lang, de),
	  ctx_get_defaulted(CTX, stemming_language, SLang, de(19)),
	  es_words_by_frequency(Es, [nostop=true,
				     text_language=Lang,
				     stemming_language=SLang,
				     stems=true,
				     relate_to_global_frequencies=GF,
%				     output_frequency=true,
				     minfreq=0.00001
				     ], Words)
	; err('No registered document')
	).

mk_semctx_prop(key_ngrams, Es, CTX, PrettyStemPairs) :-
	!,
	ctx_get_defaulted(CTX, text_language, Lang, de),
	ctx_get_defaulted(CTX, stemming_language, SLang, de(19)),
	ctx_get_defaulted(CTX, semctx_key_ngrams_ngram_max_n, MaxN, 3),
	ctx_get_defaulted(CTX, semctx_key_ngrams_ngram_min_f, MinF, 4),
	ctx_get_defaulted(CTX, semctx_key_ngrams_freq_percent, Freq, 30),
	ctx_get_defaulted(CTX, semctx_key_ngrams_maxnum, Num, 10),
	Options = [output_rating=false,
		   text_language=Lang,
		   stemming_language=SLang,
		   ngram_max_n=MaxN,
		   ngram_min_f=MinF,
		   freq_percent=Freq],
	es_key_ngrams(Es, Options, NGrams1),
	list_prefix_upto(Num, NGrams1, NGrams2),
	map_ngram_add_unstemmed(NGrams2, PrettyStemPairs).
	

map_ngram_add_unstemmed([X|Xs], [Y-X|Xs1]) :-
	( X = _Rating-S -> true ; S = X ),
	kwdata_unstem(S, Y),
	map_ngram_add_unstemmed(Xs, Xs1).
map_ngram_add_unstemmed([], []).

make_semctx_at_point(Es, Point, SCTX) :-
	main_context(CTX),
	ctx_get_defaulted(CTX, semctx_width_before, WBef, 3000),
	ctx_get_defaulted(CTX, semctx_width_after, WAft, 1000),
	es_split_at_srcpos(Es, Point, RevBefore, After),
	list_prefix_upto(WBef, RevBefore, Bef),
	list_prefix_upto(WAft, After, Aft),
	reverse_with_rest(Bef, Es1, Aft),
	mk_semctx(Es1, [key_ngrams,weighed_stems,stems], CTX, SCTX).
%	mk_semctx(Es1, [key_ngrams,words], CTX, SCTX).

show_semctx(SCTX) :-
	format_header_1([en-'Semantic Context',
			 de-'Semantischer Kontext']),
	( member(key_ngrams=PrettyStemPairs, SCTX) ->
	  format_header_2([en-'Key-NGrams and Stems',
			   de-'Schlüssel-NGrams und Stammformen']),
	  ( member(W-S, PrettyStemPairs),
	    format('~w ~t~60|~w~n', [W, S]),
	    fail
	  ; true
	  )
	; true
	),
	( member(weighed_stems=WStems, SCTX) ->
	  format_header_2([en-'Frequent Stems (Weighed)',
			   de-'Häufige Stammwörter (gewichtet)']),
	  ( member(X, WStems),
	    ( X=(F-(S-Ws)) ->
	      format('~6f ~@ ~t~60|~w~n',
		     [F, write_members_separated_by_blanks(Ws), S])
	    ; X = (S-Ws),
	      format('~@ ~t~60|~w~n', [write_members_separated_by_blanks(Ws), S])
	    ),
	    fail
	  ; true
	  )
	; true
	),
	( member(stems=Stems, SCTX) ->
	  format_header_2([en-'Frequent Stems',
			   de-'Häufige Stammwörter']),
	  ( member(X, Stems),
	    ( X=(F-(S-Ws)) ->
	      format('~6f ~@ ~t~60|~w~n',
		     [F, write_members_separated_by_blanks(Ws), S])
	    ; X = (S-Ws),
	      format('~@ ~t~60|~w~n', [write_members_separated_by_blanks(Ws), S])
	    ),
	    fail
	  ; true
	  )
	; true
	),
	( member(words=Words, SCTX) ->
	  format_header_2([en-'Frequent Words',
			   de-'Häufige Wörter']),
	  ( member(W, Words),
	    format('~w~n', [W]),
	    fail
	  ; true
	  )
	; true
	).
% 
% persons
% dates
% locations
% by frequency ?
% sub command show persons...

