%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(stemming, [stem_de_20/2,
		     stem_de_191/2,
		     stem_de_19/2,
		     stem_de_18/2]).


:- use_module(swilib(fromonto)).
:- use_module(kbset_base(textutils)).

:- encoding(utf8).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% STEMMING
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% "Porter stemming algorithm" for De, follows
%%%% http://snowball.tartarus.org/texts/germanic.html, retrieved June 2014
%%%% 
%%%% Note: Examples that seem not satisfatorily handled by the algoritum
%%%% fallend, sprechend, rasiert, spielte
%%%%
%%%% Examples not satisfactorily handled by version 2: aktuell
%%%%
%%%% These are truncated: auffallend, rasieren

%%%% 
%%%% Yields results identical to the diff.txt table supplied
%%%% at http://snowball.tartarus.org/texts/germanic.html
%%%% 
stem_de_20(W, W1) :-
	prepare(W, W0),
	w_p1_p2(W0, P1, P2),
%	w_r1_r2(W0, R1, R2), format('R1: ~w R2: ~w~n', [R1, R2]),
	step_1(W0, P1, W2),
	step_2(W2, P1, W3),
	step_3(W3, P1, P2, W4),
	cleanup(W4, W1).

%%%% 
%%%% Version 2 of the algorithm, might be better suitled for the
%%%% 19th century texts which often write e.g. Ue for Ü
%%%% 
stem_de_191(W, W1) :-
	prepare_v2(W, W0),
	w_p1_p2(W0, P1, P2),
	%% w_r1_r2(W0, R1, R2), format('R12: ~w ~w~n', [R1, R2]),
	step_1(W0, P1, W2),
	step_2(W2, P1, W3),
	step_3(W3, P1, P2, W4),
	cleanup(W4, W1).

%%%% 
%%%% Version 3: Calls version 2 after preprocessing to handle
%%%% 19th century ways of writing
%%%%
stem_de_19(W, W1) :-
	prepare_19(W, W2),
	stem_de_191(W2, W1).

%%%% 
%%%% For 18th century ways of writing
%%%%
stem_de_18(W, W1) :-
	prepare_18(W, W2),
	stem_de_191(W2, W1).
	
vowel(a).
vowel(e).
vowel(i).
vowel(o).
vowel(u).
vowel(y).
vowel('ä').
vowel('ö').
vowel('ü').

prepare(W, W1) :-
	atom_chars(W, C0),
	downcase_chars(C0, C1),
	replace_sz_1(C1, C2),
	replace_uy_1(C2, C3),
	atom_chars(W1, C3).

prepare_v2(W, W1) :-
	atom_chars(W, C0),
	downcase_chars(C0, C1),
	replace_uy_1(C1, C2),
	replace_sz_v2(C2, C3),
	atom_chars(W1, C3).

cleanup(W, W1) :-
	atom_chars(W, C1),
	cleanup_1(C1, C2),
	atom_chars(W1, C2).

cleanup_1(['ä'|Cs], [a|Cs1]) :-	!, cleanup_1(Cs, Cs1).
cleanup_1(['ö'|Cs], [o|Cs1]) :- !, cleanup_1(Cs, Cs1).
cleanup_1(['ü'|Cs], [u|Cs1]) :-	!, cleanup_1(Cs, Cs1).
cleanup_1(['Y'|Cs], [y|Cs1]) :-	!, cleanup_1(Cs, Cs1).
cleanup_1(['U'|Cs], [u|Cs1]) :-	!, cleanup_1(Cs, Cs1).
cleanup_1([C|Cs], [C|Cs1]) :- cleanup_1(Cs, Cs1).
cleanup_1([], []).

replace_uy_1([C, u, D|Cs], [C|Cs1]) :-
	vowel(C),
	vowel(D),
	!,
	replace_uy_1(['U',D|Cs], Cs1).
replace_uy_1([C, y, D|Cs], [C|Cs1]) :-
	vowel(C),
	vowel(D),
	!,
	replace_uy_1(['Y',D|Cs], Cs1).
replace_uy_1([C|Cs], [C|Cs1]) :-
	replace_uy_1(Cs, Cs1).
replace_uy_1([], []).

replace_sz_1(['ß'|Cs], [s,s|Cs1]) :-
	!,
	replace_sz_1(Cs, Cs1).
replace_sz_1([C|Cs], [C|Cs1]) :-
	replace_sz_1(Cs, Cs1).
replace_sz_1([], []).

replace_sz_v2(Cs, Cs1) :-
	replace_sz_v2(Cs, '', Cs1).
replace_sz_v2([a,e|Cs], _, ['ä'|Cs1]) :-
	!,
	replace_sz_v2(Cs, 'ä', Cs1).
replace_sz_v2([o,e|Cs], _, ['ö'|Cs1]) :-
	!,
	replace_sz_v2(Cs, 'ö', Cs1).
replace_sz_v2([u,e|Cs], C, ['ü'|Cs1]) :-
	C \= q,
	!,
	replace_sz_v2(Cs, 'ü', Cs1).
replace_sz_v2(['ß'|Cs], _, [s,s|Cs1]) :-
	!,
	replace_sz_v2(Cs, s, Cs1).
replace_sz_v2([C|Cs], _, [C|Cs1]) :-
	replace_sz_v2(Cs, C, Cs1).
replace_sz_v2([], _, []).

w_r1_r2(W, R1, R2) :-
	LIMIT = 10000,
	w_p1_p2(W, P1, P2),
	( P1 = LIMIT -> R1 = ''
	; sub_atom(W, P1, _, 0, R1)
	),
	( P2 = LIMIT -> R2 = ''
	; sub_atom(W, P2, _, 0, R2)
	).


w_p1_p2(W, P1, P2) :-
	LIMIT = 10000,
	( first_vowel(W, 0, N1), first_consonant(W, N1, O1) ->
	  P1 is max(O1+1, 3),
	  ( first_vowel(W, O1, N2),
	    first_consonant(W, N2, O2) ->
	    P2 is O2+1
	  ; P2 = LIMIT
	  )
	; P1 = LIMIT,
	  P2 = LIMIT
	).

first_consonant(W, N, N1) :-
	sub_atom(W, N, 1, _, C),
	( vowel(C) ->
	  N2 is N+1,
	  first_consonant(W, N2, N1)
	; N1 = N
	).

first_vowel(W, N, N1) :-
	sub_atom(W, N, 1, _, C),
	( vowel(C) ->
	  N1 = N
	; N2 is N+1,
	  first_vowel(W, N2, N1)
	).

step_1(W, P1, W1) :-
	suffix_1a(E),
	sub_atom(W, N, L, 0, E),
	N >= P1,
	!,
	sub_atom(W, 0, _, L, W1).
step_1(W, P1, W1) :-
	suffix_1b(E),
	sub_atom(W, N, L, 0, E),
	N >= P1,
	!,
	sub_atom(W, 0, _, L, W2),
	( sub_atom(W2, _, _, 0, niss) ->
	  sub_atom(W2, 0, _, 1, W1)
	; W1 = W2
	).
step_1(W, P1, W1) :-
	sub_atom(W, N, 1, 0, s),
	N >= P1,
	sub_atom(W, _, 1, 1, C),
	s_ending(C),
	!,
	sub_atom(W, 0, _, 1, W1).
step_1(W, _, W).


step_2(W, P1, W1) :-
	suffix_2a(E),
	sub_atom(W, N, L, 0, E),
	N >= P1,
	!,
	sub_atom(W, 0, _, L, W1).
step_2(W, P1, W1) :-
	sub_atom(W, N, 2, 0, st),
	N >= P1,
	sub_atom(W, S, 1, 2, C),
	st_ending(C),
	S >= 3,
	!,
	sub_atom(W, 0, _, 2, W1).
step_2(W, _, W).

step_3(W, _, P2, W1) :-
	suffix_3a(E),
	sub_atom(W, N, L, 0, E),
	N >= P2,
	!,
	sub_atom(W, 0, _, L, W2),
	( sub_atom(W2, N1, _, 0, ig), N1 >= P2, \+ sub_atom(W2, _, 1, 2, e) ->
	  sub_atom(W2, 0, _, 2, W1)
	; W1 = W2
	).
step_3(W, _, P2, W1) :-
	suffix_3b(E),
	sub_atom(W, N, L, 0, E),
	N >= P2,
	\+ sub_atom(W, _, 1, L, e),
	!,
	sub_atom(W, 0, _, L, W1).
step_3(W, P1, P2, W1) :-
	suffix_3c(E),
	sub_atom(W, N, L, 0, E),
	N >= P2,
	!,
	sub_atom(W, 0, _, L, W2),
	( sub_atom(W2, N1, _, 0, er), N1 >= P1 ->
	  sub_atom(W2, 0, _, 2, W1)
	; sub_atom(W2, N1, _, 0, en), N1 >= P1 ->
	  sub_atom(W2, 0, _, 2, W1)
	; W1 = W2
	).
step_3(W, _, P2, W1) :-
	suffix_3d(E),
	sub_atom(W, N, L, 0, E),
	N >= P2,
	!,
	sub_atom(W, 0, _, L, W2),
	( sub_atom(W2, N1, _, 0, lich), N1 >= P2 ->
	  sub_atom(W2, 0, _, 4, W1)
	; sub_atom(W2, N1, _, 0, ig), N1 >= P2 ->
	  sub_atom(W2, 0, _, 2, W1)
	; W1 = W2
	).
step_3(W, _, _, W).

suffix_1a(em).
suffix_1a(ern).
suffix_1a(er).

suffix_1b(e).
suffix_1b(en).
suffix_1b(es).

suffix_2a(en).
suffix_2a(er).
suffix_2a(est).

suffix_3a(end).
suffix_3a(ung).

suffix_3b(ig).
suffix_3b(ik).
suffix_3b(isch).

suffix_3c(lich).
suffix_3c(heit).

suffix_3d(keit).

s_ending(C) :- st_ending(C).
s_ending(r).

st_ending(b).
st_ending(d).
st_ending(f).
st_ending(g).
st_ending(h).
st_ending(k).
st_ending(l).
st_ending(m).
st_ending(n).
st_ending(t).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prepare_18(X, Y) :-
	prepare_19(X, Y).

prepare_19(X, Y) :-
	downcase_atom(X, X1),
	( map_v3(X1, Y) ->
	  true
	; Y = X1
	).

map_v3(verschiedne, verschiedene).
map_v3(verschiedner, verschiedener).
map_v3(verschiednen, verschiedenen).
map_v3(unsre, unsere).
map_v3(unsren, unseren).
map_v3(unsres, unseres).


/* ? oh/o t/th k/ck i/y ie/i
*/