;;;; ==========================================================================
;;;;
;;;; Copyright (C) 2014, 2015, 2016, 2020 Christoph Wernhard
;;;;
;;;; This file is part of KBSET.
;;;;
;;;; KBSET is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; KBSET is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; ==========================================================================

;;;; ==========================================================================
;;;;
;;;; Set these variables according to your installation and requirements. Can
;;;; be done without modifying this file by setting these variables before
;;;; loading it. E.g.  by statements like this in your .emacs file:
;;;;
;;;;    (setq *kbset-tmpdir* "/home/ch/tmp_kbset")
;;;;    (load "/home/ch/provers/kbset/elisp/kbset.el")
;;;;
(defvar *kbset-tmpdir* "/home/ch/tmp"
  "Directory for temporary files created and read by Kbset")

(defvar *kbset-high-timeout* 40
  "Timeout in seconds for Emacs when checking for text
highlighting markings produced by the invocation of a Prolog call
from Emacs. After this timeout, markings can always be loaded
with menu Refresh Markings or function kbset-high. For smaller
books 20 is recommendend, for books with many referenced entities
values of around 60 should be sufficient.")

;;;; ==========================================================================

(defvar kbset-high-temp-results-file 
  (concat *kbset-tmpdir* "/kbset_temp_results.txt")
  "File for temporary Kbset results.")

(defun kbset-high-spec-file (srcfile) 
  (concat *kbset-tmpdir* 
	  "/kbset_high_"
	  (file-name-sans-extension (file-name-nondirectory srcfile))
	  ".el"))

;;;; ==========================================================================

(defun kbset-info ()
  (interactive)
  (let ((kbset-info-buffer-name
	(cond ((equal (buffer-name) "*kbset-info*") "*kbset-info2*")
	      ((equal (buffer-name) "*kbset-info2*") "*kbset-info3*")
	      (t "*kbset-info*"))))
    (let ((info (kbset-info-at-point (point))))
      (save-excursion
	(get-buffer-create kbset-info-buffer-name)
	(set-buffer kbset-info-buffer-name)
	(erase-buffer)
	(insert info)
	(goto-char (point-min))
	(display-buffer kbset-info-buffer-name t)))))

(defun kbset-info-on ()
  (interactive)
  (add-hook 'post-command-hook 'kbset-info))

(defun kbset-info-off ()
  (interactive)
  (remove-hook 'post-command-hook 'kbset-info))

;;;; ==========================================================================

(defun kbset-browse-url-at-point ()
  (interactive)
  (let ((at-url
	 (save-excursion
	   (goto-char (line-beginning-position))
	   (search-forward "://" (line-end-position) t))))
    (cond (at-url
	   (browse-url-at-point))
	  ((get-buffer "*kbset-info*")
	   (save-excursion
	     (set-buffer "*kbset-info*")
	     (goto-char (point-min))
	     (search-forward "://" (point-max) t)
	     (browse-url-at-point))))))

;;;; ==========================================================================

(defun setup-kbset-map (kbset-map kbset-render-map kbset-summary-map)

  (define-key kbset-map (kbd "C-.") 'kbset-goto-next)
  (define-key kbset-map (kbd "M-.") 'kbset-goto-previous)
  (define-key kbset-map (kbd "C-/") 'kbset-browse-url-at-point)
  (define-key kbset-map (kbd "C-9") 'kbset-search-quotemark-forward)
  (define-key kbset-map (kbd "M-9") 'kbset-search-quotemark-backward)
    
  (define-key kbset-map [menu-bar kbset render]
    (cons "Render..." kbset-render-map))
  
  (define-key kbset-render-map
    [m-kbset-high-render-ner-style]
    '("Render in NER style" . kbset-render-ner-style))

  (define-key kbset-render-map
    [m-kbset-high-render-dariah-csv]
    '("Render Dariah CSV" . kbset-render-dariah-csv))

  (define-key kbset-render-map
    [m-kbset-high-render-entity-facts]
    '("Render entity/3 Prolog facts" . kbset-render-entity-facts))

  (define-key kbset-render-map
    [m-kbset-high-render-rxml]
    '("Render annotated as XML" . kbset-render-rxml))
  
  (define-key kbset-render-map
    [m-kbset-high-render-plain]
    '("Render as plain text" . kbset-render-plain))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-render-latex-annotated]
    '("Render annotated as LaTeX" . kbset-render-latex-annotated))

  (define-key kbset-map
    [menu-bar kbset separator-render]
    '(menu-item "--"))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-clear]
    '("Clear display: Markings" . kbset-high-clear))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high]
    '("Refresh display: Markings" . kbset-high))
  
  (define-key kbset-map
    [menu-bar kbset m-kbset-high-open-tempresults]
    '("Refresh display: Last result" . kbset-open-tempresults))
  
  (define-key kbset-map
    [menu-bar kbset separator-display]
    '(menu-item "--"))

  (define-key kbset-map [menu-bar kbset summary]
    (cons "Summary..." kbset-summary-map))

  (define-key kbset-summary-map
    [m-kbset-high-person-summary-alphabetical-brief]
    '("Summary: Persons - alphabetical, brief" . 
      kbset-call-person-summary-alphabetical-brief))

  (define-key kbset-summary-map
    [m-kbset-high-person-summary-alphabetical]
    '("Summary: Persons - alphabetical" . 
      kbset-call-person-summary-alphabetical))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-date-summary]
    '("Summary: Dates" . kbset-call-date-summary))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-location-summary]
    '("Summary: Locations" . kbset-call-location-summary))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-person-summary]
    '("Summary: Persons" . kbset-call-person-summary))

  (define-key kbset-map
    [menu-bar kbset separator-summary]
    '(menu-item "--"))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-semantic-context]
    '("At point: Show semantic context" . kbset-call-semantic-context))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-structure-context]
    '("At point: Show structural context" . kbset-call-structure-context))
  
  (define-key kbset-map
    [menu-bar kbset m-kbset-high-inspect-srcpos]
    '("At point: Show NER candidates" . kbset-call-inspect-srcpos))

  (define-key kbset-map
    [menu-bar kbset separator-move]
    '(menu-item "--"))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-open-url]
    '("Browse URL" . kbset-browse-url-at-point))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-previous]
    '("Goto: Previous recognized item" . kbset-goto-previous))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-next]
    '("Goto: Next recognized item" . kbset-goto-next))

  (define-key kbset-map
    [menu-bar kbset separator-process]
    '(menu-item "--"))

  (define-key kbset-map
    [menu-bar kbset m-kbset-high-process]
    '("Process source" . kbset-process-file))
)

;;;; ==========================================================================

(defvar kbset-high-table '())

(defvar kbset-high-starts '())

(defun load-kbset-high-file ()
  (load-file (kbset-high-spec-file (buffer-file-name)))
  (setq kbset-high-starts (sort (mapcar #'car kbset-high-table) #'<=)))

(defun kbset-goto-next ()
  "Go to next highlighted point and show info about it"
  (interactive)
  (goto-char (kbset-find-next (point) kbset-high-starts))
  (kbset-info))

(defun kbset-goto-previous ()
  "Go to next highlighted point and show info about it"
  (interactive)
  (goto-char (kbset-find-previous (point) (reverse kbset-high-starts)))
  (kbset-info))

(defun kbset-find-next (point points)
  (while (and points (>= point (car points)))
    (setq points (cdr points)))
  (or (car points) (beep) point))
  
(defun kbset-find-previous (point points)
  (while (and points (<= point (car points)))
    (setq points (cdr points)))
  (or (car points) (beep) point))

(defun kbset-info-at-point (pos)
  (kbset-overlays-info (overlays-at pos)))

(defun kbset-overlays-info (overlays)
  (concat (kbset-overlays-info-1 overlays '(date))
	  (kbset-overlays-info-1 overlays '(persons))
	  (kbset-overlays-info-1 overlays '(locations))
	  (kbset-overlays-info-1 overlays '(function_name))))

(defun kbset-overlays-info-1 (overlays properties)
  (apply 'concat
	 (mapcar #'(lambda (overlay)
		     (let ((prop (overlay-get overlay 'kbset-property)))
		       (cond ((member prop properties)
			      (kbset-info-string 
			       (overlay-get overlay 'kbset-info)))
			     (t ""))))
		 overlays)))

(defun kbset-info-string (kbset-info)
  (cond ((stringp kbset-info) kbset-info)
	((null kbset-info) "")
	(t (car kbset-info))))

(defface kbset-face-high-blue
  '((((class color)) :foreground "black" :background "LightSkyBlue")
    (t :inverse-video t)) "")

(defface kbset-face-high-red
  '((((class color)) :foreground "black" :background "pink1")
    (t :inverse-video t)) "")

(defface kbset-face-high-green
  '((((class color)) :foreground "black" :background "pale green")
    (t :inverse-video t)) "")

(defface kbset-face-high-orange
  '((((class color)) :foreground "black" :background "khaki")
    (t :inverse-video t)) "")

; (setq hlt-use-overlays-flag t) this is default

(defun echoose-face (key)
  (cond ((eq key 'date) 'kbset-face-high-red)
	((eq key 'persons) 'kbset-face-high-blue)
	((eq key 'locations) 'kbset-face-high-green)
	(t 'kbset-face-high-orange)))

(defun kbset-high ()
  "Load file with highlighting table and highlight accordingly."
  (interactive)
  (load-kbset-high-file)
  (kbset-high-redo))

(defun kbset-high-clear ()
  "Remove highlighting."
  (interactive)
  (save-excursion
    (remove-overlays (point-min) (point-max) 'kbset-type 'kbset-high)))

(defun kbset-high-redo ()
  "Highlight according to current highlighting table."
  (interactive)
  (kbset-high-clear)
  (dolist (row kbset-high-table)
    (let ((overlay (make-overlay (nth 0 row) (nth 1 row))))
      ;; predefined attributes:
      (overlay-put overlay 'face (echoose-face (nth 2 row)))
      (overlay-put overlay 'help-echo (nth 3 row))
      ;; own attributes
      (overlay-put overlay 'kbset-type 'kbset-high)
      (overlay-put overlay 'kbset-property (nth 2 row))
      (overlay-put overlay 'kbset-info (nth 3 row)))))

;;;; ======================================================================

(defun prolog-call (stmt)
  (prolog-ensure-process)
  (prolog-record-call stmt)
  (save-excursion
    (set-buffer "*prolog*")
    (goto-char (point-max))
    (display-buffer "*prolog*" t))
  (save-excursion
    (let* ((process (get-process "prolog")))
      (process-send-string "prolog" (concat stmt "\n"))))
  (save-excursion
    (set-buffer "*prolog*")
    (goto-char (point-max))
    (display-buffer "*prolog*" t))
  (princ (concat "Prolog call: " stmt)))

(defun prolog-call-quiet (stmt)
  (prolog-ensure-process)
  (prolog-record-call stmt)
  (save-excursion
    (set-buffer "*prolog*")
    (goto-char (point-max))
    (let* ((process (get-process "prolog")))
      (process-send-string "prolog" (concat stmt "\n")))))

(defun sit-for-quiet (seconds &optional ignore)
;  (sit-for seconds)
  (sleep-for seconds))
;  (read-event nil nil seconds))

(defun prolog-call-open-tempresults (stmt &optional firstsit timeout lastsit)
  (prolog-call-quiet stmt)
  (save-excursion
    (set-buffer "*prolog*")
    (sit-for-quiet (or firstsit 0.4))
    (let ((wait-step 0.2)
	  (wait-total (or timeout 20))
	  (wait t))
      (while (and wait (> wait-total 0))
	(sit-for-quiet wait-step)
	(setq wait-total (- wait-total wait-step))
	(cond ((equal (buffer-substring (- (point-max) 3) (point-max)) "?- ")
	       (setq wait nil))))))
  (sit-for-quiet (or lastsit 0.0))
  (kbset-open-tempresults))

(defun prolog-call-and-kbset-high (stmt &optional timeout)
  (prolog-call stmt)
  (save-excursion
    (set-buffer "*prolog*")
    ; (sleep-for 0.5)
    (sit-for 0.5 t)
    ; (sleep-for 0.4)
    (let ((wait-step 0.1)
	  (wait-total (or timeout *kbset-high-timeout*))
	  (wait t))
      (while (and wait (> wait-total 0))
	; (sleep-for wait-step)
	(sit-for wait-step t)
	(setq wait-total (- wait-total wait-step))
	(cond ((equal (buffer-substring (- (point-max) 3) (point-max)) "?- ")
	       (setq wait nil))))))
  (kbset-high))

(defun prolog-call-dontshow (stmt)
  (prolog-ensure-process)
  (prolog-record-call stmt)
  (save-excursion
    (let* ((process (get-process "prolog")))
      (process-send-string "prolog" (concat stmt "\n")))))

(defun kbset-process-file ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "process_source('" (buffer-file-name) "').")))
      (prolog-call-and-kbset-high stmt))))

(defun kbset-render-latex-annotated ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_latex_annotated('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-render-plain ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_plain('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-render-rxml ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_rxml_annotated('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-render-ner-style ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_ner_style('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-render-dariah-csv ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_dariah_csv('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-render-entity-facts ()
  (interactive)
  (save-excursion
    (let ((stmt (concat "render_entity_facts('" (buffer-file-name) "').")))
      (prolog-call stmt))))

(defun kbset-call-person-summary ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_person_summary('" 
	       (buffer-file-name) "'), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-person-summary-alphabetical ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_person_summary_alphabetical('" 
	       (buffer-file-name) "'), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-person-summary-alphabetical-brief ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_person_summary_alphabetical_brief('" 
	       (buffer-file-name) "'), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-location-summary ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_location_summary('" (buffer-file-name) "'), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-date-summary ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_date_summary('" 
	       (buffer-file-name) "'), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-inspect-srcpos ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(inspect_item_at_srcpos('" 
	       (buffer-file-name)
	       "', "
	       (number-to-string (point))
	       "), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt 0.5 10 1.0)))

(defun kbset-call-semantic-context ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_semantic_context_at_point('" 
	       (buffer-file-name)
	       "', "
	       (number-to-string (point))
	       "), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-call-structure-context ()
  (interactive)
  (display-buffer "*prolog*" t)
  (let ((stmt (concat 
	       "onto_file(show_structural_context_at_point('" 
	       (buffer-file-name)
	       "', "
	       (number-to-string (point))
	       "), '"
	       kbset-high-temp-results-file
	       "').")))
    (prolog-call-open-tempresults stmt)))

(defun kbset-open-tempresults ()
  (interactive)
  (let ((buf (get-buffer "*kbset-result*")))
    (cond (buf 
	   (unless (equal (current-buffer) buf)
	     (set-buffer buf)
	     (display-buffer "*kbset-result*" t)))
	  (t
	   (find-file-other-window kbset-high-temp-results-file)
	   (rename-buffer "*kbset-result*"))))
  (revert-buffer nil t)
  (toggle-read-only 0)
  (goto-char (point-min))
  (let ((view-read-only t))
    (toggle-read-only 1)))

(defun prolog-record-call (stmt)
  (save-excursion
    (set-buffer "*prolog*")
    (ring-insert comint-input-ring stmt)))

;;;; ==========================================================================

(defun kbset-search-quotemark-forward ()
  (interactive)
  (search-forward-regexp "\'\'\\|,,\\|``"))

(defun kbset-search-quotemark-backward ()
  (interactive)
  (search-backward-regexp "\'\'\\|,,\\|``"))

;;;; ==========================================================================

(defvar kbset-map (make-keymap) "Kbset keymap.")

(defvar kbset-render-map (make-keymap) "Kbset render submenu keymap.")

(defvar kbset-summary-map (make-keymap) "Kbset summary submenu keymap.")

(define-minor-mode kbset-minor-mode
  "Minor mode for Kbset functionality."
  nil " Kbset" kbset-map)

(add-hook 'kbset-minor-mode-hook
	  ;; expressed as hook because otherwise the "Kbset" menu
	  ;; would appear initially only as status bar menu.
    	  (lambda () 
	    (define-key-after kbset-map [menu-bar kbset]
	      (cons "Kbset" (make-sparse-keymap "kbset"))
	      'help-menu)
	    (setup-kbset-map kbset-map 
			     kbset-render-map
			     kbset-summary-map)))

