%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%% Copyright (C) 2014, 2015, 2019, 2021 Christoph Wernhard
%%%%
%%%% This file is part of KBSET.
%%%%
%%%% KBSET is free software: you can redistribute it and/or modify
%%%% it under the terms of the GNU General Public License as published by
%%%% the Free Software Foundation, either version 3 of the License, or
%%%% (at your option) any later version.
%%%% 
%%%% KBSET is distributed in the hope that it will be useful,
%%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%%% GNU General Public License for more details.
%%%% 
%%%% You should have received a copy of the GNU General Public License
%%%% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Example assistance file for Max Stirner, Geschichte der Reaction, Vol 1
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(kbset_ner(ner_kbset)).
:- initialization(update_assistance).

:- set_verbosity(50).

:- set_docdir(load_context).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Specifying Paths to Auxiliary Directories and File Locations
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- register_file_path(tmp(_), '~/tmp/').
:- register_file_path(bigdata(_), '~/space/bigdata_stirner/').

tmp_file(kbset_high_file_pattern, 'kbset_high_~w.el').

bigdata_file(gnd_source_rdf_person, 'authorities-person_lds_20201013.rdf.gz').
bigdata_file(gnd_source_rdf_subject, 'authorities-sachbegriff_lds_20201013.rdf.gz').
bigdata_file(gnd_persons_bornbefore_pattern, 'gnd_persons_bornbefore_~w').
bigdata_file(gnd_subject, 'gnd_subject').
bigdata_file(dewiki_page_links_ttl, 'dewiki-20140813-page-links.ttl.gz').
bigdata_file(dewiki_page_links_pattern, 'page_links_bornbefore_~w').
bigdata_file(geonames, 'geonames.txt').
bigdata_file(geonames_prolog, 'geonames_prolog').
bigdata_file(geonames_altnames, 'alternateNames_de_en_link.txt').
bigdata_file(yago_multilingual_instances, 'extract_deu_yagoMultilingualInstanceLabels.tsv').
bigdata_file(yago_prolog, 'yago_prolog').

:- initialization(( tmp_file(X, Y), register_file(X, tmp(Y)),  fail
		  ; true)).
:- initialization(( bigdata_file(X, Y), register_file(X, bigdata(Y)),  fail
		  ; true)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% DOWNLOADING AND PREPARING DATA
%%%%
%%%% The links point to archived copies of the original fact bases.  See
%%%% http://cs.christophwernhard.com/kbset/datasets.html for information on
%%%% them.
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prepare_data :-
	dep_ensure(all).

dep(['all'
    -items([ 'gnd_persons_bornbefore_1850.pl',
	     'gnd_subject.pl',
	     'geonames_prolog.pl',
	     'yago_prolog.pl',
	     'page_links_bornbefore_1850.pl']),
      
     'gnd_persons_bornbefore_1850.pl'
    -call(['authorities-person_lds_20201013.rdf.gz'],
	  prepare_gnd_persons_bornbefore(1850)),

     'gnd_subject.pl'
    -call(['authorities-sachbegriff_lds_20201013.rdf.gz'],
	  prepare_gnd_subject_name),

     'page_links_bornbefore_1850.pl'
    -call(['dewiki-20140813-page-links.ttl.gz', 
	   'gnd_persons_bornbefore_1850.pl'], prepare_dewiki_links(1850)),

     'geonames_prolog.pl'
    -call(['geonames.txt','alternateNames_de_en_link.txt'],
	  convert_geonames_to_prolog),
     
     'yago_prolog.pl'
    -call(['extract_deu_yagoMultilingualInstanceLabels.tsv'],
	  convert_wikinames_to_prolog),
     
     'downloads.dep'
    -items(['dewiki.dep','geonames.dep','gnd.dep','yago.dep']),

     'yago.dep'
    -items(['extract_deu_yagoMultilingualInstanceLabels.tsv']),
     
     'extract_deu_yagoMultilingualInstanceLabels.tsv'
    -shell(['yagoMultilingualInstanceLabels.tsv.7z'],
	   '7z e -so yagoMultilingualInstanceLabels.tsv.7z | grep "@deu" >extract_deu_yagoMultilingualInstanceLabels.tsv'),
     
     'yagoMultilingualInstanceLabels.tsv.7z'
    -download('http://cs.christophwernhard.com/kbset/webarchive/yago/2014-12-09/yagoMultilingualInstanceLabels.tsv.7z'),
     %%

     'dewiki.dep'
    -items(['dewiki-20140813-page-links.ttl.gz']),
     
     'dewiki-20140813-page-links.ttl.gz'
    -download('http://cs.christophwernhard.com/kbset/webarchive/dbpedia/2014-12-09/dewiki-20140813-page-links.ttl.gz'),
     
     'gnd.dep'
     -items(['authorities-person_lds_20201013.rdf.gz','authorities-sachbegriff_lds_20201013.rdf.gz']),
     
     'authorities-sachbegriff_lds_20201013.rdf.gz'
     -download('http://cs.christophwernhard.com/kbset/webarchive/gnd/2020-10-13/authorities-sachbegriff_lds_20201013.rdf.gz'),
%    -download('https://data.dnb.de/opendata/authorities-sachbegriff_lds_20201013.rdf.gz'),

     'authorities-person_lds_20201013.rdf.gz'
     -download('http://cs.christophwernhard.com/kbset/webarchive/gnd/2020-10-13/authorities-person_lds_20201013.rdf.gz'),
%    -download('https://data.dnb.de/opendata/authorities-person_lds_20201013.rdf.gz'),
     
     'geonames.dep'
    -items(['geonames.txt', 'alternateNames_de_en_link.txt']),

     'geonames.txt'
    -shell(['AT.txt', 'CH.txt', 'CZ.txt', 'DE.txt', 'FR.txt',
	    'GB.txt', 'PL.txt', 'cities15000.txt'],
	   'cat AT.txt CH.txt CZ.txt DE.txt FR.txt GB.txt PL.txt cities15000.txt | sort -u >geonames.txt'),

     'alternateNames_de_en_link.txt'
    -shell(['alternateNames.txt'],
	   'grep -P "\\tde\\t|\\ten\\t|\\tlink\\t" alternateNames.txt >alternateNames_de_en_link.txt'),

     'alternateNames.txt'-unzip('geonames_alternate_names.zip'),
     'cities15000.txt'-unzip('geonames_cities_15000.zip'),
     'AT.txt'-unzip('geonames_at.zip'),
     'CH.txt'-unzip('geonames_ch.zip'),
     'CZ.txt'-unzip('geonames_cz.zip'),
     'DE.txt'-unzip('geonames_de.zip'),
     'FR.txt'-unzip('geonames_fr.zip'),
     'GB.txt'-unzip('geonames_gb.zip'),
     'PL.txt'-unzip('geonames_pl.zip'),

     'geonames_at.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/AT.zip'),
     'geonames_ch.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/CH.zip'),
     'geonames_cz.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/CZ.zip'),
     'geonames_fr.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/FR.zip'),
     'geonames_de.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/DE.zip'),
     'geonames_gb.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/GB.zip'),
     'geonames_pl.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/PL.zip'),
     'geonames_cities_15000.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/cities15000.zip'),
     'geonames_alternate_names.zip'
    -download('http://cs.christophwernhard.com/kbset/webarchive/geonames/2019-11-13/alternateNames.zip')]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% LOADING PREPROCESSED EXTERNAL DATA INTO THE SYSTEM
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

load_data :-
	install_gnd_persons_bornbefore(1850),
	install_gnd_subjects,
	install_dewiki_links,
	install_wikinames,
	install_geonames,
	install_person_stuff,
	add_assistance_update_hook(person_ner:compile_person_assistance),
	add_assistance_update_hook(location_ner:compile_location_assistance),
	update_assistance,
	install_function_info,  %% requires persons from assistance completed
	update_assistance.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Configuring the Source Document Parsing
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- register_latex_command(xchaph, [t,c], '').
:- register_latex_command(xseca, [t,c], '').
:- register_latex_command(xinit, [t,c,c], '').
:- register_latex_command(xsic, [c,t], '').
:- register_latex_command(xsicm, [t], '').
:- register_latex_command(xsicw, [c], '').
:- register_latex_command(xnote, [c], '').
:- register_latex_command(xpage, [c], '').
:- register_latex_command(xpagep, [c], '').
:- register_latex_command(xmargin, [c], '').
:- register_latex_command(xnote, [c], '').
%% Tironean et
:- register_latex_command(etc, [c], '⁊').

kwd_section_commands([xchap,xchapn,xchaph,xchapnc]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Associating Annotation and Source Documents}
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

annotated_by(docdir('src/reaction_01_16.tex'),
	     docdir('apparat/annot_reaction_01.tex')).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% General Control of Named Entity Recognition
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% 
%%%% Which entities should be recognized
%%%% 
ner_processing_options([quotes,
			dates,
			function_persons,
			persons,
			locations]).

%%%% 
%%%% Feature lists for entity types, order in the lists determines
%%%% order of (lexical) comparison
%%%% 
entity_features(person,
		[not_explicitly_blocked,
		 explicitly_specified_in_context(_),
		 followed_by_matching_roman_number(_),
		 preceded_by_matching_first_names(_),
		 explicitly_specified,
		 preceded_by_matching_first_names_initials(_),
		 followed_by_matching_extra_names(_),
		 followed_by_matching_extension(_),
		 no_stopword,
		 no_common_substantive,
		 no_common_downcase,
		 no_common_geoname,
		 no_common_firstname,
		 already_identified_in_context,
		 linked_to_person_in_context(_),
		 referenced_by_preferred_name,
		 linked_to_many_others(_),
		 born_in_span_before_year_in_context(_, _),
		 in_wikipedia_de,
		 in_wikipedia,
		 linked_to_others(_)]).


entity_features(function_person,
		[not_explicitly_blocked,
		 followed_by_matching_specifier(_),
		 explicitly_specified_in_context(_),
		 explicitly_specified,
		 no_stopword,
		 no_common_substantive,
		 no_common_downcase,
		 no_common_geoname,
		 already_identified_in_context,
		 born_in_span_before_year_in_context(_, _),
		 linked_to_person_in_context(_),
		 in_wikipedia_de,
		 in_wikipedia,
		 linked_to_others(_)]).

entity_features(location,
		[not_explicitly_blocked,
		 explicitly_specified_in_context(_),
		 explicitly_specified,
		 no_stopword,
		 no_common_substantive,
		 no_common_downcase,
		 no_common_firstname,
		 no_common_lastname,
		 already_identified_in_context,
		 near_to_already_identified_in_context(_),
		 in_wikipedia,
		 is_country,
		 has_population_over_10000,
		 is_city_or_village,
		 population(_),
		 referenced_by_preferred_name
		 ]).

entity_features(date,
		[complete_parsed_date,
		 explicitly_defaulted_parsed_date(_,_),
		 defaulted_parsed_date(_,_),
		 incomplete_parsed_date]).

%%%% 
%%%% Features used as thresholds - to reject candidates straight away
%%%% 
%%%% Only atoms are allowed as threshold features
%%%% 
entity_threshold_features(person,
			  [not_explicitly_blocked,
			   no_stopword,
			   no_common_substantive,
			   no_common_downcase,
			   no_common_geoname,
			   no_common_firstname
			   ]).

entity_threshold_features(function_person,
			  [not_explicitly_blocked,
			   no_stopword,
			   no_common_substantive,
			   no_common_downcase,
			   no_common_geoname]).

entity_threshold_features(location,
			  [not_explicitly_blocked,
			   no_stopword,
			   no_common_substantive,
			   no_common_downcase,
			   no_common_firstname
			   ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Global Context Settings
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def_context(settings,
	    [user_language=de]).

def_context(ner_settings,
	    [person_ner_many_links_granularity=50,
	     person_ner_linked_context_window_size=5]).

def_context(semctx_settings,
	    [semctx_width_before=3000,
	     semctx_width_after=1000,
	     semctx_words_minfreq=0.002,
	     semctx_stems_minfreq=0.005,
	     semctx_key_ngrams_ngram_max_n=3,
	     semctx_key_ngrams_ngram_min_f=4,
	     semctx_key_ngrams_freq_percent=30,
	     semctx_key_ngrams_maxnum=12]).

def_context(ctx1,
	    [text_language=de,
	     text_language_variant=de(19),
	     stemming_language=de(19),
	     context_years=[1789]
	    ]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Supplementing Word Data
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- register_common_downcase('beständig').
:- register_common_downcase('blos').
:- register_common_downcase('ihrigen').
:- register_common_downcase('jedermann').
:- register_common_downcase('jemand').
:- register_common_downcase('redlichen').

:- register_common_substantive('Comite').
:- register_common_substantive('Eckstein').
:- register_common_substantive('Gabellen').
:- register_common_substantive('Grundherren').
:- register_common_substantive('Herreden').
:- register_common_substantive('Herredens').
:- register_common_substantive('Hofmeister').
:- register_common_substantive('König').
:- register_common_substantive('Königes').
:- register_common_substantive('Landmann').
:- register_common_substantive('Landmanns').
:- register_common_substantive('Landmesser').
:- register_common_substantive('Landmessers').
:- register_common_substantive('Muthwillen').
:- register_common_substantive('Pabst').
:- register_common_substantive('Pabstthums').
:- register_common_substantive('Regierer').
:- register_common_substantive('Rhetoren').
:- register_common_substantive('Römer').
:- register_common_substantive('Schmausen').
:- register_common_substantive('Ständehaus').
:- register_common_substantive('Ungelehrten').
:- register_common_substantive('Universitäts').
:- register_common_substantive('Unterrichter').
:- register_common_substantive('Warners').
:- register_common_substantive('Weingarten').
:- register_common_substantive('Werkmeister').
:- register_common_substantive('Zehnter').
:- register_common_substantive('Conseil').
:- register_common_substantive('Conseils').
:- register_common_substantive('Notarien').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Supplementing Further Entities
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- register_ext_person(cice_1725,
		       [preferredNameForThePerson
		       ='Cicé, Jean-Baptiste-Marie Champion de',
			dateOfBirth='1725-02-10',
			dateOfDeath='1805-08-06',
			gender=male,
			biographicalOrHistoricalInformation
		       =lang(de,'Erzbischof von Auxerre (1761-1790), Bischof von Auxerre'),
			url='https://fr.wikipedia.org/wiki/Jean-Baptiste-Marie_Champion_de_Cic%C3%A9']).

:- register_ext_person(lubersac_1740,
		       [preferredNameForThePerson
		       ='Lubersac, Jean-Baptiste-Joseph de',
			dateOfBirth='1740-04-15',
			dateOfDeath='1822-08-30',
			gender=male,
			biographicalOrHistoricalInformation
		       =lang(de,'Bischof von Chartres (1780-1790)'),
			url='https://fr.wikipedia.org/wiki/Jean-Baptiste-Joseph_de_Lubersac']).

:- register_ext_person(grandin_1755,
		       [preferredNameForThePerson
		       ='Grandin, François Henry Christophe',
			dateOfBirth='1755-04-16',
			dateOfDeath='1823-03-24',
			gender=male,
			name='Abbé Grandin',
			biographicalOrHistoricalInformation
		       =lang(de,'Curé d\' Ernée'),
			url='http://data.bnf.fr/14954963/']).

:- register_ext_person(tullius,
		       %% lost in the GND translation? Perhaps due to
		       %% missing dates in the GND?
		       [preferredNameForThePerson='Tullius, Servius',
			url='https://de.wikipedia.org/wiki/Servius_Tullius']).

:- register_ext_person(cethegus,
		       [preferredNameForThePerson='Cethegus, Gaius Cornelius',
			dateOfDeath='-063-12-05',
			url='https://de.wikipedia.org/wiki/Gaius_Cornelius_Cethegus_(Senator)']).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Specifying Assistance to Named Entity Recognition
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def_assistance(no_entities,
	       [no_entity(date,
			  [word='August',
			   followed_by_words=['Comte']]),
		no_entity(location,
			  [word_in=['Frankreich', 'Frankreichs',
				    'Galliens', 'Jus', 'Lan', 'Lit',
				    'Lits', 'Maire', 'Perigord',
				    'Latona', 'Germain', 'Gouy',
				    'Mallet',
				    'Dauphiné', 'Autun', 'Cours',
				    'Lafayette', 'Schilling', 'Ubi',
				    'Noailles', 'Großen',  'Groß',
				    'Sire', 'Souverain', 'Souverains',
				    'Layen', 'Non', 'Situs', 'Untern', 'Unwillens'
				    ]]),

		no_entity(person,
			  [word_in=['Delos', 'Calendis',
				    'Souverain', 'Souverains',
				    'Frankreich', 'Frankreichs',
				    'Du', 'Denys',
				    'Gallien', 'Galliens', 'Germain',
				    'Hannoverschen', 'Jam', 'Maj', 'Non',
				    'Sire', 'Situs', 'Cachet', 'Oesterreich',
				    'Alphius', 'Venedig', 'Goui', 'Gouy',
				    'Arcy', 'Mallet', 'Cours',
				    'Dauphiné', 'Oranien', 'Utrecht',
				    'Schilling', 'Bourdeaux', 'Layen',
				    'Picarden', 'Satyren', 'Servius', 'Rom',
				    'Beatus', 'Maire', 'Marquis',
				    'Silber', 'Silbers', 'Guyenne', 'Jus',
				    'France'
				   ]])]).


def_assistance(person_assistance_1,
	       [supplement(location,
			   [name='Bordeaux',
			    feature_class='P',
			    population_is_at_least='100000'],
			   [name='Bourdeaux']),
		supplement(person,
			   [name='Cazalès', dateOfBirth='1758'],
			   [name='Cazales']),
		%% Avoid collission with Spengler, Auguste = A. v. Gentz
		entity(person,
		       [name='Gentz, Friedrich von', dateOfBirth='1764']),
		entity(person,
		       [name='Duport, Adrien Jean François',
			dateOfBirth-'1759']),
		entity(person,
		       [name='Noailles, Louis Marie de',
			dateOfBirth='1756']),
		entity(person,
		       [name='Maistre, Joseph Marie de',
			dateOfBirth='1753']),
% 		%% Wikipedia link is missing in GND 2018-02
% 		%% and there is another person with last name Empedocles:
%  		entity(person,
%  		       [name='Empedokles']),
		entity(person,
		       [name='Lefranc de Pompignan, Jean Georges']),
		entity(person,
		       [name='Starcke, Johann Friedrich',
			professionOrOccupation='Drucker'],
		       [near_word_in=['Druck']]),
		entity(person,
		       [name='Rabaut Saint-Etienne, Jean Paul']),
		entity(person,
		       [name='Grégoire, Henri', dateOfBirth='1750'],
		       [near_word_in=['Abt']]),
		supplement(person,
			   [name='Grégoire, Henri', dateOfBirth='1750'],
			   [name='Gregoire, Henri']),
		supplement(person,
			   [name='Talleyrand-Périgord, Charles Maurice de'],
			   [biographicalOrHistoricalInformation
			  =lang(de,'Bischof von Autun')]),
		entity(person,
		       [name='Talleyrand-Périgord, Charles Maurice de'],
		       []),
		entity(person,
		       [name='Wilhelm III., England, König'],
		       [near_word_in=['1688']]),
		entity(person,
		       [name='Du Moulin, Charles'],
		       [near_word_in=['Coutumes']]),
		entity(person,
		       [name='Reubell, Jean-François']),
		entity(person,
		       [name='Louis XI., France, Roi'],
		       [near_word_in=['Pretention']]),
		entity(person,
		       [name='La Rochefoucauld-Liancourt, François-Alexandre-Frédéric de']),
		entity(person, %% GND has wrong date of birth: 1848
		       %% French wikipedia has 1846
		       [name='Legrand, Jérôme']),
		entity(person,
		       [name='Thouret, Jacques Guillaume']),
	        supplement(person,
			   [name='Karl X., Frankreich, König'],
			   [biographicalOrHistoricalInformation
			  =lang(de,'Graf von Artois')]),
		entity(person,
		       [name='Lally-Tolendal, Trophime-Gérard de']),
		supplement(person, %% ist das der richtige Dupont?
			   [name='Dupont de L\'Eure, Jacques-Charles'],
			   [name='Dupont']),
		entity(person,
		       [name='Dupont de L\'Eure, Jacques-Charles'],
		       [following_words=['zuerst']]),
		entity(person,
		       [name='Du Pont de Nemours, Pierre Samuel'],
		       [following_words=['Oekonomist', 'von']]),
		entity(person,
		       [name='Heinrich VIII., England, König'],
		       [near_word_in=['Rom']]),
		entity(person,
		       [name='Franz I., Frankreich, König'],
		       [near_word_in=['Heinrich']]),
		supplement(person,
			   [name='Joseph II., Heiliges Römisches Reich, Kaiser'],
			   [biographicalOrHistoricalInformation
			  =lang(de,'Herzog von Luxemburg (1765-1790)')]),
		supplement(person,
			   [name='Dulau, Jean-Marie'],
			  [biographicalOrHistoricalInformation
			  =lang(de,'Erzbischof von Arles (1775-1792)')]),
		entity(person,
		       [name='Helvétius, Claude Adrien',
			professionOrOccupation='Philosoph']),
		entity(person,
		       [name='Ivernois',
			professionOrOccupation='Schriftsteller']),
		entity(person,
		       [name='Mallet du Pan, Jacques',
			dateOfBirth='1749']),
		entity(person,
		       [name='La Rochefoucauld, Dominique de'],
		       [near_word_in=['Cardinal']]),
		entity(person,
		       [name='Lévis',
			dateOfBirth='1764',
			professionOrOccupation='Schriftsteller'],
		       [following_words=['Herzog', 'von']]),
		entity(person,
		       [name='Tacitus',
			professionOrOccupation='Historiker'],
		       [near_word_in=['Römern']]),
		entity(person,
		       [name='Tacitus',
			variantNameForThePerson='Tacitus, Römisches Reich, Kaiser'],
		       [near_word_in=['Adel']]),
		entity(person,
		       [name='Karl <der Große>'],
		       [followed_by_words=['des','Großen']]),
		entity(person,
		       [name='Locke', professionOrOccupation='Philosoph'],
		       [near_word_in=['Metaphysik', 'Philosoph']]),
		entity(person,
		       [name='Le Clerc de Juigné, Antoine Eléonore Léon'
		       %% Erzbischof von Paris in 1789
		       ])
	       ]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Specify Assistance Components that Should Take Effect
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_effective([settings,
		semctx_settings,
		ctx1,
		ner_settings,
		person_assistance_1,
		no_entities]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


